import numpy as np

###--------------------------------------------------------------------------------------------------###
###MEASURED PARAMETER STANDARDISATIONS
###--------------------------------------------------------------------------------------------------###

#Define dictionary with details pertaining to standardised measured parameters
 
standard_parameters ={  
#GAS
'O3':                    {'long_parameter_name':'ozone',                                        'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':1.0,   'extreme_lower_limit':0.0, 'extreme_upper_limit':400.0,    'extreme_upper_monthly_median':120.0,   'aqs_code':['44201'], 'airbase_name':'Ozone (air)',                             'airbase_code':'00007', 'miteco_code':'14', 'ebas_parameter_name':['ozone'],                               'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['O3'],                                          'castnet_parameter_name':'OZONE',                                  'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':['06','42'], 'japan_nies_name':['OX','O3'],           'uk_air_name':['Ozone'],                                                  'chemical_formula':'O3',      'bsc_parameter_name':'sconco3',     'identifiers':{'IUPAC_name':'ozone',                  'CAS_number':'10028-15-6', 'PubChem_CID':'24823',   'InChIKey':'CBENFWSGALASAD-UHFFFAOYSA-N', 'WIGOS_name':'O3 (ozone)',                                                         'WIGOS_number':'12164'}},
'NO':                    {'long_parameter_name':'nitrogen monoxide',                            'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':1.0,   'extreme_lower_limit':0.0, 'extreme_upper_limit':1200.0,   'extreme_upper_monthly_median':250.0,   'aqs_code':['42601'], 'airbase_name':'Nitrogen monoxide (air)',                 'airbase_code':'00038', 'miteco_code':'7',  'ebas_parameter_name':['nitrogen_monoxide'],                   'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['NO'],                                          'castnet_parameter_name':'NO',                                     'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':['02'],      'japan_nies_name':['NO'],                'uk_air_name':['Nitric oxide'],                                           'chemical_formula':'NO',      'bsc_parameter_name':'sconcno',     'identifiers':{'IUPAC_name':'nitric oxide',           'CAS_number':'10102-43-9', 'PubChem_CID':'145068',  'InChIKey':'MWUXSHHQAYIFBG-UHFFFAOYSA-N', 'WIGOS_name':'NO (nitrogen monoxide)',                                             'WIGOS_number':'395'}},
'NO2':                   {'long_parameter_name':'nitrogen dioxide',                             'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':1.0,   'extreme_lower_limit':0.0, 'extreme_upper_limit':600.0,    'extreme_upper_monthly_median':200.0,   'aqs_code':['42602'], 'airbase_name':'Nitrogen dioxide (air)',                  'airbase_code':'00008', 'miteco_code':'8',  'ebas_parameter_name':['nitrogen_dioxide'],                    'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['NO2'],                                         'castnet_parameter_name':'NO2_TRUE',                               'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':['03'],      'japan_nies_name':['NO2'],               'uk_air_name':['Nitrogen dioxide'],                                       'chemical_formula':'NO2',     'bsc_parameter_name':'sconcno2',    'identifiers':{'IUPAC_name':'nitrogen dioxide',       'CAS_number':'10102-44-0', 'PubChem_CID':'3032552', 'InChIKey':'JCXJVPUVTGWSNB-UHFFFAOYSA-N', 'WIGOS_name':'NO2 (nitrogen dioxide)',                                             'WIGOS_number':'396'}},
'SO2':                   {'long_parameter_name':'sulphur dioxide',                              'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':2.0,   'extreme_lower_limit':0.0, 'extreme_upper_limit':3000.0,   'extreme_upper_monthly_median':750.0,   'aqs_code':['42401'], 'airbase_name':'Sulphur dioxide (air)',                   'airbase_code':'00001', 'miteco_code':'1',  'ebas_parameter_name':['sulphur_dioxide'],                     'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['SO2','Sulphur Dioxide'],                       'castnet_parameter_name':{'hourly_gas':'SO2_GA','drychem':'WSO2'}, 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':['01'],      'japan_nies_name':['SO2'],               'uk_air_name':['Sulphur dioxide','gaseous sulphur dioxide','a_SO2_g'],    'chemical_formula':'SO2',     'bsc_parameter_name':'sconcso2',    'identifiers':{'IUPAC_name':'sulfur dioxide',         'CAS_number':'7446-09-5',  'PubChem_CID':'1119',    'InChIKey':'RAHZWNYVWXNFOC-UHFFFAOYSA-N', 'WIGOS_name':'SO2',                                                                'WIGOS_number':'430'}},
'CO':                    {'long_parameter_name':'carbon monoxide',                              'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':20.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':30000.0,  'extreme_upper_monthly_median':7500.0,  'aqs_code':['42101'], 'airbase_name':'Carbon monoxide (air)',                   'airbase_code':'00010', 'miteco_code':'6',  'ebas_parameter_name':['carbon_monoxide'],                     'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['CO'],                                          'castnet_parameter_name':'CO',                                     'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':['05'],      'japan_nies_name':['CO'],                'uk_air_name':['Carbon monoxide'],                                        'chemical_formula':'CO',      'bsc_parameter_name':'sconcco',     'identifiers':{'IUPAC_name':'carbon monoxide',        'CAS_number':'630-08-0',   'PubChem_CID':'281',     'InChIKey':'UGFAIRIUMAVXCW-UHFFFAOYSA-N', 'WIGOS_name':'CO',                                                                 'WIGOS_number':'284'}},
'CH4':                   {'long_parameter_name':'methane',                                      'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':20.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':50000.0,  'extreme_upper_monthly_median':5000.0,  'aqs_code':['43201'], 'airbase_name':'Methane (air)',                           'airbase_code':'00041', 'miteco_code':'',   'ebas_parameter_name':['methane'],                             'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Methane'],                                     'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':['08'],      'japan_nies_name':['CH4'],               'uk_air_name':[],                                                         'chemical_formula':'CH4',     'bsc_parameter_name':'sconcch4',    'identifiers':{'IUPAC_name':'methane',                'CAS_number':'74-82-8',    'PubChem_CID':'297',     'InChIKey':'VNWKTOKETHGBQD-UHFFFAOYSA-N', 'WIGOS_name':'CH4 (methane)',                                                      'WIGOS_number':'192'}},
'ACETYLENE':             {'long_parameter_name':'acetylene',                                    'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':[''],      'airbase_name':'Ethyne (Acetylene) (air)',                'airbase_code':'00432', 'miteco_code':'',   'ebas_parameter_name':['ethyne'],                              'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Acetylene'],                                   'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['ethyne'],                                                 'chemical_formula':'C2H2',    'bsc_parameter_name':'sconcc2h2',   'identifiers':{'IUPAC_name':'acetylene',              'CAS_number':'74-86-2',    'PubChem_CID':'6326',    'InChIKey':'HSFWRNGVRCDJHI-UHFFFAOYSA-N', 'WIGOS_name':'C2H2 (ethyne, acetylene)',                                           'WIGOS_number':'434'}},
'GLYOXAL':               {'long_parameter_name':'glyoxal',                                      'matrix':'gas', 'standard_units':'',           'axis_label':'concentration', 'minimum_resolution':'',    'extreme_lower_limit':0.0, 'extreme_upper_limit':'',       'extreme_upper_monthly_median':'',      'aqs_code':[''],      'airbase_name':'ethanedial (air)',                        'airbase_code':'00429', 'miteco_code':'',   'ebas_parameter_name':['ethanedial'],                          'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                                         'chemical_formula':'C2H2O2',  'bsc_parameter_name':'sconcglyox',  'identifiers':{'IUPAC_name':'oxaldehyde',             'CAS_number':'107-22-2',   'PubChem_CID':'7860',    'InChIKey':'LEQAOMBKQFMDFZ-UHFFFAOYSA-N', 'WIGOS_name':'C2H2O2 (glyoxal)',                                                   'WIGOS_number':'12135'}},
'ETHENE':                {'long_parameter_name':'ethene',                                       'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':['43203'], 'airbase_name':'Ethene (Ethylene) (air)',                 'airbase_code':'00430', 'miteco_code':'',   'ebas_parameter_name':['ethene'],                              'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Ethylene'],                                    'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['ethene'],                                                 'chemical_formula':'C2H4',    'bsc_parameter_name':'sconcc2h4',   'identifiers':{'IUPAC_name':'ethene',                 'CAS_number':'74-85-1',    'PubChem_CID':'6325',    'InChIKey':'VGGSQFUCUMXWEO-UHFFFAOYSA-N', 'WIGOS_name':'C2H4 (ethene)',                                                      'WIGOS_number':'436'}},
'ACETALDEHYDE':          {'long_parameter_name':'acetaldehyde',                                 'matrix':'gas', 'standard_units':'',           'axis_label':'concentration', 'minimum_resolution':'',    'extreme_lower_limit':0.0, 'extreme_upper_limit':'',       'extreme_upper_monthly_median':'',      'aqs_code':[''],      'airbase_name':'ethanal (air)',                           'airbase_code':'00427', 'miteco_code':'',   'ebas_parameter_name':['ethanal'],                             'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Acetaldehyde'],                                'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                                         'chemical_formula':'C2H4O',   'bsc_parameter_name':'sconcald2',   'identifiers':{'IUPAC_name':'acetaldehyde',           'CAS_number':'75-07-0',    'PubChem_CID':'177',     'InChIKey':'IKHGUXGNUITLKF-UHFFFAOYSA-N', 'WIGOS_name':'CH3CHO (acetaldehyde, ethanal)',                                     'WIGOS_number':'491'}},
'ETHANE':                {'long_parameter_name':'ethane',                                       'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':['43202'], 'airbase_name':'Ethane (air)',                            'airbase_code':'00428', 'miteco_code':'',   'ebas_parameter_name':['ethane'],                              'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Ethane'],                                      'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['ethane'],                                                 'chemical_formula':'C2H6',    'bsc_parameter_name':'sconcc2h6',   'identifiers':{'IUPAC_name':'ethane',                 'CAS_number':'74-84-0',    'PubChem_CID':'6324',    'InChIKey':'OTMSDBZUPAUEDD-UHFFFAOYSA-N', 'WIGOS_name':'C2H6 (ethane)',                                                      'WIGOS_number':'437'}},
'ETHANOL':               {'long_parameter_name':'ethanol',                                      'matrix':'gas', 'standard_units':'',           'axis_label':'concentration', 'minimum_resolution':'',    'extreme_lower_limit':0.0, 'extreme_upper_limit':'',       'extreme_upper_monthly_median':'',      'aqs_code':[''],      'airbase_name':'',                                        'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['ethanol'],                             'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                                         'chemical_formula':'C2H6O',   'bsc_parameter_name':'sconcetoh',   'identifiers':{'IUPAC_name':'ethanol',                'CAS_number':'64-17-5',    'PubChem_CID':'702',     'InChIKey':'LFQSCWFLJHTTHZ-UHFFFAOYSA-N', 'WIGOS_name':'CH3CH2OH (ethanol)',                                                 'WIGOS_number':'490'}},
'PROPENE':               {'long_parameter_name':'propene',                                      'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':['43205'], 'airbase_name':'Propene (air)',                           'airbase_code':'00505', 'miteco_code':'',   'ebas_parameter_name':['propene'],                             'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Propylene'],                                   'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['propene'],                                                'chemical_formula':'C3H6',    'bsc_parameter_name':'sconcc3h6',   'identifiers':{'IUPAC_name':'prop-1-ene',             'CAS_number':'115-07-1',   'PubChem_CID':'8252',    'InChIKey':'QQONPFPTGQHPMA-UHFFFAOYSA-N', 'WIGOS_name':'C3H6 (propene)',                                                     'WIGOS_number':'442'}},
'PROPANE':               {'long_parameter_name':'propane',                                      'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':['43204'], 'airbase_name':'Propane (air)',                           'airbase_code':'00503', 'miteco_code':'',   'ebas_parameter_name':['propane'],                             'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Propane'],                                     'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['propane'],                                                'chemical_formula':'C3H8',    'bsc_parameter_name':'sconcc3h8',   'identifiers':{'IUPAC_name':'propane',                'CAS_number':'74-98-6',    'PubChem_CID':'6334',    'InChIKey':'ATUOYWHBWRKTHZ-UHFFFAOYSA-N', 'WIGOS_name':'C3H8 (propane)',                                                     'WIGOS_number':'445'}},
'1,3-BUTADIENE':         {'long_parameter_name':'1,3-butadiene',                                'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':[''],      'airbase_name':'1.3 Butadiene (air)',                     'airbase_code':'00024', 'miteco_code':'',   'ebas_parameter_name':['1-3-butadiene'],                       'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['1,3-Butadiene'],                               'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['1,3-butadiene'],                                          'chemical_formula':'C4H6',    'bsc_parameter_name':'sconcc4h6',   'identifiers':{'IUPAC_name':'buta-1,3-diene',         'CAS_number':'106-99-0',   'PubChem_CID':'7845',    'InChIKey':'KAKZBPTYRLMSJV-UHFFFAOYSA-N', 'WIGOS_name':'C4H6 (1,3-butadiene, butadiene)',                                    'WIGOS_number':'446'}},
'1-BUTENE':              {'long_parameter_name':'1-butene',                                     'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':[''],      'airbase_name':'1-Butene (air)',                          'airbase_code':'06005', 'miteco_code':'',   'ebas_parameter_name':['1-butene'],                            'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['1-Butene/Isobutene'],                          'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['1-butene'],                                               'chemical_formula':'C4H8',    'bsc_parameter_name':'sconcc4h8',   'identifiers':{'IUPAC_name':'but-1-ene',              'CAS_number':'106-98-9',   'PubChem_CID':'7844',    'InChIKey':'VXNZUUAINFGPBY-UHFFFAOYSA-N', 'WIGOS_name':'C4H8 (1-butene)',                                                    'WIGOS_number':'449'}},
'ISOBUTANE':             {'long_parameter_name':'isobutane',                                    'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':[''],      'airbase_name':'i-Butane (2-methylpropane) (air)',        'airbase_code':'00447', 'miteco_code':'',   'ebas_parameter_name':['2-methylpropane'],                     'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Isobutane'],                                   'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['iso-butane'],                                             'chemical_formula':'C4H10',   'bsc_parameter_name':'sconcc4h10',  'identifiers':{'IUPAC_name':'2-methylpropane',        'CAS_number':'75-28-5',    'PubChem_CID':'6360',    'InChIKey':'NNPPMTNAJDCUHE-UHFFFAOYSA-N', 'WIGOS_name':'i-C4H10 (2-methylpropane, iso-butane)',                              'WIGOS_number':'497'}},
'ISOPRENE':              {'long_parameter_name':'isoprene',                                     'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':['43243'], 'airbase_name':'Isoprene (air)',                          'airbase_code':'00451', 'miteco_code':'',   'ebas_parameter_name':['isoprene'],                            'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Isoprene'],                                    'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['isoprene'],                                               'chemical_formula':'C5H8',    'bsc_parameter_name':'sconcisop',   'identifiers':{'IUPAC_name':'2-methylbuta-1,3-diene', 'CAS_number':'78-79-5',    'PubChem_CID':'6557',    'InChIKey':'RRHGJUQNOFWUDK-UHFFFAOYSA-N', 'WIGOS_name':'C5H8 (2-methyl-1,3-butadiene, isoprene)',                            'WIGOS_number':'464'}},
'N-PENTANE':             {'long_parameter_name':'n-pentane',                                    'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':[''],      'airbase_name':'n-Pentane (air)',                         'airbase_code':'00486', 'miteco_code':'',   'ebas_parameter_name':['n-pentane'],                           'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Pentane'],                                     'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['n-pentane'],                                              'chemical_formula':'C5H12',   'bsc_parameter_name':'sconcc5h12',  'identifiers':{'IUPAC_name':'pentane',                'CAS_number':'109-66-0',   'PubChem_CID':'8003',    'InChIKey':'OFBQJSOFQDEBGM-UHFFFAOYSA-N', 'WIGOS_name':'n-C5H12 (n-pentane)',                                                'WIGOS_number':'502'}},
'BENZENE':               {'long_parameter_name':'benzene',                                      'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':['45201'], 'airbase_name':'Benzene (air)',                           'airbase_code':'00020', 'miteco_code':'30', 'ebas_parameter_name':['benzene'],                             'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Benzene'],                                     'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['benzene','BENZENE'],                                      'chemical_formula':'C6H6',    'bsc_parameter_name':'sconcc6h6',   'identifiers':{'IUPAC_name':'benzene',                'CAS_number':'71-43-2',    'PubChem_CID':'241',     'InChIKey':'UHOVQNZJYSORNB-UHFFFAOYSA-N', 'WIGOS_name':'C6H6 (benzene)',                                                     'WIGOS_number':'476'}},
'N-HEXANE':              {'long_parameter_name':'n-hexane',                                     'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':[''],      'airbase_name':'n-Hexane (air)',                          'airbase_code':'00443', 'miteco_code':'',   'ebas_parameter_name':['n-hexane'],                            'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Hexane'],                                      'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['n-hexane'],                                               'chemical_formula':'C6H14',   'bsc_parameter_name':'sconcc6h14',  'identifiers':{'IUPAC_name':'hexane',                 'CAS_number':'110-54-3',   'PubChem_CID':'8058',    'InChIKey':'VLKZOEOYAKHREP-UHFFFAOYSA-N', 'WIGOS_name':'C6H14 (n-hexane)',                                                   'WIGOS_number':'474'}},
'TOLUENE':               {'long_parameter_name':'toluene',                                      'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':['45202'], 'airbase_name':'Toluene (air)',                           'airbase_code':'00021', 'miteco_code':'20', 'ebas_parameter_name':['toluene'],                             'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Toluene'],                                     'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['toluene'],                                                'chemical_formula':'C7H8',    'bsc_parameter_name':'sconcc7h8',   'identifiers':{'IUPAC_name':'toluene',                'CAS_number':'108-88-3',   'PubChem_CID':'1140',    'InChIKey':'YXFVVABEGXRONW-UHFFFAOYSA-N', 'WIGOS_name':'C7H8 (toluene)',                                                     'WIGOS_number':'482'}},
'XYLENE':                {'long_parameter_name':'xylene',                                       'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':[],        'airbase_name':'Xylene (air)',                            'airbase_code':'00078', 'miteco_code':'31', 'ebas_parameter_name':[''],                          'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                             'chemical_formula':'C8H10',   'bsc_parameter_name':'sconcxyl',    'identifiers':{'IUPAC_name':'dimethylbenzene',        'CAS_number':'1330-20-7',  'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'C8H10 (Sum of p-xylene and m-xylene, 1,3- and 1,4-dimethylbenzene)', 'WIGOS_number':'486'}},
'MP-XYLENE':             {'long_parameter_name':'mp-xylene',                                    'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':[],        'airbase_name':'m,p-Xylene (air)',                        'airbase_code':'00464', 'miteco_code':'', 'ebas_parameter_name':['m-p-xylene'],                          'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['m and p-Xylene'],                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                             'chemical_formula':'C8H10',   'bsc_parameter_name':'sconcmpxyl',    'identifiers':{'IUPAC_name':'dimethylbenzene',        'CAS_number':'1330-20-7',  'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'C8H10 (Sum of p-xylene and m-xylene, 1,3- and 1,4-dimethylbenzene)', 'WIGOS_number':'486'}},
'M-XYLENE':              {'long_parameter_name':'m-xylene',                                     'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':[],        'airbase_name':'m-Xylene (air)',                          'airbase_code':'00081', 'miteco_code':'', 'ebas_parameter_name':['m-xylene'],                          'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                             'chemical_formula':'C8H10',   'bsc_parameter_name':'sconcmxyl',    'identifiers':{'IUPAC_name':'dimethylbenzene',        'CAS_number':'1330-20-7',  'PubChem_CID':'7929',        'InChIKey':'',                            'WIGOS_name':'C8H10 (Sum of p-xylene and m-xylene, 1,3- and 1,4-dimethylbenzene)', 'WIGOS_number':'486'}},
'P-XYLENE':              {'long_parameter_name':'p-xylene',                                     'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':[],        'airbase_name':'p-Xylene (air)',                          'airbase_code':'00080', 'miteco_code':'', 'ebas_parameter_name':['p-xylene'],                          'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                             'chemical_formula':'C8H10',   'bsc_parameter_name':'sconcpxyl',    'identifiers':{'IUPAC_name':'1,4-xylene',        'CAS_number':'106-42-3',  'PubChem_CID':'7809',        'InChIKey':'URLKBWYHVLBVBO-UHFFFAOYSA-N',                            'WIGOS_name':'C8H10 (Sum of p-xylene and m-xylene, 1,3- and 1,4-dimethylbenzene)', 'WIGOS_number':'486'}},
'O-XYLENE':              {'long_parameter_name':'o-xylene',                                     'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':[],        'airbase_name':'o-Xylene (air)',                          'airbase_code':'00482', 'miteco_code':'', 'ebas_parameter_name':['o-xylene'],                          'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                             'chemical_formula':'C8H10',   'bsc_parameter_name':'sconcoxyl',    'identifiers':{'IUPAC_name':'1,3-xylene',        'CAS_number':'95-47-6',  'PubChem_CID':'7237',        'InChIKey':'CTQNGGLPUBDAKN-UHFFFAOYSA-N',                            'WIGOS_name':'C8H10 (Sum of p-xylene and m-xylene, 1,3- and 1,4-dimethylbenzene)', 'WIGOS_number':'486'}},
'1,2,4-TRIMETHYLBENZENE':{'long_parameter_name':'1,2,4-trimethylbenzene',                       'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':[''],      'airbase_name':'1,2,4-Trimethylbenzene (air)',            'airbase_code':'06011', 'miteco_code':'',   'ebas_parameter_name':['1-2-4-trimethylbenzene'],              'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['1,2,4-Trimethylbenzene'],                      'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['1,2,4-trimethylbenzene'],                                 'chemical_formula':'C9H12',   'bsc_parameter_name':'sconcc9h12',  'identifiers':{'IUPAC_name':'1,2,4-trimethylbenzene', 'CAS_number':'95-63-6',    'PubChem_CID':'7247',    'InChIKey':'GWHJZXXIDMPWGX-UHFFFAOYSA-N', 'WIGOS_name':'C9H12 (1,2,4-trimethylbenzene)',                                     'WIGOS_number':'487'}},
'MONOTERPENES':          {'long_parameter_name':'monoterpenes',                                 'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':100.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':500000.0, 'extreme_upper_monthly_median':50000.0, 'aqs_code':[],        'airbase_name':'',                                        'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['monoterpenes'],                        'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                                         'chemical_formula':'C10H16',  'bsc_parameter_name':'sconcc10h16', 'identifiers':{'IUPAC_name':'',                       'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                                   'WIGOS_number':''}},
'BENZO[a]PYRENE':        {'long_parameter_name':'benzo[a]pyrene',                               'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':0.01,  'extreme_lower_limit':0.0, 'extreme_upper_limit':10.0,     'extreme_upper_monthly_median':2.0,     'aqs_code':[],        'airbase_name':'Benzo(a)pyrene (air+aerosol)',            'airbase_code':'06015', 'miteco_code':'',   'ebas_parameter_name':['benzo_a_pyrene'],                      'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Benzo(a)Pyrene (BaP)','Benzo(a)Pyrene','BaP'], 'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                                         'chemical_formula':'C20H12',  'bsc_parameter_name':'sconcbap',    'identifiers':{'IUPAC_name':'benzo[a]pyrene',         'CAS_number':'50-32-8',    'PubChem_CID':'2336',    'InChIKey':'FMMWHPNWAFZXNH-UHFFFAOYSA-N', 'WIGOS_name':'',                                                                   'WIGOS_number':''}},
'NMVOC':                 {'long_parameter_name':'total non-methane volatile organic compounds', 'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':20.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':20000.0,  'extreme_upper_monthly_median':5000.0,  'aqs_code':['43102'], 'airbase_name':'',                                        'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                      'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                                         'chemical_formula':'HC',      'bsc_parameter_name':'sconcnmvoc',  'identifiers':{'IUPAC_name':'',                       'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                                   'WIGOS_number':''}},
'VOC':                   {'long_parameter_name':'total volatile organic compounds',             'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':50.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':70000.0,  'extreme_upper_monthly_median':10000.0, 'aqs_code':['43104'], 'airbase_name':'Total volatile organic compounds (air)',  'airbase_code':'00033', 'miteco_code':'',   'ebas_parameter_name':['voc_complete'],                        'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                                         'chemical_formula':'HC',      'bsc_parameter_name':'sconcvoc',    'identifiers':{'IUPAC_name':'',                       'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                                   'WIGOS_number':''}},
'NMHC':                  {'long_parameter_name':'total non-methane hydrocarbons',               'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':20.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':20000.0,  'extreme_upper_monthly_median':5000.0,  'aqs_code':[],        'airbase_name':'Total non-methane hydrocarbons (air)',    'airbase_code':'00032', 'miteco_code':'',   'ebas_parameter_name':[],                                      'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':['07'],      'japan_nies_name':['NMHC'],              'uk_air_name':[],                                                         'chemical_formula':'HC',      'bsc_parameter_name':'sconnmhc',    'identifiers':{'IUPAC_name':'',                       'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'NMHC',                                                               'WIGOS_number':'494'}},
'HC':                    {'long_parameter_name':'total hydrocarbons',                           'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':50.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':70000.0,  'extreme_upper_monthly_median':10000.0, 'aqs_code':['43101'], 'airbase_name':'',                                        'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                      'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':['09'],      'japan_nies_name':['THC','THCP','THCM'], 'uk_air_name':[],                                                         'chemical_formula':'HC',      'bsc_parameter_name':'sconchc',     'identifiers':{'IUPAC_name':'',                       'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                                   'WIGOS_number':''}},
'NH3':                   {'long_parameter_name':'ammonia',                                      'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':1.0,   'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,   'extreme_upper_monthly_median':100.0,   'aqs_code':['42604'], 'airbase_name':'Ammonia (air)',                           'airbase_code':'00035', 'miteco_code':'',   'ebas_parameter_name':['ammonia'],                             'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Ammonia','AMMONIA'],                           'castnet_parameter_name':'NH3',                                    'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['gaseous ammonia','alpha_NH3_g','delta_NH3_g','DT_NH3_g'], 'chemical_formula':'NH3',     'bsc_parameter_name':'sconcnh3',    'identifiers':{'IUPAC_name':'azane',                  'CAS_number':'7664-41-7',  'PubChem_CID':'222',     'InChIKey':'QGZKDVFQNNGYKY-UHFFFAOYSA-N', 'WIGOS_name':'NH3 (Ammonia)',                                                      'WIGOS_number':'390'}},
'HNO3':                  {'long_parameter_name':'nitric acid',                                  'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':0.1,   'extreme_lower_limit':0.0, 'extreme_upper_limit':25.0,     'extreme_upper_monthly_median':5.0,     'aqs_code':['42305'], 'airbase_name':'Gaseous nitric acid (air)',               'airbase_code':'00050', 'miteco_code':'',   'ebas_parameter_name':['nitric_acid'],                         'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Nitric Acid','NITRIC ACID'],                   'castnet_parameter_name':{'hourly_gas':'HNO3','drychem':'NHNO3'},  'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['gaseous nitric acid','a_HNO3_g'],                         'chemical_formula':'HNO3',    'bsc_parameter_name':'sconchno3',   'identifiers':{'IUPAC_name':'nitric acid',            'CAS_number':'7697-37-2',  'PubChem_CID':'944',     'InChIKey':'GRYLNZFGIOXLOG-UHFFFAOYSA-N', 'WIGOS_name':'HNO3 (nitric acid)',                                                 'WIGOS_number':'394'}},
'PAN':                   {'long_parameter_name':'peroxyacetyl nitrate',                         'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':0.1,   'extreme_lower_limit':0.0, 'extreme_upper_limit':25.0,     'extreme_upper_monthly_median':5.0,     'aqs_code':['44301'], 'airbase_name':'Peroxyacetyl nitrate (air)',              'airbase_code':'00034', 'miteco_code':'',   'ebas_parameter_name':[],                                      'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                                         'chemical_formula':'C2H3NO5', 'bsc_parameter_name':'sconcpan',    'identifiers':{'IUPAC_name':'nitro ethaneperoxoate',  'CAS_number':'2278-22-0',  'PubChem_CID':'16782',   'InChIKey':'VGQXTTSVLMQFHM-UHFFFAOYSA-N', 'WIGOS_name':'C2H3O5N (peroxyacetylnitrate, PAN)',                                 'WIGOS_number':'391'}},
'HCHO':                  {'long_parameter_name':'formaldehyde',                                 'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':0.2,   'extreme_lower_limit':0.0, 'extreme_upper_limit':100.0,    'extreme_upper_monthly_median':25.0,    'aqs_code':['43502'], 'airbase_name':'Formaldehyde (air)',                      'airbase_code':'00025', 'miteco_code':'',   'ebas_parameter_name':['methanal'],                            'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':['Formaldehyde'],                                'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':[],                                                         'chemical_formula':'CH2O',    'bsc_parameter_name':'sconchcho',   'identifiers':{'IUPAC_name':'formaldehyde',           'CAS_number':'50-00-0',    'PubChem_CID':'712',     'InChIKey':'WSFSSNUMVMOOMR-UHFFFAOYSA-N', 'WIGOS_name':'CH2O (methanal, formaldehyde)',                                      'WIGOS_number':'489'}},
'HCl':                   {'long_parameter_name':'hydrochloric acid',                            'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':0.1,   'extreme_lower_limit':0.0, 'extreme_upper_limit':25.0,     'extreme_upper_monthly_median':5.0,     'aqs_code':['42302'], 'airbase_name':'Hydrogen chloride (air)',                 'airbase_code':'00039', 'miteco_code':'',   'ebas_parameter_name':['hydrochloric_acid'],                   'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':['43'],      'japan_nies_name':['HCL'],               'uk_air_name':['gaseous hydrochloric acid','a_HCl_g'],                    'chemical_formula':'HCl',     'bsc_parameter_name':'sconchcl',    'identifiers':{'IUPAC_name':'chlorane',               'CAS_number':'7647-01-0',  'PubChem_CID':'313',     'InChIKey':'VEXZGXHMUGYJMC-UHFFFAOYSA-N', 'WIGOS_name':'HCl (hydrochloric acid)',                                            'WIGOS_number':'288'}},
'HF':                    {'long_parameter_name':'hydrofluoric acid',                            'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':1.0,   'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,   'extreme_upper_monthly_median':200.0,   'aqs_code':['42303'], 'airbase_name':'Hydrogen fluoride (air)',                 'airbase_code':'00040', 'miteco_code':'',   'ebas_parameter_name':[],                                      'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':['44'],      'japan_nies_name':['HF'],                'uk_air_name':[],                                                         'chemical_formula':'HF',      'bsc_parameter_name':'sconchf',     'identifiers':{'IUPAC_name':'fluorane',               'CAS_number':'7664-39-3',  'PubChem_CID':'14917',   'InChIKey':'KRHYYFGTRYWZRS-UHFFFAOYSA-N', 'WIGOS_name':'HF (hydrofluoric acid)',                                             'WIGOS_number':'289'}},
'H2S':                   {'long_parameter_name':'hydrogen sulphide',                            'matrix':'gas', 'standard_units':'nmol mol-1', 'axis_label':'concentration', 'minimum_resolution':1.0,   'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,   'extreme_upper_monthly_median':200.0,   'aqs_code':['42402'], 'airbase_name':'Hydrogen sulphide (air)',                 'airbase_code':'00011', 'miteco_code':'',   'ebas_parameter_name':[],                                      'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':['45'],      'japan_nies_name':['H2S'],               'uk_air_name':[],                                                         'chemical_formula':'H2S',     'bsc_parameter_name':'sconch2s',    'identifiers':{'IUPAC_name':'sulfane',                'CAS_number':'7783-06-4',  'PubChem_CID':'402',     'InChIKey':'RWSOTUBLDIXVET-UHFFFAOYSA-N', 'WIGOS_name':'',                                                                   'WIGOS_number':''}},
'Hg-GEM':                {'long_parameter_name':'gaseous elemental mercury',                    'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':'',    'extreme_lower_limit':0.0, 'extreme_upper_limit':'',       'extreme_upper_monthly_median':'',      'aqs_code':[''],      'airbase_name':'Elemental Gaseous Mercury (air+aerosol)', 'airbase_code':'4013',  'miteco_code':'',   'ebas_parameter_name':['mercury_GEM'],                         'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['elemental mercury'],                                      'chemical_formula':'Hg',      'bsc_parameter_name':'sconchggem',  'identifiers':{'IUPAC_name':'mercury',                'CAS_number':'7439-97-6',  'PubChem_CID':'23931',   'InChIKey':'QSHDDOUJBYECFT-UHFFFAOYSA-N', 'WIGOS_name':'',                                                                   'WIGOS_number':''}},
'Hg-GOM':                {'long_parameter_name':'gaseous oxidised mercury',                     'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':'',    'extreme_lower_limit':0.0, 'extreme_upper_limit':'',       'extreme_upper_monthly_median':'',      'aqs_code':[''],      'airbase_name':'reactive_mercury (air+aerosol)',          'airbase_code':'653',   'miteco_code':'',   'ebas_parameter_name':['mercury_GOM'],                         'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['reactive mercury'],                                       'chemical_formula':'Hg',      'bsc_parameter_name':'sconchggom',  'identifiers':{'IUPAC_name':'mercury',                'CAS_number':'7439-97-6',  'PubChem_CID':'23931',   'InChIKey':'QSHDDOUJBYECFT-UHFFFAOYSA-N', 'WIGOS_name':'',                                                                   'WIGOS_number':''}},
'Hg-TGM':                {'long_parameter_name':'total gaseous mercury',                        'matrix':'gas', 'standard_units':'pmol mol-1', 'axis_label':'concentration', 'minimum_resolution':'',    'extreme_lower_limit':0.0, 'extreme_upper_limit':'',       'extreme_upper_monthly_median':'',      'aqs_code':[''],      'airbase_name':'Total gaseous mercury (air+aerosol)',     'airbase_code':'4813',  'miteco_code':'',   'ebas_parameter_name':['mercury_TGM','total_gaseous_mercury'], 'ebas_matrix':'air', 'ebas_priority_units':['mol/mol','/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',                                       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],          'japan_nies_name':[],                    'uk_air_name':['total gaseous mercury'],                                  'chemical_formula':'Hg',      'bsc_parameter_name':'sconchgtgm',  'identifiers':{'IUPAC_name':'mercury',                'CAS_number':'7439-97-6',  'PubChem_CID':'23931',   'InChIKey':'QSHDDOUJBYECFT-UHFFFAOYSA-N', 'WIGOS_name':'',                                                                   'WIGOS_number':''}},
#PM
'Al':                             {'long_parameter_name':'total particulate aluminium',                                               'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['12101','14101'],         'airbase_name':'aluminium (aerosol)',            'airbase_code':'00604', 'miteco_code':'',   'ebas_parameter_name':['aluminium'],                                              'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Aluminum (Al)','Aluminum','Al'],               'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Al'],                                   'chemical_formula':'Al',     'bsc_parameter_name':'sconcal',      'identifiers':{'IUPAC_name':'aluminum',             'CAS_number':'7429-90-5',  'PubChem_CID':'5359268', 'InChIKey':'XAGFODPZIPBFFR-UHFFFAOYSA-N', 'WIGOS_name':'Al (Aluminium), TSP',                             'WIGOS_number':'674'}},
'As':                             {'long_parameter_name':'total particulate arsenic',                                                 'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':1.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':200.0,   'aqs_code':['12103','14103'],         'airbase_name':'Arsenic (aerosol)',              'airbase_code':'00018', 'miteco_code':'',   'ebas_parameter_name':['arsenic'],                                                'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Arsenic (As)','Arsenic','As'],                 'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['As'],                                   'chemical_formula':'As',     'bsc_parameter_name':'sconcas',      'identifiers':{'IUPAC_name':'arsenic',              'CAS_number':'7440-38-2',  'PubChem_CID':'5359596', 'InChIKey':'RQNWIZPPADIBDY-UHFFFAOYSA-N', 'WIGOS_name':'As (Arsenic), TSP',                               'WIGOS_number':'678'}},
'BaP':                            {'long_parameter_name':'total particulate benzo[a]pyrene',                                          'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':0.1,    'extreme_lower_limit':0.0,   'extreme_upper_limit':100.0,   'extreme_upper_monthly_median':20.0,    'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['benzo_a_pyrene'],                                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Benzo(a)Pyrene (BaP)','Benzo(a)Pyrene','BaP'], 'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'C20H12', 'bsc_parameter_name':'sconcbappm',   'identifiers':{'IUPAC_name':'benzo[a]pyrene',       'CAS_number':'50-32-8',    'PubChem_CID':'2336',    'InChIKey':'FMMWHPNWAFZXNH-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'BC':                             {'long_parameter_name':'total particulate black carbon',                                            'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':25000.0, 'extreme_upper_monthly_median':2500.0,  'aqs_code':[],                        'airbase_name':'black_carbon (aerosol)',         'airbase_code':'00391', 'miteco_code':'',   'ebas_parameter_name':['equivalent_black_carbon','equivalent_black_carbon_mass'], 'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Black carbon', 'Black Carbon (880nm)'], 'chemical_formula':'C',      'bsc_parameter_name':'sconcbc',      'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'C':                              {'long_parameter_name':'total particulate carbon',                                                  'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['total_carbon','total_carbon_corrected'],                  'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['TC(corr)','TC'],                               'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'C',      'bsc_parameter_name':'sconcc',       'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'Ca++':                           {'long_parameter_name':'total particulate calcium',                                                 'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':100.0,   'extreme_upper_monthly_median':20.0,    'aqs_code':['12111','14111'],         'airbase_name':'calcium (aerosol)',              'airbase_code':'00629', 'miteco_code':'',   'ebas_parameter_name':['calcium'],                                                'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Calcium','Ca'],                                'castnet_parameter_name':'CA',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'Ca',     'bsc_parameter_name':'sconcca',      'identifiers':{'IUPAC_name':'calcium(2+)',          'CAS_number':'14127-61-8', 'PubChem_CID':'271',     'InChIKey':'BHPQYMZQTOCNFJ-UHFFFAOYSA-N', 'WIGOS_name':'Ca++ (calcium), TSP',                             'WIGOS_number':'634'}},
'Cd':                             {'long_parameter_name':'total particulate cadmium',                                                 'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':500.0,   'extreme_upper_monthly_median':75.0,    'aqs_code':['12110','14110'],         'airbase_name':'Cadmium (aerosol)',              'airbase_code':'00014', 'miteco_code':'',   'ebas_parameter_name':['cadmium'],                                                'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Cadmium (Cd)','Cadmium','Cd'],                 'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Cd'],                                   'chemical_formula':'Cd',     'bsc_parameter_name':'sconccd',      'identifiers':{'IUPAC_name':'cadmium',              'CAS_number':'7440-43-9',  'PubChem_CID':'23973',   'InChIKey':'BDOSMKKIYDKNTQ-UHFFFAOYSA-N', 'WIGOS_name':'Cd (Cadmium), TSP',                               'WIGOS_number':'687'}},
'Cl-':                            {'long_parameter_name':'total particulate chloride',                                                'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['12191','12203','13115'], 'airbase_name':'chloride (aerosol)',             'airbase_code':'00631', 'miteco_code':'',   'ebas_parameter_name':['chloride'],                                               'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Chloride','CHLORIDE','Cl-'],                   'castnet_parameter_name':'CL',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'Cl',     'bsc_parameter_name':'sconccl',      'identifiers':{'IUPAC_name':'chloride',             'CAS_number':'16887-00-6', 'PubChem_CID':'312',     'InChIKey':'VEXZGXHMUGYJMC-UHFFFAOYSA-M', 'WIGOS_name':'Cl- (chloride), TSP',                             'WIGOS_number':'617'}},
'Co':                             {'long_parameter_name':'total particulate cobalt',                                                  'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':0.1,    'extreme_lower_limit':0.0,   'extreme_upper_limit':50.0,    'extreme_upper_monthly_median':5.0,     'aqs_code':['12113','14113'],         'airbase_name':'Cobalt (aerosol)',               'airbase_code':'00064', 'miteco_code':'',   'ebas_parameter_name':['cobalt'],                                                 'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Cobalt (Co)','Cobalt','Co'],                   'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Co'],                                   'chemical_formula':'Co',     'bsc_parameter_name':'sconccobalt',  'identifiers':{'IUPAC_name':'cobalt',               'CAS_number':'7440-48-4',  'PubChem_CID':'104730',  'InChIKey':'GUTLYIVDDKVIGB-UHFFFAOYSA-N', 'WIGOS_name':'Co (Cobalt), TSP',                                'WIGOS_number':'691'}},
'Cr':                             {'long_parameter_name':'total particulate chromium',                                                'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':1.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':500.0,   'extreme_upper_monthly_median':100.0,   'aqs_code':['12112','14112'],         'airbase_name':'Chromium (aerosol)',             'airbase_code':'00016', 'miteco_code':'',   'ebas_parameter_name':['chromium'],                                               'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Chromium (Cr)','Chromium','Cr'],               'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Cr'],                                   'chemical_formula':'Cr',     'bsc_parameter_name':'sconccr',      'identifiers':{'IUPAC_name':'chromium',             'CAS_number':'7440-47-3',  'PubChem_CID':'23976',   'InChIKey':'VYZAMTAEIAYCRO-UHFFFAOYSA-N', 'WIGOS_name':'Cr (Chromium), TSP',                              'WIGOS_number':'690'}},
'Cu':                             {'long_parameter_name':'total particulate copper',                                                  'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':1.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':750.0,   'extreme_upper_monthly_median':150.0,   'aqs_code':['12114','14114'],         'airbase_name':'Copper (aerosol)',               'airbase_code':'00073', 'miteco_code':'',   'ebas_parameter_name':['copper'],                                                 'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Copper (Cu)','Copper','Cu'],                   'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Cu'],                                   'chemical_formula':'Cu',     'bsc_parameter_name':'sconccu',      'identifiers':{'IUPAC_name':'copper',               'CAS_number':'7440-50-8',  'PubChem_CID':'23978',   'InChIKey':'RYGMFSIKBFXOCR-UHFFFAOYSA-N', 'WIGOS_name':'Cu (Copper), TSP',                                'WIGOS_number':'693'}},
'DUST':                           {'long_parameter_name':'total particulate dust',                                                    'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'sconcdu',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Dust mass concentration',                         'WIGOS_number':'727'}},
'EC':                             {'long_parameter_name':'total particulate elemental carbon',                                        'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':25000.0, 'extreme_upper_monthly_median':2500.0,  'aqs_code':[],                        'airbase_name':'Elemental carbon (aerosol)',     'airbase_code':'00771', 'miteco_code':'',   'ebas_parameter_name':['elemental_carbon'],                                       'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['EC(corr)','EC','EC_R','EC_T','EC(A)'],         'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'C',      'bsc_parameter_name':'sconcec',      'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'Fe':                             {'long_parameter_name':'total particulate iron',                                                    'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['12126','14126'],         'airbase_name':'Iron (aerosol)',                 'airbase_code':'00065', 'miteco_code':'',   'ebas_parameter_name':['iron'],                                                   'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Iron (Fe)','Iron','Fe'],                       'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Fe'],                                   'chemical_formula':'Fe',     'bsc_parameter_name':'sconcfe',      'identifiers':{'IUPAC_name':'iron',                 'CAS_number':'7439-89-6',  'PubChem_CID':'23925',   'InChIKey':'XEEYBQQBJWHFJM-UHFFFAOYSA-N', 'WIGOS_name':'Fe (Iron), TSP',                                  'WIGOS_number':'694'}},
'Hg':                             {'long_parameter_name':'total particulate mercury',                                                 'matrix':'pm', 'standard_units':'pg m-3',   'axis_label':'concentration',                          'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':30000.0, 'extreme_upper_monthly_median':3000.0,  'aqs_code':['12142','14142'],         'airbase_name':'Mercury (aerosol)',              'airbase_code':'00013', 'miteco_code':'',   'ebas_parameter_name':['mercury','mercury_PBM'],                                  'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'Hg',     'bsc_parameter_name':'sconchg',      'identifiers':{'IUPAC_name':'mercury',              'CAS_number':'7439-97-6',  'PubChem_CID':'23931',   'InChIKey':'QSHDDOUJBYECFT-UHFFFAOYSA-N', 'WIGOS_name':'Hg (Mercury), TSP',                               'WIGOS_number':'703'}},
'K+':                             {'long_parameter_name':'total particulate potassium',                                               'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':50.0,    'extreme_upper_monthly_median':10.0,    'aqs_code':['12180','14180'],         'airbase_name':'potassium (aerosol)',            'airbase_code':'00657', 'miteco_code':'',   'ebas_parameter_name':['potassium'],                                              'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Potassium','K'],                               'castnet_parameter_name':'K',             'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'K',      'bsc_parameter_name':'sconck',       'identifiers':{'IUPAC_name':'potassium(1+)',        'CAS_number':'24203-36-9', 'PubChem_CID':'813',     'InChIKey':'NPYPAHLBTDXSSS-UHFFFAOYSA-N', 'WIGOS_name':'K+ (potassium), TSP',                             'WIGOS_number':'640'}},
'Mg++':                           {'long_parameter_name':'total particulate magnesium',                                               'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':50.0,    'extreme_upper_monthly_median':10.0,    'aqs_code':['12140','14140'],         'airbase_name':'magnesium (aerosol)',            'airbase_code':'00659', 'miteco_code':'',   'ebas_parameter_name':['magnesium'],                                              'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Magnesium','Mg'],                              'castnet_parameter_name':'MG',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'Mg',     'bsc_parameter_name':'sconcmg',      'identifiers':{'IUPAC_name':'magnesium(2+)',        'CAS_number':'22537-22-0', 'PubChem_CID':'888',     'InChIKey':'JLVVSXFLKOJNIY-UHFFFAOYSA-N', 'WIGOS_name':'Mg++ (magnesium), TSP',                           'WIGOS_number':'637'}},
'Mn':                             {'long_parameter_name':'total particulate manganese',                                               'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':2.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':['12132','14132'],         'airbase_name':'Manganese (aerosol)',            'airbase_code':'00017', 'miteco_code':'',   'ebas_parameter_name':['manganese'],                                              'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Manganese (Mn)','Mn','Manganese'],             'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Mn'],                                   'chemical_formula':'Mn',     'bsc_parameter_name':'sconcmn',      'identifiers':{'IUPAC_name':'manganese',            'CAS_number':'7439-96-5',  'PubChem_CID':'23930',   'InChIKey':'PWHULOQIROXLJO-UHFFFAOYSA-N', 'WIGOS_name':'Mn (Manganese), TSP',                             'WIGOS_number':'702'}},
'MSA':                            {'long_parameter_name':'total particulate methanesulfonic acid',                                    'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':75.0,    'extreme_upper_monthly_median':25.0,    'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['methanesulfonic_acid'],                                   'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['MSA'],                                         'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'CH4O3S', 'bsc_parameter_name':'sconcmsa',     'identifiers':{'IUPAC_name':'methanesulfonic acid', 'CAS_number':'75-75-2',    'PubChem_CID':'6395',    'InChIKey':'AFVFQIVMOAPDHO-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'Na+':                            {'long_parameter_name':'total particulate sodium',                                                  'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['12184'],                 'airbase_name':'sodium (aerosol)',               'airbase_code':'00668', 'miteco_code':'',   'ebas_parameter_name':['sodium'],                                                 'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Sodium','Na+'],                                'castnet_parameter_name':'NA',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'Na',     'bsc_parameter_name':'sconcna',      'identifiers':{'IUPAC_name':'sodium(1+)',           'CAS_number':'17341-25-2', 'PubChem_CID':'923',     'InChIKey':'FKNQFGJONOIPTF-UHFFFAOYSA-N', 'WIGOS_name':'Na+ (sodium), TSP',                               'WIGOS_number':'643'}},
'NH4+':                           {'long_parameter_name':'total particulate ammonium',                                                'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['12301'],                 'airbase_name':'Particulate ammonium (aerosol)', 'airbase_code':'00045', 'miteco_code':'',   'ebas_parameter_name':['ammonium'],                                               'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Ammonium','AMMONIUM','NH4+','NH4'],            'castnet_parameter_name':'TNH4',          'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'NH4',    'bsc_parameter_name':'sconcnh4',     'identifiers':{'IUPAC_name':'azanium',              'CAS_number':'14798-03-9', 'PubChem_CID':'223',     'InChIKey':'QGZKDVFQNNGYKY-UHFFFAOYSA-O', 'WIGOS_name':'NH4+ (ammonium), TSP',                            'WIGOS_number':'647'}},
'NH4NO3':                         {'long_parameter_name':'total particulate ammonium nitrate',                                        'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['ammonium_nitrate'],                                       'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'NH4NO3', 'bsc_parameter_name':'sconcnh4no3',  'identifiers':{'IUPAC_name':'azanium;nitrate',      'CAS_number':'6484-52-2',  'PubChem_CID':'22985',   'InChIKey':'DVARTQFDIMZBAA-UHFFFAOYSA-O', 'WIGOS_name':'NH4NO3 (ammonium nitrate), TSP',                  'WIGOS_number':'648'}},
'Ni':                             {'long_parameter_name':'total particulate nickel',                                                  'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':5.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':10000.0, 'extreme_upper_monthly_median':1000.0,  'aqs_code':['12136','14136'],         'airbase_name':'Nickel (aerosol)',               'airbase_code':'00015', 'miteco_code':'',   'ebas_parameter_name':['nickel'],                                                 'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Nickel (Ni)','Nickel','Ni'],                   'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Ni'],                                   'chemical_formula':'Ni',     'bsc_parameter_name':'sconcni',      'identifiers':{'IUPAC_name':'nickel',               'CAS_number':'7440-02-0',  'PubChem_CID':'935',     'InChIKey':'PXHVJJICTQNCMI-UHFFFAOYSA-N', 'WIGOS_name':'Ni (Nickel), TSP',                                'WIGOS_number':'705'}},
'NO3-':                           {'long_parameter_name':'total particulate nitrate',                                                 'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':250.0,   'extreme_upper_monthly_median':75.0,    'aqs_code':['12306','14306'],         'airbase_name':'Particulate nitrate (aerosol)',  'airbase_code':'00046', 'miteco_code':'',   'ebas_parameter_name':['nitrate'],                                                'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Nitrate','NITRATE','NO3'],                     'castnet_parameter_name':['TNO3','WNO3'], 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'NO3',    'bsc_parameter_name':'sconcno3',     'identifiers':{'IUPAC_name':'nitrate',              'CAS_number':'14797-55-8', 'PubChem_CID':'943',     'InChIKey':'NHNBFGGVMKEFGY-UHFFFAOYSA-N', 'WIGOS_name':'NO3- (nitrate), TSP',                             'WIGOS_number':'652'}},
'OC':                             {'long_parameter_name':'total particulate organic carbon',                                          'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':25000.0, 'extreme_upper_monthly_median':2500.0,  'aqs_code':[],                        'airbase_name':'Organic carbon (aerosol)',       'airbase_code':'00772', 'miteco_code':'',   'ebas_parameter_name':['organic_carbon','organic_carbon_corrected'],              'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['OC(corr)','OC','OC_R','OC_T'],                 'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'C',      'bsc_parameter_name':'sconcoc',      'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'Pb':                             {'long_parameter_name':'total particulate lead',                                                    'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':50.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':60000.0, 'extreme_upper_monthly_median':15000.0, 'aqs_code':['12128','14128','14129'], 'airbase_name':'Lead (aerosol)',                 'airbase_code':'00012', 'miteco_code':'',   'ebas_parameter_name':['lead'],                                                   'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Lead (Pb)','Lead','Pb'],                       'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Pb'],                                   'chemical_formula':'Pb',     'bsc_parameter_name':'sconcpb',      'identifiers':{'IUPAC_name':'lead',                 'CAS_number':'7439-92-1',  'PubChem_CID':'5352425', 'InChIKey':'WABPQHHGFIMREM-UHFFFAOYSA-N', 'WIGOS_name':'Pb (Lead), TSP',                                  'WIGOS_number':'698'}},
'Se':                             {'long_parameter_name':'total particulate selenium',                                                'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['12154','14154'],         'airbase_name':'Selenium (aerosol)',             'airbase_code':'00048', 'miteco_code':'',   'ebas_parameter_name':['selenium'],                                               'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Selenium (Se)','Selenium','Se'],               'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Se'],                                   'chemical_formula':'Se',     'bsc_parameter_name':'sconcse',      'identifiers':{'IUPAC_name':'selenium',             'CAS_number':'7782-49-2',  'PubChem_CID':'6326970', 'InChIKey':'BUGBHKTXTAQXES-UHFFFAOYSA-N', 'WIGOS_name':'Se (Selenium), TSP',                              'WIGOS_number':'710'}},
'SO4--':                          {'long_parameter_name':'total particulate sulphate',                                                'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['12403','14403'],         'airbase_name':'Particulate sulphate (aerosol)', 'airbase_code':'00047', 'miteco_code':'',   'ebas_parameter_name':['sulphate_total'],                                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Sulphate','SULPHATE','SO4'],                   'castnet_parameter_name':['TSO4','NSO4'], 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'SO4',    'bsc_parameter_name':'sconcso4',     'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'SO4= (sulfate), TSP',                             'WIGOS_number':'620'}},
'SO4--_NSS':                      {'long_parameter_name':'total particulate sulphate: non-sea salt',                                  'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['sulphate_corrected'],                                     'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'SO4',    'bsc_parameter_name':'sconcso4nss',  'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'SO4--_SS':                       {'long_parameter_name':'total particulate sulphate: sea salt',                                      'matrix':'pm', 'standard_units':'ug m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['sulphate_seasalt'],                                       'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'SO4',    'bsc_parameter_name':'sconcso4ss',   'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'V':                              {'long_parameter_name':'total particulate vanadium',                                                'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':100.0,   'extreme_upper_monthly_median':20.0,    'aqs_code':['12164','14164'],         'airbase_name':'Vanadium (aerosol)',             'airbase_code':'00049', 'miteco_code':'',   'ebas_parameter_name':['vanadium'],                                               'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Vanadium (V)','Vanadium','V'],                 'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['V'],                                    'chemical_formula':'V',      'bsc_parameter_name':'sconcv',       'identifiers':{'IUPAC_name':'vanadium',             'CAS_number':'7440-62-2',  'PubChem_CID':'23990',   'InChIKey':'LEONUFNNVUYDNQ-UHFFFAOYSA-N', 'WIGOS_name':'V (Vanadium), TSP',                               'WIGOS_number':'722'}},
'Zn':                             {'long_parameter_name':'total particulate zinc',                                                    'matrix':'pm', 'standard_units':'ng m-3',   'axis_label':'concentration',                          'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':30000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['12167','14167'],         'airbase_name':'Zinc (aerosol)',                 'airbase_code':'00063', 'miteco_code':'',   'ebas_parameter_name':['zinc'],                                                   'ebas_matrix':'aerosol', 'ebas_priority_units':['/m3'], 'naps_code':['Zinc (Zn)','Zinc','Zn'],                       'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Zn'],                                   'chemical_formula':'Zn',     'bsc_parameter_name':'sconczn',      'identifiers':{'IUPAC_name':'zinc',                 'CAS_number':'7440-66-6',  'PubChem_CID':'23994',   'InChIKey':'HCHKCACWOHOZIP-UHFFFAOYSA-N', 'WIGOS_name':'Zn (Zinc), TSP',                                  'WIGOS_number':'724'}},
'AERO_ABS_COEFF_370nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 370nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco370',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_467nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 467nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco467',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_470nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 470nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco470',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_520nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 520nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco520',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_522nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 522nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco522',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_525nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 525nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco525',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_528nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 528nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco528',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_530nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 530nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco530',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_550nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 550nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco550',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_565nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 565nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco565',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_590nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 590nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco590',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_637nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 637nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco637',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_652nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 652nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco652',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_660nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 660nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco660',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_670nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 670nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco670',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_880nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 880nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco880',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'AERO_ABS_COEFF_950nm':           {'long_parameter_name':'total particulate aerosol absorption coefficient at 950nm',                 'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absco950',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, TSP',      'WIGOS_number':'318'}},  
'ABSAE_370-880nm':                {'long_parameter_name':'angstrom exponent between 370 and 880 nm for total particulate absorption', 'matrix':'pm', 'standard_units':'unitless', 'axis_label':'angstrom exponent 370-880nm',            'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absae370-880', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'ABSAE_370-950nm':                {'long_parameter_name':'angstrom exponent between 370 and 950 nm for total particulate absorption', 'matrix':'pm', 'standard_units':'unitless', 'axis_label':'angstrom exponent 370-950nm',            'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absae370-950', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'ABSAE_467-652nm':                {'long_parameter_name':'angstrom exponent between 467 and 652 nm for total particulate absorption', 'matrix':'pm', 'standard_units':'unitless', 'axis_label':'angstrom exponent 467-652nm',            'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absae467-652', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'ABSAE_467-660nm':                {'long_parameter_name':'angstrom exponent between 467 and 660 nm for total particulate absorption', 'matrix':'pm', 'standard_units':'unitless', 'axis_label':'angstrom exponent 467-660nm',            'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'absae467-660', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'AERO_LIGHT_BACKSCAT_COEFF_450nm':{'long_parameter_name':'total particulate aerosol light backscattering coefficient at 450nm',       'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsco450',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'AERO_LIGHT_BACKSCAT_COEFF_520nm':{'long_parameter_name':'total particulate aerosol light backscattering coefficient at 520nm',       'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsco520',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'AERO_LIGHT_BACKSCAT_COEFF_525nm':{'long_parameter_name':'total particulate aerosol light backscattering coefficient at 525nm',       'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsco525',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'AERO_LIGHT_BACKSCAT_COEFF_530nm':{'long_parameter_name':'total particulate aerosol light backscattering coefficient at 530nm',       'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsco530',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'AERO_LIGHT_BACKSCAT_COEFF_532nm':{'long_parameter_name':'total particulate aerosol light backscattering coefficient at 532nm',       'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsco532',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'AERO_LIGHT_BACKSCAT_COEFF_550nm':{'long_parameter_name':'total particulate aerosol light backscattering coefficient at 550nm',       'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsco550',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'AERO_LIGHT_BACKSCAT_COEFF_635nm':{'long_parameter_name':'total particulate aerosol light backscattering coefficient at 635nm',       'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsco635',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'AERO_LIGHT_BACKSCAT_COEFF_700nm':{'long_parameter_name':'total particulate aerosol light backscattering coefficient at 700nm',       'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsco700',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'AERO_LIGHT_BACKSCAT_COEFF_850nm':{'long_parameter_name':'total particulate aerosol light backscattering coefficient at 850nm',       'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsco850',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'AERO_LIGHT_SCAT_COEFF_450nm':    {'long_parameter_name':'total particulate aerosol light scattering coefficient at 450nm',           'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lsco450',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, TSP',      'WIGOS_number':'324'}},  
'AERO_LIGHT_SCAT_COEFF_520nm':    {'long_parameter_name':'total particulate aerosol light scattering coefficient at 520nm',           'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lsco520',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, TSP',      'WIGOS_number':'324'}},  
'AERO_LIGHT_SCAT_COEFF_525nm':    {'long_parameter_name':'total particulate aerosol light scattering coefficient at 525nm',           'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lsco525',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, TSP',      'WIGOS_number':'324'}},  
'AERO_LIGHT_SCAT_COEFF_530nm':    {'long_parameter_name':'total particulate aerosol light scattering coefficient at 530nm',           'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lsco530',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, TSP',      'WIGOS_number':'324'}},  
'AERO_LIGHT_SCAT_COEFF_532nm':    {'long_parameter_name':'total particulate aerosol light scattering coefficient at 532nm',           'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lsco532',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, TSP',      'WIGOS_number':'324'}},  
'AERO_LIGHT_SCAT_COEFF_550nm':    {'long_parameter_name':'total particulate aerosol light scattering coefficient at 550nm',           'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lsco550',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, TSP',      'WIGOS_number':'324'}},  
'AERO_LIGHT_SCAT_COEFF_635nm':    {'long_parameter_name':'total particulate aerosol light scattering coefficient at 635nm',           'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lsco635',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, TSP',      'WIGOS_number':'324'}},  
'AERO_LIGHT_SCAT_COEFF_700nm':    {'long_parameter_name':'total particulate aerosol light scattering coefficient at 700nm',           'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lsco700',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, TSP',      'WIGOS_number':'324'}},  
'AERO_LIGHT_SCAT_COEFF_850nm':    {'long_parameter_name':'total particulate aerosol light scattering coefficient at 850nm',           'matrix':'pm', 'standard_units':'Mm-1',     'axis_label':'coefficient',                            'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'aerosol', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lsco850',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, TSP',      'WIGOS_number':'324'}},  
'SCATAE_450-635nm':               {'long_parameter_name':'angstrom exponent between 450 and 635 nm for total particulate scattering', 'matrix':'pm', 'standard_units':'unitless', 'axis_label':'angstrom exponent 450-635nm',            'minimum_resolution':np.NaN, 'extreme_lower_limit':-2.0,  'extreme_upper_limit':4.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lsae450-635',  'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light scattering', 'WIGOS_number':'12148'}},  
'SCATAE_450-700nm':               {'long_parameter_name':'angstrom exponent between 450 and 700 nm for total particulate scattering', 'matrix':'pm', 'standard_units':'unitless', 'axis_label':'angstrom exponent 450-700nm',            'minimum_resolution':np.NaN, 'extreme_lower_limit':-2.0,  'extreme_upper_limit':4.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lsae450-700',  'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light scattering', 'WIGOS_number':'12148'}},  
'BACKSCAT_FRAC_450nm':            {'long_parameter_name':'total particulate aerosol light backscattering fraction at 450nm',          'matrix':'pm', 'standard_units':'unitless', 'axis_label':'fraction',                               'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsf450',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'BACKSCAT_FRAC_520nm':            {'long_parameter_name':'total particulate aerosol light backscattering fraction at 520nm',          'matrix':'pm', 'standard_units':'unitless', 'axis_label':'fraction',                               'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsf520',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'BACKSCAT_FRAC_525nm':            {'long_parameter_name':'total particulate aerosol light backscattering fraction at 525nm',          'matrix':'pm', 'standard_units':'unitless', 'axis_label':'fraction',                               'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsf525',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'BACKSCAT_FRAC_530nm':            {'long_parameter_name':'total particulate aerosol light backscattering fraction at 530nm',          'matrix':'pm', 'standard_units':'unitless', 'axis_label':'fraction',                               'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsf530',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'BACKSCAT_FRAC_532nm':            {'long_parameter_name':'total particulate aerosol light backscattering fraction at 532nm',          'matrix':'pm', 'standard_units':'unitless', 'axis_label':'fraction',                               'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsf532',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'BACKSCAT_FRAC_550nm':            {'long_parameter_name':'total particulate aerosol light backscattering fraction at 550nm',          'matrix':'pm', 'standard_units':'unitless', 'axis_label':'fraction',                               'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsf550',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'BACKSCAT_FRAC_635nm':            {'long_parameter_name':'total particulate aerosol light backscattering fraction at 635nm',          'matrix':'pm', 'standard_units':'unitless', 'axis_label':'fraction',                               'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsf635',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'BACKSCAT_FRAC_700nm':            {'long_parameter_name':'total particulate aerosol light backscattering fraction at 700nm',          'matrix':'pm', 'standard_units':'unitless', 'axis_label':'fraction',                               'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsf700',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'BACKSCAT_FRAC_850nm':            {'long_parameter_name':'total particulate aerosol light backscattering fraction at 850nm',          'matrix':'pm', 'standard_units':'unitless', 'axis_label':'fraction',                               'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'lbsf850',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'ASY_525nm':                      {'long_parameter_name':'total particulate asymmetry factor at 525nm',                               'matrix':'pm', 'standard_units':'unitless', 'axis_label':'asymmetry factor',                       'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':1.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'asy525',       'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'ASY_550nm':                      {'long_parameter_name':'total particulate asymmetry factor at 550nm',                               'matrix':'pm', 'standard_units':'unitless', 'axis_label':'asymmetry factor',                       'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':1.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                        'airbase_name':'',                               'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',        'ebas_priority_units':[''],    'naps_code':[],                                              'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                                       'chemical_formula':'',       'bsc_parameter_name':'asy550',       'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
#PM10
'PM10':                                {'long_parameter_name':'PM10 mass',                                                    'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['81102','85101'],            'airbase_name':'Particulate matter < 10 µm (aerosol)', 'airbase_code':'00005', 'miteco_code':'10', 'ebas_parameter_name':['pm10_mass'],                                              'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['PM10','PM2.5-10','MASS'],              'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':['10'], 'japan_nies_name':['SPM','SPMB','SPMP'], 'uk_air_name':['PM10 particulate matter (Hourly measured)','PM<sub>10</sub> particulate matter (Hourly measured)','PM10 particulate matter (Daily measured)','Daily measured PM10 (uncorrected)','Daily measured PM<sub>10</sub> (uncorrected)'], 'chemical_formula':'',       'bsc_parameter_name':'pm10',             'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle mass concentration, PM10',               'WIGOS_number':'364'}},
'PM10_Al':                             {'long_parameter_name':'PM10 aluminium',                                               'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['82101','85104'],            'airbase_name':'aluminium in PM10 (aerosol)',          'airbase_code':'05604', 'miteco_code':'',   'ebas_parameter_name':['aluminium'],                                              'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Aluminum (Al)','Aluminum','Al'],       'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'Al',     'bsc_parameter_name':'pm10al',           'identifiers':{'IUPAC_name':'aluminum',             'CAS_number':'7429-90-5',  'PubChem_CID':'5359268', 'InChIKey':'XAGFODPZIPBFFR-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_As':                             {'long_parameter_name':'PM10 arsenic',                                                 'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':1.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':200.0,   'aqs_code':['82103','85103'],            'airbase_name':'Arsenic in PM10 (aerosol)',            'airbase_code':'05018', 'miteco_code':'17', 'ebas_parameter_name':['arsenic'],                                                'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Arsenic (As)','Arsenic','As'],         'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'As',     'bsc_parameter_name':'pm10as',           'identifiers':{'IUPAC_name':'arsenic',              'CAS_number':'7440-38-2',  'PubChem_CID':'5359596', 'InChIKey':'RQNWIZPPADIBDY-UHFFFAOYSA-N', 'WIGOS_name':'As (Arsenic), PM10',                              'WIGOS_number':'677'}},
'PM10_BaP':                            {'long_parameter_name':'PM10 benzo[a]pyrene',                                          'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.1,    'extreme_lower_limit':0.0,   'extreme_upper_limit':100.0,   'extreme_upper_monthly_median':20.0,    'aqs_code':[],                           'airbase_name':'Benzo(a)pyrene in PM10 (aerosol)',     'airbase_code':'05029', 'miteco_code':'27', 'ebas_parameter_name':['benzo_a_pyrene'],                                         'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'C20H12', 'bsc_parameter_name':'pm10bap',          'identifiers':{'IUPAC_name':'benzo[a]pyrene',       'CAS_number':'50-32-8',    'PubChem_CID':'2336',    'InChIKey':'FMMWHPNWAFZXNH-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_BC':                             {'long_parameter_name':'PM10 black carbon',                                            'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':25000.0, 'extreme_upper_monthly_median':2500.0,  'aqs_code':['85313'],                    'airbase_name':'black_carbon in PM10 (aerosol)',       'airbase_code':'05391', 'miteco_code':'',   'ebas_parameter_name':['equivalent_black_carbon','equivalent_black_carbon_mass'], 'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'C',      'bsc_parameter_name':'pm10bc',           'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_C':                              {'long_parameter_name':'PM10 carbon',                                                  'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['82116'],                    'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['total_carbon','total_carbon_corrected'],                  'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['TC(corr)','TC'],                       'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':['Total Carbon-PM10'],                                                                                                                                                                                                              'chemical_formula':'C',      'bsc_parameter_name':'pm10c',            'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'Total carbon, PM10',                              'WIGOS_number':'626'}},
'PM10_Ca++':                           {'long_parameter_name':'PM10 calcium',                                                 'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':100.0,   'extreme_upper_monthly_median':20.0,    'aqs_code':['82111','85111'],            'airbase_name':'calcium in PM10 (aerosol)',            'airbase_code':'05629', 'miteco_code':'',   'ebas_parameter_name':['calcium'],                                                'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Calcium','Ca'],                        'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':['Calcium aerosol - PM10'],                                                                                                                                                                                                         'chemical_formula':'Ca',     'bsc_parameter_name':'pm10ca',           'identifiers':{'IUPAC_name':'calcium(2+)',          'CAS_number':'14127-61-8', 'PubChem_CID':'271',     'InChIKey':'BHPQYMZQTOCNFJ-UHFFFAOYSA-N', 'WIGOS_name':'Ca++ (calcium), PM10',                            'WIGOS_number':'632'}},
'PM10_Cd':                             {'long_parameter_name':'PM10 cadmium',                                                 'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':500.0,   'extreme_upper_monthly_median':75.0,    'aqs_code':['82110','85110'],            'airbase_name':'Cadmium in PM10 (aerosol)',            'airbase_code':'05014', 'miteco_code':'28', 'ebas_parameter_name':['cadmium'],                                                'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Cadmium (Cd)','Cadmium','Cd'],         'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'Cd',     'bsc_parameter_name':'pm10cd',           'identifiers':{'IUPAC_name':'cadmium',              'CAS_number':'7440-43-9',  'PubChem_CID':'23973',   'InChIKey':'BDOSMKKIYDKNTQ-UHFFFAOYSA-N', 'WIGOS_name':'Cd (Cadmium), PM10',                              'WIGOS_number':'684'}},
'PM10_Cl-':                            {'long_parameter_name':'PM10 chloride',                                                'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['82115','82203','85115'],    'airbase_name':'chloride in PM10 (aerosol)',           'airbase_code':'05631', 'miteco_code':'',   'ebas_parameter_name':['chloride'],                                               'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Chloride','CHLORIDE','Cl-'],           'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':['Chloride aerosol - PM10'],                                                                                                                                                                                                        'chemical_formula':'Cl',     'bsc_parameter_name':'pm10cl',           'identifiers':{'IUPAC_name':'chloride',             'CAS_number':'16887-00-6', 'PubChem_CID':'312',     'InChIKey':'VEXZGXHMUGYJMC-UHFFFAOYSA-M', 'WIGOS_name':'Cl- (chloride), PM10',                            'WIGOS_number':'615'}},
'PM10_Co':                             {'long_parameter_name':'PM10 cobalt',                                                  'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.1,    'extreme_lower_limit':0.0,   'extreme_upper_limit':50.0,    'extreme_upper_monthly_median':5.0,     'aqs_code':['82113','85113'],            'airbase_name':'Cobalt in PM10 (aerosol)',             'airbase_code':'05064', 'miteco_code':'',   'ebas_parameter_name':['cobalt'],                                                 'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Cobalt (Co)','Cobalt','Co'],           'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'Co',     'bsc_parameter_name':'pm10cobalt',       'identifiers':{'IUPAC_name':'cobalt',               'CAS_number':'7440-48-4',  'PubChem_CID':'104730',  'InChIKey':'GUTLYIVDDKVIGB-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_Cr':                             {'long_parameter_name':'PM10 chromium',                                                'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':1.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':500.0,   'extreme_upper_monthly_median':100.0,   'aqs_code':['82112','85112'],            'airbase_name':'Chromium in PM10 (aerosol)',           'airbase_code':'05016', 'miteco_code':'',   'ebas_parameter_name':['chromium'],                                               'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Chromium (Cr)','Chromium','Cr'],       'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'Cr',     'bsc_parameter_name':'pm10cr',           'identifiers':{'IUPAC_name':'chromium',             'CAS_number':'7440-47-3',  'PubChem_CID':'23976',   'InChIKey':'VYZAMTAEIAYCRO-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_Cu':                             {'long_parameter_name':'PM10 copper',                                                  'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':1.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':750.0,   'extreme_upper_monthly_median':150.0,   'aqs_code':['82114','85114'],            'airbase_name':'Copper in PM10 (aerosol)',             'airbase_code':'05073', 'miteco_code':'',   'ebas_parameter_name':['copper'],                                                 'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Copper (Cu)','Copper','Cu'],           'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'Cu',     'bsc_parameter_name':'pm10cu',           'identifiers':{'IUPAC_name':'copper',               'CAS_number':'7440-50-8',  'PubChem_CID':'23978',   'InChIKey':'RYGMFSIKBFXOCR-UHFFFAOYSA-N', 'WIGOS_name':'Cu (Copper), PM10',                               'WIGOS_number':'692'}},
'PM10_DUST':                           {'long_parameter_name':'PM10 dust',                                                    'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10du',           'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_EC':                             {'long_parameter_name':'PM10 elemental carbon',                                        'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':25000.0, 'extreme_upper_monthly_median':2500.0,  'aqs_code':['82307','85357','85380'],    'airbase_name':'Elemental carbon in PM10 (aerosol)',   'airbase_code':'05771', 'miteco_code':'',   'ebas_parameter_name':['elemental_carbon'],                                       'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['EC(corr)','EC','EC_R','EC_T','EC(A)'], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':['Elemental Carbon by reflectance-PM10','Elemental Carbon by transmittance-PM10'],                                                                                                                                                  'chemical_formula':'C',      'bsc_parameter_name':'pm10ec',           'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'Elemental carbon, PM10',                          'WIGOS_number':'623'}},
'PM10_Fe':                             {'long_parameter_name':'PM10 iron',                                                    'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['82126','85126'],            'airbase_name':'Iron in PM10 (aerosol)',               'airbase_code':'05065', 'miteco_code':'',   'ebas_parameter_name':['iron'],                                                   'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Iron (Fe)','Iron','Fe'],               'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'Fe',     'bsc_parameter_name':'pm10fe',           'identifiers':{'IUPAC_name':'iron',                 'CAS_number':'7439-89-6',  'PubChem_CID':'23925',   'InChIKey':'XEEYBQQBJWHFJM-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_Hg':                             {'long_parameter_name':'PM10 mercury',                                                 'matrix':'pm10', 'standard_units':'pg m-3',   'axis_label':'concentration',               'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':30000.0, 'extreme_upper_monthly_median':3000.0,  'aqs_code':['82142','85142'],            'airbase_name':'Mercury in PM10 (aerosol)',            'airbase_code':'05013', 'miteco_code':'',   'ebas_parameter_name':['mercury','mercury_PBM'],                                  'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'Hg',     'bsc_parameter_name':'pm10hg',           'identifiers':{'IUPAC_name':'mercury',              'CAS_number':'7439-97-6',  'PubChem_CID':'23931',   'InChIKey':'QSHDDOUJBYECFT-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM10_K+':                             {'long_parameter_name':'PM10 potassium',                                               'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':50.0,    'extreme_upper_monthly_median':10.0,    'aqs_code':['82180','85180','85303'],    'airbase_name':'potassium in PM10 (aerosol)',          'airbase_code':'05657', 'miteco_code':'',   'ebas_parameter_name':['potassium'],                                              'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Potassium','K'],                       'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':['Potassium aerosol - PM10'],                                                                                                                                                                                                       'chemical_formula':'K',      'bsc_parameter_name':'pm10k',            'identifiers':{'IUPAC_name':'potassium(1+)',        'CAS_number':'24203-36-9', 'PubChem_CID':'813',     'InChIKey':'NPYPAHLBTDXSSS-UHFFFAOYSA-N', 'WIGOS_name':'K+ (potassium), PM10',                            'WIGOS_number':'638'}},
'PM10_Mg++':                           {'long_parameter_name':'PM10 magnesium',                                               'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':50.0,    'extreme_upper_monthly_median':10.0,    'aqs_code':['82140','85140'],            'airbase_name':'magnesium in PM10 (aerosol)',          'airbase_code':'05659', 'miteco_code':'',   'ebas_parameter_name':['magnesium'],                                              'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Magnesium','Mg'],                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':['Magnesium aerosol - PM10'],                                                                                                                                                                                                       'chemical_formula':'Mg',     'bsc_parameter_name':'pm10mg',           'identifiers':{'IUPAC_name':'magnesium(2+)',        'CAS_number':'22537-22-0', 'PubChem_CID':'888',     'InChIKey':'JLVVSXFLKOJNIY-UHFFFAOYSA-N', 'WIGOS_name':'Mg++ (magnesium), PM10',                          'WIGOS_number':'635'}},
'PM10_Mn':                             {'long_parameter_name':'PM10 manganese',                                               'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':2.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':['82132','85132'],            'airbase_name':'Manganese in PM10 (aerosol)',          'airbase_code':'05017', 'miteco_code':'',   'ebas_parameter_name':['manganese'],                                              'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Manganese (Mn)','Mn','Manganese'],     'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'Mn',     'bsc_parameter_name':'pm10mn',           'identifiers':{'IUPAC_name':'manganese',            'CAS_number':'7439-96-5',  'PubChem_CID':'23930',   'InChIKey':'PWHULOQIROXLJO-UHFFFAOYSA-N', 'WIGOS_name':'Mn (Manganese), PM10',                            'WIGOS_number':'701'}},
'PM10_MSA':                            {'long_parameter_name':'PM10 methanesulfonic acid',                                    'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':75.0,    'extreme_upper_monthly_median':25.0,    'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['methanesulfonic_acid'],                                   'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['MSA'],                                 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'CH4O3S', 'bsc_parameter_name':'pm10msa',          'identifiers':{'IUPAC_name':'methanesulfonic acid', 'CAS_number':'75-75-2',    'PubChem_CID':'6395',    'InChIKey':'AFVFQIVMOAPDHO-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_Na+':                            {'long_parameter_name':'PM10 sodium',                                                  'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['82184','85184','85302'],    'airbase_name':'sodium in PM10 (aerosol)',             'airbase_code':'05668', 'miteco_code':'',   'ebas_parameter_name':['sodium'],                                                 'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Sodium','Na+'],                        'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':['Sodium aerosol - PM10'],                                                                                                                                                                                                          'chemical_formula':'Na',     'bsc_parameter_name':'pm10na',           'identifiers':{'IUPAC_name':'sodium(1+)',           'CAS_number':'17341-25-2', 'PubChem_CID':'923',     'InChIKey':'FKNQFGJONOIPTF-UHFFFAOYSA-N', 'WIGOS_name':'Na+ (sodium), PM10',                              'WIGOS_number':'641'}},
'PM10_NH4+':                           {'long_parameter_name':'PM10 ammonium',                                                'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['82301','85301'],            'airbase_name':'Ammonium in PM10 (aerosol)',           'airbase_code':'05045', 'miteco_code':'',   'ebas_parameter_name':['ammonium'],                                               'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Ammonium','AMMONIUM','NH4+','NH4'],    'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':['Ammonium aerosol - PM10'],                                                                                                                                                                                                        'chemical_formula':'NH4',    'bsc_parameter_name':'pm10nh4',          'identifiers':{'IUPAC_name':'azanium',              'CAS_number':'14798-03-9', 'PubChem_CID':'223',     'InChIKey':'QGZKDVFQNNGYKY-UHFFFAOYSA-O', 'WIGOS_name':'NH4+ (ammonium), PM10',                           'WIGOS_number':'645'}},
'PM10_NH4NO3':                         {'long_parameter_name':'PM10 ammonium nitrate',                                        'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['ammonium_nitrate'],                                       'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'NH4NO3', 'bsc_parameter_name':'pm10nh4no3',       'identifiers':{'IUPAC_name':'azanium;nitrate',      'CAS_number':'6484-52-2',  'PubChem_CID':'22985',   'InChIKey':'DVARTQFDIMZBAA-UHFFFAOYSA-O', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_Ni':                             {'long_parameter_name':'PM10 nickel',                                                  'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':5.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':10000.0, 'extreme_upper_monthly_median':1000.0,  'aqs_code':['82136','85136'],            'airbase_name':'Nickel in PM10 (aerosol)',             'airbase_code':'05015', 'miteco_code':'62', 'ebas_parameter_name':['nickel'],                                                 'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Nickel (Ni)','Nickel','Ni'],           'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'Ni',     'bsc_parameter_name':'pm10ni',           'identifiers':{'IUPAC_name':'nickel',               'CAS_number':'7440-02-0',  'PubChem_CID':'935',     'InChIKey':'PXHVJJICTQNCMI-UHFFFAOYSA-N', 'WIGOS_name':'Ni (Nickel), PM10',                               'WIGOS_number':'704'}},
'PM10_NO3-':                           {'long_parameter_name':'PM10 nitrate',                                                 'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':250.0,   'extreme_upper_monthly_median':75.0,    'aqs_code':['82306','85306'],            'airbase_name':'Nitrate in PM10 (aerosol)',            'airbase_code':'05046', 'miteco_code':'',   'ebas_parameter_name':['nitrate'],                                                'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Nitrate','NITRATE','NO3'],             'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':['Particulate nitrate - PM10'],                                                                                                                                                                                                     'chemical_formula':'NO3',    'bsc_parameter_name':'pm10no3',          'identifiers':{'IUPAC_name':'nitrate',              'CAS_number':'14797-55-8', 'PubChem_CID':'943',     'InChIKey':'NHNBFGGVMKEFGY-UHFFFAOYSA-N', 'WIGOS_name':'NO3- (nitrate), PM10',                            'WIGOS_number':'650'}},
'PM10_OC':                             {'long_parameter_name':'PM10 organic carbon',                                          'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':25000.0, 'extreme_upper_monthly_median':2500.0,  'aqs_code':['82305','85355','85370'],    'airbase_name':'Organic carbon in PM10 (aerosol)',     'airbase_code':'05772', 'miteco_code':'',   'ebas_parameter_name':['organic_carbon','organic_carbon_corrected'],              'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['OC(corr)','OC','OC_R','OC_T'],         'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':['Organic Carbon by reflectance-PM10','Organic Carbon by transmittance-PM10'],                                                                                                                                                      'chemical_formula':'C',      'bsc_parameter_name':'pm10oc',           'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'Carbonaceous-organic material, PM10',             'WIGOS_number':'670'}},
'PM10_Pb':                             {'long_parameter_name':'PM10 lead',                                                    'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':50.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':60000.0, 'extreme_upper_monthly_median':15000.0, 'aqs_code':['82128','85128','85129'],    'airbase_name':'Lead in PM10 (aerosol)',               'airbase_code':'05012', 'miteco_code':'19', 'ebas_parameter_name':['lead'],                                                   'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Lead (Pb)','Lead','Pb'],               'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'Pb',     'bsc_parameter_name':'pm10pb',           'identifiers':{'IUPAC_name':'lead',                 'CAS_number':'7439-92-1',  'PubChem_CID':'5352425', 'InChIKey':'WABPQHHGFIMREM-UHFFFAOYSA-N', 'WIGOS_name':'Pb (Lead), PM10',                                 'WIGOS_number':'697'}},
'PM10_Se':                             {'long_parameter_name':'PM10 selenium',                                                'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['82154','85154'],            'airbase_name':'Selenium in PM10 (aerosol)',           'airbase_code':'05048', 'miteco_code':'',   'ebas_parameter_name':['selenium'],                                               'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Selenium (Se)','Selenium','Se'],       'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'Se',     'bsc_parameter_name':'pm10se',           'identifiers':{'IUPAC_name':'selenium',             'CAS_number':'7782-49-2',  'PubChem_CID':'6326970', 'InChIKey':'BUGBHKTXTAQXES-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_SO4--':                          {'long_parameter_name':'PM10 sulphate',                                                'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['82403','85403'],            'airbase_name':'sulphate in PM10 (aerosol)',           'airbase_code':'05047', 'miteco_code':'',   'ebas_parameter_name':['sulphate_total'],                                         'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Sulphate','SULPHATE','SO4'],           'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':['Sulphate aerosol - PM10'],                                                                                                                                                                                                        'chemical_formula':'SO4',    'bsc_parameter_name':'pm10so4',          'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'SO4= (sulfate), PM10',                            'WIGOS_number':'621'}},
'PM10_SO4--_NSS':                      {'long_parameter_name':'PM10 sulphate : non-sea salt',                                 'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['sulphate_corrected'],                                     'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'SO4',    'bsc_parameter_name':'pm10so4nss',       'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_SO4--_SS':                       {'long_parameter_name':'PM10 sulphate : sea salt',                                     'matrix':'pm10', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['sulphate_seasalt'],                                       'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'SO4',    'bsc_parameter_name':'pm10so4ss',        'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_V':                              {'long_parameter_name':'PM10 vanadium',                                                'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':100.0,   'extreme_upper_monthly_median':20.0,    'aqs_code':['82164','85164'],            'airbase_name':'Vanedium in PM10 (aerosol)',           'airbase_code':'05049', 'miteco_code':'',   'ebas_parameter_name':['vanadium'],                                               'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Vanadium (V)','Vanadium','V'],         'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'V',      'bsc_parameter_name':'pm10v',            'identifiers':{'IUPAC_name':'vanadium',             'CAS_number':'7440-62-2',  'PubChem_CID':'23990',   'InChIKey':'LEONUFNNVUYDNQ-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_Zn':                             {'long_parameter_name':'PM10 zinc',                                                    'matrix':'pm10', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':30000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['82167','85167'],            'airbase_name':'Zinc in PM10 (aerosol)',               'airbase_code':'05063', 'miteco_code':'',   'ebas_parameter_name':['zinc'],                                                   'ebas_matrix':'pm10', 'ebas_priority_units':['/m3'], 'naps_code':['Zinc (Zn)','Zinc','Zn'],               'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'Zn',     'bsc_parameter_name':'pm10zn',           'identifiers':{'IUPAC_name':'zinc',                 'CAS_number':'7440-66-6',  'PubChem_CID':'23994',   'InChIKey':'HCHKCACWOHOZIP-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM10_AERO_ABS_COEFF_370nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 370nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco370',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_467nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 467nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco467',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_470nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 470nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco470',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_520nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 520nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco520',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_522nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 522nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco522',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_525nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 525nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco525',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_528nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 528nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco528',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_530nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 530nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco530',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_550nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 550nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco550',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_565nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 565nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco565',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_590nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 590nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco590',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_637nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 637nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco637',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_652nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 652nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco652',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_660nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 660nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco660',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_670nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 670nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco670',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_880nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 880nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco880',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_AERO_ABS_COEFF_950nm':           {'long_parameter_name':'PM10 aerosol absorption coefficient at 950nm',                 'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absco950',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM10',     'WIGOS_number':'317'}},  
'PM10_ABSAE_370-880nm':                {'long_parameter_name':'angstrom exponent between 370 and 880 nm for PM10 absorption', 'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'angstrom exponent 370-880nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absae370-880', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'PM10_ABSAE_370-950nm':                {'long_parameter_name':'angstrom exponent between 370 and 950 nm for PM10 absorption', 'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'angstrom exponent 370-950nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absae370-950', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'PM10_ABSAE_467-652nm':                {'long_parameter_name':'angstrom exponent between 467 and 652 nm for PM10 absorption', 'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'angstrom exponent 467-652nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absae467-652', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'PM10_ABSAE_467-660nm':                {'long_parameter_name':'angstrom exponent between 467 and 660 nm for PM10 absorption', 'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'angstrom exponent 467-660nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10absae467-660', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm':{'long_parameter_name':'PM10 aerosol light backscattering coefficient at 450nm',       'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsco450',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm':{'long_parameter_name':'PM10 aerosol light backscattering coefficient at 520nm',       'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsco520',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm':{'long_parameter_name':'PM10 aerosol light backscattering coefficient at 525nm',       'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsco525',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm':{'long_parameter_name':'PM10 aerosol light backscattering coefficient at 530nm',       'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsco530',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm':{'long_parameter_name':'PM10 aerosol light backscattering coefficient at 532nm',       'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsco532',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm':{'long_parameter_name':'PM10 aerosol light backscattering coefficient at 550nm',       'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsco550',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm':{'long_parameter_name':'PM10 aerosol light backscattering coefficient at 635nm',       'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsco635',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm':{'long_parameter_name':'PM10 aerosol light backscattering coefficient at 700nm',       'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsco700',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm':{'long_parameter_name':'PM10 aerosol light backscattering coefficient at 850nm',       'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsco850',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_AERO_LIGHT_SCAT_COEFF_450nm':    {'long_parameter_name':'PM10 aerosol light scattering coefficient at 450nm',           'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lsco450',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM10',     'WIGOS_number':'323'}},   
'PM10_AERO_LIGHT_SCAT_COEFF_520nm':    {'long_parameter_name':'PM10 aerosol light scattering coefficient at 520nm',           'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lsco520',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM10',     'WIGOS_number':'323'}},   
'PM10_AERO_LIGHT_SCAT_COEFF_525nm':    {'long_parameter_name':'PM10 aerosol light scattering coefficient at 525nm',           'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lsco525',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM10',     'WIGOS_number':'323'}},   
'PM10_AERO_LIGHT_SCAT_COEFF_530nm':    {'long_parameter_name':'PM10 aerosol light scattering coefficient at 530nm',           'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lsco530',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM10',     'WIGOS_number':'323'}},   
'PM10_AERO_LIGHT_SCAT_COEFF_532nm':    {'long_parameter_name':'PM10 aerosol light scattering coefficient at 532nm',           'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lsco532',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM10',     'WIGOS_number':'323'}},   
'PM10_AERO_LIGHT_SCAT_COEFF_550nm':    {'long_parameter_name':'PM10 aerosol light scattering coefficient at 550nm',           'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lsco550',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM10',     'WIGOS_number':'323'}},   
'PM10_AERO_LIGHT_SCAT_COEFF_635nm':    {'long_parameter_name':'PM10 aerosol light scattering coefficient at 635nm',           'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lsco635',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM10',     'WIGOS_number':'323'}},   
'PM10_AERO_LIGHT_SCAT_COEFF_700nm':    {'long_parameter_name':'PM10 aerosol light scattering coefficient at 700nm',           'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lsco700',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM10',     'WIGOS_number':'323'}},   
'PM10_AERO_LIGHT_SCAT_COEFF_850nm':    {'long_parameter_name':'PM10 aerosol light scattering coefficient at 850nm',           'matrix':'pm10', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm10', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lsco850',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM10',     'WIGOS_number':'323'}},   
'PM10_SCATAE_450-635nm':               {'long_parameter_name':'angstrom exponent between 450 and 635 nm for PM10 scattering', 'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'angstrom exponent 450-635nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':-2.0,  'extreme_upper_limit':4.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lsae450-635',  'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light scattering', 'WIGOS_number':'12148'}},  
'PM10_SCATAE_450-700nm':               {'long_parameter_name':'angstrom exponent between 450 and 700 nm for PM10 scattering', 'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'angstrom exponent 450-700nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':-2.0,  'extreme_upper_limit':4.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lsae450-700',  'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light scattering', 'WIGOS_number':'12148'}},  
'PM10_BACKSCAT_FRAC_450nm':            {'long_parameter_name':'PM10 aerosol light backscattering fraction at 450nm',          'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsf450',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_BACKSCAT_FRAC_520nm':            {'long_parameter_name':'PM10 aerosol light backscattering fraction at 520nm',          'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsf520',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_BACKSCAT_FRAC_525nm':            {'long_parameter_name':'PM10 aerosol light backscattering fraction at 525nm',          'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsf525',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_BACKSCAT_FRAC_530nm':            {'long_parameter_name':'PM10 aerosol light backscattering fraction at 530nm',          'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsf530',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_BACKSCAT_FRAC_532nm':            {'long_parameter_name':'PM10 aerosol light backscattering fraction at 532nm',          'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsf532',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_BACKSCAT_FRAC_550nm':            {'long_parameter_name':'PM10 aerosol light backscattering fraction at 550nm',          'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsf550',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_BACKSCAT_FRAC_635nm':            {'long_parameter_name':'PM10 aerosol light backscattering fraction at 635nm',          'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsf635',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_BACKSCAT_FRAC_700nm':            {'long_parameter_name':'PM10 aerosol light backscattering fraction at 700nm',          'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsf700',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_BACKSCAT_FRAC_850nm':            {'long_parameter_name':'PM10 aerosol light backscattering fraction at 850nm',          'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10lbsf850',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_ASY_525nm':                      {'long_parameter_name':'PM10 asymmetry factor at 525nm',                               'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'asymmetry factor',            'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':1.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10asy525',       'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM10_ASY_550nm':                      {'long_parameter_name':'PM10 asymmetry factor at 550nm',                               'matrix':'pm10', 'standard_units':'unitless', 'axis_label':'asymmetry factor',            'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':1.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                           'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',   'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                    'uk_air_name':[],                                                                                                                                                                                                                                 'chemical_formula':'',       'bsc_parameter_name':'pm10asy550',       'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
#PM2.5
'PM2.5':                                {'long_parameter_name':'PM2.5 mass',                                                    'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['88101','88502'],                                  'airbase_name':'Particulate matter < 2.5 m (aerosol)', 'airbase_code':'06001', 'miteco_code':'9', 'ebas_parameter_name':['pm25_mass'],                                              'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['PM2.5','MASS'],                        'castnet_parameter_name':'FINE_MASS',     'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':['12'], 'japan_nies_name':['PM25','PMFL','PMBH'], 'uk_air_name':['PM2.5 particulate matter (Hourly measured)','PM<sub>2.5</sub> particulate matter (Hourly measured)','PM2.5 particulate matter (Daily measured)','Daily measured PM2.5 (uncorrected)','Daily measured PM<sub>2.5</sub> (uncorrected)'], 'chemical_formula':'',       'bsc_parameter_name':'pm2p5',             'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle mass concentration, PM2.5',              'WIGOS_number':'366'}},
'PM2.5_Al':                             {'long_parameter_name':'PM2.5 aluminium',                                               'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['84101','88104'],                                  'airbase_name':'aluminium in PM2.5 (aerosol)',         'airbase_code':'01604', 'miteco_code':'',  'ebas_parameter_name':['aluminium'],                                              'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Aluminum (Al)','Aluminum','Al'],       'castnet_parameter_name':'AL',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'Al',     'bsc_parameter_name':'pm2p5al',           'identifiers':{'IUPAC_name':'aluminum',             'CAS_number':'7429-90-5',  'PubChem_CID':'5359268', 'InChIKey':'XAGFODPZIPBFFR-UHFFFAOYSA-N', 'WIGOS_name':'Al (Aluminium), PM2.5',                           'WIGOS_number':'673'}},
'PM2.5_As':                             {'long_parameter_name':'PM2.5 arsenic',                                                 'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':1.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':200.0,   'aqs_code':['84103','88103'],                                  'airbase_name':'Arsenic in PM2.5 (aerosol)',           'airbase_code':'01018', 'miteco_code':'',  'ebas_parameter_name':['arsenic'],                                                'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Arsenic (As)','Arsenic','As'],         'castnet_parameter_name':'ARSENIC',       'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'As',     'bsc_parameter_name':'pm2p5as',           'identifiers':{'IUPAC_name':'arsenic',              'CAS_number':'7440-38-2',  'PubChem_CID':'5359596', 'InChIKey':'RQNWIZPPADIBDY-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_BaP':                            {'long_parameter_name':'PM2.5 benzo[a]pyrene',                                          'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.1,    'extreme_lower_limit':0.0,   'extreme_upper_limit':100.0,   'extreme_upper_monthly_median':20.0,    'aqs_code':[],                                                 'airbase_name':'Benzo(a)pyrene in PM2.5 (aerosol)',    'airbase_code':'01029', 'miteco_code':'',  'ebas_parameter_name':['benzo_a_pyrene'],                                         'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'C20H12', 'bsc_parameter_name':'pm2p5bap',          'identifiers':{'IUPAC_name':'benzo[a]pyrene',       'CAS_number':'50-32-8',    'PubChem_CID':'2336',    'InChIKey':'FMMWHPNWAFZXNH-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_BC':                             {'long_parameter_name':'PM2.5 black carbon',                                            'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':25000.0, 'extreme_upper_monthly_median':2500.0,  'aqs_code':['84313','88317'],                                  'airbase_name':'black_carbon in PM2.5 (aerosol)',      'airbase_code':'01391', 'miteco_code':'',  'ebas_parameter_name':['equivalent_black_carbon','equivalent_black_carbon_mass'], 'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'C',      'bsc_parameter_name':'pm2p5bc',           'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_C':                              {'long_parameter_name':'PM2.5 carbon',                                                  'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['84312','88312'],                                  'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['total_carbon','total_carbon_corrected'],                  'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['TC(corr)','TC'],                       'castnet_parameter_name':'TCTC',          'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':['TC-PM2.5'],                                                                                                                                                                                                                            'chemical_formula':'C',      'bsc_parameter_name':'pm2p5c',            'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_Ca++':                           {'long_parameter_name':'PM2.5 calcium',                                                 'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':100.0,   'extreme_upper_monthly_median':20.0,    'aqs_code':['84111','88111'],                                  'airbase_name':'calcium in PM2.5 (aerosol)',           'airbase_code':'01629', 'miteco_code':'',  'ebas_parameter_name':['calcium'],                                                'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Calcium','Ca'],                        'castnet_parameter_name':'CA',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':['Calcium aerosol - PM2.5'],                                                                                                                                                                                                             'chemical_formula':'Ca',     'bsc_parameter_name':'pm2p5ca',           'identifiers':{'IUPAC_name':'calcium(2+)',          'CAS_number':'14127-61-8', 'PubChem_CID':'271',     'InChIKey':'BHPQYMZQTOCNFJ-UHFFFAOYSA-N', 'WIGOS_name':'Ca++ (calcium), PM2.5',                           'WIGOS_number':'633'}},
'PM2.5_Cd':                             {'long_parameter_name':'PM2.5 cadmium',                                                 'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':500.0,   'extreme_upper_monthly_median':75.0,    'aqs_code':['84110','88110'],                                  'airbase_name':'Cadmium in PM2.5 (aerosol)',           'airbase_code':'01014', 'miteco_code':'',  'ebas_parameter_name':['cadmium'],                                                'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Cadmium (Cd)','Cadmium','Cd'],         'castnet_parameter_name':'CD',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'Cd',     'bsc_parameter_name':'pm2p5cd',           'identifiers':{'IUPAC_name':'cadmium',              'CAS_number':'7440-43-9',  'PubChem_CID':'23973',   'InChIKey':'BDOSMKKIYDKNTQ-UHFFFAOYSA-N', 'WIGOS_name':'Cd (Cadmium), PM2.5',                             'WIGOS_number':'685'}},
'PM2.5_Cl-':                            {'long_parameter_name':'PM2.5 chloride',                                                'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['84115','84203','88115','88203'],                  'airbase_name':'chloride in PM2.5 (aerosol)',          'airbase_code':'01631', 'miteco_code':'',  'ebas_parameter_name':['chloride'],                                               'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Chloride','CHLORIDE','Cl-'],           'castnet_parameter_name':'CL',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':['Chloride aerosol - PM2.5'],                                                                                                                                                                                                            'chemical_formula':'Cl',     'bsc_parameter_name':'pm2p5cl',           'identifiers':{'IUPAC_name':'chloride',             'CAS_number':'16887-00-6', 'PubChem_CID':'312',     'InChIKey':'VEXZGXHMUGYJMC-UHFFFAOYSA-M', 'WIGOS_name':'Cl- (chloride), PM2.5',                           'WIGOS_number':'616'}},
'PM2.5_Co':                             {'long_parameter_name':'PM2.5 cobalt',                                                  'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.1,    'extreme_lower_limit':0.0,   'extreme_upper_limit':50.0,    'extreme_upper_monthly_median':5.0,     'aqs_code':['84113','88113'],                                  'airbase_name':'Cobalt in PM2.5 (aerosol)',            'airbase_code':'01064', 'miteco_code':'',  'ebas_parameter_name':['cobalt'],                                                 'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Cobalt (Co)','Cobalt','Co'],           'castnet_parameter_name':'CO',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'Co',     'bsc_parameter_name':'pm2p5cobalt',       'identifiers':{'IUPAC_name':'cobalt',               'CAS_number':'7440-48-4',  'PubChem_CID':'104730',  'InChIKey':'GUTLYIVDDKVIGB-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_Cr':                             {'long_parameter_name':'PM2.5 chromium',                                                'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':1.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':500.0,   'extreme_upper_monthly_median':100.0,   'aqs_code':['84112','88112'],                                  'airbase_name':'Chromium in PM2.5 (aerosol)',          'airbase_code':'01016', 'miteco_code':'',  'ebas_parameter_name':['chromium'],                                               'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Chromium (Cr)','Chromium','Cr'],       'castnet_parameter_name':'CR',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'Cr',     'bsc_parameter_name':'pm2p5cr',           'identifiers':{'IUPAC_name':'chromium',             'CAS_number':'7440-47-3',  'PubChem_CID':'23976',   'InChIKey':'VYZAMTAEIAYCRO-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_Cu':                             {'long_parameter_name':'PM2.5 copper',                                                  'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':1.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':750.0,   'extreme_upper_monthly_median':150.0,   'aqs_code':['84114','88114'],                                  'airbase_name':'Copper in PM2.5 (aerosol)',            'airbase_code':'01073', 'miteco_code':'',  'ebas_parameter_name':['copper'],                                                 'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Copper (Cu)','Copper','Cu'],           'castnet_parameter_name':'CU',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'Cu',     'bsc_parameter_name':'pm2p5cu',           'identifiers':{'IUPAC_name':'copper',               'CAS_number':'7440-50-8',  'PubChem_CID':'23978',   'InChIKey':'RYGMFSIKBFXOCR-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_DUST':                           {'long_parameter_name':'PM2.5 dust',                                                    'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5du',           'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_EC':                             {'long_parameter_name':'PM2.5 elemental carbon',                                        'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':25000.0, 'extreme_upper_monthly_median':2500.0,  'aqs_code':['84307','88307','88321','88357','88380','88381'],  'airbase_name':'Elemental carbon in PM2.5 (aerosol)',  'airbase_code':'01771', 'miteco_code':'',  'ebas_parameter_name':['elemental_carbon'],                                       'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['EC(corr)','EC','EC_R','EC_T','EC(A)'], 'castnet_parameter_name':['EHTC','ECTC'], 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':['EC-TOR-PM2.5','EC-TOT-PM2.5'],                                                                                                                                                                                                         'chemical_formula':'C',      'bsc_parameter_name':'pm2p5ec',           'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'Elemental carbon, PM2.5',                         'WIGOS_number':'625'}},
'PM2.5_Fe':                             {'long_parameter_name':'PM2.5 iron',                                                    'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['84126','88126'],                                  'airbase_name':'Iron in PM2.5 (aerosol)',              'airbase_code':'01065', 'miteco_code':'',  'ebas_parameter_name':['iron'],                                                   'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Iron (Fe)','Iron','Fe'],               'castnet_parameter_name':'FE',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'Fe',     'bsc_parameter_name':'pm2p5fe',           'identifiers':{'IUPAC_name':'iron',                 'CAS_number':'7439-89-6',  'PubChem_CID':'23925',   'InChIKey':'XEEYBQQBJWHFJM-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_Hg':                             {'long_parameter_name':'PM2.5 mercury',                                                 'matrix':'pm2.5', 'standard_units':'pg m-3',   'axis_label':'concentration',               'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':30000.0, 'extreme_upper_monthly_median':3000.0,  'aqs_code':['84142','88142'],                                  'airbase_name':'Mercury in PM2.5 (aerosol)',           'airbase_code':'01013', 'miteco_code':'',  'ebas_parameter_name':['mercury','mercury_PBM'],                                  'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'HG',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':['mercury in PM2.5'],                                                                                                                                                                                                                    'chemical_formula':'Hg',     'bsc_parameter_name':'pm2p5hg',           'identifiers':{'IUPAC_name':'mercury',              'CAS_number':'7439-97-6',  'PubChem_CID':'23931',   'InChIKey':'QSHDDOUJBYECFT-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_K+':                             {'long_parameter_name':'PM2.5 potassium',                                               'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':50.0,    'extreme_upper_monthly_median':10.0,    'aqs_code':['84180','88180','88303'],                          'airbase_name':'potassium in PM2.5 (aerosol)',         'airbase_code':'01657', 'miteco_code':'',  'ebas_parameter_name':['potassium'],                                              'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Potassium','K'],                       'castnet_parameter_name':'K',             'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':['Potassium aerosol - PM2.5'],                                                                                                                                                                                                           'chemical_formula':'K',      'bsc_parameter_name':'pm2p5k',            'identifiers':{'IUPAC_name':'potassium(1+)',        'CAS_number':'24203-36-9', 'PubChem_CID':'813',     'InChIKey':'NPYPAHLBTDXSSS-UHFFFAOYSA-N', 'WIGOS_name':'K+ (potassium), PM2.5',                           'WIGOS_number':'639'}},
'PM2.5_Mg++':                           {'long_parameter_name':'PM2.5 magnesium',                                               'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':50.0,    'extreme_upper_monthly_median':10.0,    'aqs_code':['84140','88140'],                                  'airbase_name':'magnesium in PM2.5 (aerosol)',         'airbase_code':'01659', 'miteco_code':'',  'ebas_parameter_name':['magnesium'],                                              'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Magnesium','Mg'],                      'castnet_parameter_name':'MG',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':['Magnesium aerosol - PM2.5'],                                                                                                                                                                                                           'chemical_formula':'Mg',     'bsc_parameter_name':'pm2p5mg',           'identifiers':{'IUPAC_name':'magnesium(2+)',        'CAS_number':'22537-22-0', 'PubChem_CID':'888',     'InChIKey':'JLVVSXFLKOJNIY-UHFFFAOYSA-N', 'WIGOS_name':'Mg++ (magnesium), PM2.5',                         'WIGOS_number':'636'}},
'PM2.5_Mn':                             {'long_parameter_name':'PM2.5 manganese',                                               'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':2.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':['84132','88132'],                                  'airbase_name':'Manganese in PM2.5 (aerosol)',         'airbase_code':'01017', 'miteco_code':'',  'ebas_parameter_name':['manganese'],                                              'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Manganese (Mn)','Mn','Manganese'],     'castnet_parameter_name':'MN',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'Mn',     'bsc_parameter_name':'pm2p5mn',           'identifiers':{'IUPAC_name':'manganese',            'CAS_number':'7439-96-5',  'PubChem_CID':'23930',   'InChIKey':'PWHULOQIROXLJO-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_MSA':                            {'long_parameter_name':'PM2.5 methanesulfonic acid',                                    'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':75.0,    'extreme_upper_monthly_median':25.0,    'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['methanesulfonic_acid'],                                   'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['MSA'],                                 'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'CH4O3S', 'bsc_parameter_name':'pm2p5msa',          'identifiers':{'IUPAC_name':'methanesulfonic acid', 'CAS_number':'75-75-2',    'PubChem_CID':'6395',    'InChIKey':'AFVFQIVMOAPDHO-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_Na+':                            {'long_parameter_name':'PM2.5 sodium',                                                  'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['88184','88302'],                                  'airbase_name':'sodium in PM2.5 (aerosol)',            'airbase_code':'01668', 'miteco_code':'',  'ebas_parameter_name':['sodium'],                                                 'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Sodium','Na+'],                        'castnet_parameter_name':'NA',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':['Sodium aerosol - PM2.5'],                                                                                                                                                                                                              'chemical_formula':'Na',     'bsc_parameter_name':'pm2p5na',           'identifiers':{'IUPAC_name':'sodium(1+)',           'CAS_number':'17341-25-2', 'PubChem_CID':'923',     'InChIKey':'FKNQFGJONOIPTF-UHFFFAOYSA-N', 'WIGOS_name':'Na+ (sodium), PM2.5',                             'WIGOS_number':'642'}},
'PM2.5_NH4+':                           {'long_parameter_name':'PM2.5 ammonium',                                                'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['88301'],                                          'airbase_name':'Ammonium in PM2.5 (aerosol)',          'airbase_code':'01045', 'miteco_code':'',  'ebas_parameter_name':['ammonium'],                                               'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Ammonium','AMMONIUM','NH4+','NH4'],    'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':['Ammonium aerosol - PM2.5'],                                                                                                                                                                                                            'chemical_formula':'NH4',    'bsc_parameter_name':'pm2p5nh4',          'identifiers':{'IUPAC_name':'azanium',              'CAS_number':'14798-03-9', 'PubChem_CID':'223',     'InChIKey':'QGZKDVFQNNGYKY-UHFFFAOYSA-O', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_NH4NO3':                         {'long_parameter_name':'PM2.5 ammonium nitrate',                                        'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['88344'],                                          'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['ammonium_nitrate'],                                       'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'NH4NO3', 'bsc_parameter_name':'pm2p5nh4no3',       'identifiers':{'IUPAC_name':'azanium;nitrate',      'CAS_number':'6484-52-2',  'PubChem_CID':'22985',   'InChIKey':'DVARTQFDIMZBAA-UHFFFAOYSA-O', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_Ni':                             {'long_parameter_name':'PM2.5 nickel',                                                  'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':5.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':10000.0, 'extreme_upper_monthly_median':1000.0,  'aqs_code':['84136','88136'],                                  'airbase_name':'Nickel in PM2.5 (aerosol)',            'airbase_code':'01015', 'miteco_code':'',  'ebas_parameter_name':['nickel'],                                                 'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Nickel (Ni)','Nickel','Ni'],           'castnet_parameter_name':'NI',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'Ni',     'bsc_parameter_name':'pm2p5ni',           'identifiers':{'IUPAC_name':'nickel',               'CAS_number':'7440-02-0',  'PubChem_CID':'935',     'InChIKey':'PXHVJJICTQNCMI-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_NO3-':                           {'long_parameter_name':'PM2.5 nitrate',                                                 'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':250.0,   'extreme_upper_monthly_median':75.0,    'aqs_code':['82356','84306','88306'],                          'airbase_name':'Nitrate in PM2.5 (aerosol)',           'airbase_code':'01046', 'miteco_code':'',  'ebas_parameter_name':['nitrate'],                                                'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Nitrate','NITRATE','NO3'],             'castnet_parameter_name':'NO3',           'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':['Nitrate aerosol - PM2.5','Particulate nitrate - PM2.5'],                                                                                                                                                                               'chemical_formula':'NO3',    'bsc_parameter_name':'pm2p5no3',          'identifiers':{'IUPAC_name':'nitrate',              'CAS_number':'14797-55-8', 'PubChem_CID':'943',     'InChIKey':'NHNBFGGVMKEFGY-UHFFFAOYSA-N', 'WIGOS_name':'NO3- (nitrate), PM2.5',                           'WIGOS_number':'651'}},
'PM2.5_OC':                             {'long_parameter_name':'PM2.5 organic carbon',                                          'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':25000.0, 'extreme_upper_monthly_median':2500.0,  'aqs_code':['84305','88305','88320','88355','88370', '88382'], 'airbase_name':'Organic carbon in PM2.5 (aerosol)',    'airbase_code':'01772', 'miteco_code':'',  'ebas_parameter_name':['organic_carbon','organic_carbon_corrected'],              'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['OC(corr)','OC','OC_R','OC_T'],         'castnet_parameter_name':['OHTC','OCTC'], 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':['OC-TOR-PM2.5','OC-TOT-PM2.5','Organic Carbon - PM2.5'],                                                                                                                                                                                'chemical_formula':'C',      'bsc_parameter_name':'pm2p5oc',           'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'Carbonaceous-organic material, fine',             'WIGOS_number':'671'}},
'PM2.5_Pb':                             {'long_parameter_name':'PM2.5 lead',                                                    'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':50.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':60000.0, 'extreme_upper_monthly_median':15000.0, 'aqs_code':['84128','88128'],                                  'airbase_name':'Lead in PM2.5 (aerosol)',              'airbase_code':'01012', 'miteco_code':'',  'ebas_parameter_name':['lead'],                                                   'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Lead (Pb)','Lead','Pb'],               'castnet_parameter_name':'PB',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'Pb',     'bsc_parameter_name':'pm2p5pb',           'identifiers':{'IUPAC_name':'lead',                 'CAS_number':'7439-92-1',  'PubChem_CID':'5352425', 'InChIKey':'WABPQHHGFIMREM-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_Se':                             {'long_parameter_name':'PM2.5 selenium',                                                'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['84154','88154'],                                  'airbase_name':'Selenium in PM2.5 (aerosol)',          'airbase_code':'01048', 'miteco_code':'',  'ebas_parameter_name':['selenium'],                                               'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Selenium (Se)','Selenium','Se'],       'castnet_parameter_name':'SE',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'Se',     'bsc_parameter_name':'pm2p5se',           'identifiers':{'IUPAC_name':'selenium',             'CAS_number':'7782-49-2',  'PubChem_CID':'6326970', 'InChIKey':'BUGBHKTXTAQXES-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_SO4--':                          {'long_parameter_name':'PM2.5 sulphate',                                                'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':['82453','84403','88403'],                          'airbase_name':'sulphate in PM2.5 (aerosol)',          'airbase_code':'01047', 'miteco_code':'',  'ebas_parameter_name':['sulphate_total'],                                         'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Sulphate','SULPHATE','SO4'],           'castnet_parameter_name':'SO4',           'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':['Sulphate aerosol - PM2.5'],                                                                                                                                                                                                            'chemical_formula':'SO4',    'bsc_parameter_name':'pm2p5so4',          'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'SO4= (sulfate), PM2.5',                           'WIGOS_number':'622'}},
'PM2.5_SO4--_NSS':                      {'long_parameter_name':'PM2.5 sulphate : non-sea salt',                                 'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['sulphate_corrected'],                                     'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'SO4',    'bsc_parameter_name':'pm2p5so4nss',       'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_SO4--_SS':                       {'long_parameter_name':'PM2.5 sulphate : sea salt',                                     'matrix':'pm2.5', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['sulphate_seasalt'],                                       'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'SO4',    'bsc_parameter_name':'pm2p5so4ss',        'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM2.5_V':                              {'long_parameter_name':'PM2.5 vanadium',                                                'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':100.0,   'extreme_upper_monthly_median':20.0,    'aqs_code':['84164','88164'],                                  'airbase_name':'Vanedium in PM2.5 (aerosol)',          'airbase_code':'01049', 'miteco_code':'',  'ebas_parameter_name':['vanadium'],                                               'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Vanadium (V)','Vanadium','V'],         'castnet_parameter_name':'VA',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'V',      'bsc_parameter_name':'pm2p5v',            'identifiers':{'IUPAC_name':'vanadium',             'CAS_number':'7440-62-2',  'PubChem_CID':'23990',   'InChIKey':'LEONUFNNVUYDNQ-UHFFFAOYSA-N', 'WIGOS_name':'V (Vanadium), PM2.5',                             'WIGOS_number':'721'}},
'PM2.5_Zn':                             {'long_parameter_name':'PM2.5 zinc',                                                    'matrix':'pm2.5', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':30000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['84167','88167'],                                  'airbase_name':'Zinc in PM2.5 (aerosol)',              'airbase_code':'01063', 'miteco_code':'',  'ebas_parameter_name':['zinc'],                                                   'ebas_matrix':'pm25', 'ebas_priority_units':['/m3'], 'naps_code':['Zinc (Zn)','Zinc','Zn'],               'castnet_parameter_name':'ZN',            'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'Zn',     'bsc_parameter_name':'pm2p5zn',           'identifiers':{'IUPAC_name':'zinc',                 'CAS_number':'7440-66-6',  'PubChem_CID':'23994',   'InChIKey':'HCHKCACWOHOZIP-UHFFFAOYSA-N', 'WIGOS_name':'Zn (Zinc), PM2.5',                                'WIGOS_number':'723'}},
'PM2.5_AERO_ABS_COEFF_370nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 370nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco370',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_467nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 467nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco467',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_470nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 470nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco470',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_520nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 520nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco520',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_522nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 522nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco522',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_525nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 525nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco525',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_528nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 528nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco528',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_530nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 530nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco530',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_550nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 550nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco550',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_565nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 565nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco565',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_590nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 590nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco590',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_637nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 637nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco637',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_652nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 652nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco652',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_660nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 660nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco660',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_670nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 670nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco670',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_880nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 880nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco880',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_ABS_COEFF_950nm':           {'long_parameter_name':'PM2.5 aerosol absorption coefficient at 950nm',                 'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absco950',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_ABSAE_370-880nm':                {'long_parameter_name':'angstrom exponent between 370 and 880 nm for PM2.5 absorption', 'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'angstrom exponent 370-880nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absae370-880', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'PM2.5_ABSAE_370-950nm':                {'long_parameter_name':'angstrom exponent between 370 and 950 nm for PM2.5 absorption', 'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'angstrom exponent 370-950nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absae370-950', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'PM2.5_ABSAE_467-652nm':                {'long_parameter_name':'angstrom exponent between 467 and 652 nm for PM2.5 absorption', 'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'angstrom exponent 467-652nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absae467-652', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'PM2.5_ABSAE_467-660nm':                {'long_parameter_name':'angstrom exponent between 467 and 660 nm for PM2.5 absorption', 'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'angstrom exponent 467-660nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5absae467-660', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm':{'long_parameter_name':'PM2.5 aerosol light backscattering coefficient at 450nm',       'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsco450',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm':{'long_parameter_name':'PM2.5 aerosol light backscattering coefficient at 520nm',       'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsco520',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm':{'long_parameter_name':'PM2.5 aerosol light backscattering coefficient at 525nm',       'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsco525',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm':{'long_parameter_name':'PM2.5 aerosol light backscattering coefficient at 530nm',       'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsco530',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm':{'long_parameter_name':'PM2.5 aerosol light backscattering coefficient at 532nm',       'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsco532',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm':{'long_parameter_name':'PM2.5 aerosol light backscattering coefficient at 550nm',       'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsco550',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm':{'long_parameter_name':'PM2.5 aerosol light backscattering coefficient at 635nm',       'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsco635',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm':{'long_parameter_name':'PM2.5 aerosol light backscattering coefficient at 700nm',       'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsco700',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm':{'long_parameter_name':'PM2.5 aerosol light backscattering coefficient at 850nm',       'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsco850',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}}, 
'PM2.5_AERO_LIGHT_SCAT_COEFF_450nm':    {'long_parameter_name':'PM2.5 aerosol light scattering coefficient at 450nm',           'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lsco450',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_AERO_LIGHT_SCAT_COEFF_520nm':    {'long_parameter_name':'PM2.5 aerosol light scattering coefficient at 520nm',           'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lsco520',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_AERO_LIGHT_SCAT_COEFF_525nm':    {'long_parameter_name':'PM2.5 aerosol light scattering coefficient at 525nm',           'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lsco525',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_AERO_LIGHT_SCAT_COEFF_530nm':    {'long_parameter_name':'PM2.5 aerosol light scattering coefficient at 530nm',           'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lsco530',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_AERO_LIGHT_SCAT_COEFF_532nm':    {'long_parameter_name':'PM2.5 aerosol light scattering coefficient at 532nm',           'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lsco532',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_AERO_LIGHT_SCAT_COEFF_550nm':    {'long_parameter_name':'PM2.5 aerosol light scattering coefficient at 550nm',           'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lsco550',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_AERO_LIGHT_SCAT_COEFF_635nm':    {'long_parameter_name':'PM2.5 aerosol light scattering coefficient at 635nm',           'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lsco635',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_AERO_LIGHT_SCAT_COEFF_700nm':    {'long_parameter_name':'PM2.5 aerosol light scattering coefficient at 700nm',           'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lsco700',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_AERO_LIGHT_SCAT_COEFF_850nm':    {'long_parameter_name':'PM2.5 aerosol light scattering coefficient at 850nm',           'matrix':'pm2.5', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm25', 'ebas_priority_units':['/Mm'], 'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lsco850',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_SCATAE_450-635nm':               {'long_parameter_name':'angstrom exponent between 450 and 635 nm for PM2.5 scattering', 'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'angstrom exponent 450-635nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':-2.0,  'extreme_upper_limit':4.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lsae450-635',  'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light scattering', 'WIGOS_number':'12148'}},  
'PM2.5_SCATAE_450-700nm':               {'long_parameter_name':'angstrom exponent between 450 and 700 nm for PM2.5 scattering', 'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'angstrom exponent 450-700nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':-2.0,  'extreme_upper_limit':4.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lsae450-700',  'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light scattering', 'WIGOS_number':'12148'}},  
'PM2.5_BACKSCAT_FRAC_450nm':            {'long_parameter_name':'PM2.5 aerosol light backscattering fraction at 450nm',          'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsf450',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_BACKSCAT_FRAC_520nm':            {'long_parameter_name':'PM2.5 aerosol light backscattering fraction at 520nm',          'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsf520',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_BACKSCAT_FRAC_525nm':            {'long_parameter_name':'PM2.5 aerosol light backscattering fraction at 525nm',          'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsf525',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_BACKSCAT_FRAC_530nm':            {'long_parameter_name':'PM2.5 aerosol light backscattering fraction at 530nm',          'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsf530',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_BACKSCAT_FRAC_532nm':            {'long_parameter_name':'PM2.5 aerosol light backscattering fraction at 532nm',          'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsf532',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_BACKSCAT_FRAC_550nm':            {'long_parameter_name':'PM2.5 aerosol light backscattering fraction at 550nm',          'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsf550',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_BACKSCAT_FRAC_635nm':            {'long_parameter_name':'PM2.5 aerosol light backscattering fraction at 635nm',          'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsf635',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_BACKSCAT_FRAC_700nm':            {'long_parameter_name':'PM2.5 aerosol light backscattering fraction at 700nm',          'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsf700',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_BACKSCAT_FRAC_850nm':            {'long_parameter_name':'PM2.5 aerosol light backscattering fraction at 850nm',          'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5lbsf850',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_ASY_525nm':                      {'long_parameter_name':'PM2.5 asymmetry factor at 525nm',                               'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'asymmetry factor',            'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':1.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5asy525',       'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM2.5_ASY_550nm':                      {'long_parameter_name':'PM2.5 asymmetry factor at 550nm',                               'matrix':'pm2.5', 'standard_units':'unitless', 'axis_label':'asymmetry factor',            'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':1.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],                                                 'airbase_name':'',                                     'airbase_code':'',      'miteco_code':'',  'ebas_parameter_name':[],                                                         'ebas_matrix':'',     'ebas_priority_units':[''],    'naps_code':[],                                      'castnet_parameter_name':'',              'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[],     'japan_nies_name':[],                     'uk_air_name':[],                                                                                                                                                                                                                                      'chemical_formula':'',       'bsc_parameter_name':'pm2p5asy550',       'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
#PM1
'PM1':                                {'long_parameter_name':'PM1 mass',                                                    'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':['87111'], 'airbase_name':'Particulate matter < 1 µm (aerosol)', 'airbase_code':'06002', 'miteco_code':'', 'ebas_parameter_name':['pm1_mass'],                                               'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1',             'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle mass concentration, PM1',                'WIGOS_number':'365'}},
'PM1_Al':                             {'long_parameter_name':'PM1 aluminium',                                               'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aluminium'],                                              'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Al',     'bsc_parameter_name':'pm1al',           'identifiers':{'IUPAC_name':'aluminum',             'CAS_number':'7429-90-5',  'PubChem_CID':'5359268', 'InChIKey':'XAGFODPZIPBFFR-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_As':                             {'long_parameter_name':'PM1 arsenic',                                                 'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':1.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':200.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['arsenic'],                                                'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'As',     'bsc_parameter_name':'pm1as',           'identifiers':{'IUPAC_name':'arsenic',              'CAS_number':'7440-38-2',  'PubChem_CID':'5359596', 'InChIKey':'RQNWIZPPADIBDY-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_BaP':                            {'long_parameter_name':'PM1 benzo[a]pyrene',                                          'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.1,    'extreme_lower_limit':0.0,   'extreme_upper_limit':100.0,   'extreme_upper_monthly_median':20.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['benzo_a_pyrene'],                                         'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'C20H12', 'bsc_parameter_name':'pm1bap',          'identifiers':{'IUPAC_name':'benzo[a]pyrene',       'CAS_number':'50-32-8',    'PubChem_CID':'2336',    'InChIKey':'FMMWHPNWAFZXNH-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_BC':                             {'long_parameter_name':'PM1 black carbon',                                            'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':25000.0, 'extreme_upper_monthly_median':2500.0,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['equivalent_black_carbon','equivalent_black_carbon_mass'], 'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'C',      'bsc_parameter_name':'pm1bc',           'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_C':                              {'long_parameter_name':'PM1 carbon',                                                  'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['total_carbon','total_carbon_corrected'],                  'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'C',      'bsc_parameter_name':'pm1c',            'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_Ca++':                           {'long_parameter_name':'PM1 calcium',                                                 'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':100.0,   'extreme_upper_monthly_median':20.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['calcium'],                                                'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Ca',     'bsc_parameter_name':'pm1ca',           'identifiers':{'IUPAC_name':'calcium(2+)',          'CAS_number':'14127-61-8', 'PubChem_CID':'271',     'InChIKey':'BHPQYMZQTOCNFJ-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_Cd':                             {'long_parameter_name':'PM1 cadmium',                                                 'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':500.0,   'extreme_upper_monthly_median':75.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['cadmium'],                                                'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Cd',     'bsc_parameter_name':'pm1cd',           'identifiers':{'IUPAC_name':'cadmium',              'CAS_number':'7440-43-9',  'PubChem_CID':'23973',   'InChIKey':'BDOSMKKIYDKNTQ-UHFFFAOYSA-N', 'WIGOS_name':'Cd (Cadmium), PM1',                               'WIGOS_number':'683'}},
'PM1_Cl-':                            {'long_parameter_name':'PM1 chloride',                                                'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['chloride'],                                               'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Cl',     'bsc_parameter_name':'pm1cl',           'identifiers':{'IUPAC_name':'chloride',             'CAS_number':'16887-00-6', 'PubChem_CID':'312',     'InChIKey':'VEXZGXHMUGYJMC-UHFFFAOYSA-M', 'WIGOS_name':'Cl- (chloride), PM1',                             'WIGOS_number':'614'}},
'PM1_Co':                             {'long_parameter_name':'PM1 cobalt',                                                  'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.1,    'extreme_lower_limit':0.0,   'extreme_upper_limit':50.0,    'extreme_upper_monthly_median':5.0,     'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['cobalt'],                                                 'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Co',     'bsc_parameter_name':'pm1cobalt',       'identifiers':{'IUPAC_name':'cobalt',               'CAS_number':'7440-48-4',  'PubChem_CID':'104730',  'InChIKey':'GUTLYIVDDKVIGB-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_Cr':                             {'long_parameter_name':'PM1 chromium',                                                'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':1.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':500.0,   'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['chromium'],                                               'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Cr',     'bsc_parameter_name':'pm1cr',           'identifiers':{'IUPAC_name':'chromium',             'CAS_number':'7440-47-3',  'PubChem_CID':'23976',   'InChIKey':'VYZAMTAEIAYCRO-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_Cu':                             {'long_parameter_name':'PM1 copper',                                                  'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':1.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':750.0,   'extreme_upper_monthly_median':150.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['copper'],                                                 'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Cu',     'bsc_parameter_name':'pm1cu',           'identifiers':{'IUPAC_name':'copper',               'CAS_number':'7440-50-8',  'PubChem_CID':'23978',   'InChIKey':'RYGMFSIKBFXOCR-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_DUST':                           {'long_parameter_name':'PM1 dust',                                                    'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1du',           'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_EC':                             {'long_parameter_name':'PM1 elemental carbon',                                        'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':25000.0, 'extreme_upper_monthly_median':2500.0,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['elemental_carbon'],                                       'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'C',      'bsc_parameter_name':'pm1ec',           'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'Elemental carbon, PM1',                           'WIGOS_number':'624'}},
'PM1_Fe':                             {'long_parameter_name':'PM1 iron',                                                    'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':50000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['iron'],                                                   'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Fe',     'bsc_parameter_name':'pm1fe',           'identifiers':{'IUPAC_name':'iron',                 'CAS_number':'7439-89-6',  'PubChem_CID':'23925',   'InChIKey':'XEEYBQQBJWHFJM-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_Hg':                             {'long_parameter_name':'PM1 mercury',                                                 'matrix':'pm1', 'standard_units':'pg m-3',   'axis_label':'concentration',               'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':30000.0, 'extreme_upper_monthly_median':3000.0,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['mercury','mercury_PBM'],                                  'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Hg',     'bsc_parameter_name':'pm1hg',           'identifiers':{'IUPAC_name':'mercury',              'CAS_number':'7439-97-6',  'PubChem_CID':'23931',   'InChIKey':'QSHDDOUJBYECFT-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_K+':                             {'long_parameter_name':'PM1 potassium',                                               'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':50.0,    'extreme_upper_monthly_median':10.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['potassium'],                                              'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'K',      'bsc_parameter_name':'pm1k',            'identifiers':{'IUPAC_name':'potassium(1+)',        'CAS_number':'24203-36-9', 'PubChem_CID':'813',     'InChIKey':'NPYPAHLBTDXSSS-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_Mg++':                           {'long_parameter_name':'PM1 magnesium',                                               'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':50.0,    'extreme_upper_monthly_median':10.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['magnesium'],                                              'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Mg',     'bsc_parameter_name':'pm1mg',           'identifiers':{'IUPAC_name':'magnesium(2+)',        'CAS_number':'22537-22-0', 'PubChem_CID':'888',     'InChIKey':'JLVVSXFLKOJNIY-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_Mn':                             {'long_parameter_name':'PM1 manganese',                                               'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':2.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['manganese'],                                              'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Mn',     'bsc_parameter_name':'pm1mn',           'identifiers':{'IUPAC_name':'manganese',            'CAS_number':'7439-96-5',  'PubChem_CID':'23930',   'InChIKey':'PWHULOQIROXLJO-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_MSA':                            {'long_parameter_name':'PM1 methanesulfonic acid',                                    'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':75.0,    'extreme_upper_monthly_median':25.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['methanesulfonic_acid'],                                   'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'CH4O3S', 'bsc_parameter_name':'pm1msa',          'identifiers':{'IUPAC_name':'methanesulfonic acid', 'CAS_number':'75-75-2',    'PubChem_CID':'6395',    'InChIKey':'AFVFQIVMOAPDHO-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_Na+':                            {'long_parameter_name':'PM1 sodium',                                                  'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['sodium'],                                                 'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Na',     'bsc_parameter_name':'pm1na',           'identifiers':{'IUPAC_name':'sodium(1+)',           'CAS_number':'17341-25-2', 'PubChem_CID':'923',     'InChIKey':'FKNQFGJONOIPTF-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_NH4+':                           {'long_parameter_name':'PM1 ammonium',                                                'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['ammonium'],                                               'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Ammonium aerosol - PM1'], 'chemical_formula':'NH4',    'bsc_parameter_name':'pm1nh4',          'identifiers':{'IUPAC_name':'azanium',              'CAS_number':'14798-03-9', 'PubChem_CID':'223',     'InChIKey':'QGZKDVFQNNGYKY-UHFFFAOYSA-O', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_NH4NO3':                         {'long_parameter_name':'PM1 ammonium nitrate',                                        'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['ammonium_nitrate'],                                       'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'NH4NO3', 'bsc_parameter_name':'pm1nh4no3',       'identifiers':{'IUPAC_name':'azanium;nitrate',      'CAS_number':'6484-52-2',  'PubChem_CID':'22985',   'InChIKey':'DVARTQFDIMZBAA-UHFFFAOYSA-O', 'WIGOS_name':'NH4NO3 (ammonium nitrate), PM1',                  'WIGOS_number':'649'}},
'PM1_Ni':                             {'long_parameter_name':'PM1 nickel',                                                  'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':5.0,    'extreme_lower_limit':0.0,   'extreme_upper_limit':10000.0, 'extreme_upper_monthly_median':1000.0,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['nickel'],                                                 'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Ni',     'bsc_parameter_name':'pm1ni',           'identifiers':{'IUPAC_name':'nickel',               'CAS_number':'7440-02-0',  'PubChem_CID':'935',     'InChIKey':'PXHVJJICTQNCMI-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_NO3-':                           {'long_parameter_name':'PM1 nitrate',                                                 'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':250.0,   'extreme_upper_monthly_median':75.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['nitrate'],                                                'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Nitrate aerosol - PM1'],  'chemical_formula':'NO3',    'bsc_parameter_name':'pm1no3',          'identifiers':{'IUPAC_name':'nitrate',              'CAS_number':'14797-55-8', 'PubChem_CID':'943',     'InChIKey':'NHNBFGGVMKEFGY-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_OC':                             {'long_parameter_name':'PM1 organic carbon',                                          'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':10.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':25000.0, 'extreme_upper_monthly_median':2500.0,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['organic_carbon','organic_carbon_corrected'],              'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Organic Carbon - PM1'],   'chemical_formula':'C',      'bsc_parameter_name':'pm1oc',           'identifiers':{'IUPAC_name':'carbon',               'CAS_number':'1333-86-4',  'PubChem_CID':'5462310', 'InChIKey':'OKTJSMMVPCPJKN-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_Pb':                             {'long_parameter_name':'PM1 lead',                                                    'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':50.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':60000.0, 'extreme_upper_monthly_median':15000.0, 'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['lead'],                                                   'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Pb',     'bsc_parameter_name':'pm1pb',           'identifiers':{'IUPAC_name':'lead',                 'CAS_number':'7439-92-1',  'PubChem_CID':'5352425', 'InChIKey':'WABPQHHGFIMREM-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_Se':                             {'long_parameter_name':'PM1 selenium',                                                'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['selenium'],                                               'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Se',     'bsc_parameter_name':'pm1se',           'identifiers':{'IUPAC_name':'selenium',             'CAS_number':'7782-49-2',  'PubChem_CID':'6326970', 'InChIKey':'BUGBHKTXTAQXES-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_SO4--':                          {'long_parameter_name':'PM1 sulphate',                                                'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['sulphate_total'],                                         'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':['Sulphate aerosol - PM1'], 'chemical_formula':'SO4',    'bsc_parameter_name':'pm1so4',          'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_SO4--_NSS':                      {'long_parameter_name':'PM1 sulphate : non-sea salt',                                 'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['sulphate_corrected'],                                     'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'SO4',    'bsc_parameter_name':'pm1so4nss',       'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_SO4--_SS':                       {'long_parameter_name':'PM1 sulphate : sea salt',                                     'matrix':'pm1', 'standard_units':'ug m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':150.0,   'extreme_upper_monthly_median':30.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['sulphate_seasalt'],                                       'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'SO4',    'bsc_parameter_name':'pm1so4ss',        'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_V':                              {'long_parameter_name':'PM1 vanadium',                                                'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':0.2,    'extreme_lower_limit':0.0,   'extreme_upper_limit':100.0,   'extreme_upper_monthly_median':20.0,    'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['vanadium'],                                               'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'V',      'bsc_parameter_name':'pm1v',            'identifiers':{'IUPAC_name':'vanadium',             'CAS_number':'7440-62-2',  'PubChem_CID':'23990',   'InChIKey':'LEONUFNNVUYDNQ-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_Zn':                             {'long_parameter_name':'PM1 zinc',                                                    'matrix':'pm1', 'standard_units':'ng m-3',   'axis_label':'concentration',               'minimum_resolution':20.0,   'extreme_lower_limit':0.0,   'extreme_upper_limit':30000.0, 'extreme_upper_monthly_median':5000.0,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['zinc'],                                                   'ebas_matrix':'pm1', 'ebas_priority_units':['/m3'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'Zn',     'bsc_parameter_name':'pm1zn',           'identifiers':{'IUPAC_name':'zinc',                 'CAS_number':'7440-66-6',  'PubChem_CID':'23994',   'InChIKey':'HCHKCACWOHOZIP-UHFFFAOYSA-N', 'WIGOS_name':'',                                                'WIGOS_number':''}},
'PM1_AERO_ABS_COEFF_370nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 370nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco370',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_467nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 467nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco467',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_470nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 470nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco470',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_520nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 520nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco520',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_522nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 522nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco522',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_525nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 525nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco525',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_528nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 528nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco528',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_530nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 530nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco530',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_550nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 550nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco550',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_565nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 565nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco565',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_590nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 590nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco590',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_637nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 637nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco637',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_652nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 652nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco652',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_660nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 660nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco660',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_670nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 670nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco670',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_880nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 880nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco880',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_AERO_ABS_COEFF_950nm':           {'long_parameter_name':'PM1 aerosol absorption coefficient at 950nm',                 'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_absorption_coefficient'],                         'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absco950',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light absorption coefficient, PM1',      'WIGOS_number':'316'}},  
'PM1_ABSAE_370-880nm':                {'long_parameter_name':'angstrom exponent between 370 and 880 nm for PM1 absorption', 'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'angstrom exponent 370-880nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absae370-880', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'PM1_ABSAE_370-950nm':                {'long_parameter_name':'angstrom exponent between 370 and 950 nm for PM1 absorption', 'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'angstrom exponent 370-950nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absae370-950', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'PM1_ABSAE_467-652nm':                {'long_parameter_name':'angstrom exponent between 467 and 652 nm for PM1 absorption', 'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'angstrom exponent 467-652nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absae467-652', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'PM1_ABSAE_467-660nm':                {'long_parameter_name':'angstrom exponent between 467 and 660 nm for PM1 absorption', 'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'angstrom exponent 467-660nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.6,   'extreme_upper_limit':10.0,    'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1absae467-660', 'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light absorption', 'WIGOS_number':'12149'}},  
'PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm':{'long_parameter_name':'PM1 aerosol light backscattering coefficient at 450nm',       'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsco450',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm':{'long_parameter_name':'PM1 aerosol light backscattering coefficient at 520nm',       'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsco520',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm':{'long_parameter_name':'PM1 aerosol light backscattering coefficient at 525nm',       'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsco525',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm':{'long_parameter_name':'PM1 aerosol light backscattering coefficient at 530nm',       'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsco530',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm':{'long_parameter_name':'PM1 aerosol light backscattering coefficient at 532nm',       'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsco532',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm':{'long_parameter_name':'PM1 aerosol light backscattering coefficient at 550nm',       'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsco550',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm':{'long_parameter_name':'PM1 aerosol light backscattering coefficient at 635nm',       'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsco635',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm':{'long_parameter_name':'PM1 aerosol light backscattering coefficient at 700nm',       'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsco700',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm':{'long_parameter_name':'PM1 aerosol light backscattering coefficient at 850nm',       'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-5.0,  'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_backscattering_coefficient'],               'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsco850',     'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_AERO_LIGHT_SCAT_COEFF_450nm':    {'long_parameter_name':'PM1 aerosol light scattering coefficient at 450nm',           'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lsco450',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM1',      'WIGOS_number':'322'}},   
'PM1_AERO_LIGHT_SCAT_COEFF_520nm':    {'long_parameter_name':'PM1 aerosol light scattering coefficient at 520nm',           'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lsco520',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM1',      'WIGOS_number':'322'}},   
'PM1_AERO_LIGHT_SCAT_COEFF_525nm':    {'long_parameter_name':'PM1 aerosol light scattering coefficient at 525nm',           'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lsco525',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM1',      'WIGOS_number':'322'}},   
'PM1_AERO_LIGHT_SCAT_COEFF_530nm':    {'long_parameter_name':'PM1 aerosol light scattering coefficient at 530nm',           'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lsco530',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM1',      'WIGOS_number':'322'}},   
'PM1_AERO_LIGHT_SCAT_COEFF_532nm':    {'long_parameter_name':'PM1 aerosol light scattering coefficient at 532nm',           'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lsco532',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM1',      'WIGOS_number':'322'}},   
'PM1_AERO_LIGHT_SCAT_COEFF_550nm':    {'long_parameter_name':'PM1 aerosol light scattering coefficient at 550nm',           'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lsco550',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM1',      'WIGOS_number':'322'}},   
'PM1_AERO_LIGHT_SCAT_COEFF_635nm':    {'long_parameter_name':'PM1 aerosol light scattering coefficient at 635nm',           'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lsco635',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM1',      'WIGOS_number':'322'}},   
'PM1_AERO_LIGHT_SCAT_COEFF_700nm':    {'long_parameter_name':'PM1 aerosol light scattering coefficient at 700nm',           'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lsco700',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM1',      'WIGOS_number':'322'}},   
'PM1_AERO_LIGHT_SCAT_COEFF_850nm':    {'long_parameter_name':'PM1 aerosol light scattering coefficient at 850nm',           'matrix':'pm1', 'standard_units':'Mm-1',     'axis_label':'coefficient',                 'minimum_resolution':np.NaN, 'extreme_lower_limit':-10.0, 'extreme_upper_limit':5000.0,  'extreme_upper_monthly_median':500.0,   'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['aerosol_light_scattering_coefficient'],                   'ebas_matrix':'pm1', 'ebas_priority_units':['/Mm'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lsco850',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Particle light scattering coefficient, PM1',      'WIGOS_number':'322'}},   
'PM1_SCATAE_450-635nm':               {'long_parameter_name':'angstrom exponent between 450 and 635 nm for PM1 scattering', 'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'angstrom exponent 450-635nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':-2.0,  'extreme_upper_limit':4.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lsae450-635',  'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light scattering', 'WIGOS_number':'12148'}},  
'PM1_SCATAE_450-700nm':               {'long_parameter_name':'angstrom exponent between 450 and 700 nm for PM1 scattering', 'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'angstrom exponent 450-700nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':-2.0,  'extreme_upper_limit':4.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lsae450-700',  'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'Ångström exponent for particle light scattering', 'WIGOS_number':'12148'}},  
'PM1_BACKSCAT_FRAC_450nm':            {'long_parameter_name':'PM1 aerosol light backscattering fraction at 450nm',          'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsf450',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_BACKSCAT_FRAC_520nm':            {'long_parameter_name':'PM1 aerosol light backscattering fraction at 520nm',          'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsf520',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_BACKSCAT_FRAC_525nm':            {'long_parameter_name':'PM1 aerosol light backscattering fraction at 525nm',          'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsf525',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_BACKSCAT_FRAC_530nm':            {'long_parameter_name':'PM1 aerosol light backscattering fraction at 530nm',          'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsf530',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_BACKSCAT_FRAC_532nm':            {'long_parameter_name':'PM1 aerosol light backscattering fraction at 532nm',          'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsf532',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_BACKSCAT_FRAC_550nm':            {'long_parameter_name':'PM1 aerosol light backscattering fraction at 550nm',          'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsf550',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_BACKSCAT_FRAC_635nm':            {'long_parameter_name':'PM1 aerosol light backscattering fraction at 635nm',          'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsf635',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_BACKSCAT_FRAC_700nm':            {'long_parameter_name':'PM1 aerosol light backscattering fraction at 700nm',          'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsf700',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_BACKSCAT_FRAC_850nm':            {'long_parameter_name':'PM1 aerosol light backscattering fraction at 850nm',          'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':0.5,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1lbsf850',      'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_ASY_525nm':                      {'long_parameter_name':'PM1 asymmetry factor at 525nm',                               'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'asymmetry factor',            'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':1.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1asy525',       'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
'PM1_ASY_550nm':                      {'long_parameter_name':'PM1 asymmetry factor at 550nm',                               'matrix':'pm1', 'standard_units':'unitless', 'axis_label':'asymmetry factor',            'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,   'extreme_upper_limit':1.0,     'extreme_upper_monthly_median':np.NaN,  'aqs_code':[],        'airbase_name':'',                                    'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':[],                                                         'ebas_matrix':'',    'ebas_priority_units':[''],    'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[],                         'chemical_formula':'',       'bsc_parameter_name':'pm1asy550',       'identifiers':{'IUPAC_name':'',                     'CAS_number':'',           'PubChem_CID':'',        'InChIKey':'',                            'WIGOS_name':'',                                                'WIGOS_number':''}},  
#PRECIP
'Al_PREC':       {'long_parameter_name':'aluminium in precipitaton',              'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':5.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':20000.0, 'extreme_upper_monthly_median':2000.0, 'aqs_code':['65342'], 'airbase_name':'aluminium (precip)',          'airbase_code':'00605', 'miteco_code':'', 'ebas_parameter_name':['aluminium'],            'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Al',     'bsc_parameter_name':'precal',     'identifiers':{'IUPAC_name':'aluminum',             'CAS_number':'7429-90-5',  'PubChem_CID':'5359268', 'InChIKey':'XAGFODPZIPBFFR-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'As_PREC':       {'long_parameter_name':'arsenic in precipitaton',                'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,  'aqs_code':['65339'], 'airbase_name':'Arsenic (precip)',            'airbase_code':'02018', 'miteco_code':'', 'ebas_parameter_name':['arsenic'],              'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'As',     'bsc_parameter_name':'precas',     'identifiers':{'IUPAC_name':'arsenic',              'CAS_number':'7440-38-2',  'PubChem_CID':'5359596', 'InChIKey':'RQNWIZPPADIBDY-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'BaP_PREC':      {'long_parameter_name':'benzo[a]pyrene in precipitaton',         'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':'',   'extreme_lower_limit':'',  'extreme_upper_limit':'',      'extreme_upper_monthly_median':'',     'aqs_code':[],        'airbase_name':'Benzo(a)pyrene (precip)',     'airbase_code':'00029', 'miteco_code':'', 'ebas_parameter_name':['benzo_a_pyrene'],       'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'C20H12', 'bsc_parameter_name':'precbap',    'identifiers':{'IUPAC_name':'benzo[a]pyrene',       'CAS_number':'50-32-8',    'PubChem_CID':'2336',    'InChIKey':'FMMWHPNWAFZXNH-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Ca++_PREC':     {'long_parameter_name':'calcium in precipitaton',                'matrix':'precip', 'standard_units':'mg l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,  'aqs_code':['65314'], 'airbase_name':'calcium (precip)',            'airbase_code':'00630', 'miteco_code':'', 'ebas_parameter_name':['calcium'],              'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Ca',     'bsc_parameter_name':'precca',     'identifiers':{'IUPAC_name':'calcium(2+)',          'CAS_number':'14127-61-8', 'PubChem_CID':'271',     'InChIKey':'BHPQYMZQTOCNFJ-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Cd_PREC':       {'long_parameter_name':'cadmium in precipitaton',                'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,  'aqs_code':['65332'], 'airbase_name':'Cadmium (precip)',            'airbase_code':'02014', 'miteco_code':'', 'ebas_parameter_name':['cadmium'],              'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Cd',     'bsc_parameter_name':'preccd',     'identifiers':{'IUPAC_name':'cadmium',              'CAS_number':'7440-43-9',  'PubChem_CID':'23973',   'InChIKey':'BDOSMKKIYDKNTQ-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Cl-_PREC':      {'long_parameter_name':'chloride in precipitaton',               'matrix':'precip', 'standard_units':'mg l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0,  'extreme_upper_monthly_median':200.0,  'aqs_code':['65316'], 'airbase_name':'chloride (precip)',           'airbase_code':'00632', 'miteco_code':'', 'ebas_parameter_name':['chloride'],             'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Cl',     'bsc_parameter_name':'preccl',     'identifiers':{'IUPAC_name':'chloride',             'CAS_number':'16887-00-6', 'PubChem_CID':'312',     'InChIKey':'VEXZGXHMUGYJMC-UHFFFAOYSA-M', 'WIGOS_name':'', 'WIGOS_number':''}},
'Co_PREC':       {'long_parameter_name':'cobalt in precipitaton',                 'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,  'aqs_code':['65113'], 'airbase_name':'Cobalt (precip)',             'airbase_code':'02064', 'miteco_code':'', 'ebas_parameter_name':['cobalt'],               'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Co',     'bsc_parameter_name':'preccobalt', 'identifiers':{'IUPAC_name':'cobalt',               'CAS_number':'7440-48-4',  'PubChem_CID':'104730',  'InChIKey':'GUTLYIVDDKVIGB-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Cr_PREC':       {'long_parameter_name':'chromium in precipitaton',               'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,  'aqs_code':['65112'], 'airbase_name':'Chromium (precip)',           'airbase_code':'02016', 'miteco_code':'', 'ebas_parameter_name':['chromium'],             'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Cr',     'bsc_parameter_name':'preccr',     'identifiers':{'IUPAC_name':'chromium',             'CAS_number':'7440-47-3',  'PubChem_CID':'23976',   'InChIKey':'VYZAMTAEIAYCRO-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Cu_PREC':       {'long_parameter_name':'copper in precipitaton',                 'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0,  'extreme_upper_monthly_median':200.0,  'aqs_code':['65333'], 'airbase_name':'Copper (precip)',             'airbase_code':'02073', 'miteco_code':'', 'ebas_parameter_name':['copper'],               'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Cu',     'bsc_parameter_name':'preccu',     'identifiers':{'IUPAC_name':'copper',               'CAS_number':'7440-50-8',  'PubChem_CID':'23978',   'InChIKey':'RYGMFSIKBFXOCR-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Fe_PREC':       {'long_parameter_name':'iron in precipitaton',                   'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':5.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':20000.0, 'extreme_upper_monthly_median':2000.0, 'aqs_code':['65334'], 'airbase_name':'Iron (precip)',               'airbase_code':'02065', 'miteco_code':'', 'ebas_parameter_name':['iron'],                 'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Fe',     'bsc_parameter_name':'precfe',     'identifiers':{'IUPAC_name':'iron',                 'CAS_number':'7439-89-6',  'PubChem_CID':'23925',   'InChIKey':'XEEYBQQBJWHFJM-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Hg_PREC':       {'long_parameter_name':'mercury in precipitaton',                'matrix':'precip', 'standard_units':'ng l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0,  'extreme_upper_monthly_median':200.0,  'aqs_code':['65329'], 'airbase_name':'Mercury (precip)',            'airbase_code':'02013', 'miteco_code':'', 'ebas_parameter_name':['mercury'],              'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Hg',     'bsc_parameter_name':'prechg',     'identifiers':{'IUPAC_name':'mercury',              'CAS_number':'7439-97-6',  'PubChem_CID':'23931',   'InChIKey':'QSHDDOUJBYECFT-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'K+_PREC':       {'long_parameter_name':'potassium in precipitaton',              'matrix':'precip', 'standard_units':'mg l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,  'aqs_code':['65312'], 'airbase_name':'potassium (precip)',          'airbase_code':'00658', 'miteco_code':'', 'ebas_parameter_name':['potassium'],            'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'K',      'bsc_parameter_name':'preck',      'identifiers':{'IUPAC_name':'potassium(1+)',        'CAS_number':'24203-36-9', 'PubChem_CID':'813',     'InChIKey':'NPYPAHLBTDXSSS-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Mg++_PREC':     {'long_parameter_name':'magnesium in precipitaton',              'matrix':'precip', 'standard_units':'mg l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,  'aqs_code':['65313'], 'airbase_name':'magnesium (precip)',          'airbase_code':'00660', 'miteco_code':'', 'ebas_parameter_name':['magnesium'],            'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Mg',     'bsc_parameter_name':'precmg',     'identifiers':{'IUPAC_name':'magnesium(2+)',        'CAS_number':'22537-22-0', 'PubChem_CID':'888',     'InChIKey':'JLVVSXFLKOJNIY-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Mn_PREC':       {'long_parameter_name':'manganese in precipitaton',              'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0,  'extreme_upper_monthly_median':200.0,  'aqs_code':['65335'], 'airbase_name':'Manganese (precip)',          'airbase_code':'02017', 'miteco_code':'', 'ebas_parameter_name':['manganese'],            'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Mn',     'bsc_parameter_name':'precmn',     'identifiers':{'IUPAC_name':'manganese',            'CAS_number':'7439-96-5',  'PubChem_CID':'23930',   'InChIKey':'PWHULOQIROXLJO-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'MSA_PREC':      {'long_parameter_name':'methanesulfonic acid in precipitaton',   'matrix':'precip', 'standard_units':'mg l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0,  'extreme_upper_monthly_median':200.0,  'aqs_code':[],        'airbase_name':'',                            'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['methanesulfonic_acid'], 'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'CH4O3S', 'bsc_parameter_name':'precmsa',    'identifiers':{'IUPAC_name':'methanesulfonic acid', 'CAS_number':'75-75-2',    'PubChem_CID':'6395',    'InChIKey':'AFVFQIVMOAPDHO-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Na+_PREC':      {'long_parameter_name':'sodium in precipitaton',                 'matrix':'precip', 'standard_units':'mg l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,  'aqs_code':['65311'], 'airbase_name':'sodium (precip)',             'airbase_code':'00669', 'miteco_code':'', 'ebas_parameter_name':['sodium'],               'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Na',     'bsc_parameter_name':'precna',     'identifiers':{'IUPAC_name':'sodium(1+)',           'CAS_number':'17341-25-2', 'PubChem_CID':'923',     'InChIKey':'FKNQFGJONOIPTF-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'NH4+_PREC':     {'long_parameter_name':'ammonium in precipitaton',               'matrix':'precip', 'standard_units':'mg l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,  'aqs_code':['65318'], 'airbase_name':'ammonium (precip)',           'airbase_code':'00664', 'miteco_code':'', 'ebas_parameter_name':['ammonium'],             'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'NH4',    'bsc_parameter_name':'precnh4',    'identifiers':{'IUPAC_name':'azanium',              'CAS_number':'14798-03-9', 'PubChem_CID':'223',     'InChIKey':'QGZKDVFQNNGYKY-UHFFFAOYSA-O', 'WIGOS_name':'', 'WIGOS_number':''}},
'NH4NO3_PREC':   {'long_parameter_name':'ammonium nitrate in precipitaton',       'matrix':'precip', 'standard_units':'mg l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,  'aqs_code':[],        'airbase_name':'',                            'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['ammonium_nitrate'],     'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'NH4NO3', 'bsc_parameter_name':'precnh4no3', 'identifiers':{'IUPAC_name':'azanium;nitrate',      'CAS_number':'6484-52-2',  'PubChem_CID':'22985',   'InChIKey':'DVARTQFDIMZBAA-UHFFFAOYSA-O', 'WIGOS_name':'', 'WIGOS_number':''}},
'Ni_PREC':       {'long_parameter_name':'nickel in precipitaton',                 'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0,  'extreme_upper_monthly_median':200.0,  'aqs_code':['65336'], 'airbase_name':'Nickel (precip)',             'airbase_code':'02015', 'miteco_code':'', 'ebas_parameter_name':['nickel'],               'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Ni',     'bsc_parameter_name':'precni',     'identifiers':{'IUPAC_name':'nickel',               'CAS_number':'7440-02-0',  'PubChem_CID':'935',     'InChIKey':'PXHVJJICTQNCMI-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'NO3-_PREC':     {'long_parameter_name':'nitrate in precipitaton',                'matrix':'precip', 'standard_units':'mg l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,  'aqs_code':['65321'], 'airbase_name':'nitrate (precip)',            'airbase_code':'00666', 'miteco_code':'', 'ebas_parameter_name':['nitrate'],              'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'NO3',    'bsc_parameter_name':'precno3',    'identifiers':{'IUPAC_name':'nitrate',              'CAS_number':'14797-55-8', 'PubChem_CID':'943',     'InChIKey':'NHNBFGGVMKEFGY-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Pb_PREC':       {'long_parameter_name':'lead in precipitaton',                   'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0,  'extreme_upper_monthly_median':200.0,  'aqs_code':['65337'], 'airbase_name':'Lead (precip)',               'airbase_code':'02012', 'miteco_code':'', 'ebas_parameter_name':['lead'],                 'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Pb',     'bsc_parameter_name':'precpb',     'identifiers':{'IUPAC_name':'lead',                 'CAS_number':'7439-92-1',  'PubChem_CID':'5352425', 'InChIKey':'WABPQHHGFIMREM-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Se_PREC':       {'long_parameter_name':'selenium in precipitaton',               'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':1000.0,  'extreme_upper_monthly_median':100.0,  'aqs_code':[],        'airbase_name':'',                            'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['selenium'],             'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Se',     'bsc_parameter_name':'precse',     'identifiers':{'IUPAC_name':'selenium',             'CAS_number':'7782-49-2',  'PubChem_CID':'6326970', 'InChIKey':'BUGBHKTXTAQXES-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'SO4--_PREC':    {'long_parameter_name':'sulphate in precipitaton',               'matrix':'precip', 'standard_units':'mg l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0,  'extreme_upper_monthly_median':200.0,  'aqs_code':['65322'], 'airbase_name':'sulphate (precip)',           'airbase_code':'00719', 'miteco_code':'', 'ebas_parameter_name':['sulphate_total'],       'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'SO4',    'bsc_parameter_name':'precso4',    'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'', 'WIGOS_number':''}},
'SO4--_NSS_PREC':{'long_parameter_name':'sulphate: non-sea salt in precipitaton', 'matrix':'precip', 'standard_units':'mg l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0,  'extreme_upper_monthly_median':200.0,  'aqs_code':[],        'airbase_name':'sulphate_corrected (precip)', 'airbase_code':'00720', 'miteco_code':'', 'ebas_parameter_name':['sulphate_corrected'],   'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'SO4',    'bsc_parameter_name':'precso4nss', 'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'', 'WIGOS_number':''}},
'SO4--_SS_PREC': {'long_parameter_name':'sulphate: sea salt in precipitaton',     'matrix':'precip', 'standard_units':'mg l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0,  'extreme_upper_monthly_median':200.0,  'aqs_code':[],        'airbase_name':'',                            'airbase_code':'',      'miteco_code':'', 'ebas_parameter_name':['sulphate_seasalt'],     'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'SO4',    'bsc_parameter_name':'precso4ss',  'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'', 'WIGOS_number':''}},
'V_PREC':        {'long_parameter_name':'vanadium in precipitaton',               'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0,  'extreme_upper_monthly_median':200.0,  'aqs_code':['65164'], 'airbase_name':'vanadium (precip)',           'airbase_code':'00728', 'miteco_code':'', 'ebas_parameter_name':['vanadium'],             'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'V',      'bsc_parameter_name':'precv',      'identifiers':{'IUPAC_name':'vanadium',             'CAS_number':'7440-62-2',  'PubChem_CID':'23990',   'InChIKey':'LEONUFNNVUYDNQ-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Zn_PREC':       {'long_parameter_name':'zinc in precipitaton',                   'matrix':'precip', 'standard_units':'ug l-1', 'axis_label':'concentration', 'minimum_resolution':1.0,  'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0,  'extreme_upper_monthly_median':200.0,  'aqs_code':['65338'], 'airbase_name':'Zinc (precip)',               'airbase_code':'02063', 'miteco_code':'', 'ebas_parameter_name':['zinc'],                 'ebas_matrix':'precip', 'ebas_priority_units':['/m3','/l'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Zn',     'bsc_parameter_name':'preczn',     'identifiers':{'IUPAC_name':'zinc',                 'CAS_number':'7440-66-6',  'PubChem_CID':'23994',   'InChIKey':'HCHKCACWOHOZIP-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
#WETDEP
'Al_WETDEP':       {'long_parameter_name':'aluminium wet deposition flux',              'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['aluminium'],            'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Al',     'bsc_parameter_name':'wetal',     'identifiers':{'IUPAC_name':'aluminum',             'CAS_number':'7429-90-5',  'PubChem_CID':'5359268', 'InChIKey':'XAGFODPZIPBFFR-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'As_WETDEP':       {'long_parameter_name':'arsenic wet deposition flux',                'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['arsenic'],              'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'As',     'bsc_parameter_name':'wetas',     'identifiers':{'IUPAC_name':'arsenic',              'CAS_number':'7440-38-2',  'PubChem_CID':'5359596', 'InChIKey':'RQNWIZPPADIBDY-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'BaP_PREC':        {'long_parameter_name':'benzo[a]pyrene wet deposition flux',         'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':'',  'extreme_lower_limit':'',  'extreme_upper_limit':'',     'extreme_upper_monthly_median':'',    'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['benzo_a_pyrene'],       'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'C20H12', 'bsc_parameter_name':'wetbap',    'identifiers':{'IUPAC_name':'benzo[a]pyrene',       'CAS_number':'50-32-8',    'PubChem_CID':'2336',    'InChIKey':'FMMWHPNWAFZXNH-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Ca++_WETDEP':     {'long_parameter_name':'calcium wet deposition flux',                'matrix':'wetdep', 'standard_units':'ug m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['calcium'],              'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Ca',     'bsc_parameter_name':'wetca',     'identifiers':{'IUPAC_name':'calcium(2+)',          'CAS_number':'14127-61-8', 'PubChem_CID':'271',     'InChIKey':'BHPQYMZQTOCNFJ-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Cd_WETDEP':       {'long_parameter_name':'cadmium wet deposition flux',                'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['cadmium'],              'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Cd',     'bsc_parameter_name':'wetcd',     'identifiers':{'IUPAC_name':'cadmium',              'CAS_number':'7440-43-9',  'PubChem_CID':'23973',   'InChIKey':'BDOSMKKIYDKNTQ-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Cl-_WETDEP':      {'long_parameter_name':'chloride wet deposition flux',               'matrix':'wetdep', 'standard_units':'ug m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['chloride'],             'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Cl',     'bsc_parameter_name':'wetcl',     'identifiers':{'IUPAC_name':'chloride',             'CAS_number':'16887-00-6', 'PubChem_CID':'312',     'InChIKey':'VEXZGXHMUGYJMC-UHFFFAOYSA-M', 'WIGOS_name':'', 'WIGOS_number':''}},
'Co_WETDEP':       {'long_parameter_name':'cobalt wet deposition flux',                 'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['cobalt'],               'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Co',     'bsc_parameter_name':'wetcobalt', 'identifiers':{'IUPAC_name':'cobalt',               'CAS_number':'7440-48-4',  'PubChem_CID':'104730',  'InChIKey':'GUTLYIVDDKVIGB-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Cr_WETDEP':       {'long_parameter_name':'chromium wet deposition flux',               'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['chromium'],             'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Cr',     'bsc_parameter_name':'wetcr',     'identifiers':{'IUPAC_name':'chromium',             'CAS_number':'7440-47-3',  'PubChem_CID':'23976',   'InChIKey':'VYZAMTAEIAYCRO-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Cu_WETDEP':       {'long_parameter_name':'copper wet deposition flux',                 'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['copper'],               'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Cu',     'bsc_parameter_name':'wetcu',     'identifiers':{'IUPAC_name':'copper',               'CAS_number':'7440-50-8',  'PubChem_CID':'23978',   'InChIKey':'RYGMFSIKBFXOCR-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Fe_WETDEP':       {'long_parameter_name':'iron wet deposition flux',                   'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['iron'],                 'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Fe',     'bsc_parameter_name':'wetfe',     'identifiers':{'IUPAC_name':'iron',                 'CAS_number':'7439-89-6',  'PubChem_CID':'23925',   'InChIKey':'XEEYBQQBJWHFJM-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Hg_WETDEP':       {'long_parameter_name':'mercury wet deposition flux',                'matrix':'wetdep', 'standard_units':'pg m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['mercury'],              'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Hg',     'bsc_parameter_name':'wethg',     'identifiers':{'IUPAC_name':'mercury',              'CAS_number':'7439-97-6',  'PubChem_CID':'23931',   'InChIKey':'QSHDDOUJBYECFT-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'K+_WETDEP':       {'long_parameter_name':'potassium wet deposition flux',              'matrix':'wetdep', 'standard_units':'ug m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['potassium'],            'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'K',      'bsc_parameter_name':'wetk',      'identifiers':{'IUPAC_name':'potassium(1+)',        'CAS_number':'24203-36-9', 'PubChem_CID':'813',     'InChIKey':'NPYPAHLBTDXSSS-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Mg++_WETDEP':     {'long_parameter_name':'magnesium wet deposition flux',              'matrix':'wetdep', 'standard_units':'ug m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['magnesium'],            'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Mg',     'bsc_parameter_name':'wetmg',     'identifiers':{'IUPAC_name':'magnesium(2+)',        'CAS_number':'22537-22-0', 'PubChem_CID':'888',     'InChIKey':'JLVVSXFLKOJNIY-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Mn_WETDEP':       {'long_parameter_name':'manganese wet deposition flux',              'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['manganese'],            'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Mn',     'bsc_parameter_name':'wetmn',     'identifiers':{'IUPAC_name':'manganese',            'CAS_number':'7439-96-5',  'PubChem_CID':'23930',   'InChIKey':'PWHULOQIROXLJO-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'MSA_WETDEP':      {'long_parameter_name':'methanesulfonic acid wet deposition flux',   'matrix':'wetdep', 'standard_units':'ug m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['methanesulfonic_acid'], 'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'CH4O3S', 'bsc_parameter_name':'wetmsa',    'identifiers':{'IUPAC_name':'methanesulfonic acid', 'CAS_number':'75-75-2',    'PubChem_CID':'6395',    'InChIKey':'AFVFQIVMOAPDHO-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Na+_WETDEP':      {'long_parameter_name':'sodium wet deposition flux',                 'matrix':'wetdep', 'standard_units':'ug m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['sodium'],               'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Na',     'bsc_parameter_name':'wetna',     'identifiers':{'IUPAC_name':'sodium(1+)',           'CAS_number':'17341-25-2', 'PubChem_CID':'923',     'InChIKey':'FKNQFGJONOIPTF-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'NH4+_WETDEP':     {'long_parameter_name':'ammonium wet deposition flux',               'matrix':'wetdep', 'standard_units':'ug m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['ammonium'],             'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'NH4',    'bsc_parameter_name':'wetnh4',    'identifiers':{'IUPAC_name':'azanium',              'CAS_number':'14798-03-9', 'PubChem_CID':'223',     'InChIKey':'QGZKDVFQNNGYKY-UHFFFAOYSA-O', 'WIGOS_name':'', 'WIGOS_number':''}},
'NH4NO3_WETDEP':   {'long_parameter_name':'ammonium nitrate wet deposition flux',       'matrix':'wetdep', 'standard_units':'ug m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['ammonium_nitrate'],     'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'NH4NO3', 'bsc_parameter_name':'wetnh4no3', 'identifiers':{'IUPAC_name':'azanium;nitrate',      'CAS_number':'6484-52-2',  'PubChem_CID':'22985',   'InChIKey':'DVARTQFDIMZBAA-UHFFFAOYSA-O', 'WIGOS_name':'', 'WIGOS_number':''}},
'Ni_WETDEP':       {'long_parameter_name':'nickel wet deposition flux',                 'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['nickel'],               'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Ni',     'bsc_parameter_name':'wetni',     'identifiers':{'IUPAC_name':'nickel',               'CAS_number':'7440-02-0',  'PubChem_CID':'935',     'InChIKey':'PXHVJJICTQNCMI-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'NO3-_WETDEP':     {'long_parameter_name':'nitrate wet deposition flux',                'matrix':'wetdep', 'standard_units':'ug m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['nitrate'],              'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'NO3',    'bsc_parameter_name':'wetno3',    'identifiers':{'IUPAC_name':'nitrate',              'CAS_number':'14797-55-8', 'PubChem_CID':'943',     'InChIKey':'NHNBFGGVMKEFGY-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Pb_WETDEP':       {'long_parameter_name':'lead wet deposition flux',                   'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['lead'],                 'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Pb',     'bsc_parameter_name':'wetpb',     'identifiers':{'IUPAC_name':'lead',                 'CAS_number':'7439-92-1',  'PubChem_CID':'5352425', 'InChIKey':'WABPQHHGFIMREM-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Se_WETDEP':       {'long_parameter_name':'selenium wet deposition flux',               'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['selenium'],             'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Se',     'bsc_parameter_name':'wetse',     'identifiers':{'IUPAC_name':'selenium',             'CAS_number':'7782-49-2',  'PubChem_CID':'6326970', 'InChIKey':'BUGBHKTXTAQXES-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'SO4--_WETDEP':    {'long_parameter_name':'sulphate wet deposition flux',               'matrix':'wetdep', 'standard_units':'ug m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['sulphate_total'],       'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'SO4',    'bsc_parameter_name':'wetso4',    'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'', 'WIGOS_number':''}},
'SO4--_NSS_WETDEP':{'long_parameter_name':'sulphate: non-sea salt wet deposition flux', 'matrix':'wetdep', 'standard_units':'ug m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['sulphate_corrected'],   'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'SO4',    'bsc_parameter_name':'wetso4nss', 'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'', 'WIGOS_number':''}},
'SO4--_SS_WETDEP': {'long_parameter_name':'sulphate: sea salt wet deposition flux',     'matrix':'wetdep', 'standard_units':'ug m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['sulphate_seasalt'],     'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'SO4',    'bsc_parameter_name':'wetso4ss',  'identifiers':{'IUPAC_name':'sulfate',              'CAS_number':'14808-79-8', 'PubChem_CID':'1117',    'InChIKey':'QAOWNCQODCNURD-UHFFFAOYSA-L', 'WIGOS_name':'', 'WIGOS_number':''}},
'V_WETDEP':        {'long_parameter_name':'vanadium wet deposition flux',               'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['vanadium'],             'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'V',      'bsc_parameter_name':'wetv',      'identifiers':{'IUPAC_name':'vanadium',             'CAS_number':'7440-62-2',  'PubChem_CID':'23990',   'InChIKey':'LEONUFNNVUYDNQ-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
'Zn_WETDEP':       {'long_parameter_name':'zinc wet deposition flux',                   'matrix':'wetdep', 'standard_units':'ng m-2 min-1', 'axis_label':'flux', 'minimum_resolution':1.0, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2000.0, 'extreme_upper_monthly_median':200.0, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['zinc'],                 'ebas_matrix':'wetdep', 'ebas_priority_units':['/m2'], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'Zn',     'bsc_parameter_name':'wetzn',     'identifiers':{'IUPAC_name':'zinc',                 'CAS_number':'7440-66-6',  'PubChem_CID':'23994',   'InChIKey':'HCHKCACWOHOZIP-UHFFFAOYSA-N', 'WIGOS_name':'', 'WIGOS_number':''}},
#METEO
'WND_DIR_10M':   {'long_parameter_name':'10m wind direction',                                                                                                               'matrix':'meteo', 'standard_units':'angular degrees', 'axis_label':'direction',    'minimum_resolution':15.0,   'extreme_lower_limit':1.0,     'extreme_upper_limit':360.0,    'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'',        'noaa_isd_parameter_name':'direction_angle',    'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'dir10',    'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Horizontal wind direction at specified distance from reference surface', 'WIGOS_number':'12005'}}, 
'WND_SPD_10M':   {'long_parameter_name':'10m wind speed',                                                                                                                   'matrix':'meteo', 'standard_units':'m s-1',           'axis_label':'speed',        'minimum_resolution':1.0,    'extreme_lower_limit':0.0,     'extreme_upper_limit':900.0,    'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'',        'noaa_isd_parameter_name':'speed',              'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'spd10',    'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Horizontal wind speed at specified distance from reference surface',     'WIGOS_number':'12006'}},  
'TEMP_2M':       {'long_parameter_name':'2m air temperature',                                                                                                               'matrix':'meteo', 'standard_units':'K',               'axis_label':'temperature',  'minimum_resolution':1.0,    'extreme_lower_limit':-658.85, 'extreme_upper_limit':891.15,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'TEMP_2M', 'noaa_isd_parameter_name':'air_temp',           'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'t2',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Air temperature (at specified distance from reference surface)',         'WIGOS_number':'224'}},  
'RH_2M':         {'long_parameter_name':'average relative humidity',                                                                                                        'matrix':'meteo', 'standard_units':'unitless',        'axis_label':'humidity (%)', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,     'extreme_upper_limit':1000.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'',        'noaa_isd_parameter_name':'avg_rh',             'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'rh2',      'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Relative humidity (with respect to water)',                              'WIGOS_number':'12249'}},  
'SST':           {'long_parameter_name':'sea surface temperature',                                                                                                          'matrix':'meteo', 'standard_units':'K',               'axis_label':'temperature',  'minimum_resolution':1.0,    'extreme_lower_limit':223.15,  'extreme_upper_limit':723.15,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'',        'noaa_isd_parameter_name':'sst',                'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'sst',      'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                                       'WIGOS_number':''}},  
'DEW_PT_2M':     {'long_parameter_name':'2m dew point',                                                                                                                     'matrix':'meteo', 'standard_units':'K',               'axis_label':'temperature',  'minimum_resolution':1.0,    'extreme_lower_limit':-708.85, 'extreme_upper_limit':641.15,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'',        'noaa_isd_parameter_name':'dpoint_temp',        'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'td2',      'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Dew-point temperature (at specified distance from reference surface)',   'WIGOS_number':'225'}},  
'PRES_2M':       {'long_parameter_name':'2m atmospheric pressure',                                                                                                          'matrix':'meteo', 'standard_units':'hPa',             'axis_label':'pressure',     'minimum_resolution':10.0,   'extreme_lower_limit':450.0,   'extreme_upper_limit':10900.0,  'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'',        'noaa_isd_parameter_name':'station_p',          'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'pshltr',   'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Atmospheric pressure',                                                   'WIGOS_number':'216'}},  
'SLP':           {'long_parameter_name':'sea level pressure',                                                                                                               'matrix':'meteo', 'standard_units':'hPa',             'axis_label':'pressure',     'minimum_resolution':10.0,   'extreme_lower_limit':8600.0,  'extreme_upper_limit':10900.0,  'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'',        'noaa_isd_parameter_name':'sea_level_pressure', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'slp',      'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                                       'WIGOS_number':''}},  
'PREC_ACCUM':    {'long_parameter_name':'liquid precipitation accumulation',                                                                                                'matrix':'meteo', 'standard_units':'mm',              'axis_label':'accumulation', 'minimum_resolution':1.0,    'extreme_lower_limit':0.0,     'extreme_upper_limit':9998.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'',        'noaa_isd_parameter_name':'prec_depth',         'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'acprec',   'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                                       'WIGOS_number':''}},  
'SNOW_ACCUM':    {'long_parameter_name':'snow accumulation',                                                                                                                'matrix':'meteo', 'standard_units':'cm',              'axis_label':'accumulation', 'minimum_resolution':1.0,    'extreme_lower_limit':0.0,     'extreme_upper_limit':1200.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'',        'noaa_isd_parameter_name':'snow_acc',           'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'acsnow',   'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                                       'WIGOS_number':''}},  
'SNOW_DEPTH':    {'long_parameter_name':'snow depth',                                                                                                                       'matrix':'meteo', 'standard_units':'cm',              'axis_label':'depth',        'minimum_resolution':1.0,    'extreme_lower_limit':0.0,     'extreme_upper_limit':1200.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'',        'noaa_isd_parameter_name':'si',                 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'si',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                                       'WIGOS_number':''}},  
'CEILING_HEIGHT':{'long_parameter_name':'the height above ground level of the lowest cloud or obscuring phenomena layer aloft with 5/8 or more summation total sky cover.', 'matrix':'meteo', 'standard_units':'m',               'axis_label':'height',       'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,     'extreme_upper_limit':22000.0,  'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'',        'noaa_isd_parameter_name':'ceiling',            'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'cldbot',   'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Cloud base height',                                                      'WIGOS_number':'531'}},  
'VIS_DIST':      {'long_parameter_name':'visibility distance',                                                                                                              'matrix':'meteo', 'standard_units':'m',               'axis_label':'distance',     'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,     'extreme_upper_limit':160000.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'',        'noaa_isd_parameter_name':'distance',           'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vdist',    'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Visibility',                                                             'WIGOS_number':'230'}},  
'CLOUD_CVG_FRAC':{'long_parameter_name':'cloud coverage fraction, taken as maximum cloud coverage across cloud layers',                                                     'matrix':'meteo', 'standard_units':'unitless',        'axis_label':'coverage (%)', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,     'extreme_upper_limit':100.0,    'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'',        'noaa_isd_parameter_name':'cloud_frac',         'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'cfracmax', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Cloud amount',                                                           'WIGOS_number':'179'}},  
#TOTAL AOD 
'AOD_380nm':           {'long_parameter_name':'aerosol optical depth at 380nm',                                    'matrix':'aod', 'standard_units':'unitless', 'axis_label':'aerosol optical depth',       'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['aerosol_optical_depth'], 'ebas_matrix':'aerosol', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_380nm'],                    'aeronet_matrix':['directsun'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'od380aero',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol optical depth, TSP',                  'WIGOS_number':'325'}},  
'AOD_440nm':           {'long_parameter_name':'aerosol optical depth at 440nm',                                    'matrix':'aod', 'standard_units':'unitless', 'axis_label':'aerosol optical depth',       'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['aerosol_optical_depth'], 'ebas_matrix':'aerosol', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_440nm'],                    'aeronet_matrix':['directsun'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'od440aero',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol optical depth, TSP',                  'WIGOS_number':'325'}},  
'AOD_500nm':           {'long_parameter_name':'aerosol optical depth at 500nm',                                    'matrix':'aod', 'standard_units':'unitless', 'axis_label':'aerosol optical depth',       'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['aerosol_optical_depth'], 'ebas_matrix':'aerosol', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Total_AOD_500nm[tau_a]'],       'aeronet_matrix':['oneill'],    'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'od500aero',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol optical depth, TSP',                  'WIGOS_number':'325'}},  
'AOD_500nm_COARSE':    {'long_parameter_name':'coarse mode aerosol optical depth at 500nm',                        'matrix':'aod', 'standard_units':'unitless', 'axis_label':'aerosol optical depth',       'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[],                        'ebas_matrix':'',        'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Coarse_Mode_AOD_500nm[tau_c]'], 'aeronet_matrix':['oneill'],    'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'od500aerocoarse', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                            'WIGOS_number':''}},  
'AOD_500nm_FINE':      {'long_parameter_name':'fine mode aerosol optical depth at 500nm',                          'matrix':'aod', 'standard_units':'unitless', 'axis_label':'aerosol optical depth',       'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[],                        'ebas_matrix':'',        'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Fine_Mode_AOD_500nm[tau_f]'],   'aeronet_matrix':['oneill'],    'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'od500aerofine',   'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                            'WIGOS_number':''}},  
'FINE_MODE_FRAC_500nm':{'long_parameter_name':'fine mode aerosol optical depth fraction at 500nm',                 'matrix':'aod', 'standard_units':'unitless', 'axis_label':'fraction',                    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':1.0,  'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[],                        'ebas_matrix':'',        'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['FineModeFraction_500nm[eta]'],  'aeronet_matrix':['oneill'],    'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'fm500frac',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                            'WIGOS_number':''}},  
'AOD_550nm':           {'long_parameter_name':'aerosol optical depth at 550nm',                                    'matrix':'aod', 'standard_units':'unitless', 'axis_label':'aerosol optical depth',       'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['aerosol_optical_depth'], 'ebas_matrix':'aerosol', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'',                               'aeronet_matrix':'',            'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'od550aero',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol optical depth, TSP',                  'WIGOS_number':'325'}},  
'AOD_675nm':           {'long_parameter_name':'aerosol optical depth at 675nm',                                    'matrix':'aod', 'standard_units':'unitless', 'axis_label':'aerosol optical depth',       'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['aerosol_optical_depth'], 'ebas_matrix':'aerosol', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_675nm'],                    'aeronet_matrix':['directsun'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'od675aero',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol optical depth, TSP',                  'WIGOS_number':'325'}},  
'AOD_870nm':           {'long_parameter_name':'aerosol optical depth at 870nm',                                    'matrix':'aod', 'standard_units':'unitless', 'axis_label':'aerosol optical depth',       'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['aerosol_optical_depth'], 'ebas_matrix':'aerosol', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_870nm'],                    'aeronet_matrix':['directsun'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'od870aero',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol optical depth, TSP',                  'WIGOS_number':'325'}},  
'AOD_1020nm':          {'long_parameter_name':'aerosol optical depth at 1020nm',                                   'matrix':'aod', 'standard_units':'unitless', 'axis_label':'aerosol optical depth',       'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':['aerosol_optical_depth'], 'ebas_matrix':'aerosol', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_1020nm'],                   'aeronet_matrix':['directsun'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'od1020aero',      'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol optical depth, TSP',                  'WIGOS_number':'325'}},  
'AE_440-870nm':        {'long_parameter_name':'angstrom exponent between 440 and 870nm for aerosol optical depth', 'matrix':'aod', 'standard_units':'unitless', 'axis_label':'angstrom exponent 440-870nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':4.0,  'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[],                        'ebas_matrix':'',        'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['440-870_Angstrom_Exponent'],    'aeronet_matrix':['directsun'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'ae440-870aero',   'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Ångström exponent for aerosol optical depth', 'WIGOS_number':'12150'}},  
#EXTINCTION AOD
'EXTAOD_440nm':           {'long_parameter_name':'extinction aerosol optical depth at 440nm',                                      'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'extinction AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_Extinction-Total[440nm]'],                  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extod440aero',        'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol optical depth, TSP',                      'WIGOS_number':'325'}},  
'EXTAOD_440nm_COARSE':    {'long_parameter_name':'extinction coarse mode aerosol optical depth at 440nm',                          'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'extinction AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_Extinction-Coarse[440nm]'],                 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extod440aerocoarse',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                'WIGOS_number':''}},  
'EXTAOD_440nm_FINE':      {'long_parameter_name':'extinction fine mode aerosol optical depth at 440nm',                            'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'extinction AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_Extinction-Fine[440nm]'],                   'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extod440aerofine',    'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                'WIGOS_number':''}},  
'EXTAOD_550nm':           {'long_parameter_name':'extinction aerosol optical depth at 550nm',                                      'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'extinction AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'',                                               'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extod550aero',        'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol optical depth, TSP',                      'WIGOS_number':'325'}},  
'EXTAOD_675nm':           {'long_parameter_name':'extinction aerosol optical depth at 675nm',                                      'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'extinction AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_Extinction-Total[675nm]'],                  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extod675aero',        'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol optical depth, TSP',                      'WIGOS_number':'325'}},  
'EXTAOD_675nm_COARSE':    {'long_parameter_name':'extinction coarse mode aerosol optical depth at 675nm',                          'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'extinction AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_Extinction-Coarse[675nm]'],                 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extod675aerocoarse',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                'WIGOS_number':''}},  
'EXTAOD_675nm_FINE':      {'long_parameter_name':'extinction fine mode aerosol optical depth at 675nm',                            'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'extinction AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_Extinction-Fine[675nm]'],                   'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extod675aerofine',    'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                'WIGOS_number':''}},  
'EXTAOD_870nm':           {'long_parameter_name':'extinction aerosol optical depth at 870nm',                                      'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'extinction AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_Extinction-Total[870nm]'],                  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extod870aero',        'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol optical depth, TSP',                      'WIGOS_number':'325'}},  
'EXTAOD_870nm_COARSE':    {'long_parameter_name':'extinction coarse mode aerosol optical depth at 870nm',                          'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'extinction AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_Extinction-Coarse[870nm]'],                 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extod870aerocoarse',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                'WIGOS_number':''}},  
'EXTAOD_870nm_FINE':      {'long_parameter_name':'extinction fine mode aerosol optical depth at 870nm',                            'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'extinction AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_Extinction-Fine[870nm]'],                   'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extod870aerofine',    'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                'WIGOS_number':''}},  
'EXTAOD_1020nm':          {'long_parameter_name':'extinction aerosol optical depth at 1020nm',                                     'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'extinction AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_Extinction-Total[1020nm]'],                 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extod1020aero',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol optical depth, TSP',                      'WIGOS_number':'325'}},  
'EXTAOD_1020nm_COARSE':   {'long_parameter_name':'extinction coarse mode aerosol optical depth at 1020nm',                         'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'extinction AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_Extinction-Coarse[1020nm]'],                'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extod1020aerocoarse', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                'WIGOS_number':''}},  
'EXTAOD_1020nm_FINE':     {'long_parameter_name':'extinction fine mode aerosol optical depth at 1020nm',                           'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'extinction AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['AOD_Extinction-Fine[1020nm]'],                  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extod1020aerofine',   'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                'WIGOS_number':''}},  
'EXTAE_440-675nm':        {'long_parameter_name':'angstrom exponent between 440nm and 675nm for extinction aerosol optical depth', 'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'angstrom exponent 440-675nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':np.NaN, 'extreme_upper_limit':np.NaN,  'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'',                                               'aeronet_matrix':'',             'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extae440-675aero',    'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Ångström exponent for aerosol optical depth',     'WIGOS_number':'12150'}},  
'EXTAE_440-870nm':        {'long_parameter_name':'angstrom exponent between 440nm and 870nm for extinction aerosol optical depth', 'matrix':'extaod', 'standard_units':'unitless', 'axis_label':'angstrom exponent 440-870nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':4.0,  'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Extinction_Angstrom_Exponent_440-870nm-Total'], 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'extae440-870aero',    'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Ångström exponent for aerosol optical depth',     'WIGOS_number':'12150'}},  
#ABSORPTION AOD 
'ABSAOD_440nm':   {'long_parameter_name':'absorption aerosol optical depth at 440nm',                                       'matrix':'absaod', 'standard_units':'unitless', 'axis_label':'absorption AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,  'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Absorption_AOD[440nm]'],                  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'absod440aero',     'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol Absorption Optical Depth',                       'WIGOS_number':'314'}},  
'ABSAOD_550nm':   {'long_parameter_name':'absorption aerosol optical depth at 550nm',                                       'matrix':'absaod', 'standard_units':'unitless', 'axis_label':'absorption AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,  'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'',                                         'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'absod550aero',     'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol Absorption Optical Depth',                       'WIGOS_number':'314'}},  
'ABSAOD_675nm':   {'long_parameter_name':'absorption aerosol optical depth at 675nm',                                       'matrix':'absaod', 'standard_units':'unitless', 'axis_label':'absorption AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,  'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Absorption_AOD[675nm]'],                  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'absod675aero',     'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol Absorption Optical Depth',                       'WIGOS_number':'314'}},  
'ABSAOD_870nm':   {'long_parameter_name':'absorption aerosol optical depth at 870nm',                                       'matrix':'absaod', 'standard_units':'unitless', 'axis_label':'absorption AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,  'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Absorption_AOD[870nm]'],                  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'absod870aero',     'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol Absorption Optical Depth',                       'WIGOS_number':'314'}},  
'ABSAOD_1020nm':  {'long_parameter_name':'absorption aerosol optical depth at 1020nm',                                      'matrix':'absaod', 'standard_units':'unitless', 'axis_label':'absorption AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,  'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Absorption_AOD[1020nm]'],                 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'absod1020aero',    'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Aerosol Absorption Optical Depth',                       'WIGOS_number':'314'}},  
'ABSAE_440-675nm':{'long_parameter_name':'angstrom exponent between 440nm and 675nm for absorption aerosol optical depth',  'matrix':'absaod', 'standard_units':'unitless', 'axis_label':'angstrom exponent 440-675nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':np.NaN,  'extreme_upper_limit':np.NaN,  'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'',                                         'aeronet_matrix':'',             'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'absae440-675aero', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Ångström exponent for aerosol absorption optical depth', 'WIGOS_number':'12151'}},  
'ABSAE_440-870nm':{'long_parameter_name':'angstrom exponent between 440nm and 870nm for absorption aerosol optical depth',  'matrix':'absaod', 'standard_units':'unitless', 'axis_label':'angstrom exponent 440-870nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0,  'extreme_upper_limit':4.0,  'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Absorption_Angstrom_Exponent_440-870nm'], 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'absae440-870aero', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Ångström exponent for aerosol absorption optical depth', 'WIGOS_number':'12151'}},  
'ABSAE_675-870nm':{'long_parameter_name':'angstrom exponent between 675nm and 870nm for absorption aerosol optical depth',  'matrix':'absaod', 'standard_units':'unitless', 'axis_label':'angstrom exponent 675-870nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':np.NaN, 'extreme_upper_limit':np.NaN,  'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'',                                         'aeronet_matrix':'',             'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'absae675-870aero', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Ångström exponent for aerosol absorption optical depth', 'WIGOS_number':'12151'}},  
'ABSWDAE':        {'long_parameter_name':'wavelength dependence of the absorption aerosol optical depth angstrom exponent', 'matrix':'absaod', 'standard_units':'unitless', 'axis_label':'wavelength dependence',       'minimum_resolution':np.NaN, 'extreme_lower_limit':np.NaN, 'extreme_upper_limit':np.NaN,  'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'',                                         'aeronet_matrix':'',             'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'abswdaeaero',      'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                                       'WIGOS_number':''}},  
#ABSORPTION AOD - BLACK CARBON
'BC_ABSAOD_440nm':   {'long_parameter_name':'black carbon absorption aerosol optical depth at 440nm',                                      'matrix':'absaod-bc', 'standard_units':'unitless', 'axis_label':'absorption AOD',              'minimum_resolution':np.NaN, 'extreme_lower_limit':np.NaN,  'extreme_upper_limit':np.NaN, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':[''], 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'absod440bc',     'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'BC_ABSAE_440-870nm':{'long_parameter_name':'angstrom exponent between 440nm and 870nm for black carbon absorption aerosol optical depth', 'matrix':'absaod-bc', 'standard_units':'unitless', 'axis_label':'angstrom exponent 440-870nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':np.NaN,  'extreme_upper_limit':np.NaN,  'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':[''], 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'absae440-870bc', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
#ABSORPTION AOD - BROWN CARBON
'BrC_ABSAOD_440nm':{'long_parameter_name':'brown carbon absorption aerosol optical depth at 440nm', 'matrix':'absaod-brc', 'standard_units':'unitless', 'axis_label':'absorption AOD', 'minimum_resolution':np.NaN, 'extreme_lower_limit':np.NaN, 'extreme_upper_limit':np.NaN, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'absod440brc', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
#SCATTERING AOD 
'SCATAE_440-675nm':{'long_parameter_name':'angstrom exponent between 440nm and 675nm for scattering aerosol optical depth', 'matrix':'scataod', 'standard_units':'unitless', 'axis_label':'angstrom exponent 440-675nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':np.NaN, 'extreme_upper_limit':np.NaN, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'', 'aeronet_matrix':'', 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'lsae440-675aero', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
#SSA
'SSA_440nm':      {'long_parameter_name':'single scattering albedo at 440nm',                                      'matrix':'ssa', 'standard_units':'unitless', 'axis_label':'single scattering albedo',    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':1.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Single_Scattering_Albedo[440nm]'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'sca440aero',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle single scattering albedo',           'WIGOS_number':'12146'}},  
'SSA_550nm':      {'long_parameter_name':'single scattering albedo at 550nm',                                      'matrix':'ssa', 'standard_units':'unitless', 'axis_label':'single scattering albedo',    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':1.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'',                                   'aeronet_matrix':'',             'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'sca550aero',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle single scattering albedo',           'WIGOS_number':'12146'}},  
'SSA_675nm':      {'long_parameter_name':'single scattering albedo at 675nm',                                      'matrix':'ssa', 'standard_units':'unitless', 'axis_label':'single scattering albedo',    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':1.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Single_Scattering_Albedo[675nm]'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'sca675aero',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle single scattering albedo',           'WIGOS_number':'12146'}},  
'SSA_870nm':      {'long_parameter_name':'single scattering albedo at 870nm',                                      'matrix':'ssa', 'standard_units':'unitless', 'axis_label':'single scattering albedo',    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':1.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Single_Scattering_Albedo[870nm]'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'sca870aero',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle single scattering albedo',           'WIGOS_number':'12146'}},  
'SSA_1020nm':     {'long_parameter_name':'single scattering albedo at 1020nm',                                     'matrix':'ssa', 'standard_units':'unitless', 'axis_label':'single scattering albedo',    'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':1.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Single_Scattering_Albedo[1020nm]'], 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'sca1020aero',      'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle single scattering albedo',           'WIGOS_number':'12146'}},  
'SSAAE_440-675nm':{'long_parameter_name':'angstrom exponent between 440nm and 675nm for single scattering albedo', 'matrix':'ssa', 'standard_units':'unitless', 'axis_label':'angstrom exponent 440-675nm', 'minimum_resolution':np.NaN, 'extreme_lower_limit':np.NaN, 'extreme_upper_limit':np.NaN, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':'',                                   'aeronet_matrix':'',             'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'scaae440-675aero', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'',                                            'WIGOS_number':''}},  
#ASYMMETRY/SPHERICITY FACTORS
'ASY_440nm':        {'long_parameter_name':'asymmetry factor at 440nm',              'matrix':'asy', 'standard_units':'unitless', 'axis_label':'asymmetry factor',  'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Asymmetry_Factor-Total[440nm]'],   'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'asy440aero',        'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'ASY_440nm_COARSE': {'long_parameter_name':'coarse mode asymmetry factor at 440nm',  'matrix':'asy', 'standard_units':'unitless', 'axis_label':'asymmetry factor',  'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Asymmetry_Factor-Coarse[440nm]'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'asy440aerocoarse',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'ASY_440nm_FINE':   {'long_parameter_name':'fine mode asymmetry factor at 440nm',    'matrix':'asy', 'standard_units':'unitless', 'axis_label':'asymmetry factor',  'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Asymmetry_Factor-Fine[440nm]'],    'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'asy440aerofine',    'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'ASY_675nm':        {'long_parameter_name':'asymmetry factor at 675nm',              'matrix':'asy', 'standard_units':'unitless', 'axis_label':'asymmetry factor',  'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Asymmetry_Factor-Total[675nm]'],   'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'asy675aero',        'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'ASY_675nm_COARSE': {'long_parameter_name':'coarse mode asymmetry factor at 675nm',  'matrix':'asy', 'standard_units':'unitless', 'axis_label':'asymmetry factor',  'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Asymmetry_Factor-Coarse[675nm]'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'asy675aerocoarse',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'ASY_675nm_FINE':   {'long_parameter_name':'fine mode asymmetry factor at 675nm',    'matrix':'asy', 'standard_units':'unitless', 'axis_label':'asymmetry factor',  'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Asymmetry_Factor-Fine[675nm]'],    'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'asy675aerofine',    'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'ASY_870nm':        {'long_parameter_name':'asymmetry factor at 870nm',              'matrix':'asy', 'standard_units':'unitless', 'axis_label':'asymmetry factor',  'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Asymmetry_Factor-Total[870nm]'],   'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'asy870aero',        'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'ASY_870nm_COARSE': {'long_parameter_name':'coarse mode asymmetry factor at 870nm',  'matrix':'asy', 'standard_units':'unitless', 'axis_label':'asymmetry factor',  'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Asymmetry_Factor-Coarse[870nm]'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'asy870aerocoarse',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'ASY_870nm_FINE':   {'long_parameter_name':'fine mode asymmetry factor at 870nm',    'matrix':'asy', 'standard_units':'unitless', 'axis_label':'asymmetry factor',  'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Asymmetry_Factor-Fine[870nm]'],    'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'asy870aerofine',    'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'ASY_1020nm':       {'long_parameter_name':'asymmetry factor at 1020nm',             'matrix':'asy', 'standard_units':'unitless', 'axis_label':'asymmetry factor',  'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Asymmetry_Factor-Total[1020nm]'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'asy1020aero',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'ASY_1020nm_COARSE':{'long_parameter_name':'coarse mode asymmetry factor at 1020nm', 'matrix':'asy', 'standard_units':'unitless', 'axis_label':'asymmetry factor',  'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Asymmetry_Factor-Coarse[1020nm]'], 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'asy1020aerocoarse', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'ASY_1020nm_FINE':  {'long_parameter_name':'fine mode asymmetry factor at 1020nm',   'matrix':'asy', 'standard_units':'unitless', 'axis_label':'asymmetry factor',  'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0,   'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Asymmetry_Factor-Fine[1020nm]'],   'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'asy1020aerofine',   'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'SPH':              {'long_parameter_name':'sphericity factor',                      'matrix':'asy', 'standard_units':'unitless', 'axis_label':'sphericity factor', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':100.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Sphericity_Factor(%)'],            'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'sphaero',           'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
#REFRACTIVE INDICES
'RIN_REAL_440nm': {'long_parameter_name':'real part of the refractive index at 440nm',       'matrix':'rin', 'standard_units':'unitless', 'axis_label':'refractive index', 'minimum_resolution':np.NaN, 'extreme_lower_limit':1.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Refractive_Index-Real_Part[440nm]'],       'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'rinreal440',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'RIN_REAL_675nm': {'long_parameter_name':'real part of the refractive index at 675nm',       'matrix':'rin', 'standard_units':'unitless', 'axis_label':'refractive index', 'minimum_resolution':np.NaN, 'extreme_lower_limit':1.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Refractive_Index-Real_Part[675nm]'],       'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'rinreal675',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'RIN_REAL_870nm': {'long_parameter_name':'real part of the refractive index at 870nm',       'matrix':'rin', 'standard_units':'unitless', 'axis_label':'refractive index', 'minimum_resolution':np.NaN, 'extreme_lower_limit':1.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Refractive_Index-Real_Part[870nm]'],       'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'rinreal870',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'RIN_REAL_1020nm':{'long_parameter_name':'real part of the refractive index at 1020nm',      'matrix':'rin', 'standard_units':'unitless', 'axis_label':'refractive index', 'minimum_resolution':np.NaN, 'extreme_lower_limit':1.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Refractive_Index-Real_Part[1020nm]'],      'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'rinreal1020', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'RIN_IMAG_440nm': {'long_parameter_name':'imaginary part of the refractive index at 440nm',  'matrix':'rin', 'standard_units':'unitless', 'axis_label':'refractive index', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':0.1, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Refractive_Index-Imaginary_Part[440nm]'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'rinimag440',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'RIN_IMAG_675nm': {'long_parameter_name':'imaginary part of the refractive index at 675nm',  'matrix':'rin', 'standard_units':'unitless', 'axis_label':'refractive index', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':0.1, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Refractive_Index-Imaginary_Part[675nm]'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'rinimag675',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'RIN_IMAG_870nm': {'long_parameter_name':'imaginary part of the refractive index at 870nm',  'matrix':'rin', 'standard_units':'unitless', 'axis_label':'refractive index', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':0.1, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Refractive_Index-Imaginary_Part[870nm]'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'rinimag870',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'RIN_IMAG_1020nm':{'long_parameter_name':'imaginary part of the refractive index at 1020nm', 'matrix':'rin', 'standard_units':'unitless', 'axis_label':'refractive index', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':0.1, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['Refractive_Index-Imaginary_Part[1020nm]'], 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'rinimag1020', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
#TOTAL VOLUME CONCENTRATION
'VCONC':       {'long_parameter_name':'normalised total volume concentration (dV(r)/dln(r))',             'matrix':'vconc', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['VolC-T'], 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaero',       'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
'VCONC_COARSE':{'long_parameter_name':'normalised total coarse mode volume concentration (dV(r)/dln(r))', 'matrix':'vconc', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['VolC-C'], 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerocoarse', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}}, 
'VCONC_FINE':  {'long_parameter_name':'normalised total fine mode volume concentration (dV(r)/dln(r))',   'matrix':'vconc', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':20.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['VolC-F'], 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerofine',   'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'', 'WIGOS_number':''}},  
#SIZE DISTRIBUTION
'VCONC_BIN1': {'long_parameter_name':'normalised volume concentration at 0.05um (dV(r)/dln(r))',      'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['0.050000'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin1',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN2': {'long_parameter_name':'normalised volume concentration at 0.065604um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['0.065604'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin2',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN3': {'long_parameter_name':'normalised volume concentration at 0.086077um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['0.086077'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin3',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN4': {'long_parameter_name':'normalised volume concentration at 0.086077um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['0.112939'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin4',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN5': {'long_parameter_name':'normalised volume concentration at 0.148184um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['0.148184'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin5',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN6': {'long_parameter_name':'normalised volume concentration at 0.194429um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['0.194429'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin6',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN7': {'long_parameter_name':'normalised volume concentration at 0.255105um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['0.255105'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin7',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN8': {'long_parameter_name':'normalised volume concentration at 0.334716um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['0.334716'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin8',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN9': {'long_parameter_name':'normalised volume concentration at 0.439173um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['0.439173'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin9',  'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN10':{'long_parameter_name':'normalised volume concentration at 0.576227um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['0.576227'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin10', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN11':{'long_parameter_name':'normalised volume concentration at 0.756052um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['0.756052'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin11', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN12':{'long_parameter_name':'normalised volume concentration at 0.991996um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['0.991996'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin12', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN13':{'long_parameter_name':'normalised volume concentration at 1.301571um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['1.301571'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin13', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN14':{'long_parameter_name':'normalised volume concentration at 1.707757um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['1.707757'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin14', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN15':{'long_parameter_name':'normalised volume concentration at 2.240702um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['2.240702'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin15', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN16':{'long_parameter_name':'normalised volume concentration at 2.939966um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['2.939966'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin16', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN17':{'long_parameter_name':'normalised volume concentration at 3.857452um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['3.857452'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin17', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN18':{'long_parameter_name':'normalised volume concentration at 5.061260um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['5.061260'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin18', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN19':{'long_parameter_name':'normalised volume concentration at 6.640745um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['6.640745'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin19', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN20':{'long_parameter_name':'normalised volume concentration at 8.713145um (dV(r)/dln(r))',  'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['8.713145'],  'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin20', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN21':{'long_parameter_name':'normalised volume concentration at 11.432287um (dV(r)/dln(r))', 'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['11.432287'], 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin21', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}},  
'VCONC_BIN22':{'long_parameter_name':'normalised volume concentration at 15.00um (dV(r)/dln(r))',     'matrix':'size', 'standard_units':'um3 um-2', 'axis_label':'volume concentration', 'minimum_resolution':np.NaN, 'extreme_lower_limit':0.0, 'extreme_upper_limit':2.0, 'extreme_upper_monthly_median':np.NaN, 'aqs_code':[], 'airbase_name':'', 'airbase_code':'', 'miteco_code':'', 'ebas_parameter_name':[], 'ebas_matrix':'', 'ebas_priority_units':[''], 'naps_code':[], 'castnet_parameter_name':'', 'noaa_isd_parameter_name':'', 'aeronet_parameter_name':['15.000000'], 'aeronet_matrix':['almucantar'], 'japan_nies_code':[], 'japan_nies_name':[], 'uk_air_name':[], 'chemical_formula':'', 'bsc_parameter_name':'vconcaerobin22', 'identifiers':{'IUPAC_name':'', 'CAS_number':'', 'PubChem_CID':'', 'InChIKey':'', 'WIGOS_name':'Particle volume size distribution', 'WIGOS_number':'12153'}}  
}

###--------------------------------------------------------------------------------------------------###
###NETWORK STANDARDISATIONS
###--------------------------------------------------------------------------------------------------###

#Define dictionary with details pertaining to standardised networks.
#information regarding the hierarchy of preferred data levels is detailed, by data network, and returns the ranking of a specific data level for data of a specific network.
#Data with a data level of a high defined ranking, is preferentially selected over data with a low data level ranking, when there is data duplication.       

standard_networks = {
'AERONET_v3_lev1.5':{'file_storage_type':'all_per_day',        'external_metadata_files':['AERONET_META.csv'],                                          'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'2.0':3, '1.5':2, '1.0':1}},    
'AERONET_v3_lev2.0':{'file_storage_type':'all_per_day',        'external_metadata_files':['AERONET_META.csv'],                                          'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'2.0':3, '1.5':2, '1.0':1}},    
'BJMEMC':           {'file_storage_type':'all_per_day',        'external_metadata_files':['BJMEMC_META_YYYYMMDD.csv'],                                  'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},  
'CANADA_NAPS':      {'file_storage_type':'all_per_year',       'external_metadata_files':['CANADA_NAPS_META.csv'],                                      'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},
'CAPMoN':           {'file_storage_type':'all_per_year',       'external_metadata_files':[],                                                            'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'3.0':2,'2.0':1}},
'CHILE_SINCA':      {'file_storage_type':'site_per_multiyear', 'external_metadata_files':['CHILE_SINCA_META.txt'],                                      'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},
'CNEMC':            {'file_storage_type':'all_per_day',        'external_metadata_files':['CNEMC_META_YYYY.csv'],                                       'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},  
'EANET':            {'file_storage_type':'site_per_year',      'external_metadata_files':[],                                                            'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},
'EBAS':             {'file_storage_type':'site_per_multiyear', 'external_metadata_files':['EBAS_META_MANUAL.csv'],                                      'metadata_data_join_variable_name':'reference', 'data_level_hierarchy':{'2.0':4, '1.5':3, '1.0':2, '0.0':1}},
'EEA_AIRBASE':      {'file_storage_type':'site_per_year',      'external_metadata_files':[],                                                            'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},
'EEA_AQ_eReporting':{'file_storage_type':'site_per_year',      'external_metadata_files':['EEA_AQ_eReporting_META.csv','EEA_AQ_eReporting_AIDE_D.csv'], 'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'E1a':2,'E2a':1}},
'JAPAN_NIES':       {'file_storage_type':'all_per_year',       'external_metadata_files':['JAPAN_NIES_META_YYYY.txt'],                                  'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},
'MEXICO_CDMX':      {'file_storage_type':'all_per_year',       'external_metadata_files':['MEXICO_CDMX_META.csv'],                                      'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},  
'MITECO':           {'file_storage_type':'all_per_year',       'external_metadata_files':[],                                                            'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'official':2, 'unofficial':1}}, 
'NOAA_ISD':         {'file_storage_type':'site_per_year',      'external_metadata_files':['isd-history.csv'],                                           'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'variable':1}}, 
'SEARCH':           {'file_storage_type':'all_per_year',       'external_metadata_files':[],                                                            'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},
'UK_AIR':           {'file_storage_type':'all_per_year',       'external_metadata_files':['UK_AIR_META.csv'],                                           'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},    
'US_EPA_AirNow_DOS':{'file_storage_type':'site_per_year',      'external_metadata_files':['US_EPA_AirNow_DOS_META.csv'],                                'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},
'US_EPA_AQS':       {'file_storage_type':'all_per_year',       'external_metadata_files':['aqs_monitors.csv','aqs_sites.csv'],                          'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},
'US_EPA_CASTNET':   {'file_storage_type':'all_per_year',       'external_metadata_files':['CASTNET_META.csv'],                                          'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'variable':1}},
'US_NADP_AMNet':    {'file_storage_type':'all',                'external_metadata_files':['US_NADP_AMNet_META.csv'],                                    'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},    
'US_NADP_AMoN':     {'file_storage_type':'all',                'external_metadata_files':['US_NADP_AMoN_META.csv'],                                     'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},    
'WMO_WDCGG':        {'file_storage_type':'site_per_multiyear', 'external_metadata_files':[''],                                                          'metadata_data_join_variable_name':'',          'data_level_hierarchy':{'none':1}},
'GHOST':            {},
'GHOST-PUBLIC':     {}
}

###--------------------------------------------------------------------------------------------------###
###TEMPORAL RESOLUTION STANDARDISATIONS
###--------------------------------------------------------------------------------------------------###

#Define dictionary multiple variables associated with standard temporal resolution output codes 

standard_temporal_resolutions = {
'H':  {'temporal_resolution_datetime_frequency':'H',  'temporal_resolution_path':'hourly',                'descriptive_temporal_resolution':'hours'},
'3H': {'temporal_resolution_datetime_frequency':'3H', 'temporal_resolution_path':'3hourly',               'descriptive_temporal_resolution':'hours'},
'6H': {'temporal_resolution_datetime_frequency':'6H', 'temporal_resolution_path':'6hourly',               'descriptive_temporal_resolution':'hours'},
'D':  {'temporal_resolution_datetime_frequency':'D',  'temporal_resolution_path':'daily',                 'descriptive_temporal_resolution':'days'},
'M':  {'temporal_resolution_datetime_frequency':'MS', 'temporal_resolution_path':'monthly',               'descriptive_temporal_resolution':'months'},
'HI': {'temporal_resolution_datetime_frequency':'H',  'temporal_resolution_path':'hourly_instantaneous',  'descriptive_temporal_resolution':'hours'},
'3HI':{'temporal_resolution_datetime_frequency':'3H', 'temporal_resolution_path':'3hourly_instantaneous', 'descriptive_temporal_resolution':'hours'},
'6HI':{'temporal_resolution_datetime_frequency':'6H', 'temporal_resolution_path':'6hourly_instantaneous', 'descriptive_temporal_resolution':'hours'}
}

###--------------------------------------------------------------------------------------------------###
###DATA VARIABLE STANDARDISATIONS
###--------------------------------------------------------------------------------------------------###

#Define dictionary with details pertaining to standardised data variables.
#These data variables include time, the measured parameter variable, as well as variables which inform of the quality of each separate measurement.

def get_standard_data(parameter_details):

    standard_data = {
    #TIME
    'time':                                             {'value':[], 'standard_name':'time',                                              'long_name':'time',                                                 'units':'minutes since 0001-01-01 00:00:00', 'data_type':np.uint32,  'string_format':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Integer time in minutes since 0001-01-01 00:00 UTC. Time given refers to the start of the time window the measurement is representative of (temporal resolution).'},        

    #MEASUREMENT VALUES
    parameter_details['bsc_parameter_name']:            {'value':[], 'standard_name':parameter_details['long_parameter_name'],            'long_name':parameter_details['long_parameter_name'],               'units':parameter_details['standard_units'], 'data_type':np.float32, 'string_format':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Measured value of surface {} for the stated temporal resolution.'.format(parameter_details['long_parameter_name'])}, 

    #PARAMETERS INFORMING QUALITY OF MEASUREMENT 
    'temporal_resolution':                              {'value':[], 'standard_name':'temporal resolution',                               'long_name':'temporal measurement resolution',                      'units':'minutes',                           'data_type':np.uint16,  'string_format':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Temporal resolution that a specific measurement is representative of. Reported as an integer string (in minutes). If measured resolution is instantaneous, the given integer is 1.'}, 
    'reported_lower_limit_of_detection_per_measurement':{'value':[], 'standard_name':'reported lower limit of detection per measurement', 'long_name':'reported lower limit of detection per measurement',    'units':parameter_details['standard_units'], 'data_type':np.float32, 'string_format':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Reported lower limit of detection of measurement methodology, for a specific measurement, in {}.'.format(parameter_details['standard_units'])}, 
    'reported_upper_limit_of_detection_per_measurement':{'value':[], 'standard_name':'reported upper limit of detection per measurement', 'long_name':'reported upper limit of detection per measurement',    'units':parameter_details['standard_units'], 'data_type':np.float32, 'string_format':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Reported upper limit of detection of measurement methodology, for a specific measurement, in {}.'.format(parameter_details['standard_units'])}, 
    'reported_uncertainty_per_measurement':             {'value':[], 'standard_name':'reported measurement uncertainty per measurement',  'long_name':'reported measurement uncertainty per measurement',     'units':parameter_details['standard_units'], 'data_type':np.float32, 'string_format':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Reported measurement uncertainty (±) of methodology, for a specific measurement. In principal this refers to the inherent uncertainty on every measurement as a function of the quadratic addition of the accuracy and precision metrics (at the same confidence interval), but is often reported incosistently e.g. being solely determined from random errors (i.e. just the measurement precision). This is given in absolute terms in {}.'.format(parameter_details['standard_units'])},
    'derived_uncertainty_per_measurement':              {'value':[], 'standard_name':'derived measurement uncertainty per measurement',   'long_name':'derived measurement uncertainty per measurement',      'units':parameter_details['standard_units'], 'data_type':np.float32, 'string_format':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Derived measurement uncertainty (±) of methodology, for a specific measurement. This is calculated through the quadratic addition of reported (or if not available, documented) accuracy and precision metrics. This is given in absolute terms in {}.'.format(parameter_details['standard_units'])},
    'day_night_code':                                   {'value':[], 'standard_name':'day/night code',                                    'long_name':'day/night code per measurement',                       'units':'unitless',                          'data_type':np.uint8,   'string_format':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Binary indication if measurement was made during the day or night. Day=0, Night=1. The classification is made by calculating the solar elevation angle for a latitude/longitude/measurement height at a mid-measurement window timestamp. If the solar elevation angle is > 0, it is classed as daytime, otherwise it is nightime. Classification is 255 if cannot be made.'}, 
    'weekday_weekend_code':                             {'value':[], 'standard_name':'weekday/weekend code',                              'long_name':'weekday/weekend code per measurement',                 'units':'unitless',                          'data_type':np.uint8,   'string_format':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Binary indication if measurement was made during the weekday or weekend. Weekday=0, Weekend=1. The classification is made by evaluating if the local-time of a mid-measurement window timestamp falls on a weekday or on the weekend. Classification is 255 if cannot be made.'}, 
    'season_code':                                      {'value':[], 'standard_name':'season code',                                       'long_name':'season code per measurement',                          'units':'unitless',                          'data_type':np.uint8,   'string_format':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Code decreeing if measurement was made during the Spring, Summer, Autumn or Winter Seasons. Spring=0, Summer=1, Autumn=2, Winter=3. The classification is made by evaluating which season the local-time of a mid-measurement window timestamp falls in.'}, 
    'flag':                                             {'value':[], 'standard_name':'flags',                                             'long_name':'data reporter provided standardised flags',            'units':'unitless',                          'data_type':object,     'string_format':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'List of associated data flag codes per measurement, indicating the data quality of a specific measurement, provided by the data reporter. Fill value code of 255.'},
    'qa':                                               {'value':[], 'standard_name':'qa',                                                'long_name':'qa standardised data flags',                           'units':'unitless',                          'data_type':object,     'string_format':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'List of derived quality assurance flag codes per measurement, informing on the quality associated with each observation, determined by multiple scientific quality control/assurance checks. Fill value code of 255.'},
    }

    return standard_data

###--------------------------------------------------------------------------------------------------###
###METADATA VARIABLE STANDARDISATIONS
###--------------------------------------------------------------------------------------------------###

#Define dictionary with details pertaining to standardised metadata variables.
#These metadata variables refers to information which provides qualitative information about measurements, typically across large swathes of time.

def get_standard_metadata(parameter_details):

    standard_metadata = {    
    #VERSION OF GHOST
    'GHOST_version':{'value':[], 'standard_name':'GHOST version', 'long_name':'GHOST version', 'units':'unitless', 'data_type':object, 'string_format':'upper_short', 'metadata_type':'STATION MISCELLANEOUS', 'identifiers':{'WIGOS_name':''}, 'description':'Version of GHOST.'}, 
 
    #NETWORK STATION REFERENCE 
    'station_reference':{'value':[], 'standard_name':'station reference', 'long_name':'station reference identifier', 'units':'unitless', 'data_type':object, 'string_format':'short', 'metadata_type':'STATION MISCELLANEOUS', 'identifiers':{'WIGOS_name':''}, 'description':'reference ID for station.'}, 

    #NETWORK PROVIDED STATION INFORMATION
    'WIGOS_station_identifier':         {'value':[], 'standard_name':'WIGOS station identifier',          'long_name':'WIGOS station identifier',                        'units':'unitless',                   'data_type':object,     'string_format':'short',        'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'WIGOS station identifier (WSI).'},
    'station_timezone':                 {'value':[], 'standard_name':'station timezone',                  'long_name':'station timezone',                                'units':'unitless',                   'data_type':object,     'string_format':'short',        'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Name of the local timezone that the measuring station is located in. This is automatically generated using Timezone Finder Python package (taking longitude and latitude as inputs).'},
    'latitude':                         {'value':[], 'standard_name':'latitude',                          'long_name':'latitude',                                        'units':'decimal degrees North',      'data_type':np.float64, 'string_format':np.NaN,         'metadata_type':'STATION POSITION',        'identifiers':{'WIGOS_name':''},                          'description':'Geodetic latitude of measuring instrument, in decimal degrees North, following the stated horizontal datum.'},
    'longitude':                        {'value':[], 'standard_name':'longitude',                         'long_name':'longitude',                                       'units':'decimal degrees East',       'data_type':np.float64, 'string_format':np.NaN,         'metadata_type':'STATION POSITION',        'identifiers':{'WIGOS_name':''},                          'description':'Geodetic longitude of measuring instrument, in decimal degrees East, following the stated horizontal datum.'},
    'altitude':                         {'value':[], 'standard_name':'altitude',                          'long_name':'altitude relative to mean sea level',             'units':'m',                          'data_type':np.float32, 'string_format':np.NaN,         'metadata_type':'STATION POSITION',        'identifiers':{'WIGOS_name':''},                          'description':'Altitude of the ground level at the station, relative to the stated vertical datum, in metres.'},
    'sampling_height':                  {'value':[], 'standard_name':'sampling height',                   'long_name':'sampling height relative to ground',              'units':'m',                          'data_type':np.float32, 'string_format':np.NaN,         'metadata_type':'STATION POSITION',        'identifiers':{'WIGOS_name':''},                          'description':'Height above the ground of the inlet/instrument/sampler, in metres.'},
    'measurement_altitude':             {'value':[], 'standard_name':'measurement altitude',              'long_name':'measurement altitude relative to mean sea level', 'units':'m',                          'data_type':np.float32, 'string_format':np.NaN,         'metadata_type':'STATION POSITION',        'identifiers':{'WIGOS_name':''},                          'description':'Altitude of the inlet/instrument/sampler, relative to the stated vertical datum, in metres.'},
    'ellipsoid':                        {'value':[], 'standard_name':'ellipsoid',                         'long_name':'ellipsoid',                                       'units':'unitless',                   'data_type':object,     'string_format':'upper_short',  'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':'CoordinateReferenceSystem'}, 'description':'The ellipsoidal model of the earth used as a basis for 2D and 3D geographic coordinate systems.'},
    'horizontal_datum':                 {'value':[], 'standard_name':'horizontal datum',                  'long_name':'horizontal datum',                                'units':'unitless',                   'data_type':object,     'string_format':'upper_short',  'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':'CoordinateReferenceSystem'}, 'description':"Name of the horizontal datum used in defining geodetic latitudes and longitudes on the Earth's surface. The datum is set when positioning an ellipsoid model of the Earth to an anchor point. If not explicitely stated then this is assumed to be 'World Geodetic System 1984'."}, 
    'vertical_datum':                   {'value':[], 'standard_name':'vertical datum',                    'long_name':'vertical datum',                                  'units':'unitless',                   'data_type':object,     'string_format':'upper_short',  'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':'CoordinateReferenceSystem'}, 'description':"Name of the vertical datum used to define vertical elevation on the Earth. The datum is a surface of zero elevation to which other heights can be reference against. If not explicitely stated then this is assumed to be 'tidal - mean sea level'."}, 
    'projection':                       {'value':[], 'standard_name':'projection',                        'long_name':'projection',                                      'units':'unitless',                   'data_type':object,     'string_format':'upper_short',  'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':"Name of the projected coordinate system of the original provided station position x, y coordinates. If the original coordinates are not projected, then this is set as 'geographic'."}, 
    'distance_to_building':             {'value':[], 'standard_name':'distance to building',              'long_name':'distance to the nearest building',                'units':'m',                          'data_type':np.float32, 'string_format':np.NaN,         'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Distance to the nearest building of the inlet/instrument/sampler, in metres.'},
    'distance_to_kerb':                 {'value':[], 'standard_name':'distance to kerb',                  'long_name':'distance to the street kerb',                     'units':'m',                          'data_type':np.float32, 'string_format':np.NaN,         'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Distance to the street kerb of the inlet/instrument/sampler, in metres.'},
    'distance_to_junction':             {'value':[], 'standard_name':'distance to junction',              'long_name':'distance to the nearest road junction',           'units':'m',                          'data_type':np.float32, 'string_format':np.NaN,         'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Distance to the nearest road junction of the inlet/instrument/sampler, in metres.'},
    'distance_to_source':               {'value':[], 'standard_name':'distance to source',                'long_name':'distance to the main emission source',            'units':'km',                         'data_type':np.float32, 'string_format':np.NaN,         'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Distance to the main emission source (see variable: "main_emission_source") of the inlet/instrument/sampler, in kilometres.'},
    'street_width':                     {'value':[], 'standard_name':'street width',                      'long_name':'width of the street',                             'units':'m',                          'data_type':np.float32, 'string_format':np.NaN,         'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Width of the street where measurements are being made (if applicable), in metres.'},
    'street_type':                      {'value':[], 'standard_name':'street type',                       'long_name':'type of street',                                  'units':'unitless',                   'data_type':object,     'string_format':'title_short',  'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Type of street where measurements are being made (if applicable).'},
    'daytime_traffic_speed':            {'value':[], 'standard_name':'daytime traffic speed',             'long_name':'average daytime speed of passing traffic',        'units':'km hr-1',                    'data_type':np.float32, 'string_format':np.NaN,         'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Average daytime speed of the passing traffic where measurements are being made (if applicable), in kilometres per hour.'},
    'daily_passing_vehicles':           {'value':[], 'standard_name':'daily passing vehicles',            'long_name':'average daily number of passing vehicles',        'units':'unitless',                   'data_type':np.float32, 'string_format':np.NaN,         'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Average number of vehicles passing daily.'},
    'data_level':                       {'value':[], 'standard_name':'data level',                        'long_name':'data level',                                      'units':'unitless',                   'data_type':object,     'string_format':'short',        'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Data level of data reported. This varies per network. If data level is variable per measurement, and not static per reported file, then this is set as "variable". If there is no reported data level this is set as "none"'},
    'climatology':                      {'value':[], 'standard_name':'climatology',                       'long_name':'climatology',                                     'units':'unitless',                   'data_type':object,     'string_format':'upper_short',  'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Name of the climatology of which the observations pertain to.'},
    'station_name':                     {'value':[], 'standard_name':'station name',                      'long_name':'station name',                                    'units':'unitless',                   'data_type':object,     'string_format':'title_short',  'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Name of station where the measurement was conducted.'},
    'city':                             {'value':[], 'standard_name':'city',                              'long_name':'city',                                            'units':'unitless',                   'data_type':object,     'string_format':'title_short',  'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Name of the city the station is located in.'},
    'country':                          {'value':[], 'standard_name':'country',                           'long_name':'country',                                         'units':'unitless',                   'data_type':object,     'string_format':'title_short',  'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':'TerritoryName'},             'description':'Name of the country the station is located in.'},
    'administrative_country_division_1':{'value':[], 'standard_name':'administrative country division 1', 'long_name':'administrative country division 1',               'units':'unitless',                   'data_type':object,     'string_format':'title_short',  'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Name of the first (i.e. largest) country administrative division in which the station lies, e.g. countries within soverign state, state, province, county etc. These are defined for the purposes of managing of land and the affairs of people. This is automatically generated using Reverse Geocoder Python package (taking longitude and latitude as inputs).'},
    'administrative_country_division_2':{'value':[], 'standard_name':'administrative country division 2', 'long_name':'administrative country division 2',               'units':'unitless',                   'data_type':object,     'string_format':'title_short',  'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Name of the second (i.e. second largest) country administrative division in which the station lies, e.g. countries within soverign state, state, province, county etc. These are defined for the purposes of managing of land and the affairs of people. This is automatically generated using Reverse Geocoder Python package (taking longitude and latitude as inputs).'},
    'population':                       {'value':[], 'standard_name':'population',                        'long_name':'population',                                      'units':'unitless',                   'data_type':np.float32, 'string_format':np.NaN,         'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'Population size of the nearest urban settlement.'},
    'representative_radius':            {'value':[], 'standard_name':'representative_radius',             'long_name':'representative_radius',                           'units':'km',                         'data_type':np.float32, 'string_format':np.NaN,         'metadata_type':'STATION CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'Representativeness'},        'description':'Radius of representativity of the measurements made (i.e. for what distance scale around the sampling point would the measurements be very similar?), given in kilometres.  A quantitative version of the "measurement_scale" classification.'},
    'network':                          {'value':[], 'standard_name':'network',                           'long_name':'data network',                                    'units':'unitless',                   'data_type':object,     'string_format':'short',        'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'The name of the network which reports data for the specific station in question.'},
    'associated_networks':              {'value':[], 'standard_name':'other associated networks',         'long_name':'other associated networks',                       'units':'unitless',                   'data_type':object,     'string_format':'long',         'metadata_type':'STATION MISCELLANEOUS',   'identifiers':{'WIGOS_name':''},                          'description':'String pair of associated network name and station reference. Format: network1:station_reference1;network2:station_reference2'},

    #NETWORK PROVIDED STANDARDISED CLASSIFICATIONS
    'area_classification':   {'value':[], 'standard_name':'area classification',    'long_name':'standardised network provided area classification',    'units':'unitless', 'data_type':object, 'string_format':'lower_short', 'metadata_type':'STATION CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Standardised network provided classification, describing type of area a measurement station is situated in.'},                
    'station_classification':{'value':[], 'standard_name':'station classification', 'long_name':'standardised network provided station classification', 'units':'unitless', 'data_type':object, 'string_format':'lower_short', 'metadata_type':'STATION CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Standardised network provided classification, categorising the type of air measured by a station.'},                
    'main_emission_source':  {'value':[], 'standard_name':'main emission source',   'long_name':'standardised network provided main emission source',   'units':'unitless', 'data_type':object, 'string_format':'lower_short', 'metadata_type':'STATION CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Standardised network provided classification, describing the main emission source influencing air measured at a station.'},                
    'land_use':              {'value':[], 'standard_name':'land use',               'long_name':'standardised network provided land use type',          'units':'unitless', 'data_type':object, 'string_format':'lower_short', 'metadata_type':'STATION CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Standardised network provided classification, describing the dominant land use in the area of the reporting station.'},                
    'terrain':               {'value':[], 'standard_name':'terrain',                'long_name':'standardised network provided terrain type',           'units':'unitless', 'data_type':object, 'string_format':'lower_short', 'metadata_type':'STATION CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Standardised network provided classification, describing the dominant terrain in the area of the reporting station.'},                 
    'measurement_scale':     {'value':[], 'standard_name':'measurement scale',      'long_name':'standardised network provided measurement scale',      'units':'unitless', 'data_type':object, 'string_format':'lower_short', 'metadata_type':'STATION CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Standardised network provided classification, a denotation of the geographic scope of the air quality measurements made.'},                         

    #GLOBALLY GRIDDED CLASSIFICATIONS
    'ESDAC_Iwahashi_landform_classification':           {'value':[], 'standard_name':'ESDAC Iwahashi landform classification',            'long_name':'ESDAC Iwahashi landform classification',                      'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'European Soil Data Centre (ESDAC) Iwahashi landform classification. The classification presents relief classes which are classified using an unsupervised nested-means algorithms and a three part geometric signature. Slope gradient, surface texture and local convexity are calculated based on the SRTM30 digital elevation model, within a given window size and classified according to the inherent data set properties. This is a dynamic landform classification method. Native resolution of 0.0083 x 0.0083 degrees. A correction for coastal sites is made: if the native class is "water", then the modal classification of the neighbouring grid boxes is used instead (lowest code kept preferentially in case of a tie). If the site is truly an "ocean" site, all the surrounding gridcells will be water also, and therefore the class will be maintained as "water".'},
    'ESDAC_modal_Iwahashi_landform_classification_5km': {'value':[], 'standard_name':'ESDAC modal Iwahashi landform classification 5km',  'long_name':'ESDAC modal Iwahashi landform classification in 5km radius',  'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'Modal European Soil Data Centre (ESDAC) Iwahashi landform classification in radius of 5km around station location.'},
    'ESDAC_modal_Iwahashi_landform_classification_25km':{'value':[], 'standard_name':'ESDAC modal Iwahashi landform classification 25km', 'long_name':'ESDAC modal Iwahashi landform classification in 25km radius', 'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'Modal European Soil Data Centre (ESDAC) Iwahashi landform classification in radius of 25km around station location.'},
    'ESDAC_Meybeck_landform_classification':            {'value':[], 'standard_name':'ESDAC Meybeck landform classification',             'long_name':'ESDAC Meybeck landform classification',                       'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'European Soil Data Centre (ESDAC) Meybeck landform classification. The classification presents relief classes which are calculated based on the relief roughness. Roughness and elevation are classified based on a digital elevation model according to static thresholds, with a given window size. This is a static landform classification method. Native resolution of 0.0083 x 0.0083 degrees. A correction for coastal sites is made: if the native class is "water", then the modal classification of the neighbouring grid boxes is used instead (lowest code kept preferentially in case of a tie). If the site is truly an "ocean" site, all the surrounding gridcells will be water also, and therefore the class will be maintained as "water".'},
    'ESDAC_modal_Meybeck_landform_classification_5km':  {'value':[], 'standard_name':'ESDAC modal Meybeck landform classification 5km',   'long_name':'ESDAC modal Meybeck landform classification in 5km radius',   'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'Modal European Soil Data Centre (ESDAC) Meybeck landform classification in radius of 5km around station location.'},
    'ESDAC_modal_Meybeck_landform_classification_25km': {'value':[], 'standard_name':'ESDAC modal Meybeck landform classification 25km',  'long_name':'ESDAC modal Meybeck landform classification in 25km radius',  'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'Modal European Soil Data Centre (ESDAC) Meybeck landform classification in radius of 25km around station location.'},
    'GHSL_settlement_model_classification':             {'value':[], 'standard_name':'GHSL settlement model classification',              'long_name':'GHSL settlement model classification',                        'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'Global Human Settlement Layer (GHSL) settlement model classification (technical label: GHS_SMOD_POPMT_GLOBE_R2019A). The classification delineates and classify settlement typologies via a logic of population size, population and built-up area densities as a refinement of the ‘degree of urbanization’ method described by EUROSTAT. The classification is derived by using the GHS_POP_MT_GLOBE_R2019A and GHS_BUILT_LDSMT_GLOBE_R2018A products. The GHS Settlement Model grid is an improvement of the GHS Settlement Grid (R2016A) introducing a more detailed classification of settlements in two levels, also called ‘refined degree of urbanization’. The Settlement Model is provided at detailed level (Second Level - L2). The First Level, as a porting of the Degree of Urbanization adopted by EUROSTAT can be obtained aggregating L2. Native resolution of 1.0 x 1.0 kilometres.'},   
    'GHSL_modal_settlement_model_classification_5km':   {'value':[], 'standard_name':'GHSL modal settlement model classification 5km',    'long_name':'GHSL modal settlement model classification in 5km radius',    'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'Modal Global Human Settlement Layer (GHSL) settlement model classification in radius of 5km around station location.'},   
    'GHSL_modal_settlement_model_classification_25km':  {'value':[], 'standard_name':'GHSL modal settlement model classification 25km',   'long_name':'GHSL modal settlement model classification in 25km radius',   'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'Modal Global Human Settlement Layer (GHSL) settlement model classification in radius of 25km around station location.'},   
    'Joly-Peuch_classification_code':                   {'value':[], 'standard_name':'Joly-Peuch classification code',                    'long_name':'Joly-Peuch classification code',                              'units':'unitless', 'data_type':np.float32, 'string_format':np.NaN,        'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'Joly-Peuch European classification code (range of 1-10) designed to objectively stratify stations between those displaying rural and urban signatures (most rural == 1, most urban == 10). This classification is objectively made per species. The species that this is done for are: O3, NO2, SO2, CO, PM10, PM2.5. See reference here: https://www.sciencedirect.com/science/article/abs/pii/S1352231011012088'},        
    'Koppen-Geiger_classification':                     {'value':[], 'standard_name':'Koppen-Geiger classification',                      'long_name':'Koppen-Geiger classification',                                'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'ClimateZone'},      'description':'Koppen-Geiger classification, classifying the global climates into 5 main groups (30 total groups with subcategories). Native resolution of 0.0083 x 0.0083 degrees. A correction for coastal sites is made: if the native class is "water", then the modal classification of the neighbouring grid boxes is used instead (lowest code kept preferentially in case of a tie). If the site is truly an "ocean" site, all the surrounding gridcells will be water also, and therefore the class will be maintained as "water". See citation: Beck, H.E., N.E. Zimmermann, T.R. McVicar, N. Vergopolan, A. Berg, E.F. Wood: Present and future Köppen-Geiger climate classification maps at 1-km resolution, Nature Scientific Data, 2018.'},
    'Koppen-Geiger_modal_classification_5km':           {'value':[], 'standard_name':'Koppen-Geiger modal classification 5km',            'long_name':'Koppen-Geiger classification',                                'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'ClimateZone'},      'description':'Modal Koppen-Geiger classification in radius of 5km around station location.'},
    'Koppen-Geiger_modal_classification_25km':          {'value':[], 'standard_name':'Koppen-Geiger modal classification 25km',           'long_name':'Koppen-Geiger classification',                                'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'ClimateZone'},      'description':'Modal Koppen-Geiger classification in radius of 25km around station location.'},
    'MODIS_MCD12C1_v6_IGBP_land_use':                   {'value':[], 'standard_name':'MODIS MCD12C1 v6 IGBP land use',                    'long_name':'MODIS MCD12C1 v6 IGBP land use',                              'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'SurfaceCoverIGBP'}, 'description':'Majority land use class from the Moderate Resolution Imaging Spectroradiometer (MODIS) Land Cover Climate Modeling Grid (CMG) MCD12C1 version 6, using the International Geosphere-Biosphere Programme (IGBP) classification. Native resolution of 0.05 x 0.05 degrees. See dataset user guide here: https://lpdaac.usgs.gov/documents/101/MCD12_User_Guide_V6.pd. A correction for coastal sites is made: if the native class is "water bodies", then the modal classification of the neighbouring grid boxes is used instead (lowest code kept preferentially in case of a tie). If the site is truly an "ocean" site, all the surrounding gridcells will be water also, and therefore the class will be maintained as "water".'},
    'MODIS_MCD12C1_v6_modal_IGBP_land_use_5km':         {'value':[], 'standard_name':'MODIS MCD12C1 v6 IGBP modal land use 5km',          'long_name':'MODIS MCD12C1 v6 IGBP modal land use in 5km radius',          'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'SurfaceCoverIGBP'}, 'description':'Modal land use in radius of 5km around the station location from the Moderate Resolution Imaging Spectroradiometer (MODIS) Land Cover Climate Modeling Grid (CMG) MCD12C1 version 6, using the International Geosphere-Biosphere Programme (IGBP) classification.'},
    'MODIS_MCD12C1_v6_modal_IGBP_land_use_25km':        {'value':[], 'standard_name':'MODIS MCD12C1 v6 IGBP modal land use 25km',         'long_name':'MODIS MCD12C1 v6 IGBP modal land use in 25km radius',         'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'SurfaceCoverIGBP'}, 'description':'Modal land use in radius of 25km around the station location from the Moderate Resolution Imaging Spectroradiometer (MODIS) Land Cover Climate Modeling Grid (CMG) MCD12C1 version 6, using the International Geosphere-Biosphere Programme (IGBP) classification.'},
    'MODIS_MCD12C1_v6_UMD_land_use':                    {'value':[], 'standard_name':'MODIS MCD12C1 v6 UMD land use',                     'long_name':'MODIS MCD12C1 v6 UMD land use',                               'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'SurfaceCoverUMD'},  'description':'Majority land use class from the Moderate Resolution Imaging Spectroradiometer (MODIS) Land Cover Climate Modeling Grid (CMG) MCD12C1 version 6, using the University of Maryland (UMD) classification. Native resolution of 0.05 x 0.05 degrees. See dataset user guide here: https://lpdaac.usgs.gov/documents/101/MCD12_User_Guide_V6.pd. A correction for coastal sites is made: if the native class is "water bodies", then the modal classification of the neighbouring grid boxes is used instead (lowest code kept preferentially in case of a tie). If the site is truly an "ocean" site, all the surrounding gridcells will be water also, and therefore the class will be maintained as "water".'},
    'MODIS_MCD12C1_v6_modal_UMD_land_use_5km':          {'value':[], 'standard_name':'MODIS MCD12C1 v6 UMD modal land use 5km',           'long_name':'MODIS MCD12C1 v6 UMD modal land use in 5km radius',           'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'SurfaceCoverUMD'},  'description':'Modal land use in radius of 5km around the station location from the Moderate Resolution Imaging Spectroradiometer (MODIS) Land Cover Climate Modeling Grid (CMG) MCD12C1 version 6, using the University of Maryland (UMD) classification.'},
    'MODIS_MCD12C1_v6_modal_UMD_land_use_25km':         {'value':[], 'standard_name':'MODIS MCD12C1 v6 UMD modal land use 25km',          'long_name':'MODIS MCD12C1 v6 UMD modal land use in 25km radius',          'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'SurfaceCoverUMD'},  'description':'Modal land use in radius of 25km around the station location from the Moderate Resolution Imaging Spectroradiometer (MODIS) Land Cover Climate Modeling Grid (CMG) MCD12C1 version 6, using the University of Maryland (UMD) classification.'},
    'MODIS_MCD12C1_v6_LAI':                             {'value':[], 'standard_name':'MODIS MCD12C1 v6 LAI',                              'long_name':'MODIS MCD12C1 v6 Leaf Area Index',                            'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'SurfaceCoverLAI'},  'description':'Majority Leaf Area Index class from the Moderate Resolution Imaging Spectroradiometer (MODIS) Land Cover Climate Modeling Grid (CMG) MCD12C1 version 6. Native resolution of 0.05 x 0.05 degrees. See dataset user guide here: https://lpdaac.usgs.gov/documents/101/MCD12_User_Guide_V6.pd. A correction for coastal sites is made: if the native class is "water bodies", then the modal classification of the neighbouring grid boxes is used instead (lowest code kept preferentially in case of a tie). If the site is truly an "ocean" site, all the surrounding gridcells will be water also, and therefore the class will be maintained as "water".'},
    'MODIS_MCD12C1_v6_modal_LAI_5km':                   {'value':[], 'standard_name':'MODIS MCD12C1 v6 LAI modal 5km',                    'long_name':'MODIS MCD12C1 v6 modal Leaf Area Index in 5km radius',        'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'SurfaceCoverLAI'},  'description':'Modal Leaf Area Index in radius of 5km around the station location from the Moderate Resolution Imaging Spectroradiometer (MODIS) Land Cover Climate Modeling Grid (CMG) MCD12C1 version 6.'},
    'MODIS_MCD12C1_v6_modal_LAI_25km':                  {'value':[], 'standard_name':'MODIS MCD12C1 v6 LAI modal 25km',                   'long_name':'MODIS MCD12C1 v6 modal Leaf Area Index in 25km radius',       'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'SurfaceCoverLAI'},  'description':'Modal Leaf Area Index in radius of 25km around the station location from the Moderate Resolution Imaging Spectroradiometer (MODIS) Land Cover Climate Modeling Grid (CMG) MCD12C1 version 6.'},
    'WMO_region':                                       {'value':[], 'standard_name':'WMO region',                                        'long_name':'WMO region code',                                             'units':'unitless', 'data_type':object,     'string_format':'title_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':'WMORegion'},        'description':'World Meteorological Organization (WMO) region of station. The available regions are: Africa, Asia, South America, "Northern America, Central America and the Caribbean", South-West Pacific, Europe and Antarctica.'},
    'WWF_TEOW_terrestrial_ecoregion':                   {'value':[], 'standard_name':'WWF TEOW terrestrial ecoregion',                    'long_name':'WWF TEOW terrestrial ecoregion',                              'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'Terrestrial Ecoregions of the World (TEOW) World Wildlife Foundation (WWF) classification. There are 825 terrestrial ecoregions. Ecoregions are relatively large units of land containing distinct assemblages of natural communities and species, with boundaries that approximate the original extent of natural communities prior to major land-use change. See citation: Olson, D. M., Dinerstein, E., Wikramanayake, E. D., Burgess, N. D., Powell, G. V. N., Underwood, E. C., DAmico, J. A., Itoua, I., Strand, H. E., Morrison, J. C., Loucks, C. J., Allnutt, T. F., Ricketts, T. H., Kura, Y., Lamoreux, J. F., Wettengel, W. W., Hedao, P., Kassem, K. R. 2001. Terrestrial ecoregions of the world: a new map of life on Earth. Bioscience 51(11):933-938.'},
    'WWF_TEOW_biogeographical_realm':                   {'value':[], 'standard_name':'WWF TEOW biogeographical realm',                    'long_name':'WWF TEOW biogeographical realm',                              'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'Terrestrial Ecoregions of the World (TEOW) World Wildlife Foundation (WWF) classification. There are 8 biogeographical realms. See citation: Olson, D. M., Dinerstein, E., Wikramanayake, E. D., Burgess, N. D., Powell, G. V. N., Underwood, E. C., DAmico, J. A., Itoua, I., Strand, H. E., Morrison, J. C., Loucks, C. J., Allnutt, T. F., Ricketts, T. H., Kura, Y., Lamoreux, J. F., Wettengel, W. W., Hedao, P., Kassem, K. R. 2001. Terrestrial ecoregions of the world: a new map of life on Earth. Bioscience 51(11):933-938.'},
    'WWF_TEOW_biome':                                   {'value':[], 'standard_name':'WWF TEOW biome',                                    'long_name':'WWF TEOW biome',                                              'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'Terrestrial Ecoregions of the World (TEOW) World Wildlife Foundation (WWF) classification. There are 14 biomes. See citation: Olson, D. M., Dinerstein, E., Wikramanayake, E. D., Burgess, N. D., Powell, G. V. N., Underwood, E. C., DAmico, J. A., Itoua, I., Strand, H. E., Morrison, J. C., Loucks, C. J., Allnutt, T. F., Ricketts, T. H., Kura, Y., Lamoreux, J. F., Wettengel, W. W., Hedao, P., Kassem, K. R. 2001. Terrestrial ecoregions of the world: a new map of life on Earth. Bioscience 51(11):933-938.'},    
    'UMBC_anthrome_classification':                     {'value':[], 'standard_name':'UMBC anthrome classification',                      'long_name':'UMBC anthrome classification',                                'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'University of Maryland Baltimore County (UMBC) anthrome classification, describing the anthropogenic land use (for the year 2000). There are 20 distinct classifications. Native resolution of 0.0833 x 0.0833 degrees. A correction for coastal sites is made: if the native anthrome class is "water", then the modal classification of the neighbouring grid boxes is used instead (lowest code kept preferentially in case of a tie). If the site is truly an "ocean" site, all the surrounding gridcells will be water also, and therefore the class will be maintained as "water".'},
    'UMBC_modal_anthrome_classification_5km':           {'value':[], 'standard_name':'UMBC modal anthrome classification 5km',            'long_name':'UMBC modal anthrome classification in 5km radius',            'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'University of Maryland Baltimore County (UMBC) modal anthrome classification in radius of 5km around station location.'},
    'UMBC_modal_anthrome_classification_25km':          {'value':[], 'standard_name':'UMBC modal anthrome classification 25km',           'long_name':'UMBC modal anthrome classification in 25km radius',           'units':'unitless', 'data_type':object,     'string_format':'lower_short', 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''},                 'description':'University of Maryland Baltimore County (UMBC) modal anthrome classification in radius of 25km around station location.'},
          
    #GLOBALLY GRIDDED PRODUCTS
    'EDGAR_v4.3.2_annual_average_BC_emissions':                        {'value':[], 'standard_name':'EDGAR v4.3.2 annual average BC emissions',                         'long_name':'EDGAR v4.3.2 annual average black carbon emissions',                                                          'units':'kg m-2 s-1', 'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'EDGAR v4.3.2 annual average BC emissions, in kilograms per squared metre per second. Native resolution of 0.1 x 0.1 degrees.'},         
    'EDGAR_v4.3.2_annual_average_CO_emissions':                        {'value':[], 'standard_name':'EDGAR v4.3.2 annual average CO emissions',                         'long_name':'EDGAR v4.3.2 annual average carbon monoxide emissions',                                                       'units':'kg m-2 s-1', 'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'EDGAR v4.3.2 annual average CO emissions, in kilograms per squared metre per second. Native resolution of 0.1 x 0.1 degrees.'},         
    'EDGAR_v4.3.2_annual_average_NH3_emissions':                       {'value':[], 'standard_name':'EDGAR v4.3.2 annual average NH3 emissions',                        'long_name':'EDGAR v4.3.2 annual average ammonia emissions',                                                               'units':'kg m-2 s-1', 'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'EDGAR v4.3.2 annual average NH3 emissions, in kilograms per squared metre per second. Native resolution of 0.1 x 0.1 degrees.'},         
    'EDGAR_v4.3.2_annual_average_NMVOC_emissions':                     {'value':[], 'standard_name':'EDGAR v4.3.2 annual average NMVOC emissions',                      'long_name':'EDGAR v4.3.2 annual average non-methane volatile organic compound emissions',                                 'units':'kg m-2 s-1', 'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'EDGAR v4.3.2 annual average NMVOC emissions, in kilograms per squared metre per second. Native resolution of 0.1 x 0.1 degrees.'},         
    'EDGAR_v4.3.2_annual_average_NOx_emissions':                       {'value':[], 'standard_name':'EDGAR v4.3.2 annual average NOx emissions',                        'long_name':'EDGAR v4.3.2 annual average emissions of nitrogen oxides',                                                    'units':'kg m-2 s-1', 'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'EDGAR v4.3.2 annual average NOx emissions, in kilograms per squared metre per second. Native resolution of 0.1 x 0.1 degrees.'},         
    'EDGAR_v4.3.2_annual_average_OC_emissions':                        {'value':[], 'standard_name':'EDGAR v4.3.2 annual average OC emissions',                         'long_name':'EDGAR v4.3.2 annual average organic carbon emissions',                                                        'units':'kg m-2 s-1', 'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'EDGAR v4.3.2 annual average OC emissions, in kilograms per squared metre per second. Native resolution of 0.1 x 0.1 degrees.'},         
    'EDGAR_v4.3.2_annual_average_PM10_emissions':                      {'value':[], 'standard_name':'EDGAR v4.3.2 annual average PM10 emissions',                       'long_name':'EDGAR v4.3.2 annual average PM10 emissions',                                                                  'units':'kg m-2 s-1', 'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'EDGAR v4.3.2 annual average PM10 emissions, in kilograms per squared metre per second. Native resolution of 0.1 x 0.1 degrees.'},         
    'EDGAR_v4.3.2_annual_average_biogenic_PM2.5_emissions':            {'value':[], 'standard_name':'EDGAR v4.3.2 annual average biogenic PM2.5 emissions',             'long_name':'EDGAR v4.3.2 annual average biogenic PM2.5 emissions',                                                        'units':'kg m-2 s-1', 'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'EDGAR v4.3.2 annual average biogenic PM2.5 emissions, in kilograms per squared metre per second. Native resolution of 0.1 x 0.1 degrees.'},         
    'EDGAR_v4.3.2_annual_average_fossilfuel_PM2.5_emissions':          {'value':[], 'standard_name':'EDGAR v4.3.2 annual average fossil fuel PM2.5 emissions',          'long_name':'EDGAR v4.3.2 annual average fossil fuel PM2.5 emissions',                                                     'units':'kg m-2 s-1', 'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'EDGAR v4.3.2 annual average fossil fuel PM2.5 emissions, in kilograms per squared metre per second. Native resolution of 0.1 x 0.1 degrees.'},         
    'EDGAR_v4.3.2_annual_average_SO2_emissions':                       {'value':[], 'standard_name':'EDGAR v4.3.2 annual average SO2 emissions',                        'long_name':'EDGAR v4.3.2 annual average sulphur dioxide emissions',                                                       'units':'kg m-2 s-1', 'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'EDGAR v4.3.2 annual average SO2 emissions, in kilograms per squared metre per second. Native resolution of 0.1 x 0.1 degrees.'},         
    'ASTER_v3_altitude':                                               {'value':[], 'standard_name':'ASTER v3 altitude',                                                'long_name':'ASTER v3 altitude, relative to EGM96 geoid datum',                                                            'units':'m',          'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Altitude from ASTER v3 digital elevation model, relative to EGM96 geoid vertical datum, in metres. The dataset was generated using 1,880,306 Level-1A scenes (taken from the NASA TERRA spacecraft) acquired between March 1, 2000 and November 30, 2013. The ASTER GDEM was created by stacking all individual cloud-masked scene DEMs and non-cloud-masked scene DEMs, then applying various algorithms to remove abnormal data. A statistical approach is not always effective for anomaly removal in areas with a limited number of images. Several existing reference DEMs were used to replace residual anomalies caused by the insufficient number of stacked scenes. In addition to ASTER GDEM, the ASTER Global Water Body Database (ASTWBD) was generated as a by-product to correct elevation values of water body surfaces like sea, rivers, and lakes. The ASTWBD was applied to GDEM to provide proper elevation values for water body surfaces. The sea and lake have a flattened elevation value. The river has a stepped-down elevation value from the upper stream to the lower stream. Native resolution of 1 arc second ~= 30m at the equator.'},         
    'ETOPO1_altitude':                                                 {'value':[], 'standard_name':'ETOPO1 altitude',                                                  'long_name':'ETOPO1 altitude, relative to sea level datum',                                                                'units':'m',          'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Altitude from ETOPO1 digital elevation model, relative to sea level vertical datum, in metres. Over Antarctica and Greenland the elevation given is on top of the ice sheets. Native resolution of 1 arc minute. A correction for coastal sites is made: if the derived altitude is <= -5 m, the maximum altitude of the neighbouring grid boxes will be used instead. If all neighbouring grid boxes have altitudes <= -5 m, the original value will be retained.'},         
    'ETOPO1_max_altitude_difference_5km':                              {'value':[], 'standard_name':'ETOPO1 max altitude difference 5km',                               'long_name':'ETOPO1 maximum altitude difference between the ETOPO1_altitude and all ETOPO1 altitudes in 5km radius',       'units':'m',          'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Altitude difference between the ETOPO1_altitude, and the minimum ETOPO1 altitude in a radius of 5km around the station location, in metres.'},         
    'GHSL_built_up_area_density':                                      {'value':[], 'standard_name':'GHSL built up area density',                                       'long_name':'GHSL built up area density',                                                                                  'units':'%',          'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Global Human Settlement Layer (GHSL) built up area density (technical label: GHS_BUILT_LDSMT_GLOBE_R2018A), in units of built-up area percent per gridcell (0-100). The product is a multitemporal information layer on built-up presence as derived from Landsat image collections (GLS1975, GLS1990, GLS2000, and ad-hoc Landsat 8 collection 2013/2014). Native resolution of 0.25 x 0.25 kilometres.'},   
    'GHSL_average_built_up_area_density_5km':                          {'value':[], 'standard_name':'GHSL average built up area density 5km',                           'long_name':'GHSL average built up area density in 5km radius',                                                            'units':'%',          'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Global Human Settlement Layer (GHSL) average built up area density in a radius of 5km around the station location.'},   
    'GHSL_average_built_up_area_density_25km':                         {'value':[], 'standard_name':'GHSL average built up area density 25km',                          'long_name':'GHSL average built up area density in 25km radius',                                                           'units':'%',          'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Global Human Settlement Layer (GHSL) average built up area density in a radius of 25km around the station location.'},   
    'GHSL_max_built_up_area_density_5km':                              {'value':[], 'standard_name':'GHSL max built up area density 5km',                               'long_name':'GHSL max built up area density in 5km radius',                                                                'units':'%',          'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Global Human Settlement Layer (GHSL) max built up area density in a radius of 5km around the station location.'},   
    'GHSL_max_built_up_area_density_25km':                             {'value':[], 'standard_name':'GHSL max built up area density 25km',                              'long_name':'GHSL max built up area density in 25km radius',                                                               'units':'%',          'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Global Human Settlement Layer (GHSL) max built up area density in a radius of 25km around the station location.'},   
    'GHSL_population_density':                                         {'value':[], 'standard_name':'GHSL population density',                                          'long_name':'GHSL population density',                                                                                     'units':'xx km–2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Global Human Settlement Layer (GHSL) population density (technical label: GHS_POP_MT_GLOBE_R2019A), in populus per squared kilometre. It depicts the distribution of population, expressed as the number of people per cell. Residential population estimates for target years 1975, 1990, 2000 and 2015 provided by CIESIN GPWv4.10 were disaggregated from census or administrative units to grid cells, informed by the distribution and density of built-up as mapped in the GHSL global layer per corresponding epoch. Native resolution of 0.25 x 0.25 kilometres.'},   
    'GHSL_average_population_density_5km':                             {'value':[], 'standard_name':'GHSL average population density 5km',                              'long_name':'GHSL average population density in 5km radius',                                                               'units':'xx km–2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Global Human Settlement Layer (GHSL) average population density in a radius of 5km around the station location.'},   
    'GHSL_average_population_density_25km':                            {'value':[], 'standard_name':'GHSL average population density 25km',                             'long_name':'GHSL average population density in 25km radius',                                                              'units':'xx km–2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Global Human Settlement Layer (GHSL) average population density in a radius of 25km around the station location.'},   
    'GHSL_max_population_density_5km':                                 {'value':[], 'standard_name':'GHSL max population density 5km',                                  'long_name':'GHSL max population density in 5km radius',                                                                   'units':'xx km–2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Global Human Settlement Layer (GHSL) max population density in a radius of 5km around the station location.'},   
    'GHSL_max_population_density_25km':                                {'value':[], 'standard_name':'GHSL max population density 25km',                                 'long_name':'GHSL max population density in 25km radius',                                                                  'units':'xx km–2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Global Human Settlement Layer (GHSL) max population density in a radius of 25km around the station location.'},   
    'GPW_population_density':                                          {'value':[], 'standard_name':'GPW population density',                                           'long_name':'GPW population density',                                                                                      'units':'xx km–2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Gridded Population of the World (GPW) population density, in populus per squared kilometre, from either version 3 and 4 of the provided gridded datasets, dependent on the data year: v3 (1990-2000), v4 (2000-2015). Native resolution of 0.04166 x 0.04166 for v3 data; native resolution of 0.0083 x 0.0083 degrees for v4 data.'},         
    'GPW_average_population_density_5km':                              {'value':[], 'standard_name':'GPW average population density 5km',                               'long_name':'GPW average population density in 5km radius',                                                                'units':'xx km–2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Gridded Population of the World (GPW) average population density in a radius of 5km around the station location.'},         
    'GPW_average_population_density_25km':                             {'value':[], 'standard_name':'GPW average population density 25km',                              'long_name':'GPW average population density in 25km radius',                                                               'units':'xx km–2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Gridded Population of the World (GPW) average population density in a radius of 25km around the station location.'},         
    'GPW_max_population_density_5km':                                  {'value':[], 'standard_name':'GPW max population density 5km',                                   'long_name':'GPW average population density in 5km radius',                                                                'units':'xx km–2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Gridded Population of the World (GPW) maximum population density in a radius of 5km around the station location.'},         
    'GPW_max_population_density_25km':                                 {'value':[], 'standard_name':'GPW max population density 25km',                                  'long_name':'GPW average population density in 25km radius',                                                               'units':'xx km–2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Gridded Population of the World (GPW) maximum population density in a radius of 25km around the station location.'},         
    'NOAA-DMSP-OLS_v4_nighttime_stable_lights':                        {'value':[], 'standard_name':'NOAA-DMSP-OLS v4 nighttime stable lights',                         'long_name':'NOAA DMSP-OLS version 4 nighttime stable lights',                                                             'units':'unitless',   'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'National Oceanic and Atmospheric Administration (NOAA), Defense Meteorological Satellite Program - Operational Linescane System (DMSP-OLS) version 4 nighttime stable lights. Native resolution of 0.0083 x 0.0083 degrees. The values represent a brightness index ranging from 0 to 63. The sensor saturates at a value of 63.'},         
    'NOAA-DMSP-OLS_v4_average_nighttime_stable_lights_5km':            {'value':[], 'standard_name':'NOAA-DMSP-OLS v4 average nighttime stable lights 5km',             'long_name':'NOAA DMSP-OLS version 4 average nighttime stable lights in 5km radius',                                       'units':'unitless',   'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'National Oceanic and Atmospheric Administration (NOAA), Defense Meteorological Satellite Program - Operational Linescane System (DMSP-OLS) version 4 average nighttime stable lights in 5km radius around the station location. The values represent a brightness index ranging from 0 to 63. The sensor saturates at a value of 63.'},         
    'NOAA-DMSP-OLS_v4_average_nighttime_stable_lights_25km':           {'value':[], 'standard_name':'NOAA-DMSP-OLS v4 average nighttime stable lights 25km',            'long_name':'NOAA DMSP-OLS version 4 average nighttime stable lights in 25km radius',                                      'units':'unitless',   'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'National Oceanic and Atmospheric Administration (NOAA), Defense Meteorological Satellite Program - Operational Linescane System (DMSP-OLS) version 4 average nighttime stable lights in 25km radius around the station location. The values represent a brightness index ranging from 0 to 63. The sensor saturates at a value of 63.'},         
    'NOAA-DMSP-OLS_v4_max_nighttime_stable_lights_5km':                {'value':[], 'standard_name':'NOAA-DMSP-OLS v4 max nighttime stable lights 5km',                 'long_name':'NOAA DMSP-OLS version 4 maximum nighttime stable lights in 5km radius',                                       'units':'unitless',   'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'National Oceanic and Atmospheric Administration (NOAA), Defense Meteorological Satellite Program - Operational Linescane System (DMSP-OLS) version 4 maximum nighttime stable lights in 5km radius around the station location. The values represent a brightness index ranging from 0 to 63. The sensor saturates at a value of 63.'},         
    'NOAA-DMSP-OLS_v4_max_nighttime_stable_lights_25km':               {'value':[], 'standard_name':'NOAA-DMSP-OLS v4 max nighttime stable lights 25km',                'long_name':'NOAA DMSP-OLS version 4 maximum nighttime stable lights in 25km radius',                                      'units':'unitless',   'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'National Oceanic and Atmospheric Administration (NOAA), Defense Meteorological Satellite Program - Operational Linescane System (DMSP-OLS) version 4 maximum nighttime stable lights in 25km radius around the station location. The values represent a brightness index ranging from 0 to 63. The sensor saturates at a value of 63.'},         
    'OMI_level3_column_annual_average_NO2':                            {'value':[], 'standard_name':'OMI level3 column annual average NO2',                             'long_name':'OMI level3 column annual average nitrogen dioxide',                                                           'units':'xx cm-2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'AURA Ozone monitoring instrument (OMI) level3 column annual average NO2, in molecules per squared centimetres. Native resolution of 0.25 x 0.25 degrees.'},         
    'OMI_level3_column_cloud_screened_annual_average_NO2':             {'value':[], 'standard_name':'OMI level3 column cloud screened annual average NO2',              'long_name':'OMI level3 column cloud screened annual average nitrogen dioxide',                                            'units':'xx cm-2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'AURA Ozone monitoring instrument (OMI) level3 column cloud screened (where cloud fraction is less than 30 percent) annual average NO2, in molecules per squared centimetres. Native resolution of 0.25 x 0.25 degrees.'},         
    'OMI_level3_tropospheric_column_annual_average_NO2':               {'value':[], 'standard_name':'OMI level3 tropospheric column annual average NO2',                'long_name':'OMI level3 tropospheric column annual average nitrogen dioxide',                                              'units':'xx cm-2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'AURA Ozone monitoring instrument (OMI) level3 tropospheric column annual average NO2, in molecules per squared centimetres. Native resolution of 0.25 x 0.25 degrees.'},         
    'OMI_level3_tropospheric_column_cloud_screened_annual_average_NO2':{'value':[], 'standard_name':'OMI level3 tropospheric column cloud screened annual average NO2', 'long_name':'OMI level3 tropospheric column cloud screened annual average nitrogen dioxide',                               'units':'xx cm-2',    'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'AURA Ozone monitoring instrument (OMI) level3 tropospheric column cloud screened (where cloud fraction is less than 30 percent) annual average NO2, in molecules per squared centimetres. Native resolution of 0.25 x 0.25 degrees.'},         
    'GSFC_coastline_proximity':                                        {'value':[], 'standard_name':'GSFC coastline proximity',                                         'long_name':'GSFC proximity to the coastline',                                                                             'units':'km',         'data_type':np.float32, 'string_format':np.NaN, 'metadata_type':'GLOBALLY GRIDDED CLASSIFICATIONS', 'identifiers':{'WIGOS_name':''}, 'description':'Proximity to the coastline provided by the NASA Goddard Space Flight Center (GSFC) Ocean Color Group, in kilometres, produced using the Generic Mapping Tools package. Native resolution of 0.01 x 0.01 degrees. Negative distances represent locations over land (including land-locked bodies of water), while positive distances represent locations over the ocean. There is an uncertainty of up to 1 km in the computed distance at any given point.'},         

    #MEASUREMENT INFORMATION
    'primary_sampling_type':                                   {'value':[], 'standard_name':'primary sampling type',                                    'long_name':'standardised primary sampling type',                                    'units':'unitless',                          'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':'SamplingStrategy'},          'description':'Standardised primary sampling type.'},
    'primary_sampling_instrument_name':                        {'value':[], 'standard_name':'primary sampling instrument name',                         'long_name':'standardised primary sampling instrument name',                         'units':'unitless',                          'data_type':object,     'string_format':'short',       'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Standardised name of the primary sampling instrument (if no specific instrument is used, or known, this is the standardised primary sampling type).'},  
    'primary_sampling_instrument_documented_flow_rate':        {'value':[], 'standard_name':'primary sampling instrument documented flow rate',         'long_name':'primary sampling instrument documented sampling flow rate',             'units':'l min-1',                           'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Volume (litres) of fluid which passes to the primary sampling instrument, per unit time (minutes), as given in instrumental manual/documentation. Can be a range: e.g. 1.0-3.0.'},  
    'primary_sampling_instrument_reported_flow_rate':          {'value':[], 'standard_name':'primary sampling instrument reported flow rate',           'long_name':'primary sampling instrument reported sampling flow rate',               'units':'l min-1',                           'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Volume (litres) of fluid which passes to the primary sampling instrument, per unit time (minutes), as given in metadata. Can be a range: e.g. 1.0-3.0.'},
    'primary_sampling_process_details':                        {'value':[], 'standard_name':'primary sampling process details',                         'long_name':'primary sampling process details',                                      'units':'unitless',                          'data_type':object,     'string_format':'lower_long',  'metadata_type':np.NaN,                            'identifiers':{'WIGOS_name':''},                          'description':'Miscellaneous details regarding assumptions made in the standardisation of the primary sampling type/instrument.'},
    'primary_sampling_instrument_manual_name':                 {'value':[], 'standard_name':'primary sampling instrument manual name',                  'long_name':'primary sampling instrument manual name',                               'units':'unitless',                          'data_type':object,     'string_format':'short',       'metadata_type':np.NaN,                            'identifiers':{'WIGOS_name':''},                          'description':'Path to the location in the esarchive of the manual for the specific primary sampling instrument.'},
    'primary_sampling_further_details':                        {'value':[], 'standard_name':'primary sampling further details',                         'long_name':'primary sampling further details',                                      'units':'unitless',                          'data_type':object,     'string_format':'lower_long',  'metadata_type':np.NaN,                            'identifiers':{'WIGOS_name':''},                          'description':'Further associated details regarding the specifics of the primary sampling instrument/type.'},
    'sample_preparation_types':                                {'value':[], 'standard_name':'sample preparation types',                                 'long_name':'standardised sample preparation types',                                 'units':'unitless',                          'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':'SampleTreatment'},           'description':'Standardised sample preparation types utilised in the measurement process. Multiple types are separated by ";".'},               
    'sample_preparation_techniques':                           {'value':[], 'standard_name':'sample preparation techniques',                            'long_name':'standardised specific preparation techniques',                          'units':'unitless',                          'data_type':object,     'string_format':'short',       'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Standardised sample preparation techniques utilised in the measurement process. Multiple names are separated by ";".'},               
    'sample_preparation_process_details':                      {'value':[], 'standard_name':'sample preparation process details',                       'long_name':'sample preparation process details',                                    'units':'unitless',                          'data_type':object,     'string_format':'lower_long',  'metadata_type':np.NaN,                            'identifiers':{'WIGOS_name':''},                          'description':'Miscellaneous details regarding assumptions made in the standardisation of the sample preparation types/techniques. Multiple details specific to different types are separated by ";".'},
    'sample_preparation_further_details':                      {'value':[], 'standard_name':'sample preparation further details',                       'long_name':'sample preparation further details',                                    'units':'unitless',                          'data_type':object,     'string_format':'lower_long',  'metadata_type':np.NaN,                            'identifiers':{'WIGOS_name':''},                          'description':'Further associated details regarding the specifics of the sample preparation types/techniques. Multiple details specific to different types are separated by ";".'},
    'measurement_methodology':                                 {'value':[], 'standard_name':'measurement methodology',                                  'long_name':'standardised measurement methodology',                                  'units':'unitless',                          'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':'ObservingMethodAtmosphere'}, 'description':'Standardised name of the measurement methodology.'},               
    'measuring_instrument_name':                               {'value':[], 'standard_name':'measuring instrument name',                                'long_name':'standardised measuring instrument name',                                'units':'unitless',                          'data_type':object,     'string_format':'short',       'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Standardised name of the measuring instrument.'},  
    'measuring_instrument_sampling_type':                      {'value':[], 'standard_name':'measuring instrument sampling type',                       'long_name':'standardised sampling type of the measuring instrument',                'units':'unitless',                          'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':'SamplingStrategy'},          'description':'Standardised name of the measuring instrument sampling type.'},     
    'measuring_instrument_documented_flow_rate':               {'value':[], 'standard_name':'measuring instrument documented flow rate',                'long_name':'measuring instrument documented flow rate',                             'units':'l min-1',                           'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Volume (litres) of fluid which passes to the measuring instrument, per unit time (minutes), as given in instrumental manual/documentation. Can be a range: e.g. 1.0-3.0.'}, 
    'measuring_instrument_reported_flow_rate':                 {'value':[], 'standard_name':'measuring instrument reported flow rate',                  'long_name':'measuring instrument reported flow rate',                               'units':'l min-1',                           'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Volume (litres) of fluid which passes to the measuring instrument, per unit time (minutes), as given in metadata. Can be a range: e.g. 1.0-3.0.'},  
    'measuring_instrument_process_details':                    {'value':[], 'standard_name':'measuring instrument process details',                     'long_name':'measuring instrument process details',                                  'units':'unitless',                          'data_type':object,     'string_format':'lower_long',  'metadata_type':np.NaN,                            'identifiers':{'WIGOS_name':''},                          'description':'Miscellaneous details regarding assumptions made in the standardisation of the measurement methodology/instrument.'},
    'measuring_instrument_manual_name':                        {'value':[], 'standard_name':'measuring instrument manual name',                         'long_name':'measuring instrument manual name',                                      'units':'unitless',                          'data_type':object,     'string_format':'short',       'metadata_type':np.NaN,                            'identifiers':{'WIGOS_name':''},                          'description':'Path to the location in the esarchive of the manual for the specific measuring instrument.'},
    'measuring_instrument_further_details':                    {'value':[], 'standard_name':'measuring instrument further details',                     'long_name':'measuring instrument further details',                                  'units':'unitless',                          'data_type':object,     'string_format':'lower_long',  'metadata_type':np.NaN,                            'identifiers':{'WIGOS_name':''},                          'description':'Further associated details regarding the specifics of the measurement methodology/instrument.'},
    'measuring_instrument_reported_units':                     {'value':[], 'standard_name':'measuring instrument reported units',                      'long_name':'measuring instrument reported measurement units',                       'units':'unitless',                          'data_type':object,     'string_format':'short',       'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Units that the measured parameter are natively reported in.'}, 
    'measuring_instrument_reported_lower_limit_of_detection':  {'value':[], 'standard_name':'measuring instrument reported lower limit of detection',   'long_name':'measuring instrument reported lower limit of detection',                'units':parameter_details['standard_units'], 'data_type':np.float32, 'string_format':np.NaN,        'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Lower limit of detection of measurement methodology, as given in metadata.'},
    'measuring_instrument_documented_lower_limit_of_detection':{'value':[], 'standard_name':'measuring instrument documented lower limit of detection', 'long_name':'measuring instrument documented lower limit of detection',              'units':parameter_details['standard_units'], 'data_type':np.float32, 'string_format':np.NaN,        'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Lower limit of detection of measurement methodology, as given in the instrumental manual/documentation.'}, 
    'measuring_instrument_reported_upper_limit_of_detection':  {'value':[], 'standard_name':'measuring instrument reported upper limit of detection',   'long_name':'measuring instrument reported upper limit of detection',                'units':parameter_details['standard_units'], 'data_type':np.float32, 'string_format':np.NaN,        'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Upper limit of detection of measurement methodology, as given in metadata.'},
    'measuring_instrument_documented_upper_limit_of_detection':{'value':[], 'standard_name':'measuring instrument documented upper limit of detection', 'long_name':'measuring instrument documented upper limit of detection',              'units':parameter_details['standard_units'], 'data_type':np.float32, 'string_format':np.NaN,        'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Upper limit of detection of measurement methodology, as given in the instrumental manual/documentation.'}, 
    'measuring_instrument_reported_uncertainty':               {'value':[], 'standard_name':'measuring instrument reported measurement uncertainty',    'long_name':'measuring instrument reported measurement uncertainty',                 'units':parameter_details['standard_units'], 'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Measurement uncertainty (±), as given in metadata. In principal this refers to the inherent uncertainty on every measurement as a function of the quadratic addition of the accuracy and precision metrics (at the same confidence interval), but is often reported incosistently e.g. being solely determined from random errors (i.e. just the measurement precision). It can be given in absolute terms; as a percentage; the greater of either an absolute value or percentage (i.e. 25.0/0.5%); or a percentage quantity after a fixed limit (i.e. 0.5%>=50).'},
    'measuring_instrument_documented_uncertainty':             {'value':[], 'standard_name':'measuring instrument documented measurement uncertainty',  'long_name':'measuring instrument documented measurement uncertainty',               'units':parameter_details['standard_units'], 'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Measurement uncertainty (±), as given in the instrumental manual/documentation. In principal this refers to the inherent uncertainty on every measurement as a function of the quadratic addition of the accuracy and precision metrics (at the same confidence interval), but is often reported incosistently e.g. being solely determined from random errors (i.e. just the measurement precision). This can be given in absolute terms; as a percentage; the greater of either an absolute value or percentage (i.e. 25.0/0.5%); or a percentage quantity after a fixed limit (i.e. 0.5%>=50).'},
    'measuring_instrument_reported_accuracy':                  {'value':[], 'standard_name':'measuring instrument reported measurement accuracy',       'long_name':'measuring instrument reported measurement accuracy',                    'units':parameter_details['standard_units'], 'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Measurement accuracy (±), as given in metadata. Accuracy describes the difference between the measurement and the actual value of the part that is measured. It includes: Bias (a measure of the difference between the true value and the observed value of a part -- If the “true” value is unknown, it can be calculated by averaging several measurements with the most accurate measuring equipment available) and Linearity (a measure of how the size of the part affects the bias of a measurement system -- It is the difference in the observed bias values through the expected range of measurement). This can be given as in absolute terms; as a percentage; the greater of either an absolute value or percentage (i.e. 25.0/0.5%); or a percentage quantity after a fixed limit (i.e. 0.5%>=50).'}, 
    'measuring_instrument_documented_accuracy':                {'value':[], 'standard_name':'measuring instrument documented measurement accuracy',     'long_name':'measuring instrument documented measurement accuracy',                  'units':parameter_details['standard_units'], 'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Measurement accuracy (±), as given in the instrumental manual/documentation. Accuracy describes the difference between the measurement and the actual value of the part that is measured. It includes: Bias (a measure of the difference between the true value and the observed value of a part -- If the “true” value is unknown, it can be calculated by averaging several measurements with the most accurate measuring equipment available) and Linearity (a measure of how the size of the part affects the bias of a measurement system -- It is the difference in the observed bias values through the expected range of measurement). This can be given as in absolute terms; as a percentage; the greater of either an absolute value or percentage (i.e. 25.0/0.5%); or a percentage quantity after a fixed limit (i.e. 0.5%>=50).'}, 
    'measuring_instrument_reported_precision':                 {'value':[], 'standard_name':'measuring instrument reported measurement precision',      'long_name':'measuring instrument reported measurement precision',                   'units':parameter_details['standard_units'], 'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Measurement precision (±), as given in metadata. Precision describes the variation you see when you measure the same part repeatedly with the same device. It includes the following two types of variation: Repeatability (variation due to the measuring device -- it is the variation observed when the same operator measures the same part repeatedly with the same device) and Reproducibility (variation due to the operators and the interaction between operator and part -- It is the variation of the bias observed when different operators measure the same parts using the same device). This can be given as in absolute terms; as a percentage; the greater of either an absolute value or percentage (i.e. 25.0/0.5%); or a percentage quantity after a fixed limit (i.e. 0.5%>=50).'}, 
    'measuring_instrument_documented_precision':               {'value':[], 'standard_name':'measuring instrument documented measurement precision',    'long_name':'measuring instrument documented measurement precision',                 'units':parameter_details['standard_units'], 'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Measurement precision (±), as given in instrumental manual/documentation. Precision describes the variation you see when you measure the same part repeatedly with the same device. It includes the following two types of variation: Repeatability (variation due to the measuring device -- it is the variation observed when the same operator measures the same part repeatedly with the same device) and Reproducibility (variation due to the operators and the interaction between operator and part -- It is the variation of the bias observed when different operators measure the same parts using the same device). This can be given as in absolute terms; as a percentage; the greater of either an absolute value or percentage (i.e. 25.0/0.5%); or a percentage quantity after a fixed limit (i.e. 0.5%>=50).'}, 
    'measuring_instrument_reported_zero_drift':                {'value':[], 'standard_name':'measuring instrument reported zero drift',                 'long_name':'measuring instrument reported zero drift',                              'units':parameter_details['standard_units'], 'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Zero drift of measuring instrument per unit of time, as given in metadata. Zero drift (or baseline drift) refers to the shifting of the whole calibration by the same amount caused by slippage or due to undue warming up of the electronic circuits. It is reported as the maximum possible drift per unit of time in absolute terms; as a percentage; the greater of either an absolute value or percentage (i.e. 25.0/0.5%/day); or a percentage quantity after a fixed limit (i.e. 0.5%>=50/day).'}, 
    'measuring_instrument_documented_zero_drift':              {'value':[], 'standard_name':'measuring instrument documented zero drift',               'long_name':'measuring instrument documented zero drift',                            'units':parameter_details['standard_units'], 'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Zero drift of measuring instrument per unit of time, as given in instrumental manual/documentation. Zero drift (or baseline drift) refers to the shifting of the whole calibration by the same amount caused by slippage or due to undue warming up of the electronic circuits. It is reported as the maximum possible drift per unit of time in absolute terms; as a percentage; the greater of either an absolute value or percentage (i.e. 25.0/0.5%/day); or a percentage quantity after a fixed limit (i.e. 0.5%>=50/day).'}, 
    'measuring_instrument_reported_span_drift':                {'value':[], 'standard_name':'measuring instrument reported span drift',                 'long_name':'measuring instrument reported span drift',                              'units':parameter_details['standard_units'], 'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Span drift of measuring instrument per unit of time, as given in metadata. Span drift (or sensitivity drift) refers to when there is proportional change in the indication of an instrument all along the upward scale, hence higher calibrations end up being shifted more than lower calibrations. It is reported as the maximum possible drift per unit of time in absolute terms; as a percentage; the greater of either an absolute value or percentage (i.e. 25.0/0.5%/day); or a percentage quantity after a fixed limit (i.e. 0.5%>=50/day).'}, 
    'measuring_instrument_documented_span_drift':              {'value':[], 'standard_name':'measuring instrument documented span drift',               'long_name':'measuring instrument documented span drift',                            'units':parameter_details['standard_units'], 'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Span drift of measuring instrument per unit of time, as given in instrumental manual/documentation. Span drift (or sensitivity drift) refers to when there is proportional change in the indication of an instrument all along the upward scale, hence higher calibrations end up being shifted more than lower calibrations. It is reported as the maximum possible drift per unit of time in absolute terms; as a percentage; the greater of either an absolute value or percentage (i.e. 25.0/0.5%/day); or a percentage quantity after a fixed limit (i.e. 0.5%>=50/day).'}, 
    'measuring_instrument_reported_zonal_drift':               {'value':[], 'standard_name':'measuring instrument reported zonal drift',                'long_name':'measuring instrument reported zonal drift',                             'units':parameter_details['standard_units'], 'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Zonal drift of measuring instrument per unit of time, as given in metadata. Zonal drift refers to when drift occurs only over a portion of the full scale or span of an instrument, while the remaining portion of the scale remains unaffected. It is reported as the maximum possible drift per unit of time in absolute terms; as a percentage; the greater of either an absolute value or percentage (i.e. 25.0/0.5%/day); or a percentage quantity after a fixed limit (i.e. 0.5%>=50/day).'}, 
    'measuring_instrument_documented_zonal_drift':             {'value':[], 'standard_name':'measuring instrument documented zonal drift',              'long_name':'measuring instrument documented zonal drift',                           'units':parameter_details['standard_units'], 'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Zonal drift of measuring instrument per unit of time, as given in instrumental manual/documentation. Zonal drift refers to when drift occurs only over a portion of the full scale or span of an instrument, while the remaining portion of the scale remains unaffected. It is reported as the maximum possible drift per unit of time in absolute terms; as a percentage; the greater of either an absolute value or percentage (i.e. 25.0/0.5%/day); or a percentage quantity after a fixed limit (i.e. 0.5%>=50/day).'}, 
    'measuring_instrument_reported_measurement_resolution':    {'value':[], 'standard_name':'measuring instrument reported measurement resolution',     'long_name':'measuring instrument reported measurement resolution',                  'units':parameter_details['standard_units'], 'data_type':np.float32, 'string_format':np.NaN,        'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Measurement resolution, as given in metadata. The measurement resolution is defined as the smallest change or increment in the measured quantity that the instrument can detect. However it is often reported inconsistently, often being simply the number of digits an instrument can display, which does not relate to the actual physical resolution of the instrument.'}, 
    'measuring_instrument_documented_measurement_resolution':  {'value':[], 'standard_name':'measuring instrument documented measurement resolution',   'long_name':'measuring instrument documented measurement resolution',                'units':parameter_details['standard_units'], 'data_type':np.float32, 'string_format':np.NaN,        'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Measurement resolution, as given in instrumental manual/documentation. The measurement resolution is defined as the smallest change or increment in the measured quantity that the instrument can detect. However it is often reported inconsistently, often being simply the number of digits an instrument can display, which does not relate to the actual physical resolution of the instrument.'}, 
    'measuring_instrument_reported_absorption_cross_section':  {'value':[], 'standard_name':'measuring instrument reported absorption cross section',   'long_name':'measuring instrument reported absorption cross section',                'units':'cm2',                               'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Assumed molecule cross-section for parameter being measured (in cm2/molecule), as given in metadata. This field is only used for parameters being measured using optical methods, where a molecule cross section is assumed for processing the measurement values. Physically it is the effective area of the molecule that photon needs to traverse in order to be absorbed. The larger the absorption cross section, the easier it is to photoexcite the molecule. Can be a range: e.g. 1e-15-1.5e-15.'}, 
    'measuring_instrument_documented_absorption_cross_section':{'value':[], 'standard_name':'measuring instrument documented absorption cross section', 'long_name':'measuring instrument documented absorption cross section',              'units':'cm2',                               'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Assumed molecule cross-section for parameter being measured (in cm2/molecule), as given in instrumental manual/documentation. This field is only used for parameters being measured using optical methods, where a molecule cross section is assumed for processing the measurement values. Physically it is the effective area of the molecule that photon needs to traverse in order to be absorbed. The larger the absorption cross section, the easier it is to photoexcite the molecule. Can be a range: e.g. 1e-15-1.5e-15.'}, 
    'measuring_instrument_inlet_information':                  {'value':[], 'standard_name':'measuring instrument inlet information',                   'long_name':'measuring instrument measurement inlet information',                    'units':'unitless',                          'data_type':object,     'string_format':'lower_long',  'metadata_type':np.NaN,                            'identifiers':{'WIGOS_name':''},                          'description':'Description of sampling inlet of the measuring instrument.'}, 
    'measuring_instrument_calibration_scale':                  {'value':[], 'standard_name':'measuring instrument calibration scale',                   'long_name':'measuring instrument calibration scale name',                           'units':'unitless',                          'data_type':object,     'string_format':'upper_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'Name of calibration scale used for the calibration of the measuring instrument.'},
    'network_provided_volume_standard_temperature':            {'value':[], 'standard_name':'network provided volume standard temperature',             'long_name':'network provided volume standard temperature',                          'units':'K',                                 'data_type':np.float64, 'string_format':np.NaN,        'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'The temperature (in Kelvin) associated with the volume of the sampled gas (which varies with temperature and pressure). This volume is typically normalised in-instrument to a standard temperature and pressure. These standard values typically follow network/continental/global standards (e.g. European Union) for the measured component. If no in-instrument normalisation is done then the reported temperature should be reported as the internal temperature of the instrument (i.e. the measurement conditions). If no numbers are reported explicitly per measurement, then the sample gas temperature is assumed to be the known network standard temperature for the measured component.'},  
    'network_provided_volume_standard_pressure':               {'value':[], 'standard_name':'network provided volume standard pressure',                'long_name':'network provided volume standard pressure',                             'units':'hPa',                               'data_type':np.float64, 'string_format':np.NaN,        'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'The pressure (in hPa) associated with the volume of the sampled gas (which varies with temperature and pressure). This volume is typically normalised in-instrument to a standard temperature and pressure. These standard values typically follow network/continental/global standards (e.g. European Union) for the measured component. If no in-instrument normalisation is done then the reported pressure should be reported as the internal pressure of the instrument (i.e. the measurement conditions). If no numbers are reported explicitly per measurement, then the sample gas pressure is assumed to be the known network standard pressure for the measured component.'},  
    'retrieval_algorithm':                                     {'value':[], 'standard_name':'retrieval algorithm',                                      'long_name':'retrieval algorithm',                                                   'units':'unitless',                          'data_type':object,     'string_format':'lower_short', 'metadata_type':'MEASUREMENT PROCESS INFORMATION', 'identifiers':{'WIGOS_name':''},                          'description':'The name of the retrieval algorithm. Remote sensing algorithms are used to retrieve the aerosol optical properties (as aerosol optical depths or single scattering albedo among others) using remote-sensing radiances for multiple wavelengths from ground stations or on satellite platforms. Each algorithm is particularly designed considering the characteristics of the sensor and other ancillary information.'},  
    
    #CONTACT INFORMATION
    'principal_investigator_name':         {'value':[], 'standard_name':'principal investigator name',          'long_name':'principal investigator name',          'units':'unitless', 'data_type':object, 'string_format':'title_short', 'metadata_type':'STATION MISCELLANEOUS', 'identifiers':{'WIGOS_name':''},                   'description':'Full name of the principal scientific investigator for the specific reported data.'}, 
    'principal_investigator_institution':  {'value':[], 'standard_name':'principal investigator institution',   'long_name':'principal investigator institution',   'units':'unitless', 'data_type':object, 'string_format':'short',       'metadata_type':'STATION MISCELLANEOUS', 'identifiers':{'WIGOS_name':''},                   'description':'Institution of the principal scientific investigator for the specific reported data.'}, 
    'principal_investigator_email_address':{'value':[], 'standard_name':'principal investigator email address', 'long_name':'principal investigator email address', 'units':'unitless', 'data_type':object, 'string_format':'short',       'metadata_type':'STATION MISCELLANEOUS', 'identifiers':{'WIGOS_name':''},                   'description':'Email address of the principal scientific investigator for the specific reported data.'}, 
    'contact_name':                        {'value':[], 'standard_name':'contact name',                         'long_name':'contact name',                         'units':'unitless', 'data_type':object, 'string_format':'title_short', 'metadata_type':'STATION MISCELLANEOUS', 'identifiers':{'WIGOS_name':''},                   'description':'Full name of the principal data contact for the specific reported data.'}, 
    'contact_institution':                 {'value':[], 'standard_name':'contact institution',                  'long_name':'contact institution',                  'units':'unitless', 'data_type':object, 'string_format':'short',       'metadata_type':'STATION MISCELLANEOUS', 'identifiers':{'WIGOS_name':'ProgramAffiliation'}, 'description':'Institution of the principal data contact for the specific reported data.'}, 
    'contact_email_address':               {'value':[], 'standard_name':'contact email address',                'long_name':'contact email address',                'units':'unitless', 'data_type':object, 'string_format':'short',       'metadata_type':'STATION MISCELLANEOUS', 'identifiers':{'WIGOS_name':''},                   'description':'Email address of the principal data contact for the specific reported data.'}, 

    #TIMESTAMPS
    'meta_update_stamp':  {'value':[], 'standard_name':'metdata update timestamp', 'long_name':'metdata update timestamp', 'units':'unitless', 'data_type':np.uint32, 'string_format':np.NaN, 'metadata_type':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Time stamp of metadata updates in integer minutes from 0001-01-01 00:00 UTC.'}, 
    'data_download_stamp':{'value':[], 'standard_name':'data download timestamp',  'long_name':'data download timestamp',  'units':'unitless', 'data_type':np.uint32, 'string_format':np.NaN, 'metadata_type':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Time stamp of date/time of data download in integer minutes from 0001-01-01 00:00 UTC.'}, 
    'data_revision_stamp':{'value':[], 'standard_name':'data revision timestamp',  'long_name':'data revision timestamp',  'units':'unitless', 'data_type':np.uint32, 'string_format':np.NaN, 'metadata_type':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Time stamp of date/time of the last data revision in integer minutes from 0001-01-01 00:00 UTC.'}, 

    #FURTHER DETAIL
    'network_sampling_details':     {'value':[], 'standard_name':'network sampling details',      'long_name':'network sampling details',      'units':'unitless', 'data_type':object, 'string_format':'lower_long', 'metadata_type':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Extra details provided by the reporting network about the sampling methods employed.'},
    'network_uncertainty_details':  {'value':[], 'standard_name':'network uncertainty details',   'long_name':'network uncertainty details',   'units':'unitless', 'data_type':object, 'string_format':'lower_long', 'metadata_type':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Extra details provided by the reporting network about the uncertainties involved with the measurement methods employed.'},
    'network_maintenance_details':  {'value':[], 'standard_name':'network maintenance details',   'long_name':'network maintenance details',   'units':'unitless', 'data_type':object, 'string_format':'lower_long', 'metadata_type':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Extra details provided by the reporting network about the operational maintenance done at the station.'},
    'network_qa_details':           {'value':[], 'standard_name':'network qa details',            'long_name':'network qa details',            'units':'unitless', 'data_type':object, 'string_format':'lower_long', 'metadata_type':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Extra details provided by the reporting network about the in-network quality assurance of measurements.'},
    'network_miscellaneous_details':{'value':[], 'standard_name':'network miscellaneous details', 'long_name':'network miscellaneous details', 'units':'unitless', 'data_type':object, 'string_format':'lower_long', 'metadata_type':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Extra miscellaneous details provided by the reporting network.'},
    
    #DATA LICENCE
    'data_licence':{'value':[], 'standard_name':'data licence', 'long_name':'data licence', 'units':'unitless', 'data_type':object, 'string_format':'long', 'metadata_type':np.NaN, 'identifiers':{'WIGOS_name':'DataPolicy'}, 'description':'Information pertaining to the data licence governing the redistribution/publication of the ingested network data.'},
    
    #WARNINGS
    'process_warnings':{'value':[], 'standard_name':'process warnings', 'long_name':'process warnings', 'units':'unitless', 'data_type':object, 'string_format':'lower_long', 'metadata_type':np.NaN, 'identifiers':{'WIGOS_name':''}, 'description':'Warnings accumulated through GHOST processing regarding the data that should be considered.'}    
    }

    return standard_metadata

###--------------------------------------------------------------------------------------------------###
###DATA REPORTER PROVIDED DATA FLAG STANDARDISATIONS
###--------------------------------------------------------------------------------------------------###

#Define dictionary associating standardised data reporter provided data flag names with unique data flag codes

#These standardised data flags refer specifically to information/or lack of information provided by the data provider to fill the 'flag' field,
#and not to post-processing/quality control of observations (given separately by the ‘qa’ field).
    
standard_data_flag_name_to_data_flag_code = {

#Basic Data Flags
#-----------------------------------------------------
'Valid Data': 0,

'Preliminary Data': 1,

'Missing Data': 2, 

'Invalid Data - Unspecified': 3,

'Un-Flagged Data': 4,
#-----------------------------------------------------


#Estimated Data Flags 
#-----------------------------------------------------
'Estimated Data - Unspecified': 10,

'Estimated Data - Measured Negative Value': 11,

'Estimated Data - No Value Detected': 12,

'Estimated Data - Value Below Detection Limit': 13,

'Estimated Data - Value Above Detection Limit': 14,

'Estimated Data - Value Substituted from Secondary Monitor': 15,

'Estimated Data - Multiple Parameters Aggregated': 16,
#-----------------------------------------------------


#Extreme/Irregular Data Flags
#-----------------------------------------------------
'Extreme/Irregular Data - Unspecified': 20,

'Data Does Not Meet Internal Network Quality Control Criteria': 21,

'High Variability of Data': 22,

'Irregular Data Manually Screened and Accepted': 23,

'Irregular Data Manually Screened and Rejected': 24,

'Negative Value': 25,

'No Value Detected': 26,
    
'Reconstructed/Recalculated Data': 27,

'Value Close to Detection Limit': 28,

'Value Below Acceptable Range': 29,

'Value Above Acceptable Range': 30,

'Value Below Detection Limit': 31,

'Value Above Detection Limit': 32,

#-----------------------------------------------------


#Measurement Issue Data Flags
#-----------------------------------------------------
'Measurement Issue - Unspecified': 40,

'Chemical Issue': 41,

'Erroneous Sampling Operation': 42,

'Extreme Internal Instrument Meteorological Conditions': 43,

'Extreme Ambient Laboratory Meteorological Conditions': 44,

'Extreme External Meteorological Conditions': 45,

'Extreme Sample Transport Conditions': 46,

'Invalid Flow Rate': 47,

'Human Error': 48,

'Matrix Effect': 49,

'Mechanical Issue/Non-Operational Equipment': 50,

'No Technician': 51,

'Operational Maintenance Check Issue': 52,

'Physical Issue With Filter': 53,

'Power Failure': 54,

'Sample Diluted for Analysis': 55,

'Unmeasured Key Meteorological Parameter': 56,

'Sample Not Analysed': 57,

#-----------------------------------------------------


#Operational Maintenance Data Flags
#-----------------------------------------------------
'Operational Maintenance - Unspecified': 60,

'Calibration': 61,

'Accuracy Check': 62,

'Blank Check': 63,

'Detection Limits Check': 64,

'Precision Check': 65,

'Retention Time Check': 66,

'Span Check': 67,

'Zero Check': 68,

'Instrumental Inspection': 69,

'Instrumental Repair': 70,

'Quality Control Audit': 71,
#-----------------------------------------------------


#Data Formatting/Processing Issue Data Flags
#-----------------------------------------------------
'Data Formatting/Processing Issue': 80,

'Corrected Data Formatting/Processing Issue': 81,
#-----------------------------------------------------


#Aggregation/Representation Flags
#-----------------------------------------------------
'Aggregation/Representation Issue - Unspecified': 90,

'Data Window Completeness < 90%': 91,

'Data Window Completeness < 75%': 92,

'Data Window Completeness < 66%': 93,

'Data Window Completeness < 50%': 94,

'Data Window Completeness < 25%': 95,

'>= 75% of Measurements in Window Below Detection Limit': 96,

'>= 50% of Measurements in Window Below Detection Limit': 97,
#-----------------------------------------------------


# Weather Condition Flags
#-----------------------------------------------------
'No Significant Weather': 100,

'Precipitation - Unspecified Intensity': 101,

'Precipitation - Light': 102,

'Precipitation - Moderate': 103,

'Precipitation - Heavy': 104,

'Drizzle - Unspecified Intensity': 105,

'Drizzle - Light': 106,

'Drizzle - Moderate': 107,

'Drizzle - Heavy': 108,

'Freezing Drizzle - Unspecified Intensity': 109,

'Freezing Drizzle - Light': 110,

'Freezing Drizzle - Moderate': 111,

'Freezing Drizzle - Heavy': 112,

'Rain - Unspecified Intensity': 113,

'Rain - Light': 114,

'Rain - Moderate': 115,

'Rain - Heavy': 116,

'Rain Shower/s - Unspecified Intensity': 117,

'Rain Shower/s - Light': 118,

'Rain Shower/s - Moderate': 119,

'Rain Shower/s - Heavy': 120,

'Freezing Rain - Unspecified Intensity': 121,

'Freezing Rain - Light': 122,

'Freezing Rain - Moderate': 123,

'Freezing Rain - Heavy': 124,

'Freezing Rain Shower/s - Unspecified Intensity': 125,

'Freezing Rain Shower/s - Light': 126,

'Freezing Rain Shower/s - Moderate': 127,

'Freezing Rain Shower/s - Heavy': 128,

'Snow - Unspecified Intensity': 129,

'Snow - Light': 130,

'Snow - Moderate': 131,

'Snow - Heavy': 132,

'Snow Shower/s - Unspecified Intensity': 133,

'Snow Shower/s - Light': 134,

'Snow Shower/s - Moderate': 135,

'Snow Shower/s - Heavy': 136,

'Hail - Unspecified Intensity': 137,

'Hail - Light': 138,

'Hail - Moderate': 139,

'Hail - Heavy': 140,

'Hail Shower/s - Unspecified Intensity': 141,

'Hail Shower/s - Light': 142,

'Hail Shower/s - Moderate': 143,

'Hail Shower/s - Heavy': 144,

'Ice Pellets - Unspecified Intensity': 145,

'Ice Pellets - Light': 146,

'Ice Pellets - Moderate': 147,

'Ice Pellets - Heavy': 148,

'Ice Pellets Shower/s - Unspecified Intensity': 149,

'Ice Pellets Shower/s - Light': 150,

'Ice Pellets Shower/s - Moderate': 151,

'Ice Pellets Shower/s - Heavy': 152,

'Snow Pellets - Unspecified Intensity': 153,

'Snow Pellets - Light': 154,

'Snow Pellets - Moderate': 155,

'Snow Pellets - Heavy': 156,

'Snow Pellets Shower/s - Unspecified Intensity': 157,

'Snow Pellets Shower/s - Light': 158,

'Snow Pellets Shower/s - Moderate': 159,

'Snow Pellets Shower/s - Heavy': 160,

'Snow Grains - Unspecified Intensity': 161,

'Snow Grains - Light': 162,

'Snow Grains - Moderate': 163,

'Snow Grains - Heavy': 164,

'Diamond Dust - Unspecified Intensity': 165,

'Diamond Dust - Light': 166,

'Diamond Dust - Moderate': 167,

'Diamond Dust - Heavy': 168,

'Glaze': 169,

'Rime': 170,

'Thunderstorm': 171,

'Funnel Cloud/s': 172,

'Squalls': 173,

'Tropical Cyclone (Cyclone/Hurricane/Typhoon)': 174,

'Duststorm': 175,

'Sandstorm': 176,

'Dust/Sand Whirls': 177,

'High Winds': 178,
#-----------------------------------------------------


# Atmospheric Obscuration/Local Area Contamination Flags
#-----------------------------------------------------

'No Atmospheric Obscuration': 180,

'Atmospheric Obscuration - Unknown': 181,

'Dust': 182,

'Blowing Dust': 183,

'Drifting Dust': 184,

'Sand': 185,

'Blowing Sand': 186,

'Drifting Sand': 187,

'Blowing Snow': 188,

'Drifting Snow': 189,

'Fog': 190,

'Freezing Fog': 191,

'Ground Fog': 192,

'Ice Fog': 193,

'Haze': 194,

'Mist': 195,

'Sea Spray': 196,

'Smoke': 197,

'Volcanic Ash': 198,

'No Local Contamination': 199,

'Local Contamination - Unspecified': 200,

'Agricultural Contamination': 201,

'Bird-Dropping Contamination': 202,

'Construction Contamination': 203,

'Industrial Contamination': 204,

'Insect Contamination': 205,

'Internal Laboratory/Instrument Contamination': 206,

'Pollen/Leaf Contamination': 207,

'Traffic Contamination': 208,
#-----------------------------------------------------


#Exceptional Event Data Flags
#-----------------------------------------------------
'Exceptional Event - Unspecified': 210,

#Natural Events
#---------------------------

'Seismic Activity': 211,

'Stratospheric Ozone Intrusion': 212,

'Volcanic Eruptions': 213, 

'Wildfire': 214,

#Anthropogenically Induced Events
#---------------------------
'Chemical Spill/Industrial Accident': 220,

'Cleanup After a Major Disaster': 221,

'Demolition': 222,

'Fireworks': 223,
    
'Infrequent Large Gathering': 224,

'Terrorist Act': 225,
#-----------------------------------------------------


# Meteorological Measurement Flags
#-----------------------------------------------------

'Visibility Distance Unlimited': 230,

'Ceiling Height Unlimited': 231
#-----------------------------------------------------

}

#Define dictionary associating standardised simplified data reporter provided data flag names with unique data flag codes

#These simplified data flags refer specifically to information/or lack of information provided by the data provider to fill the 'flag_simple' field.
#The template for these flags is taken from WaterML2, see http://codes.wmo.int/wmdr/_WaterML2_0

standard_simple_data_flag_name_to_simple_data_flag_code = {
#The data is an estimate only, not a direct measurement. 
'estimate': 0,
#The data has been examined and represents a reliable measurement. 
'good':1,
#The data is missing.
'missing':2,
#The data should be considered as low quality and may have been rejected. 
'poor':3,
#The data should be treated as suspect.
'suspect':4,
#The data has not been checked by any qualitative or quantitative method. 
'unchecked':5

}

###--------------------------------------------------------------------------------------------------###
###QUALITY ASSURANCE FLAGS STANDARDISATIONS
###--------------------------------------------------------------------------------------------------###

#Define dictionary associating standardised quality assurance (QA) flag names with unique QA flag codes

standard_QA_name_to_QA_code = {

#Most Common QA Flags
#-----------------------------------------------------
#Missing Measurement
#Measurement is missing (i.e. NaN).
'Missing Measurement': 0,

#Infinite Value
#Value is infinite -- happens when data values are outside of the range that the float32 data type can handle (-3.4E+38 to +3.4E+38).
'Infinite Value': 1,

#Negative Measurement
#Measurement is negative in absolute terms.
'Negative Measurement': 2,

#Zero Measurement
#Have measurement equal to zero.
'Zero Measurement': 3,

#Not Maximum Data Quality Level
#Measurement is not of highest data quality level available from data provider
'Not Maximum Data Quality Level': 4,

#Preliminary Data
#Measurement has been flagged by data provider to be preliminary in nature (i.e. pending review)
'Preliminary Data': 5,

#Invalid Data Provider Flags - GHOST Decreed
#Measurements are associated with data quality flags given by the data provider which have been decreed by the GHOST project architects to suggest the measurements are associated with substantial uncertainty/bias
'Invalid Data Provider Flags - GHOST Decreed': 6,

#Invalid Data Provider Flags - Network Decreed
#Measurements are associated with data quality flags given by the data provider which have been decreed by the reporting network to suggest the measurements are associated with substantial uncertainty/bias 
'Invalid Data Provider Flags - Network Decreed': 7,

#No Valid Data to Average
#After screening by key QA flags, no valid data remains to average in temporal window
'No Valid Data to Average': 8,

#-----------------------------------------------------


#Measurement Process Flags
#-----------------------------------------------------
#Methodology Not Mapped
#The measurement methodology used has not yet been mapped to standardised dictionaries of measurement methodologies.
'Methodology Not Mapped': 10,

#Assumed Primary Sampling
#A level of assumption has been made in determining the primary sampling type.
'Assumed Primary Sampling': 11,

#Assumed Sample Preparation
#A level of assumption has been made in determining the sample preparation.
'Assumed Sample Preparation': 12,

#Assumed Measurement Methodology
#A level of assumption has been made in determining the measurement methodology. 
'Assumed Measurement Methodology': 13,

#Unknown Primary Sampling Type
#The specific name of the primary sampling type is unknown.
'Unknown Primary Sampling Type': 14,

#Unknown Primary Sampling Instrument
#The specific name of the primary sampling instrument is unknown.
'Unknown Primary Sampling Instrument': 15,

#Unknown Sample Preparation Type
#The specific name of the sample preparation type is unknown.
'Unknown Sample Preparation Type': 16,

#Unknown Sample Preparation Technique
#The specific name of the sample preparation technique is unknown.
'Unknown Sample Preparation Technique': 17, 

#Unknown Measurement Method
#The specific name of the measurement method is unknown.
'Unknown Measurement Method': 18, 

#Unknown Measuring Instrument
#The specific name of measuring instrument is unknown. 
'Unknown Measuring Instrument': 19,

#Erroneous Primary Sampling
#The primary sampling is not appropriate to prepare the specific parameter for subsequent measurement.
'Erroneous Primary Sampling': 20,

#Erroneous Sample Preparation
#The sample preparation is not appropriate to prepare the specific parameter for subsequent measurement.
'Erroneous Sample Preparation': 21,

#Erroneous Measurement Method
#The measurement methodology used is not known to be able to measure the specific parameter. Only do check when known (or have assumed method).
'Erroneous Measurement Methodology': 22,

#Invalid QA Measurement Method 
#The specific measurement methodology has been decreed not to conform to QA standards as the method is not sufficiently proven/ subject to substantial biases/uncertainty. Only do check when known (or have assumed method).
'Invalid QA Measurement Methodology': 23,

#Corrected Parameter
#The measurements for a specific parameter have been corrected, or are of significantly higher quality than other types of measurements
'Corrected Parameter': 24,
#-----------------------------------------------------


#Sample Gas Volume Flags
#-----------------------------------------------------
#Sample Gas Volume - Network Standard
#The volume of the sampled gas (which varies with temperature and pressure) is assumed to be for a known standard network defined temperature and pressure for a component. This happens when no sample gas temperature/pressure are reported explicitly per measurement.
'Sample Gas Volume - Network Standard': 30,

#Sample Gas Volume - Unknown
#The volume of the sampled gas (which varies with temperature and pressure) is unknown. This happens when no sample gas temperature/pressure are reported explicitly per measurement AND there is no known standard network defined temperature and pressure for a component.
'Sample Gas Volume - Unknown': 31,

#Unit Conversion - Network Standard Sample Gas Volume Assumption
#Unit conversion has been done assuming the volume of the sampled gas (which varies with temperature and pressure) is for a known standard network defined temperature and pressure for a component.
'Unit Conversion - Network Standard Sample Gas Volume Assumption': 32,

#Unit Conversion - Educated Guess Sample Gas Volume Assumption
#Unit conversion has been done making an educated guess at the sampled gas temperature/pressure measurements are associated with.
'Unit Conversion - Educated Guess Sample Gas Volume Assumption': 33,
#-----------------------------------------------------


#Positional Metadata Doubt
#-----------------------------------------------------
#Station Position Doubt - DEM Decreed
#There exists significant doubt about the accuracy of the station position, as when cross-referencing the reported station altitude with the altitude from the ASTER v3 DEM (30 metre resolution) the absolute difference returned is >= 50 metres.
'Station Position Doubt - DEM Decreed': 40,

#Station Position Doubt - Manually Decreed
#There exists significant doubt about the accuracy of the station position, determined from empirical/word of mouth evidence
'Station Position Doubt - Manually Decreed': 41,
#-----------------------------------------------------


#Data Product Flags
#-----------------------------------------------------
#Data Product
#Data is a product that has been calculated from multiple components
'Data Product': 45,

#Insufficient Data to Calculate Data Product
#There is insufficient valid data required to calculate data product
'Insufficient Data to Calculate Data Product': 46,
#-----------------------------------------------------


#Local Condition Flags
#-----------------------------------------------------
#Local Precipitation
#There is some form of precipitation present at the measuring location at the time of measurement, reported by the data providers
'Local Precipitation': 50,

#Local Extreme Weather
#There is some form of extreme weather (heavy precipitation, funnel clouds, squalls, tropical cyclones, duststorm, sandstorm, dust/sand whirls, high winds) present at the measuring location at the time of measurement, reported by the data providers
'Local Extreme Weather': 51,

#Local Atmospheric Obscuration
#There is some form of atmospheric obscuration present at the measuring location at the time of measurement, reported by the data providers
'Local Atmospheric Obscuration': 52,

#Local Contamination
#There is some form of local contamination present at the measuring location at the time of measurement, reported by the data providers
'Local Contamination': 53,

#Local Exceptional Event
#There is some kind of exceptional event (both natural and anthropogenically induced) active at the measuring location at the time of measurement, reported by the data providers
'Local Exceptional Event': 54,


#Timezone Flags
#-----------------------------------------------------
#Non-Integer Local Timezone (relative to UTC)
#Local timezone has been determined to be non-integer for time of measurement, relative to UTC.
'Non-Integer Local Timezone (relative to UTC)': 60,

#Timezone Doubt
#Significant doubt exists regarding the timezone of the reported data 
'Timezone Doubt':61,


#Limit of Detection Flags
#-----------------------------------------------------
#Below Documented Lower Limit of Detection
#Measurement is below or equal to the instrumental documented lower limit of detection.
'Below Documented Lower Limit of Detection': 70,

#Below Reported Lower Limit of Detection
#Measurement is below or equal to the network reported lower limit of detection. 
'Below Reported Lower Limit of Detection': 71,

#Below Preferential Lower Limit of Detection
#Measurement is below or equal to the preferential lower limit of detection (reported, and then documented). 
'Below Preferential Lower Limit of Detection': 72,

#Above Documented Upper Limit of Detection
#Measurement is above or equal to the instrumental documented upper limit of detection. 
'Above Documented Upper Limit of Detection': 73,

#Above Reported Upper Limit of Detection
#Measurement is above or equal to the network reported upper limit of detection. 
'Above Reported Upper Limit of Detection': 74,

#Above Preferential Upper Limit of Detection
#Measurement is above or equal to the preferential upper limit of detection (reported, and then documented). 
'Above Preferential Upper Limit of Detection': 75,


#Measurement Resolution Flags
#-----------------------------------------------------
#Insufficient Measurement Resolution - Documented
#The documented resolution of measurement is coarser than a set limit (variable by measured parameter).
'Insufficient Measurement Resolution - Documented': 80,

#Insufficient Measurement Resolution - Reported
#The reported resolution of measurement is coarser than a set limit (variable by measured parameter).
'Insufficient Measurement Resolution - Reported': 81,

#Insufficient Measurement Resolution - Preferential
#The preferential resolution of measurement (reported, and then documented) is coarser than a set limit (variable by measured parameter).
'Insufficient Measurement Resolution - Preferential': 82,

#Insufficient Measurement Resolution - Empirical
#The resolution of measurement is analysed month by month. If the minimum difference between observations is coarser than a set limit (variable by measured parameter), measurements are flagged.
'Insufficient Measurement Resolution - Empirical': 83,


#Recurring Value Flags
#-----------------------------------------------------
#Persistent Recurring Values - 5/6
#Check for persistently recurring values. Check is done by using a moving window of 9 measurements. If 5/6 (i.e. 83.33%) of values in the window are the same then the entire window is flagged.
'Persistent Recurring Values - 5/6': 90,

#Persistent Recurring Values - 9/12
#Check for persistently recurring values. Check is done by using a moving window of 12 measurements. If 9/12 (i.e. 75%) of values in the window are the same, then the entire window is flagged.
'Persistent Recurring Values - 9/12': 91,

#Persistent Recurring Values - 16/24
#Check for persistently recurring values. Check is done by using a moving window of 24 measurements. If 16/24 (i.e. 66.66%) of values in the window are the same, then the entire window is flagged.
'Persistent Recurring Values - 16/24': 92,


#Monthly Fractional Unique Value Flags
#-----------------------------------------------------
#Monthly Fractional Unique Values <= 1% 
#Fraction of unique values in a UTC month is <= 1% of total valid values in period 
'Monthly Fractional Unique Values <= 1%': 100,

#Monthly Fractional Unique Values <= 5% 
#Fraction of unique values in a UTC month is <= 5% of total valid values in period 
'Monthly Fractional Unique Values <= 5%': 101,

#Monthly Fractional Unique Values <= 10% 
#Fraction of unique values in a UTC month is <= 10% of total valid values in period 
'Monthly Fractional Unique Values <= 10%': 102,

#Monthly Fractional Unique Values <= 30% 
#Fraction of unique values in a UTC month is <= 30% of total valid values in period 
'Monthly Fractional Unique Values <= 30%': 103,

#Monthly Fractional Unique Values <= 50% 
#Fraction of unique values in a UTC month is <= 50% of total valid values in period 
'Monthly Fractional Unique Values <= 50%': 104,

#Monthly Fractional Unique Values <= 70% 
#Fraction of unique values in a UTC month is <= 70% of total valid values in period 
'Monthly Fractional Unique Values <= 70%': 105,

#Monthly Fractional Unique Values <= 90% 
#Fraction of unique values in a UTC month is <= 90% of total valid values in period 
'Monthly Fractional Unique Values <= 90%': 106,


#Data Outlier Flags
#-----------------------------------------------------
#Data Outlier - Exceeds Scientifically Decreed Lower/Upper Limit
#The measured value is below or greater than scientifically feasible lower/upper limits (variable by parameter).
'Data Outlier - Exceeds Scientifically Decreed Lower/Upper Limit': 110,

#Data Outlier - Monthly Median Exceeds Scientifically Decreed Upper Limit 
#The median of the measurements in a month is greater than a scientifically feasible limit (variable by parameter).
'Data Outlier - Monthly Median Exceeds Scientifically Decreed Upper Limit': 111,

#Data Outlier - Network Decreed
#Data has been reported to be an outlier through data flags by the network data reporters (and not manually checked and verified as valid).
'Data Outlier - Network Decreed': 112,

#Data Outlier - Manually Decreed
#Data has been found and decreed manually to be an outlier.
'Data Outlier - Manually Decreed': 113,

#Possible Data Outlier - Monthly Adjusted Boxplot
#Measured value exceeds adjusted boxplot inner fence (lower or upper) of monthly data, therefore is a possible data outlier.
'Possible Data Outlier - Monthly Adjusted Boxplot': 114,

#Probable Data Outlier - Monthly Adjusted Boxplot
#Measured value exceeds adjusted boxplot outer fence (lower or upper) of monthly data, therefore is a probable data outlier.
'Probable Data Outlier - Monthly Adjusted Boxplot': 115,

#Monthly Distribution Consistency 
#QA flags which describe how consistent a monthly distribution is with other distributions for the same month, across the years.  
#This is done by calculating how consistent a monthly distribution of measurements is with the distributions for the same month of the neighbouring 5 years (backwards and forwards in time), as well as calculating how this metric of consistency differs from all other calculated metrics for the same month, across the time series.
#Kernel density estimation (kde) is used to estimate the probability density function (PDF) for the month in question, as well as the neighbouring year months.  
#The intersection is then taken between the month in question's PDF and the surrounding months PDF's, for which 1.0 - the median intersection value is taken as the metric of consistency (perfect intersection = 0.0, no intersection = 1.0). 
#Only the neighbouring 5 years are used, to limit the impact of any long-term trends biasing the comparison. 
#Additionally determined is to what extent does the median PDF intersection metric per month differ from the typical (median) intersection score for the same month across the entire timerseries (0.0 = perfectly consistent or better than the median intersection score, 1.0 = intersection score is as far below as possible the median intersection score). 
#This metric provides a weighting for the multi-annual variance of the monthly distributions. If there is a low intersection score, but there are typically always low intersection scores due to high variance across the years, then the difference score will be close to 0.0. Conversely, if there is a low intersection score, but there are typically always high intersection scores, then the difference score will be closer to 1.0, as it is outlying.           
#These metrics are then both summed to give a total classification score of the normality of the monthly distribution (2.0 = Extremely different from typical case. 0.0 = Extremely consistent with typical case).
#The classification score is split into 10 zones (in range increments of 0.2) from the most consistent monthly distributions in Zone 1 (score of 0.0 to 0.2), to the least consistent monthly distributions in Zone 10 (score of 1.8 to 2.0).
#All months for which a total classification score can be determined are flagged with a flag for the respective classification score zone.
#If 2/3, 4/6, 8/12 consecutive months are classed as zone 6 or higher, then it is suspected there is a systematic reason for the non-typical distributions, and the entire periods are flagged. 
#-----------------------------------------------------
#Monthly Distribution Consistency - Zone 1
'Monthly Distribution Consistency - Zone 1': 120,

#Monthly Distribution Consistency - Zone 2
'Monthly Distribution Consistency - Zone 2': 121,

#Monthly Distribution Consistency - Zone 3
'Monthly Distribution Consistency - Zone 3': 122,

#Monthly Distribution Consistency - Zone 4
'Monthly Distribution Consistency - Zone 4': 123,

#Monthly Distribution Consistency - Zone 5
'Monthly Distribution Consistency - Zone 5': 124,

#Monthly Distribution Consistency - Zone 6
'Monthly Distribution Consistency - Zone 6': 125,

#Monthly Distribution Consistency - Zone 7
'Monthly Distribution Consistency - Zone 7': 126,

#Monthly Distribution Consistency - Zone 8
'Monthly Distribution Consistency - Zone 8': 127,

#Monthly Distribution Consistency - Zone 9
'Monthly Distribution Consistency - Zone 9': 128,

#Monthly Distribution Consistency - Zone 10
'Monthly Distribution Consistency - Zone 10': 129,

#Monthly Distribution Consistency - Unclassified
'Monthly Distribution Consistency - Unclassified': 130,

#Systematic Inconsistent Monthly Distributions - 2/3 Months >= Zone 6
#2 out of 3 months' distributions are classed as Zone 6 or higher, suggesting there is potentially systematic reason for the inconsistent distributions across the 3 months.
'Systematic Inconsistent Monthly Distributions - 2/3 Months >= Zone 6': 131,

#Systematic Inconsistent Monthly Distributions - 4/6 Months >= Zone 6
#4 out of 6 months' distributions are classed as Zone 6 or higher, suggesting there is potentially systematic reason for the inconsistent distributions across the 6 months.
'Systematic Inconsistent Monthly Distributions - 4/6 Months >= Zone 6': 132,

#Systematic Inconsistent Monthly Distributions - 8/12 Months >= Zone 6
#8 out of 12 months' distributions are classed as Zone 6 or higher, suggesting there is potentially systematic reason for the inconsistent distributions across the 12 months.
'Systematic Inconsistent Monthly Distributions - 8/12 Months >= Zone 6': 133,

#-----------------------------------------------------


}

###--------------------------------------------------------------------------------------------------###
###DEFINE DEFAULT DATA FLAGS / QA FLAGS FOR PROVIDENTIA
###--------------------------------------------------------------------------------------------------###

providentia_defaults = {
"flag": [],

"qa_standard": ["Missing Measurement", "Infinite Value", "Negative Measurement",
"Invalid Data Provider Flags - GHOST Decreed", "No Valid Data to Average",
"Erroneous Primary Sampling", "Erroneous Sample Preparation", "Erroneous Measurement Methodology", 
"Below Preferential Lower Limit of Detection", "Above Preferential Upper Limit of Detection", 
"Insufficient Measurement Resolution - Preferential", "Insufficient Measurement Resolution - Empirical", 
"Data Outlier - Exceeds Scientifically Decreed Lower/Upper Limit", 
"Data Outlier - Monthly Median Exceeds Scientifically Decreed Upper Limit", 
"Data Outlier - Network Decreed", "Data Outlier - Manually Decreed", 
"Probable Data Outlier - Monthly Adjusted Boxplot",
"Systematic Inconsistent Monthly Distributions - 4/6 Months >= Zone 6",
"Systematic Inconsistent Monthly Distributions - 8/12 Months >= Zone 6"],

"qa_non_negative": ["Missing Measurement", "Infinite Value", 
"Invalid Data Provider Flags - GHOST Decreed", "No Valid Data to Average",
"Erroneous Primary Sampling", "Erroneous Sample Preparation", "Erroneous Measurement Methodology",
"Below Preferential Lower Limit of Detection", "Above Preferential Upper Limit of Detection",
"Insufficient Measurement Resolution - Preferential", "Insufficient Measurement Resolution - Empirical",
"Data Outlier - Exceeds Scientifically Decreed Lower/Upper Limit",
"Data Outlier - Monthly Median Exceeds Scientifically Decreed Upper Limit",
"Data Outlier - Network Decreed", "Data Outlier - Manually Decreed",
"Probable Data Outlier - Monthly Adjusted Boxplot",
"Systematic Inconsistent Monthly Distributions - 4/6 Months >= Zone 6",
"Systematic Inconsistent Monthly Distributions - 8/12 Months >= Zone 6"]
}

###--------------------------------------------------------------------------------------------------###
###DEFINE METADATA WARNING MODIFICATION FIELDS 
###--------------------------------------------------------------------------------------------------###

#define fields for which a warning will be wrote upon modification, associated with the names of the fields as they should be written in the text warning.

metadata_modification_field_to_warning_name = {'latitude':                        'Latitude',
                                               'longitude':                       'Longitude',
                                               'altitude':                        'Altitude',
                                               'sampling_height':                 'Sampling Height',
                                               'measurement_altitude':            'Measurement Altitude',
                                               'area_classification':             'Area Classification',
                                               'station_classification':          'Station Classification',
                                               'main_emission_source':            'Main Emission Source',
                                               'land_use':                        'Land Use',
                                               'terrain':                         'Terrain',
                                               'measurement_scale':               'Measurement Scale',
                                               'representative_radius':           'Representative Radius',
                                               'primary_sampling_type':           'Primary Sampling Type',
                                               'primary_sampling_instrument_name':'Primary Sampling Instrument Name',
                                               'sample_preparation_types':        'Sample Preparation Types',
                                               'sample_preparation_techniques':   'Sample Preparation Techniques',
                                               'measurement_methodology':         'Measurement Methodology',
                                               'measuring_instrument_name':       'Measuring Instrument Name'
                                              }

###--------------------------------------------------------------------------------------------------###
###MEASUREMENT PROCESS STANDARDISATIONS
###--------------------------------------------------------------------------------------------------###

#----------------------------------------------------------------------------------------#
#STANDARD SAMPLING INFORMATION
#----------------------------------------------------------------------------------------#
#define dictionary that contains information pertaining to standardised sampling types and specific sampling instrumentation

#The "sampling type" refers to type of sampling used by either the measurement instrument, or for 2 (or more) step-processes, in the sample preparation phase 

standard_sampling_types = {
#------------------------------------
#low volume continuous/high volume continuous

#Ambient air is continuously drawn in by a pump of some description. 

#The rate that air is sampled can be particularly important for the measurements of particulate matter.
#Many particulate matter measurements utilise a 2-step process, where standalone samplers first sample and filter air, and then the filtered products are measured by separate instrument.
#The standalone samplers can be separated into 2 discrete groups: low volume samplers & high volume samplers.

#Many instruments also have built-in continuous samplers. These are in-almost all cases low-volume, when not otherwise stated.

#--------
#low volume continuous

#Ambient air is continuously drawn in using a low volume sampler (typically sampling < ~24,000 litres of air over a 24-hour period)
#Some of these samplers have in-built filters, designed to specifically retain certain compounds.

'low volume continuous':{
    'instruments':{
        'Andersen RAAS10 100':                {'sample_preparation_type':'filter',           'isolated_parameters':['PM10'],                           'documented_flow_rate':'16.67',           'instrument_further_details':'Not many details but what there is, is here: https://www.scribd.com/document/404806002/Thermo-High-Volume-Sampler-Brochure'},
        'Andersen RAAS2.5 100':               {'sample_preparation_type':'filter',           'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'16.67',           'instrument_further_details':'Not many details but what there is, is here: https://www.scribd.com/document/404806002/Thermo-High-Volume-Sampler-Brochure'},
        'Andersen RAAS10 200':                {'sample_preparation_type':'filter',           'isolated_parameters':['PM10'],                           'documented_flow_rate':'16.67',           'instrument_further_details':'Not many details but what there is, is here: https://www.scribd.com/document/404806002/Thermo-High-Volume-Sampler-Brochure'},
        'Andersen RAAS2.5 200':               {'sample_preparation_type':'filter',           'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'16.67',           'instrument_further_details':'Not many details but what there is, is here: https://www.scribd.com/document/404806002/Thermo-High-Volume-Sampler-Brochure'},
        'Andersen RAAS10 300':                {'sample_preparation_type':'filter',           'isolated_parameters':['PM10'],                           'documented_flow_rate':'16.67',           'instrument_further_details':'Not many details but what there is, is here: https://www.scribd.com/document/404806002/Thermo-High-Volume-Sampler-Brochure'},
        'Andersen RAAS2.5 300':               {'sample_preparation_type':'filter',           'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'16.67',           'instrument_further_details':'Not many details but what there is, is here: https://www.scribd.com/document/404806002/Thermo-High-Volume-Sampler-Brochure'},
        'Andersen RAAS2.5 Nylon':             {'sample_preparation_type':'filter;denuder',   'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'7.3',             'instrument_further_details':'See details here: https://bgi.mesalabs.com/wp-content/uploads/sites/35/2014/12/andersen.pdf and https://www3.epa.gov/ttnamti1/files/ambient/pm25/spec/pm25pict.pdf'},       
        'Andersen RAAS2.5 Quartz':            {'sample_preparation_type':'filter',           'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'7.3',             'instrument_further_details':'See details here: https://bgi.mesalabs.com/wp-content/uploads/sites/35/2014/12/andersen.pdf and https://www3.epa.gov/ttnamti1/files/ambient/pm25/spec/pm25pict.pdf'},       
        'Andersen RAAS2.5 Teflon':            {'sample_preparation_type':'filter',           'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'16.67',           'instrument_further_details':'See details here: https://bgi.mesalabs.com/wp-content/uploads/sites/35/2014/12/andersen.pdf and https://www3.epa.gov/ttnamti1/files/ambient/pm25/spec/pm25pict.pdf'},    
        'APM PMS-104':                        {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5'],                                                             'instrument_further_details':'See more detail here: http://apm.co.kr/sub2/view.php?id=20&ca_id=&page='},
        'Atmoservice PNS3D15':                {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],                                                       'instrument_further_details':'See more detail here: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6096885/'},
        'BGI frmOmni':                        {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],             'documented_flow_rate':'5.0',             'instrument_manual_name':'BGI_frmOmni_Manual.pdf'},
        'BGI PQ100':                          {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],             'documented_flow_rate':'2.0-25.0',        'instrument_manual_name':'BGI_PQ100_Specs.pdf'},
        'BGI PQ200':                          {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],             'documented_flow_rate':'10.0-20.0',       'instrument_manual_name':'BGI_PQ200_Specs.pdf'},       
        'Comde Derenda LVS 3.1':              {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],             'documented_flow_rate':'16.6667-58.3333', 'instrument_manual_name':'Comde_Derenda_MVS6.1_Specs.pdf'},
        'Comde Derenda LVS 3.2':              {'sample_preparation_type':'filter',                                                                                                               'instrument_further_details':'Confirmation it is a sampling instrument here: http://dd.eionet.europa.eu/vocabularyconcept/aq/samplingequipment/derendaLVS-3-2/view?facet=HTML+Representation'},
        'Comde Derenda MVS 6.1':              {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],             'documented_flow_rate':'16.6667-91.6667', 'instrument_manual_name':'Comde_Derenda_MVS6.1_Specs.pdf'},
        'Comde Derenda PNS 16T':              {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],             'documented_flow_rate':'16.6667-91.6667', 'instrument_manual_name':'Comde_Derenda_PNS_16T_Specs.pdf'},
        'Digitel DPA-14':                     {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],             'documented_flow_rate':'5.0-50.0',        'instrument_manual_name':'Digitel_DPA-14_Specs.pdf'},
        'Elecos APM-1':                       {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],                                                       'instrument_further_details':'See more detail here: https://uk-air.defra.gov.uk/assets/documents/reports/empire/quarg/quarg_11.pdf', 'primary_sampling_process_details':'Make assumption sampling instrument is low volume continuous.', 'primary_sampling_assumption':True},
        'General Metal Works 9200':           {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5'],                                                             'instrument_further_details':'A little detail here: https://www.tandfonline.com/doi/pdf/10.1080/00022470.1985.10465915'},             
        'IMPROVE Module A':                   {'sample_preparation_type':'filter',           'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'22.8',            'instrument_further_details':'See more information here: https://vista.cira.colostate.edu/Improve/improve-program/ and http://vista.cira.colostate.edu/improve/wp-content/uploads/2020/02/IMPROVE-TSA-Report-2019.pdf'}, 
        'IMPROVE Module B':                   {'sample_preparation_type':'filter;denuder',   'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'22.8',            'instrument_further_details':'See more information here: https://vista.cira.colostate.edu/Improve/improve-program/ and http://vista.cira.colostate.edu/improve/wp-content/uploads/2020/02/IMPROVE-TSA-Report-2019.pdf'}, 
        'IMPROVE Module C':                   {'sample_preparation_type':'filter',           'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'22.8',            'instrument_further_details':'See more information here: https://vista.cira.colostate.edu/Improve/improve-program/ and http://vista.cira.colostate.edu/improve/wp-content/uploads/2020/02/IMPROVE-TSA-Report-2019.pdf'}, 
        'IMPROVE Module D':                   {'sample_preparation_type':'filter',           'isolated_parameters':['PM10'],                           'documented_flow_rate':'16.9',            'instrument_further_details':'See more information here: https://vista.cira.colostate.edu/Improve/improve-program/ and http://vista.cira.colostate.edu/improve/wp-content/uploads/2020/02/IMPROVE-TSA-Report-2019.pdf'}, 
        'IVL P-Model':                        {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],                                                       'instrument_further_details':'Not much detail provided, but believe this sampler uses different sampling inlets which are then named as P-Model S10 (for PM10) P-Model S2.5 (for PM2.5). Group different referenced "S" names all under "IVL P-Model". See detail here for S10 version: https://www.aces.su.se/reflab/wp-content/uploads/2016/11/ACES_Report_4.pdf; See detail here for S2.5 version: https://pubs.rsc.org/en/content/articlelanding/2017/em/c7em00122c#!divAbstract'},
        'KNF IP20-T':                         {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],             'documented_flow_rate':'11.5',            'instrument_further_details':'See more detail here: https://www.overstocklabequipment.com/product-p/un811ktp.htm'},
        'Leckel SEQ 47-50':                   {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],             'documented_flow_rate':'16.66-50',        'instrument_manual_name':'Leckel_SEQ_47-50_Specs.pdf'},
        'Leckel LVS3':                        {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],             'documented_flow_rate':'16.66-38.33',     'instrument_manual_name':'Leckel_LVS3_MVS6_Specs.pdf'},
        'Leckel MVS6':                        {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],             'documented_flow_rate':'38.33-58.33',     'instrument_manual_name':'Leckel_LVS3_MVS6_Specs.pdf'},
        'Lipinski AGT24':                     {'sample_preparation_type':'filter',                                                                                                               'instrument_further_details':'Not many details but what there is, is here: http://dd.eionet.europa.eu/vocabularyconcept/aq/samplingequipment/LipinskiAGT24/view?facet=HTML+Representation', 'primary_sampling_assumption':True, 'primary_sampling_process_details':'Make assumption sampling instrument is low volume continuous.'},
        'MCZ LVS1':                           {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5'],                   'documented_flow_rate':'16.66-38.33',     'instrument_manual_name':'MCZ_LVS1_Specs.pdf'},
        'MCZ LVS16':                          {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5'],                   'documented_flow_rate':'16.66-38.33',     'instrument_manual_name':'MCZ_LVS16_Specs.pdf'},
        'Met One E-FRM-DC':                   {'sample_preparation_type':'filter',                                                                     'documented_flow_rate':'16.67',           'instrument_manual_name':'Met_One_E-FRM-DC_Specs.pdf'},                                                                                                                                                                                                                                                                                                                   
        'Met One E-SEQ-FRM':                  {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5'],                   'documented_flow_rate':'16.67',           'instrument_manual_name':'Met_One_E-SEQ-FRM_Specs.pdf'},        
        'Met One SASS':                       {'sample_preparation_type':'filter',           'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'6.7',             'instrument_manual_name':'Met_One_SASS_Specs.pdf'},        
        'Met One SUPER SASS':                 {'sample_preparation_type':'filter',           'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'6.7',             'instrument_manual_name':'Met_One_SASS_Specs.pdf'},        
        'NILU EK':                            {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','SO2'],                     'documented_flow_rate':'10.0',            'instrument_further_details':'See detail here: http://products.nilu.no/ProductsDivision/AirSamplers/SequentialAirSamplerTypeEK.aspx'},
        'NILU FK':                            {'sample_preparation_type':'filter',           'isolated_parameters':['NO2','SO2','PM10','PM2.5','PM1'], 'documented_flow_rate':'0.5-5.0',         'instrument_further_details':'See more detail here:https://www.nilu.no/wp-content/uploads/dnn/38-2000-bs.pdf'},
        'NILU SS2000':                        {'sample_preparation_type':'filter',           'isolated_parameters':['NO2','SO2'],                      'documented_flow_rate':'0.5-1.0',         'instrument_manual_name':'NILU_SS2000_Specs.pdf'},
        'Oregon DEQ':                         {'sample_preparation_type':'filter',           'isolated_parameters':['PM10'],                                                                     'instrument_further_details':'See more details here: https://books.google.es/books?id=Dfk0AQAAMAAJ&pg=RA8-PA2&lpg=RA8-PA2&dq=RFPS-0389-071&source=bl&ots=JCh3fbWuGH&sig=ACfU3U2FkxE-ScRAxPwvtfhgp8YRdhYDqA&hl=en&sa=X&ved=2ahUKEwj1i4apxqv2AhX1iv0HHcliASsQ6AF6BAgREAM#v=onepage&q=RFPS-0389-071&f=false'},
        'Riemer Autocan 2/4':                 {'sample_preparation_type':'filter',                                                                                                               'instrument_further_details':'See more details here: http://fp7.actris.eu/Portals/97/deliverables/PU/WP4_D4.7_M32.pdf'},
        'Riemer SPF4':                        {'sample_preparation_type':'filter',           'isolated_parameters':['SO2','NH3','HNO3'],                                                         'instrument_further_details':'See more detail here: http://www.riemer-mt.com/low.volume.php'},
        'Rivm Universal Sampler':             {'sample_preparation_type':'sorbent trapping',                                                           'documented_flow_rate':'0.1-0.65',        'instrument_further_details':'See more details here: https://www.rivm.nl/bibliotheek/rapporten/723101055.pdf'},
        'Sierra Andersen 244 Dichot':         {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5'],                   'documented_flow_rate':'16.67',           'instrument_further_details':'See more details here: https://www.tandfonline.com/doi/pdf/10.1080/02786829108959500 and https://www.tandfonline.com/doi/pdf/10.1080/02786820116873'},        
        'Sierra Andersen 246B Dichot':        {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5'],                                                             'instrument_further_details':'See more details here: https://acp.copernicus.org/articles/6/1865/2006/acp-6-1865-2006.pdf and https://www.tandfonline.com/doi/pdf/10.1080/10473289.2000.10464163'},
        'Sierra Andersen 254':                {'sample_preparation_type':'filter',           'isolated_parameters':['PM10'],                           'documented_flow_rate':'11.3',            'instrument_further_details':'See more details here: https://books.google.es/books?id=yCDGFlP3toUC&pg=PA164&lpg=PA164&dq=sierra+andersen+254+sampler&source=bl&ots=-vKaKy844g&sig=ACfU3U0D7VsfZVSXiQnHikhWA7FBV2NT0A&hl=en&sa=X&ved=2ahUKEwjb87LTvqv2AhXNR_EDHRuWDwYQ6AF6BAgREAM#v=onepage&q=sierra%20andersen%20254%20sampler&f=false'},     
        'TCR Terora Skypost Gas':             {'sample_preparation_type':'sorbent trapping',                                                           'documented_flow_rate':'0.5-10.0',        'instrument_manual_name':'TCR_Tecora_Skypost_Gas_Specs.pdf'},
        'TCR Terora Skypost PM FG':           {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5'],                   'documented_flow_rate':'10.0-50.0',       'instrument_manual_name':'TCR_Tecora_Skypost_PM_Specs.pdf'},
        'TCR Tecora Sentinel':                {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],                                                       'instrument_manual_name':'TCR_Tecora_Sentinel_Specs.pdf', 'primary_sampling_assumption':True, 'primary_sampling_process_details':'Make assumption sampling instrument is low volume continuous.'},
        'Thermo Andersen FH95-KF':            {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],             'documented_flow_rate':'38.33',           'instrument_further_details':'Not many details but what there is, is here: https://core.ac.uk/download/pdf/38621632.pdf'},
        'Thermo Partisol 2000':               {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5'],                   'documented_flow_rate':'16.6667',         'instrument_manual_name':'Thermo_Partisol_2000_Manual.pdf'},
        'Thermo Partisol 2025':               {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5','PM1'],             'documented_flow_rate':'16.6667',         'instrument_manual_name':'Thermo_Partisol_2025_Specs.pdf'},
        'Thermo Partisol 2000i':              {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5'],                   'documented_flow_rate':'5-18',            'instrument_manual_name':'Thermo_Partisol_2000i_Manual.pdf'},
        'Thermo Partisol 2025i':              {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5'],                   'documented_flow_rate':'10-19',           'instrument_manual_name':'Thermo_Partisol_2025i_Specs.pdf.pdf'},
        'Thermo Partisol 2300':               {'sample_preparation_type':'filter',           'isolated_parameters':['SO2','PM10','PM2.5'],             'documented_flow_rate':'5-18',            'instrument_manual_name':'Thermo_Partisol_2300_Manual.pdf.pdf'},
        'Tisch Environmental Wilbur':         {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5'],                   'documented_flow_rate':'16.6667',         'instrument_manual_name':'Tisch_Environmental_Wilbur_Manual.pdf'},
        'URG MASS 400':                       {'sample_preparation_type':'filter;denuder',   'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'16.67',           'instrument_further_details':'See more details here: https://bgi.mesalabs.com/wp-content/uploads/sites/35/2014/12/urgtrng.pdf'},                
        'URG MASS 450':                       {'sample_preparation_type':'filter',           'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'16.67',           'instrument_further_details':'See more details here: https://bgi.mesalabs.com/wp-content/uploads/sites/35/2014/12/urgtrng.pdf'},                
        'URG 3000N':                          {'sample_preparation_type':'filter',           'isolated_parameters':['PM2.5'],                          'documented_flow_rate':'22.0',            'instrument_further_details':'See more details here: http://www.urgcorp.com/systems/manual-sampling-systems/urg-3000n-carbon-sampler'},
        'Wedding and Associates Dichot':      {'sample_preparation_type':'filter',           'isolated_parameters':['PM10','PM2.5'],                                                             'instrument_further_details':'Also called General Metal Works 9200. Not many details but what there is, is here: https://books.google.es/books?id=hEoPEAAAQBAJ&pg=PA27&lpg=PA27&dq=wedding+dichot+pm+sampling&source=bl&ots=brIKJImj44&sig=ACfU3U3QztrWJllkDy1ONu5AWKUk19FE6Q&hl=en&sa=X&ved=2ahUKEwjl6Zj7x6v2AhVNgf0HHUTtBlYQ6AF6BAgsEAM#v=onepage&q=wedding%20dichot%20pm%20sampling&f=false'},             
        'Xonteck 910':                        {'sample_preparation_type':'filter',                                                                     'documented_flow_rate':'0.02-0.1',        'instrument_manual_name':'Xonteck_910_Specs.pdf'},
        'Xonteck 924':                        {'sample_preparation_type':'filter',                                                                     'documented_flow_rate':'0.5-30.0',        'instrument_manual_name':'Xonteck_924_Specs.pdf'},
        'Xonteck 926':                        {'sample_preparation_type':'filter',                                                                                                               'instrument_further_details':'Not many details, but what there is, is here: https://buyandsell.gc.ca/cds/public/2015/04/15/de1e4f53d40b1a81d93ada3347956ae0/k8a21-14-0099qa1_fre.pdf'} 
    }
},

#--------
#high volume continuous

#Ambient air is continuously drawn in using a high volume sampler instrumentation (typically sampling > ~100,000 litres of air over a 24-hour period)
#Some of these samplers have in-built filters, designed to specifically retain certain compounds.

'high volume continuous':{
    'instruments':{
        'Digitel DH-77':             {'sample_preparation_type':'filter', 'isolated_parameters':['PM10','PM2.5','PM1'], 'documented_flow_rate':'100-1000',         'instrument_manual_name':'Digitel_DH-77_Specs.pdf'},
        'Digitel DHA-80':            {'sample_preparation_type':'filter', 'isolated_parameters':['PM10','PM2.5','PM1'], 'documented_flow_rate':'100-1000',         'instrument_manual_name':'Digitel_DHA-80_Manual.pdf'},
        'Ecotech 3000':              {'sample_preparation_type':'filter', 'isolated_parameters':['PM10','PM2.5'],       'documented_flow_rate':'750-1600',         'instrument_manual_name':'Ecotech_3000_Manual.pdf'},    
        'General Metal Works 9000':  {'sample_preparation_type':'filter', 'isolated_parameters':['PM10'],                                                          'instrument_further_details':'See more details here: https://www.tandfonline.com/doi/pdf/10.1080/00022470.1986.10466131'},
        'Graseby Andersen GMW 1200': {'sample_preparation_type':'filter',                                                                                          'instrument_further_details':'Not many details but what there is, is here: https://ww2.arb.ca.gov/sites/default/files/barcu/regact/aaqspm/revfro.pdf'},                 
        'Graseby Andersen GMW 321-B':{'sample_preparation_type':'filter',                                                                                          'instrument_further_details':'Not many details but what there is, is here: https://ww2.arb.ca.gov/sites/default/files/barcu/regact/aaqspm/revfro.pdf'},                 
        'Graseby Andersen GMW 321-C':{'sample_preparation_type':'filter',                                                                                          'instrument_further_details':'Not many details but what there is, is here: https://ww2.arb.ca.gov/sites/default/files/barcu/regact/aaqspm/revfro.pdf'},                 
        'MCV CAV-A/mb':              {'sample_preparation_type':'filter', 'isolated_parameters':['PM10','PM2.5'],       'documented_flow_rate':'333.333-1333.333', 'instrument_further_details':'See more details here: http://www.mcvsa.com/Productes/Atmosfera/CaptadordealtovolumenMCVCAVAmb/tabid/114/Default.aspx'},
        'MCV CAV-A/msb':             {'sample_preparation_type':'filter', 'isolated_parameters':['PM10','PM2.5'],       'documented_flow_rate':'250-666.667',      'instrument_further_details':'See more details here: http://www.mcvsa.com/Productes/Atmosfera/CaptadordealtovolumensecuencialMCVCAVAMSb/tabid/139/language/es-ES/Default.aspx'},
        'Sierra 305-2000':           {'sample_preparation_type':'filter', 'isolated_parameters':['PM10','PM2.5','PM1'],                                            'instrument_further_details':'See more details here: https://pubs.acs.org/doi/pdf/10.1021/es00095a726'},
        'Sierra Andersen 321A':      {'sample_preparation_type':'filter', 'isolated_parameters':['PM10'],                                                          'instrument_further_details':'See more details here: https://www.tandfonline.com/doi/pdf/10.1080/00022470.1986.10466131'},    
        'Thermo MFC-PM10':           {'sample_preparation_type':'filter', 'isolated_parameters':['PM10'],               'documented_flow_rate':'566,34-1699.01',   'instrument_manual_name':'Thermo_MFC-PM10_Specs.pdf'},
        'Thermo VFC-PM10':           {'sample_preparation_type':'filter', 'isolated_parameters':['PM10'],               'documented_flow_rate':'1132.67',          'instrument_manual_name':'Thermo_VFC-PM10_Specs.pdf'},
        'Tisch Environmental 6000':  {'sample_preparation_type':'filter', 'isolated_parameters':['PM10'],               'documented_flow_rate':'1130',             'instrument_manual_name':'Tisch_Environmental_6000_Manual.pdf'},          
        'Tisch Environmental 6070':  {'sample_preparation_type':'filter', 'isolated_parameters':['PM10'],               'documented_flow_rate':'1020-1240',        'instrument_manual_name':'Tisch_Environmental_6070_Specs.pdf'},     
        'Tisch Environmental 6070D': {'sample_preparation_type':'filter', 'isolated_parameters':['PM10'],               'documented_flow_rate':'1020-1240',        'instrument_manual_name':'Tisch_Environmental_6070D_Specs.pdf'},     
        'Tisch Environmental 6070DX':{'sample_preparation_type':'filter', 'isolated_parameters':['PM10'],               'documented_flow_rate':'1020-1240',        'instrument_manual_name':'Tisch_Environmental_6070D_Specs.pdf'},
        'Tisch Environmental 6070V': {'sample_preparation_type':'filter', 'isolated_parameters':['PM10'],               'documented_flow_rate':'1132.67-1245.94',  'instrument_manual_name':'Tisch_Environmental_6070V_Specs.pdf'},     
        'Wedding and Associates PM': {'sample_preparation_type':'filter', 'isolated_parameters':['PM10','PM2.5'],                                                  'instrument_further_details':'See more details here: https://www.nature.com/articles/7500138.pdf?origin=ppub and https://www3.epa.gov/ttnamti1/files/ambient/criteria/AMTIC%20List%20Dec%202016-2.pdf'}
        }
},

#------------------------------------
#injection

#The measuring instrument is injected with a limited quantity of air
#The limited injected quantity is typically pre-processed to aid the detection of a specific species.

'injection':{
    'instruments':{}
},

#------------------------------------
#continuous injection

#The measuring instrument is periodically injected with limited quantities of air.
#The injected quantities can either be from continuous automated collection (assumed to be of a low volume), or from pre-processed loaded samples, which are periodically injected.

'continuous injection':{
    'instruments':{}
},

#------------------------------------
#passive              

#Air is not drawn in, rather the air sampled is the ambient air which interacts with measurement apparatus

'passive':{
    'instruments':{}
},

#------------------------------------
#remote            

#The measuring instrument does not actively sample air, but uses advanced optical techniques to measure parameters in air over long distances

'remote':{
    'instruments':{}
},

#------------------------------------
#manual              
#No instrument is used to determine measured values, they are determined manually 
#e.g. this applies for some colorimetric methods where measured values are derived manually via the colour of reagent after reaction with a compound of interest

'manual':{
    'instruments':{}
},

#------------------------------------
#unknown

'unknown':{
    'instruments':{}
}
}
    

#----------------------------------------------------------------------------------------#
#STANDARD SAMPLE PREPARATION INFORMATION
#----------------------------------------------------------------------------------------#

#define dictionary that contains information pertaining to standardised sample preparation types
#Sample Preparation refers to all preparations made to the sample before the ultimate measurement 

standard_sampling_preparation_types = {
#------------------------------------
#flask

#Sample is collected/stored in measurement flasks/canisters from ambient air, or filled by pump
#The canisters can be opened for a short instantaneous window, or can be opened periodically over a long window to get a longer representative sample of air. 

'flask':{
    'specific_techniques':{}
},

#------------------------------------
#bag

#Sample is collected/stored in gas sampling bags (typically teflon) from ambient air, or filled by pump.
#These bags are a cheap alternative to canisters, with much reduced stability times.  

'bag':{
    'specific_techniques':{}
},


#------------------------------------
#preconcentration       

#This refers to the process of concentrating a sample before analysis, so that trace components won't be overlooked. 
#This is done typically through absorption of the sample onto a cooled, sorbent-packed trap before thermal desorption to transfer very quickly the sample to the analytical system. 
#Can be more than 1 trap, with different adsorbents.

'preconcentration':{
    'specific_techniques':{}
},

#------------------------------------
#filter                 

#Air is passed through a filtering system, selectively retaining compound(s) of interest

'filter':{
    'specific_techniques':{}
},

#------------------------------------
#filter pack

#Air is passed through a filter pack, selectively retaining compound(s) of interest
#The filter packs can contain different filters to target the retention of different compounds. 

'filter pack':{
    'specific_techniques':{
        '1-Stage Filter Pack':{},
        '2-Stage Filter Pack':{},
        '3-Stage Filter Pack':{},
        '4-Stage Filter Pack':{}
    }
},

#------------------------------------
#denuder

#Air is passed through a denuder before analysis to selectively retain compound(s) of interest 
#a denuder is cylindrical or annular conduit or tube internally coated with a reagent that selectively reacts with a stable flow of gas

'denuder':{
    'specific_techniques':{
        'CEH DELTA':  {'isolated_parameters':['NH3','HNO3','SO2','HCl','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'],'preparation_further_details':'See specifications here: https://www.ceh.ac.uk/services/delta-active-sampler-system'},
        'Riemer DEN2':{                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    'preparation_further_details':'See here for confirmation of technique: https://ebas-submit.nilu.no/Submit-Data/Data-Reporting/Comments/Modules/all-instrument-types-manufacturers-models'},
        'UBA Olaf':   {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    'preparation_further_details':'See here for confirmation of technique: https://ebas-submit.nilu.no/Submit-Data/Data-Reporting/Comments/Category/Aerosol/Aerosol-chemical-speciation/inorganic-aerosol-chemistry/instrument-manufacturer'}
        }
},

#------------------------------------
#sorbent trapping           

#Sample is passed through a sorbent material to trap and retain the compound(s) of interest.
#Diffusive samplers are a specific example of the utilisation of sorbent trapping.

'sorbent trapping':{
    'specific_techniques':{

#--------
#diffusive sampler

#Diffussive samplers are a specific utilisation of sorbent trapping to passively trap species over long time periods, typically in urban areas.
#Diffusive samplers are an inexpensive method for measuring atmospheric composition over periods from one day to several weeks.
#These samplers operate on the principle of a sorbent trapping and retaining a compound of interest. 
#Sorption often takes place through reaction with a reagent (i.e. typically TEA for NO2).  
#After a suitable sample time, the trapped compound can then be analysed using colorimetry/spectrophotometry/ion chromatography/GC methods 

#Diffusive samplers come in various forms, e.g: Palmes, Passam, Ogawa Badge, and Radiello radial sampler tubes.
#The way these tubes differ is associated with the way they sample air and deal with variations in meteorological conditions. 

#This method is typically associated with very large uncertainties. 
#These uncertainties mainly come from:
#1. The non-specificity of trapped gases (i.e. HONO and PAN are readily uptaken when attempting to trap NO2).
#2. Variability in temperature, relative humidity, and air velocity directly impacting upon uptake rate of the desired gas, and effectiveness of the sorbent.
#Uncertainties associated with this methodology are typically reported between 10-35%: http://publications.jrc.ec.europa.eu/repository/bitstream/JRC51106/reqno_jrc51106_eur_23793.pdf[1].pdf

#Unless additional information is given, make assumption all diffusive samplers trap gases through sorbent trapping, followed by subsequent analysis through ion chromatography (by injection).

        'Diffusive Sampler':{'preparation_further_details':'See Reference for NO2: https://ec.europa.eu/jrc/en/publication/eur-scientific-and-technical-research-reports/review-application-diffusive-samplers-measurement-nitrogen-dioxide-ambient-air-european; reference for NH3:https://www.sciencedirect.com/science/article/pii/S1352231012001136; see reference for VOCs: http://www.siremlab.com/pages/wms/pdf/Engineering-Issue-Paper-2015-Passive-Samplers-for-Investigations-of-Air-Quality.pdf'}
    }
},

#------------------------------------
#reagent reaction

#Air is reacted with a liquid/solid chemical reagent to allow subsequent measurement of a specific compound

#The variety of types of passive sampling phases are numerous. These are described in detail below. 
#The name of the passive sampling method is given as the instrument name.
#However, typically only passive sampling phase information is given, with no detail about the analysis phase. 
#Unless additional information is given, make assumption all passive sampling methods involve chemical reaction with a reagent, followed by subsequent analysis through spectrophotometry (by injection).

'reagent reaction':{
    'specific_techniques':{

#--------
#Griess-Saltzman

#NO2 reacts with water producing nitrous acid.
#This is reacted with an aromatic amine, to form a diazonium salt.
#Next, the solution is reacted with organic coupling agent (modified Saltzman reagent - 1960): 0.5% sulfanilic acid, 5% acetic acid, 0.005% NEDA(N-(1-Naphthyl)ethylenediamine dihydrochloride), to form a deeply coloured azo-dye.
#If an extended sampling time is required, there may be some bleaching by SO2. Saltzman recommended the addition of acetone to the reagent to prevent this.
#The conversion of NO2 to azo-dye is not quantitative, therefore a factor is introduced to represent the conversion efficiency under a given set of experimental conditions.
#The absorbance of the solution is then measured spectrophotometrically.
#Bennett (1930) gave stated a slighted altered reagent for the specific measurement of NO.
#Nitrate can also be measured by initial reduction to nitrite: https://www.sciencedirect.com/science/article/abs/pii/S1089860300903197

        'Griess-Saltzman':{'isolated_parameters':['NO','NO2','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-'], 'preparation_further_details':'See References: Colorimetric Azo Dye Methods for the Atmospheric Analysis of Nitrogen Dioxide; Historical Development, Szonntagh, E.L, 1979; https://www3.epa.gov/ttn/naaqs/standards/nox/previous/1982-NOx-AQCD.pdf'},

#--------
#Lyshkow 

#A modification of the Griess-Saltzman method, using a modified reagent to produce an azo-dye: 0.15% Sulfanilamide, 1.5% tartaric acid, 0.005% NEDA, 0.005%  2-Naphthol-3, 6-disulphonic acid disodium salt. 
#The absorbance of the solution is measured spectrophotometrically, as with Griess-Saltzman.
#The reagent as with the Griess-Saltzman method can be slightly modified for the specific measurement of NO.

        'Lyshkow':{'isolated_parameters':['NO','NO2'], 'preparation_further_details':'See References: Colorimetric Azo Dye Methods for the Atmospheric Analysis of Nitrogen Dioxide; Historical Development, Szonntagh, E.L, 1979; https://www3.epa.gov/ttn/naaqs/standards/nox/previous/1982-NOx-AQCD.pdf'},

#--------
#Jacobs-Hochheiser 

#Designed by Jacobs and Hochheiser (1958) to avoid the azo-dye bleaching by SO2 that occurs in the Griess-Saltzman method.
#NO2 is absorbed in a solution of sodium hydroxide.
#Butanol is added as a surfactant to improve gas transfer.
#After sampling any sulfite from absorbed SO2 is oxidised with hydrogen peroxide.
#Next, the solution is acidified with phosphoric acid.
#A diazotizer and coupling agent are then added to produce an azo-dye.
#The absorbance of the solution is then measured spectrophotometrically.
    
        'Jacobs-Hochheiser':{'isolated_parameters':['NO2'], 'preparation_further_details':'See References: Colorimetric Azo Dye Methods for the Atmospheric Analysis of Nitrogen Dioxide; Historical Development, Szonntagh, E.L, 1979; https://www3.epa.gov/ttn/naaqs/standards/nox/previous/1982-NOx-AQCD.pdf'},

#--------
#Sodium Arsenite 

#NO2 is absorbed in an alkaline solution of sodium arsenite.
#Any sulfite in the solution is oxidised with peroxide.
#The solution is then the solution is acidified with phosphoric acid.
#An azo-dye is then formed through addition of sulfanilamide and NEDA.

        'Sodium Arsenite':{'isolated_parameters':['NO2'], 'preparation_further_details':'See details: https://www3.epa.gov/ttn/naaqs/standards/nox/previous/1982-NOx-AQCD.pdf'},

#--------
#TEA 

#NO2 is absorbed in a solution of triethanolamine and n-butanol surfactant.
#The solution is then reacted with the modified Saltzman reagent (1960) to form a deeply coloured azo-dye.
#The absorbance of the solution is then measured spectrophotometrically.

        'TEA':{'isolated_parameters':['NO2'], 'preparation_further_details':'See References: Comparison of Several Methods for Nitrogen Dioxide and Sulfur Dioxide in Metro Manila Air, Quirit, L.L, Hernandez, K.N, Lee, B.J, 2008; https://www3.epa.gov/ttn/naaqs/standards/nox/previous/1982-NOx-AQCD.pdf'},

#--------
#TGS-ANSA 

#Ambient air is bubbled through a solution containing triethanolamine, guaiacol and sodium metabilsulfite.
#Through this process, NO2 is converted to nitrite ion.
#8-Anilinonaphthalene-1-sulfonic acid ammonium salt (ANSA) and sulfanilamide are then added to the solution.
#The absorbance of the solution is then measured spectrophotometrically (at 550nm).

        'TGS-ANSA':{'isolated_parameters':['NO2'], 'preparation_further_details':'See details: https://www3.epa.gov/ttn/naaqs/standards/nox/previous/1982-NOx-AQCD.pdf'},

#--------
#Sodium Phenolate 

#A solution of alkaline phenol and hypochlorite react with ammonia to form indophenol blue.
#The solution is buffered at a pH of 9.5 with a borate buffer in order to decrease hydrolysis of cyanates and organic nitrogen compounds.
#The sample is then is then distilled into a solution of boric acid.
#The blue color formed is intensified with sodium nitroprusside and measured colorimetrically. 

        'Sodium Phenolate':{'isolated_parameters':['NH3'], 'preparation_further_details':'See further details:https://www.epa.gov/sites/production/files/2015-06/documents/epa-350.1.pdf'},

#--------
#Nessler 

#A dechlorinating agent is added to distilled water until the distillate shows no trace of ammonia. 
#NaOH is added to the distilled water until the pH is 9.5.     
#NH3 is absorbed from ambient air into the solution.
#The solution is buffered at a pH of 9.5 with a borate buffer in order to decrease hydrolysis of cyanates and organic nitrogen compounds.
#The sample is then is then distilled into a solution of boric acid.
#The distillate is then reacted with Nessler reagent.
#The absorbance of the solution is then measured spectrophotometrically (at 425nm).

        'Nessler':{'isolated_parameters':['NH3'], 'preparation_further_details':'See further details:https://www.umass.edu/mwwp/pdf/epa350_2NH3titration.pdf'},

#--------
#pararosaniline (also referred to as the West and Gaeke Method)

#SO2 is absorbed from air in a solution of potasium tetrachloromercuate. 
#A dichlorosulfitomercuate complex, which resists oxidation in air, is formed. This complex is stable to strong oxidants (i.e. ozone, NOy).
#This complex is then reacted with pararosaniline and formaldehyde to form intensely coloured C19H19N3 methyl sulphonic acid. 
#The absorbance of the solution is then measured spectrophotometrically.

        'Pararosaniline':{'isolated_parameters':['SO2'], 'preparation_further_details':'Also called West and Gaeke Method. See further details:https://nepis.epa.gov/Exe/ZyNET.exe/9101GLZ9.txt?ZyActionD=ZyDocument&Client=EPA&Index=Prior%20to%201976&Docs=&Query=&Time=&EndTime=&SearchMethod=1&TocRestrict=n&Toc=&TocEntry=&QField=&QFieldYear=&QFieldMonth=&QFieldDay=&UseQField=&IntQFieldOp=0&ExtQFieldOp=0&XmlQuery=&File=D%3A%5CZYFILES%5CINDEX%20DATA%5C70THRU75%5CTXT%5C00000022%5C9101GLZ9.txt&User=anonymous&Password=anonymous&SortMethod=h%7C-&MaximumDocuments=1&FuzzyDegree=0&ImageQuality=r75g8/r75g8/x150y150g16/i425&Display=hpfr&DefSeekPage=x&SearchBack=ZyActionL&Back=ZyActionS&BackDesc=Results%20page&MaximumPages=1&ZyEntry=12'},

#--------
#hydrogen peroxide 
#SO2 is absorbed from air in a solution of 3% hydrogen peroxide.
#This produces stable, non-volatile sulphuric acid. 
#SO3 is filtered in an 80% solution of isopropyl alcohol.
#Aliquots from the impinger are analysed by titration with barium perchlorate and 2% thorin indicator to the pink-orange end point.

        'Hydrogen Peroxide':{'isolated_parameters':['SO2'], 'preparation_further_details':'See further details:https://nepis.epa.gov/Exe/ZyNET.exe/9101GLZ9.txt?ZyActionD=ZyDocument&Client=EPA&Index=Prior%20to%201976&Docs=&Query=&Time=&EndTime=&SearchMethod=1&TocRestrict=n&Toc=&TocEntry=&QField=&QFieldYear=&QFieldMonth=&QFieldDay=&UseQField=&IntQFieldOp=0&ExtQFieldOp=0&XmlQuery=&File=D%3A%5CZYFILES%5CINDEX%20DATA%5C70THRU75%5CTXT%5C00000022%5C9101GLZ9.txt&User=ANONYMOUS&Password=anonymous&SortMethod=h%7C-&MaximumDocuments=1&FuzzyDegree=0&ImageQuality=r75g8/r75g8/x150y150g16/i425&Display=hpfr&DefSeekPage=x&SearchBack=ZyActionL&Back=ZyActionS&BackDesc=Results%20page&MaximumPages=1&ZyEntry=17 ; https://www.tandfonline.com/doi/pdf/10.1080/00022470.1972.10469715'},

#--------
#potassium iodide
#O3 is absorbed from air in a solution of 2-3% potassium iodide, oxidising iodide to iodine, and developing colour.
#The solution is nutralised with a phosphate buffer.
#The absorbance of the solution is then measured spectrophotometrically at 365nm

        'Potassium Iodide':{'isolated_parameters':['O3'], 'preparation_further_details':'See further details:https://pdf.sciencedirectassets.com/271545/1-s2.0-S0142941800X01250/1-s2.0-0142941886900644/main.pdf?X-Amz-Security-Token=IQoJb3JpZ2luX2VjEAkaCXVzLWVhc3QtMSJHMEUCIQCRe2LilIGl2GGU%2FI%2BT2%2FDeDlC23y2XTUqzHNUh06hH%2FAIgArOdObV9OOJbs2k8k%2Btog0QL0salyD1rv6mUNhQCWCgqgwQIgv%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FARAEGgwwNTkwMDM1NDY4NjUiDI0ukTnsnuSlQTJYPyrXA8srrqtssGyTMSEx%2F2CSBjHjxRwUs2jmCjv35Sj54rEu1NruQInW1GA0u6pimuc5R4t7VJISQsSm4Ri1xQiQRtpoXMsS5dI2uikJSPEdzSzhPUxPo2ZcZ7qJTMKFgX9QEM74xEkSgajhjlnoxYbVkhAbcv%2F6Xb2pZVC51hIk7mHsu4%2F1wm2csaRjHA4qSV6ZwM%2FEAkd9SS0WMu%2Bpd2bkzK%2BBo2yMPfOCT2byWGsN79EojDx58S4tMw0SbCNGTXJUVxgVojwKJjfXdG7tNMa9cZ5MVlLI7cn9P53oLLzZ4zUxKw9lb5bXiWgcJnq2K8nNLNjsXLRRpVpwnqPUURXCph4QvhgddqoY5LMD0exkdGSjwDLNKiroJg9U%2FY%2BrGgqw94uCBh8c3HtYsYSw51bVaJzQrydKXK1MLBXdUoQ83XKAr%2BijmOnqeM3AuHBlivjhpPUdh6zv8Ht3eqXGS%2BPzjb8bH3YzHhvCxmJ2CLyPLKlGbnFmDKMNJddHGWNA%2BzeGmgjR373PYfDn%2BKS9xTyUshjTGI%2FIByJ9l%2Fa7FwHaxYdq6ut%2BkgjDyl09cl4fOUbFDmmc9XA5g8%2Bgbr9MQBhhulwhDMNYrx6wyzHQ33gFKHYZbW%2FOc%2FMO6DDO58SRBjqlATF%2FNDc1eK%2F%2F7Y1I1e4oIbQse15glxyK6aJxZbnx%2FBL73pxpcENU5muolWVBEf6xGZHHtybizA7aQhIJFu4GGWlqag0BHLL1ReO32YWVjGyOA51QgWDJ2jBLnQAY5eyB5ERenFX%2F2Wq%2FYq3%2Fb9sX%2B8Xhlmy1SCBCyM7L%2FRDUe7u6Xj8AUVAe0pzlQ0VyFhmP0bflPVOx5IWiH9F7gbaXCPg0HCztUg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20220316T014927Z&X-Amz-SignedHeaders=host&X-Amz-Expires=300&X-Amz-Credential=ASIAQ3PHCVTYTDXSGOHV%2F20220316%2Fus-east-1%2Fs3%2Faws4_request&X-Amz-Signature=a752694daefc19f5bfdc585ffd8cb55a868e9e294321b95d6cd69a96b2c3f923&hash=b811b42b1065d62feb6694ed69687934242d41a217a23c1296ef5a12e7be1a66&host=68042c943591013ac2b2430a89b270f6af2c76d8dfd086a07176afe7c76c2c61&pii=0142941886900644&tid=spdf-084cbae4-ddf3-4274-b05c-61e8272c1eff&sid=3aefe3ca3835c34cb59a61f580fc5904d275gxrqb&type=client&ua=4c00045c00575753560250&rr=6ec9d5172b193682'},

#------
#detection tube
    
#Gas detection tubes operate on a chemical reaction between the vapour-phase compound and a liquid or solid detecting reagent, which is supported on an inert matrix. 
#The most common types of reactions are:
#   -Acid-base reactions These include reactions of acidic gases like HCl and HF with bases, and reaction of alkaline vapors such as ammonia with an acid in the tube. A dye present in the tube changes color as the pH changes on exposure to the vapors.
#   -Reduction-oxidation (Red-ox) reactions: These generate an oxidized or reduced compound, which has a different color. The chlorine tube uses oxidative coupling of colorless o-toluidine to form an orange azo-dye. White di-iodine pentoxide is reduced by CO and many organic vapors to form deep brown-colored iodine. Orange chromium (VI) is reduced by many organic compounds to form brown or green-colored Cr(III) compounds.
#   -Ligand-exchange reactions: These generate new complexes that are more colored than the starting reagents. The most notable is the conversion of white lead acetate to brown-black lead sulfide in the detection of H2S. In the case of phosphine, the exchange of PH3 for the chlorine ligand of HgCl2 releases HCl, which then causes a pH- dependent dye-color change.
#   -Pre-layers or Pre-tubes: These are used to condition the sample by controlling humidity, removing interferences, or transforming the analyte to another detectable compound. Examples include drying agents in NH3 and HCl tubes, organic removal by charcoal or oxidation in selective CO tubes, and oxidation of NO to NO2 in the nitrogen oxides tube.

#Most detection tubes are length-of-stain types. In these tubes, the reaction of the gas with the supported reagent is fast, compared to the transport of the bulk air sample through the tube. Therefore, all of the detected vapors are reacted within the tube. As a result, there is not a strong dependence of the readings on the rate at which the gas is sampled. However, a very high flow rate can cause some smearing to a high reading. 
#Conversely, low flow rates are less likely to affect the stain length, but can give low readings by concentrating the colored products in a shorter section of the tube. In cases of flow extremes, errors outside the standard 25% accuracy can be produced.
#Detection tubes are usually calibrated using piston hand pumps. The flow during a single pump stroke initially rises sharply and then decays exponentially. The best accuracy is therefore obtained when the flow through the tube mimics this profile.
#The advantages of detection tubes over other analytical methods are simplicity of use, rapid response, low cost, and very low maintenance. Because tubes are typically pre-calibrated, no calibration equipment is necessary. 
#Errors are prevented by directly marking the calibration information on each tube, and accuracy is further ensured by controlling the volume of air sampled. 
#Air sampling can also be performed using piston pumps, which latch into a precisely defined position to fix the volume. These pumps pull a strong vacuum initially and thus create substantially higher flowrate than the bellows pumps. Piston pumps generate a high flow initially followed by an approximately exponential decay, whereas bellows pumps provide a more steady flow initially followed by the slow decay. 
#The difference in flow patterns means that the pumps cannot be interchanged between types. For example, piston pumps sometime cause a smearing of the color stain when used on tubes originally developed for bellows pumps. This occurs because the higher flow rates do not allow enough contact time to give sharp endpoints when a piston pump is used.   

#Uncertainty on measurements is typically high with this method, typically reported between 10-25 %: https://www.honeywellanalytics.com/~/media/honeywell-analytics/products/colorimetric-gas-detection-tubes/documents/hagastubeshandbook667802enlr.pdf?la=en

#Unless additional information is given, make assumption all detection tubes trap and produce characteristic colour through reagent reaction, followed by subsequent analysis through colorimetry (manual).
        'Detection Tube':{'preparation_further_details':'See details: https://www.honeywellanalytics.com/~/media/honeywell-analytics/products/colorimetric-gas-detection-tubes/documents/hagastubeshandbook667802enlr.pdf?la=en'}
    }
},

#------------------------------------
#intermediate measurement

#A measurement is made by instrument/manually which feeds later into the ultimate measurement  

#---No specific sampling preparation name
'intermediate measurement':{
    'specific_techniques':{}
},

#------------------------------------
#unknown

'unknown':{
    'specific_techniques':{}
}
}


#----------------------------------------------------------------------------------------#
#STANDARD MEASUREMENT METHOD INFORMATION
#----------------------------------------------------------------------------------------#

#define dictionary that contains information pertaining to standardised measurement methodologies and specific instrumentation
#the "measurement methodology" refers to the name of the method used by the measuring instrument to determine the measured values for a parameter)

standard_measurement_methodologies = {
#------------------------------------
#ultraviolet photometry

#Operates on the principle that a specific species efficiently absorbs light at a known wavelength in the UV range. This is the case for ozone, at 253.65nm.
#The degree to which the UV light is absorbed by a specific species is directly related to the species concentration as described by tha (I/Io = e−KLC; K = molecular absorption coefficient at STP (308 cm-1 atm-1 for O3), L = optical path length of cell, C = species concentration , I = light intensity of sample gas, Io = light intensity of sample without measured species (reference gas) ).

#Typical measurement operation for ozone:
#1. Ambient air is continuously drawn through the analyser by a vacuum pump. 
#2. Sample is split into two flow paths. One path incorporates a scrubber which selectively removes ozone, while the other path does not. 
#3. Both flow paths are connected to a solenoid valve, which switches at a fixed interval to allow either the sample gas stream (I), or the scrubbed, reference gas stream (Io), to flow through a quartz tube (cell) of accurately known length. 
#4. Typically a mercury vapour lamp at one end of the quartz cell produces a monochromatic beam of ultraviolet light at 254 nm. A vacuum diode at the opposite end of the cell measures the intensity of transmitted light (I/Io). 
#5. The intensities of 254 nm UV light transmitted through the sample (I) and ozone-free reference gas streams (Io) are related to the concentration (C) of ozone in the sample gas stream according to the Beer-Lambert Law.
#6. The analyzer’s microprocessor calculates the absolute concentration in molecules cm-3, and is then converted to a mixing ratio using in-instrument measured sample temperature and pressure.

#Known interferences: Gaseous hydrocarbons with strong absorption at 254 nm, such as aromatic hydrocarbons (i.e., benzene and substituted benzene rings).

#The greatest source of uncertainty regarding this measurement type is the cross sectional value used, for which the conventional value has been called into question: see further details: https://www.bipm.org/en/bipm/chemistry/gas-metrology/ozone/ozone-cross-section.html

'ultraviolet photometry':{'abbreviation':'UVP', 'sampling_type':'low volume continuous', 'measured_parameters':['O3'], 'qa_accepted_parameters':['O3'], 
    'instruments':{
        '2B 202':                     {'documented_lower_limit_of_detection':'3.0',  'documented_upper_limit_of_detection':'250000.0', 'documented_precision':'1.5/2.0%',  'documented_accuracy':'1.5/2.0%',  'documented_zero_drift':'2.0/day',    'documented_span_drift':'1.0%/day',       'documented_resolution':'0.1',   'documented_absorption_cross_section':'1.15e-17',   'documented_flow_rate':'1.0',     'instrument_manual_name':'2B_202_Manual.pdf'},
        '2B 205':                     {'documented_lower_limit_of_detection':'2.0',  'documented_upper_limit_of_detection':'250000.0', 'documented_precision':'1.0/2.0%',  'documented_accuracy':'1.0/2.0%',  'documented_zero_drift':'1.0/day',    'documented_span_drift':'1.0%/day',       'documented_resolution':'0.1',   'documented_absorption_cross_section':'1.15e-17',   'documented_flow_rate':'1.8',     'instrument_manual_name':'2B_205_Manual.pdf'},
        '2B 211':                     {'documented_lower_limit_of_detection':'1.0',  'documented_upper_limit_of_detection':'2000.0',   'documented_precision':'0.5/1.0%',  'documented_accuracy':'1.0/2.0%',  'documented_zero_drift':'1.0/day',    'documented_span_drift':'0.5%/day',       'documented_resolution':'0.1',   'documented_absorption_cross_section':'1.15e-17',   'documented_flow_rate':'2.0',     'instrument_manual_name':'2B_211_Manual.pdf'},
        '2B 106-L':                   {'documented_lower_limit_of_detection':'3.0',  'documented_upper_limit_of_detection':'100000.0', 'documented_precision':'1.5/2.0%',  'documented_accuracy':'1.5/2.0%',  'documented_zero_drift':'3.0/day',    'documented_span_drift':'1.0%/day',       'documented_resolution':'0.1',   'documented_absorption_cross_section':'1.15e-17',   'documented_flow_rate':'1.0',     'instrument_manual_name':'2B_106-L_Manual.pdf'},
        '2B 106-OEM-L':               {'documented_lower_limit_of_detection':'3.0',  'documented_upper_limit_of_detection':'100000.0', 'documented_precision':'1.5/2.0%',  'documented_accuracy':'1.5/2.0%',  'documented_zero_drift':'3.0/day',    'documented_span_drift':'1.0%/day',       'documented_resolution':'0.1',   'documented_absorption_cross_section':'1.15e-17',   'documented_flow_rate':'1.0',     'instrument_manual_name':'2B_106-L_Manual.pdf'},
        '2B POM':                     {'documented_lower_limit_of_detection':'3.0',  'documented_upper_limit_of_detection':'10000.0',  'documented_precision':'1.5/2.0%',  'documented_accuracy':'1.5/2.0%',  'documented_zero_drift':'2.0/day',    'documented_span_drift':'1.0%/day',       'documented_resolution':'0.1',   'documented_absorption_cross_section':'1.15e-17',   'documented_flow_rate':'0.8',     'instrument_manual_name':'2B_POM_Manual.pdf'},
        'Airpointer-O3':              {'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'20000.0',  'documented_precision':'1.0',       'documented_accuracy':'1.0%>=100', 'documented_zero_drift':'1.0/day',    'documented_span_drift':'1.0/1.0%/day',                                                                                        'documented_flow_rate':'1.0',     'instrument_manual_name':'Airpointer_Specs.pdf'},
        'CSI 3100':                   {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:https://www.healtheffects.org/system/files/ResearchReport-70.pdf'},
        'Dasibi 1003-AH':             {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Dasibi 1003-PC':             {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Dasibi 1003-RS':             {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Dasibi 1006-AHJ':            {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://conf.montreal-protocol.org/meeting/orm/9orm/presession/Information%20Documents%20are%20available%20in%20English%20onl/Indonesia.pdf'},
        'Dasibi 1007-AHJ':            {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:https://ebcrpa.jamstec.go.jp/atmoscomp/obsdata/mondy.html'},
        'Dasibi 1008-AH':             {'documented_lower_limit_of_detection':'2.0',  'documented_upper_limit_of_detection':'1000.0',   'documented_precision':'1.0',       'documented_accuracy':'1.0',                                                                                                                        'documented_absorption_cross_section':'1.1476e-17', 'documented_flow_rate':'2.0',     'instrument_manual_name':'Dasibi_1008_Manual.pdf'},
        'Dasibi 1008-HC':             {'documented_lower_limit_of_detection':'2.0',  'documented_upper_limit_of_detection':'1000.0',   'documented_precision':'1.0',       'documented_accuracy':'1.0',                                                                                                                        'documented_absorption_cross_section':'1.1476e-17', 'documented_flow_rate':'2.0',     'instrument_manual_name':'Dasibi_1008_Manual.pdf'},
        'Dasibi 1008-PC':             {'documented_lower_limit_of_detection':'2.0',  'documented_upper_limit_of_detection':'1000.0',   'documented_precision':'1.0',       'documented_accuracy':'1.0',                                                                                                                        'documented_absorption_cross_section':'1.1476e-17', 'documented_flow_rate':'2.0',     'instrument_manual_name':'Dasibi_1008_Manual.pdf'},
        'Dasibi 1008-PS':             {'documented_lower_limit_of_detection':'2.0',  'documented_upper_limit_of_detection':'1000.0',   'documented_precision':'1.0',       'documented_accuracy':'1.0',                                                                                                                        'documented_absorption_cross_section':'1.1476e-17', 'documented_flow_rate':'2.0',     'instrument_manual_name':'Dasibi_1008_Manual.pdf'},
        'Dasibi 1008-RS':             {'documented_lower_limit_of_detection':'2.0',  'documented_upper_limit_of_detection':'1000.0',   'documented_precision':'1.0',       'documented_accuracy':'1.0',                                                                                                                        'documented_absorption_cross_section':'1.1476e-17', 'documented_flow_rate':'2.0',     'instrument_manual_name':'Dasibi_1008_Manual.pdf'},
        'Dasibi 1108 W/GEN':          {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:https://books.google.es/books?id=SuIsDwAAQBAJ&pg=PA48&lpg=PA48&dq=dasibi+1108&source=bl&ots=tnXwozqtBT&sig=ACfU3U0bA3efWduqeNFD2g2zk-RqP7cQFg&hl=en&sa=X&ved=2ahUKEwjRk8bX2f7fAhUD4OAKHSr1A5wQ6AEwDHoECAAQAQ#v=onepage&q=dasibi%201108&f=false'},
        'DKK-TOA GUX-113E':           {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:https://books.google.es/books?id=KpQhwCLgVgAC&pg=PA209&lpg=PA209&dq=GUX-113E&source=bl&ots=e1rn8FsfBV&sig=Sd0wYb5DAjG4GZBrpsFPAuZ6Uqo&hl=en&sa=X&ved=0ahUKEwjbivvTq67bAhVDuRQKHY50ASYQ6AEIMjAB#v=onepage&q=GUX-113E&f=false'},
        'DKK-TOA GUX-313E':           {'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'1000.0',                                                                          'documented_zero_drift':'1.0/day',    'documented_span_drift':'1.0%/day',                                                                                                                              'instrument_manual_name':'DKK-TOA_GUX-313E_Specs.pdf'} ,
        'Dylec 1100':                 {'documented_lower_limit_of_detection':'0.1',  'documented_upper_limit_of_detection':'500.0',    'documented_precision':'0.5%',      'documented_accuracy':'0.5%',      'documented_zero_drift':'0.5%/month', 'documented_span_drift':'0.5%/month',                                                                                          'documented_flow_rate':'1.5-2.0', 'instrument_further_details':'Specifications online:https://translate.google.com/translate?hl=en&sl=ja&u=http://www.dylec.co.jp/01ozone/m1100.html&prev=search'},
        'Dylec 1150':                 {'documented_lower_limit_of_detection':'1.0',  'documented_upper_limit_of_detection':'500.0',    'documented_precision':'0.5%',      'documented_accuracy':'0.5%',      'documented_zero_drift':'0.5%/month', 'documented_span_drift':'0.5%/month',                                                                                          'documented_flow_rate':'1.5-2.0', 'instrument_further_details':'Specifications online:https://translate.google.com/translate?hl=en&sl=ja&u=http://www.dylec.co.jp/01ozone/m1100.html&prev=search'},
        'Dylec 1200':                 {                                              'documented_upper_limit_of_detection':'20000.0',  'documented_precision':'0.5%',      'documented_accuracy':'0.5%',      'documented_zero_drift':'0.5%/month', 'documented_span_drift':'0.5%/month',                                                                                          'documented_flow_rate':'1.5-2.0', 'instrument_further_details':'Specifications online:https://translate.google.com/translate?hl=en&sl=ja&u=http://www.dylec.co.jp/01ozone/m1200.html&prev=search'},
        'Ebara EG-2001F':             {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here: http://www.data.jma.go.jp/gmd/env/ghg_obs/en/station/station_minamitorishima.html'},
        'Ebara EG-2001FTP':           {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here: http://www.data.jma.go.jp/gmd/env/ghg_obs/en/station/station_minamitorishima.html'},
        'Ebara EG-3000F':             {                                              'documented_upper_limit_of_detection':'1000.0',   'documented_precision':'5.0',       'documented_accuracy':'5.0',       'documented_zero_drift':'5.0/month',  'documented_span_drift':'5.0/month',                                                                                           'documented_flow_rate':'1.5',     'instrument_manual_name':'Ebara_EG-3000F_Manual.pdf'},
        'Ecotech EC9810':             {'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'20000.0',  'documented_precision':'1.0/1.0%',                                     'documented_zero_drift':'1.0/day',    'documented_span_drift':'0.5%/day',       'documented_resolution':'0.001', 'documented_absorption_cross_section':'1.1476e-17', 'documented_flow_rate':'0.5',     'instrument_manual_name':'Ecotech_EC9810_Manual.pdf'},
        'Ecotech Serinus10':          {'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'20000.0',  'documented_precision':'0.5/0.2%',  'documented_accuracy':'1.0%',      'documented_zero_drift':'0.3/day',    'documented_span_drift':'0.3/0.5%/7days',                                                                                      'documented_flow_rate':'0.5',     'instrument_manual_name':'Ecotech_Serinus10_Specs.pdf'},
        'Environics 300':             {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Environnement SA O331M':     {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:https://www.geo.fu-berlin.de/met/ag/trumf/Lehre/Lehrveranstaltungen_alt/Dok_2006/Vorlesung_SchwarzesDreieck_20060503.pdf'},
        'Environnement SA O341M':     {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Environnement SA O342e':     {'documented_lower_limit_of_detection':'0.2',  'documented_upper_limit_of_detection':'10000.0',                                      'documented_accuracy':'1.0%',      'documented_zero_drift':'0.5/day',    'documented_span_drift':'0.5%/day',                                                                                            'documented_flow_rate':'1.0',     'instrument_manual_name':'EnvironnementSA_O342e_Specs.pdf'},
        'Environnement SA O342M':     {'documented_lower_limit_of_detection':'0.4',  'documented_upper_limit_of_detection':'10000.0',                                      'documented_accuracy':'1.0%',      'documented_zero_drift':'1.0/week',   'documented_span_drift':'1.0%/week',                                                                                           'documented_flow_rate':'1.0',     'instrument_manual_name':'EnvironnementSA_O342M_Specs.pdf'},
        'Horiba APOA300':             {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:https://books.google.es/books?id=qmWXBwAAQBAJ&pg=PA88&lpg=PA88&dq=horiba+apna+300&source=bl&ots=dBQJ0TEai1&sig=plLm6SAUIZwuIhy37JzI8nZXccQ&hl=en&sa=X&ved=2ahUKEwihxOS9uvDfAhWGnhQKHYzoAPcQ6AEwCHoECAIQAQ#v=onepage&q=horiba%20apna%20300&f=false'},
        'Horiba APOA350':             {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:https://books.google.es/books?id=WqvzBQAAQBAJ&pg=SL13-PA35&lpg=SL13-PA35&dq=Horiba+APOA350&source=bl&ots=Ltdp9qBqrc&sig=qKEZbhp9Vd0UPEbMIL6Al8CKITM&hl=en&sa=X&ved=0ahUKEwjcjJKai9bbAhXH7BQKHa_gDvQ4ChDoAQgtMAE#v=onepage&q=Horiba%20APOA350&f=false'},
        'Horiba APOA360':             {'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'10000.0',  'documented_precision':'1.0%',      'documented_accuracy':'1.0%',      'documented_zero_drift':'0.5/day',    'documented_span_drift':'0.5/day',                                                                                             'documented_flow_rate':'0.8',     'instrument_manual_name':'Horiba_APMA360_Specs.pdf'},
        'Horiba APOA370':             {'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'100000.0', 'documented_precision':'1.0%',      'documented_accuracy':'1.0%',      'documented_zero_drift':'0.5/day',    'documented_span_drift':'0.5/day',                                                                                             'documented_flow_rate':'0.7',     'instrument_manual_name':'Horiba_APOA370_Specs.pdf'},
        'MCV 48AUV':                  {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://www.academia.edu/18561546/Temporal_patterns_of_surface_ozone_levels_in_different_habitats_of_the_North_Western_Mediterranean_basin'},
        'PCI LC-12':                  {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Philips K50094':             {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/philipsK50094/view?facet=HTML+Representation'},   
        'Philips K50110':             {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:https://www.google.co.uk/url?sa=t&rct=j&q=&esrc=s&source=web&cd=&ved=2ahUKEwjciLvpw_zzAhUWxBQKHRpLB784FBAWegQIAhAB&url=https%3A%2F%2Fhamtplats.miljodatasamverkan.se%2Fvalid%2Fmallar%2Fluft%2FUppgifter_ny_m%25C3%25A4tstation_2020_1.xlsx&usg=AOvVaw1uId5jGqZrElo-oGwsNGUc'},   
        'Sabio 6030':                 {'documented_lower_limit_of_detection':'0.4',  'documented_upper_limit_of_detection':'10000.0',  'documented_precision':'0.5%',      'documented_accuracy':'0.5%',      'documented_zero_drift':'1.0/month',  'documented_span_drift':'1.0%/month',                                                                                          'documented_flow_rate':'0.5-1.0', 'instrument_manual_name':'Sabio_6030_Specs.pdf'},
        'Seres OZ 2000G':             {'documented_lower_limit_of_detection':'1.0',  'documented_upper_limit_of_detection':'10000.0',                                      'documented_accuracy':'1.0%',      'documented_zero_drift':'1.5/week',   'documented_span_drift':'1.0%/month',                                                                                          'documented_flow_rate':'1.0',     'instrument_manual_name':'Seres_OZ2000G_Specs.pdf'},
        'SIR S-5014':                 {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here: http://www.ecomission.co.za/products/model_s_5014.htm'},
        'Tanabyte 722':               {'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'10000.0',  'documented_precision':'1.0',       'documented_accuracy':'1.0%',      'documented_zero_drift':'1.0%/week',  'documented_span_drift':'1.0%/week',      'documented_resolution':'0.1',                                                       'documented_flow_rate':'0.5-2.0', 'instrument_manual_name':'Tanabyte_72x_Specs.pdf'},
        'Tanabyte 723':               {'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'10000.0',  'documented_precision':'1.0',       'documented_accuracy':'1.0%',      'documented_zero_drift':'1.0%/week',  'documented_span_drift':'1.0%/week',      'documented_resolution':'0.1',                                                       'documented_flow_rate':'0.5-2.0', 'instrument_manual_name':'Tanabyte_72x_Specs.pdf'},
        'Tanabyte 724':               {'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'10000.0',  'documented_precision':'1.0',       'documented_accuracy':'1.0%',      'documented_zero_drift':'1.0%/week',  'documented_span_drift':'1.0%/week',      'documented_resolution':'0.1',                                                       'documented_flow_rate':'0.5-2.0', 'instrument_manual_name':'Tanabyte_72x_Specs.pdf'},
        'Tanabyte 725':               {'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'10000.0',  'documented_precision':'1.0',       'documented_accuracy':'1.0%',      'documented_zero_drift':'1.0%/week',  'documented_span_drift':'1.0%/week',      'documented_resolution':'0.1',                                                       'documented_flow_rate':'0.5-2.0', 'instrument_manual_name':'Tanabyte_72x_Specs.pdf'},
        'Tanabyte 726':               {'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'10000.0',  'documented_precision':'1.0',       'documented_accuracy':'1.0%',      'documented_zero_drift':'1.0%/week',  'documented_span_drift':'1.0%/week',      'documented_resolution':'0.1',                                                       'documented_flow_rate':'0.5-2.0', 'instrument_manual_name':'Tanabyte_72x_Specs.pdf'},
        'Tanabyte 734':               {'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'10000.0',  'documented_precision':'1.0',       'documented_accuracy':'1.0%',      'documented_zero_drift':'1.0%/week',  'documented_span_drift':'1.0%/week',      'documented_resolution':'0.1',                                                       'documented_flow_rate':'0.5-2.0', 'instrument_manual_name':'Tanabyte_72x_Specs.pdf'},
        'Teledyne 400':               {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},   
        'Teledyne 400A':              {'documented_lower_limit_of_detection':'0.6',  'documented_upper_limit_of_detection':'10000.0',  'documented_precision':'0.5%',      'documented_accuracy':'1.0%',      'documented_zero_drift':'1.0/day',    'documented_span_drift':'1.0%/day',                                        'documented_absorption_cross_section':'1.1476e-17', 'documented_flow_rate':'0.8',     'instrument_manual_name':'Teledyne_400A_Manual.pdf'},   
        'Teledyne 400E':              {'documented_lower_limit_of_detection':'0.6',  'documented_upper_limit_of_detection':'10000.0',  'documented_precision':'0.5%',      'documented_accuracy':'1.0%',      'documented_zero_drift':'1.0/day',    'documented_span_drift':'1.0%/day',                                        'documented_absorption_cross_section':'1.1476e-17', 'documented_flow_rate':'0.8',     'instrument_manual_name':'Teledyne_400E_Manual.pdf'},   
        'Teledyne ML8810':            {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Teledyne ML9810':            {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Teledyne ML9810A':           {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/ML9810A/view'},
        'Teledyne ML9810B':           {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Teledyne ML9811':            {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Teledyne ML9812':            {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Teledyne T204-O3':           {'documented_lower_limit_of_detection':'2.0',  'documented_upper_limit_of_detection':'1000.0',   'documented_precision':'0.5%>=100', 'documented_accuracy':'1.0%',      'documented_zero_drift':'1.0%/day',   'documented_span_drift':'1.0%/day',                                        'documented_absorption_cross_section':'1.1476e-17', 'documented_flow_rate':'0.8',     'instrument_manual_name':'Teledyne_T204_Manual.pdf'},
        'Teledyne T400':              {'documented_lower_limit_of_detection':'0.4',  'documented_upper_limit_of_detection':'10000.0',  'documented_precision':'0.5%>=100', 'documented_accuracy':'1.0%',      'documented_zero_drift':'1.0/day',    'documented_span_drift':'1.0%/day',                                                                                            'documented_flow_rate':'0.8',     'instrument_manual_name':'Teledyne_T400_Manual.pdf'},
        'Thermo 49':                  {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Thermo 49C':                 {'documented_lower_limit_of_detection':'1.0',  'documented_upper_limit_of_detection':'200000.0', 'documented_precision':'1.0',       'documented_accuracy':'1.0%',      'documented_zero_drift':'1.0/day',    'documented_span_drift':'1.0%/month',                                      'documented_absorption_cross_section':'1.1476e-17', 'documented_flow_rate':'1.0-3.0', 'instrument_manual_name':'Thermo_49C_Manual.pdf'},   
        'Thermo 49CPS':               {'documented_lower_limit_of_detection':'1.0',  'documented_upper_limit_of_detection':'5000.0',                                       'documented_accuracy':'1.0%',      'documented_zero_drift':'100.0/day',  'documented_span_drift':'1.0%/day',                                                                                            'documented_flow_rate':'1.0-3.0', 'instrument_manual_name':'Thermo_49CPS_Specs.pdf'},
        'Thermo 49i':                 {'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'200000.0', 'documented_precision':'1.0',       'documented_accuracy':'1.0%',      'documented_zero_drift':'1.0/day',    'documented_span_drift':'1.0%/month',                                      'documented_absorption_cross_section':'1.1476e-17', 'documented_flow_rate':'1.0-3.0', 'instrument_manual_name':'Thermo_49i_Manual.pdf,Thermo_49i_Specs.pdf'},
        'Thermo 49w':                 {                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:https://www.rivm.nl/en/Documents_and_publications/Scientific/Reports/1993/juni/Acceptance_procedure_and_testresults_of_the_ozone_analyzers_model_49W_made_by_Thermo_Environmental_Instruments_on_behalf_of_the_Dutch_National_Air_Quality_Monitoring_Network'},
        'Wedding and Associates 1010':{                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet photometry here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'}
    }
},

#------------------------------------
#visible photometry

#Operates on the principle that a specific species efficiently absorbs light at a known wavelength in the visible range. This is the case for NO2, at 405nm
#The degree to which the visible light is absorbed by a specific species is directly related to the species concentration as described by the Beer-Lambert Law (C = 1/Lσ * ln(Io/I) ; σ = absorption cross section (6.06×10-19 cm2 molec-1 for NO2 at 405nm), L = optical path length of cell, C = species concentration , I = light intensity of sample gas, Io = light intensity of sample without measured species (reference gas) ).

#Typical measurement operation for NO2 (and indirectly NO):
#1. Ambient air is continuously drawn through the instrument by a vacuum pump. 
#2. An NO2 scrubber valve alternately bypasses and sends the sample air through a heated NO2 scrubber to remove all NO2 in the sample. 
#3. The NO2-scrubbed/unscrubbed air alternately passes through the optical cell (of known length) and the cell flow meter. Alternate switching of the NO2 scrubber valve once every 5 seconds allows the measurement of a light intensity in the absence of NO2 and presence of NO2. 
#4. A light-emitting diode (LED) emits a monochromatic beam of light at 405nm. A photodiode at the other end of the cell measures the intensity of transmitted light (I/Io).
#5. The intensities of 405 nm light transmitted through the sample (I) and NO2-free reference gas streams (Io) are related to the concentration (C) of ozone in the sample gas stream according to the Beer-Lambert Law.
#6. The analyser's microprocessor calculates the absolute concentration in molecules cm-3, and is then converted to a mixing ratio using in-instrument measured sample temperature and pressure.
#7. NO is measured indirectly by bypassing the NO2 scrubber and measuring the light intensity while adding (I) or not adding (Io) ozone to convert NO to NO2 according to the well-known reaction: NO + O3 → NO2 + O2

#Known interferences: Water vapour, small particles (< 5 um)

'visible photometry':{'abbreviation':'VP', 'sampling_type':'low volume continuous', 'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO','NO2','NOx'], 
    'instruments':{
        '2B 405nm':{'documented_lower_limit_of_detection':'1.0', 'documented_upper_limit_of_detection':{'NO':'2000.0','NO2':'10000.0'}, 'documented_precision':'0.5/0.5%', 'documented_resolution':'0.1', 'documented_absorption_cross_section':'6.06e-19', 'instrument_manual_name':'2B_405nm_Manual.pdf'}
    }
},
#--------------------------------------------------------------------------------------###
#ethylene chemiluminescence 

#Chemiluminescence is the emission of light (luminescence), as the result of a chemical reaction. Chemiluminescence occurs as a result of the reaction of ozone with ethylene, leading to an excited molecule.
#The return to a fundamental electronic state of the excited molecules is made by luminous radiation in a specific spectrum, which can be measured.
#The concentration of sample ozone is directly proportional to the intensity of light emitted.
#The broadband emission is detected using a photomultiplier tube (at 440 nm for ethylene + ozone).

#Known interferences: water vapour

'ethylene chemiluminescence':{'abbreviation':'ECL', 'sampling_type':'low volume continuous', 'measured_parameters':['O3'], 'qa_accepted_parameters':['O3'], 
    'instruments':{
        'Beckman 950A':    {                                                                                             'instrument_further_details':'No manual or specifications available but confirmation it is uses ethylene chemiluminescence here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Bendix 8002':     {'documented_lower_limit_of_detection':'1.0',  'documented_upper_limit_of_detection':'200.0', 'instrument_further_details':'No manual or specifications available but confirmation it is uses ethylene chemiluminescence here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf ; details about limit of detection:https://books.google.es/books?id=G0EYUql3_98C&pg=PA283&lpg=PA283&dq=thermo+43a&source=bl&ots=k2RUJXlAVJ&sig=yIxG08AoSyuqedtS6dp-O7O04X4&hl=en&sa=X&ved=0ahUKEwj17qqB3OfbAhVCVRQKHanFDlo4ChDoAQgoMAA#v=onepage&q=thermo%2043a&f=false'},
        'CSI 2000':        {                                                                                             'instrument_further_details':'No manual or specifications available but confirmation it is uses ethylene chemiluminescence here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'McMillan 1100-1': {                                                                                             'instrument_further_details':'No manual or specifications available but confirmation it is uses ethylene chemiluminescence here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'McMillan 1100-2': {                                                                                             'instrument_further_details':'No manual or specifications available but confirmation it is uses ethylene chemiluminescence here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'McMillan 1100-3': {                                                                                             'instrument_further_details':'No manual or specifications available but confirmation it is uses ethylene chemiluminescence here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Meloy 0A325-2R':  {                                                                                             'instrument_further_details':'No manual or specifications available but confirmation it is uses ethylene chemiluminescence here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Meloy 0A350-2R':  {                                                                                             'instrument_further_details':'No manual or specifications available but confirmation it is uses ethylene chemiluminescence here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'},
        'Teledyne ML8410E':{                                                                                             'instrument_further_details':'No manual or specifications available but confirmation it is uses ethylene chemiluminescence here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'}
    }
},

#--------------------------------------------------------------------------------------###
#eosin Y chemiluminescence 

#Chemiluminescence is the emission of light (luminescence), as the result of a chemical reaction. Chemiluminescence occurs as a result of the reaction of ozone with eosin-Y, leading to an excited molecule.
#The return to a fundamental electronic state of the excited molecules is made by luminous radiation in a specific spectrum, which can be measured.
#The concentration of sample ozone is directly proportional to the intensity of light emitted.
#The broadband emission is detected using a photomultiplier tube (at ~560 nm for eosin-Y + ozone).

#Known interferences: water vapour

'eosin Y chemiluminescence':{'abbreviation':'EYCL', 'sampling_type':'low volume continuous', 'measured_parameters':['O3'], 'qa_accepted_parameters':['O3'], 
    'instruments':{
        'Luminox LOZ-3':{'instrument_further_details':'No manual or specifications available but confirmation it is uses eosin-Y chemiluminescence here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'}
    }
},

#--------------------------------------------------------------------------------------###
#rhodamine B chemiluminescence 

#Chemiluminescence is the emission of light (luminescence), as the result of a chemical reaction. Chemiluminescence occurs as a result of the reaction of ozone with rhodamine B, leading to an excited molecule.
#The return to a fundamental electronic state of the excited molecules is made by luminous radiation in a specific spectrum, which can be measured.
#The concentration of sample ozone is directly proportional to the intensity of light emitted.
#The broadband emission is detected using a photomultiplier tube (at ~580 nm for rhodamine B + ozone).

#Known interferences: water vapour

'rhodamine B chemiluminescence':{'abbreviation':'RBC', 'sampling_type':'low volume continuous', 'measured_parameters':['O3'], 'qa_accepted_parameters':['O3'], 
    'instruments':{
        'Philips PW9771': {'instrument_further_details':'No manual or specifications available but confirmation it is uses rhodamine B chemiluminescence here:http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf'}
    }
},

#------------------------------------
#chemiluminescence (internal molybdenum converter)

#Chemiluminescence is the emission of light (luminescence), as the result of a chemical reaction. Chemiluminescence occurs as a result of the reaction of NO with ozone (NO+O3 --> NO2*+O2).
#The return to a fundamental electronic state of the excited NO2* molecules is made by luminous radiation in a 600-3000 nm spectrum (NO2* --> NO2 + hv), which can be measured.
#This method is designed specifically to directly measure NO directly (and NOx/NO2 indirectly). O3 can also be measured by inverting the measurement technique (using canister of NO for reaction).

#Typical measurement operation for NO (and NOx/NO2 indirectly):
#1. Sample air is drawn into the reaction cell via two separate (alternating) paths; the NO and NOx channels.
#2. NO in the first path (NO channel) reacts with ozone (from ozone generator) according to the following reaction: NO+O3 --> NO2*+O2.
#3. The excited NO2 molecules quickly returns to the ground state, releasing the excess energy. This release takes the form of a quantum of light (hv). The distribution of wavelengths for these quanta range between 600 and 3000 nm, with a peak at about 1200 nm (NO2* --> NO2 +hv (1200nm))
#4. Luminescence from the reaction is detected using a photomultiplier tube (PMT). Photons enter the PMT and strike a negatively charged photo cathode causing it to emit electrons. These electrons are accelerated by an applied high voltage and multiplied through a sequence of similar acceleration steps (dynodes) until a useable current signal is generated. The more light present (in this case photons given off by the chemiluminescent reaction described above), the more current is produced. Therefore the more NO present in the reaction cell the more current is produced by the PMT.
#5. Before entering the PMT, light is passed through a high-pass optical filter only transparent to wavelengths of light above 645nm. The narrowness of this band of sensitivity allows avoids the potential of extraneous light and radiation that might interfere with the measurement (e.g. some oxides of sulfur can also be chemiluminescent emitters when in contact with O3 but give off light at much shorter wavelengths (usually around 260nm to 480nm)).
#6. All things being constant (temperature, pressure, amount of ozone present, etc.), the relationship between the amount of NO present in the reaction cell and the amount of light emitted from the reaction is very linear. If more NO is present, more IR light is produced. 
#7. In addition, sometimes the excited NO2 collides with other gaseous molecules in the reaction cell chamber or even the molecules of the reaction cell walls and transfers its excess energy to this collision partner (represented by M) without emitting any light at all. In fact, by far the largest portion of the excited NO2 returns to the ground state this way, leaving only a few percent yield of usable chemiluminescence (NO2* + M --> NO2 + M). The probability of a collision between the NO2* molecule and a collision partner M increases proportionally with the reaction cell pressure. This non-radiating collision with the NO2* molecules is usually referred to as third body quenching. Even under the best conditions only about 20% of the NO2 that is formed by the key chemiluminescence reaction is in the excited state. In order to maximize chemiluminescence, the reaction cell is maintained at reduced pressure (thereby reducing the amount of available collision partners) and is supplied with a large, constant excess of ozone (about 3000-5000 ppm) from the internal ozone generator.
#8. The second path (NOx channel) travels through a delay loop and a heated molybdenum converter (at 315degC). The heated molybdenum reacts with NO2 in the sample gas and produces a NO gas and a variety of molybdenum. (xNO2 + yMo --> xNO + MyOz (at 315degC)).
#9. Once the NO2 in the sample gas has been converted to NO, it is routed to the reaction cell where it undergoes the chemiluminescence reaction and is measured by the PMT.
#10. By converting the NO2 in the sample gas into NO, the analyser indirectly measures total NOx content of the sample gas (i.e. the NO present + the converted NO2 present). 
#11. By switching the sample gas stream in and out of the molybdenum converter every 6 - 10 seconds, the analyser is able to quasi-continuously measure both the NO and the total NOx content.
#12. The NO2 concentration is not directly measured but calculated by subtracting the known NO content of the sample gas from the known NOx content.

#Known interferences: Water vapour (above 20 ppmv), 3rd body quenching (CO2, SOx), other NOy species converted to NO by molybdenum converter (PAN, ethyl nitrate, ethyl nitrite, HONO, HNO3, methyl nitrate, n-propyl nitrate, n-butyl nitrate, nitrocresol, NH3), other species undergoing chemiluminescence with O3 (SOx).

#Not QA accepted measurement technique for NO2/NOx due to bias in measuring NO2, due to molybdenum converters.

'chemiluminescence (internal molybdenum converter)':{'abbreviation':'CL(IMC)', 'sampling_type':'low volume continuous', 'measured_parameters':['NO','NO2','NOx','O3'], 'qa_accepted_parameters':['NO','O3'], 
    'instruments':{
        'Airpointer-NOx':        {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.4',  'documented_upper_limit_of_detection':'20000.0',                                       'documented_precision':'1.0/1.0%', 'documented_accuracy':'1.0%>=100.0', 'documented_zero_drift':'0.4/day',   'documented_span_drift':'1.0%>=100/day',                                                 'documented_flow_rate':'1.0',     'instrument_manual_name':'Airpointer_Specs.pdf', 'measuring_assumption':True},
        'Beckman 952A':          {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://www.tandfonline.com/doi/pdf/10.1080/00022470.1979.10470816?needAccess=true', 'measuring_assumption':True},
        'Bendix 8101-B':         {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://www.tandfonline.com/doi/pdf/10.1080/00022470.1980.10465149', 'measuring_assumption':True},
        'Bendix 8101-C':         {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://www.ce.berkeley.edu/sites/default/files/assets/users/kirchstetter/Kirchstetter%20et%20al%2C%20Nitrous%20Acid%20in%20Motor%20Vehicle%20Exhaust%2C%20EST%201996.pdf', 'measuring_assumption':True},
        'CSI 1600':              {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://nepis.epa.gov/Exe/ZyNET.exe/30003UBG.txt?ZyActionD=ZyDocument&Client=EPA&Index=1986%20Thru%201990&Docs=&Query=&Time=&EndTime=&SearchMethod=1&TocRestrict=n&Toc=&TocEntry=&QField=&QFieldYear=&QFieldMonth=&QFieldDay=&UseQField=&IntQFieldOp=0&ExtQFieldOp=0&XmlQuery=&File=D%3A%5CZYFILES%5CINDEX%20DATA%5C86THRU90%5CTXT%5C00000006%5C30003UBG.txt&User=ANONYMOUS&Password=anonymous&SortMethod=h%7C-&MaximumDocuments=1&FuzzyDegree=0&ImageQuality=r75g8/r75g8/x150y150g16/i425&Display=hpfr&DefSeekPage=x&SearchBack=ZyActionL&Back=ZyActionS&BackDesc=Results%20page&MaximumPages=1&ZyEntry=1', 'measuring_assumption':True},
        'Dasibi 2108':           {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter:https://books.google.es/books?id=-I7pCAAAQBAJ&pg=PA372&lpg=PA372&dq=DASIBI+2108+manual&source=bl&ots=ZwNR_ShKjO&sig=yUbYYfnJXqD8eAJovJ6pNjaa468&hl=en&sa=X&ved=0ahUKEwj_jL_SxqzbAhXrB8AKHXvKDkUQ6AEIWTAI#v=onepage&q=DASIBI%202108%20manual&f=false', 'measuring_assumption':True},
        'DKK-TOA GLN-114E':      {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'cannot find converter documentation but states instrument is previous version of DKK-TOA GLN-314E here: https://www.toadkk.co.jp/english/product/env/saleend/ap.html. This uses internal molybdenum converter, so assume same here', 'measuring_assumption':True},
        'DKK-TOA GLN-314E':      {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.4',  'documented_upper_limit_of_detection':'2000.0',                                                                                                                'documented_zero_drift':'0.6/day',   'documented_span_drift':'1.0%/day',                                                                                        'instrument_manual_name':'DKK-TOA_GLN-314E_Specs.pdf'},
        'Ecophysics CLD700AL':   {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'1.0',  'documented_upper_limit_of_detection':{'NO':'100000.0','NO2':'1000.0','NOx':'1000.0'},                                    'documented_accuracy':'1.0%',        'documented_zero_drift':'0.0',                                                                                                'documented_flow_rate':'0.7',     'instrument_manual_name':'Ecophysics_CLDAL700_specs.pdf'},
        'Ecophysics CLD770AL':   {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but information is found about converter here:https://www.atmos-chem-phys-discuss.net/acp-2014-187/acpd-14-10135-2014.pdf'},
        'Ecotech EC9841T':       {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.05', 'documented_upper_limit_of_detection':'2000.0',                                        'documented_precision':'0.1/0.5%',                                      'documented_zero_drift':'0.1/day',   'documented_span_drift':'1.0%/day',                                                      'documented_flow_rate':'0.64',    'instrument_manual_name':'Ecotech_EC9841T_Specs.pdf', 'measuring_instrument_further_details':'Converter information not given in Ecotech_EC9841T_Specs.pdf, but found online: https://books.google.es/books?id=PknpYLJjsgIC&pg=PT101&lpg=PT101&dq=ecotech+EC9841T+molybdenum&source=bl&ots=p34J8Jais1&sig=smuKt3IiSRlcdvCjufYUF1Wh2iQ&hl=en&sa=X&ved=0ahUKEwiSwNqUoefaAhVMORQKHX5kBrcQ6AEIKTAA#v=onepage&q=ecotech%20EC9841T%20molybdenum&f=false'},
        'Ecotech Serinus40':     {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.4',  'documented_upper_limit_of_detection':'20000.0',                                       'documented_precision':'0.4/0.5%', 'documented_accuracy':'1.0%',        'documented_zero_drift':'0.4/day',   'documented_span_drift':'1.0%/7days',                                                    'documented_flow_rate':'0.3',     'instrument_manual_name':'Ecotech_Serinus40_Manual.pdf'},
        'Environnement SA AC30M':{'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://helda.helsinki.fi/bitstream/handle/10138/1157/cont71.pdf?sequence=3&isAllowed=y', 'measuring_assumption':True},
        'Environnement SA AC31M':{'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.35', 'documented_upper_limit_of_detection':'10000.0',                                                                          'documented_accuracy':'1.0%',        'documented_zero_drift':'1.0/week',  'documented_span_drift':'1.0%/day',                                                      'documented_flow_rate':'0.6',     'instrument_manual_name':'EnvironnementSA_AC31M_Specs.pdf'},
        'Environnement SA AC32M':{'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.4',  'documented_upper_limit_of_detection':'50000.0',                                       'documented_precision':'1.0%',     'documented_accuracy':'1.0%',        'documented_zero_drift':'1.0/day',   'documented_span_drift':'1.0%/7days',                                                    'documented_flow_rate':'0.66',    'instrument_manual_name':'EnvironnementSA_AC32M_Manual.pdf'},
        'Environnement SA AC32e':{'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.2',  'documented_upper_limit_of_detection':'10000.0',                                       'documented_precision':'1.0%',     'documented_accuracy':'1.0%',        'documented_zero_drift':'1.0/day',   'documented_span_drift':'1.0/day',                                                       'documented_flow_rate':'0.66',    'instrument_manual_name':'EnvironnementSA_AC32e_Specs.pdf'},
        'Horiba APNA300':        {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://books.google.es/books?id=qmWXBwAAQBAJ&pg=PA88&lpg=PA88&dq=horiba+apna+300&source=bl&ots=dBQJ0TEai1&sig=plLm6SAUIZwuIhy37JzI8nZXccQ&hl=en&sa=X&ved=2ahUKEwihxOS9uvDfAhWGnhQKHYzoAPcQ6AEwCHoECAIQAQ#v=onepage&q=horiba%20apna%20300&f=false', 'measuring_assumption':True},
        'Horiba APNA350':        {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: http://www.wseas.us/e-library/conferences/2009/timisoara/SSE2/SSE2-11.pdf', 'measuring_assumption':True},
        'Horiba APNA360':        {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'10000.0',                                       'documented_precision':'1.0%',     'documented_accuracy':'1.0%',        'documented_zero_drift':'0.5/day',   'documented_span_drift':'0.5/day',                                                       'documented_flow_rate':'0.8',     'instrument_manual_name':'Horiba_APNA360_Manual.pdf'},
        'Horiba APNA365':        {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.1',  'documented_upper_limit_of_detection':'100.0',                                         'documented_precision':'2.0%',     'documented_accuracy':'2.0%',        'documented_zero_drift':'2.0%/day',  'documented_span_drift':'2.0%<20.0/day 1.0%>=20.0/day',                                  'documented_flow_rate':'1.6',     'instrument_further_details':'Details found online:http://www.biowell.sk/partner/horiba/apna_365_NOx.html'},
        'Horiba APNA370':        {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'10000.0',                                       'documented_precision':'1.0%',     'documented_accuracy':'1.0%',        'documented_zero_drift':'0.5/day',   'documented_span_drift':'0.5/day',                                                       'documented_flow_rate':'0.8',     'instrument_manual_name':'Horiba_APNA370_Manual.pdf'},
        'MCV 30QL':              {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://studylib.es/doc/4716520/expedient-de-contractació-núm.-tipus-de-contracte--servei...', 'measuring_assumption':True},
        'Meloy NA530R':          {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://books.google.es/books?id=yCDGFlP3toUC&pg=PA248&lpg=PA248&dq=MELOY+NA530R&source=bl&ots=-uI9Lv61Zl&sig=Av_D54Gtv1qh8qVrAa4qLde7_cw&hl=en&sa=X&ved=0ahUKEwjg1q2-8uvaAhUGhiwKHcCxAFYQ6AEIKzAB#v=onepage&q=MELOY%20NA530R&f=false', 'measuring_assumption':True},
        'Philips K50034':        {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://www.google.co.uk/url?sa=t&rct=j&q=&esrc=s&source=web&cd=11&ved=2ahUKEwipieOHk73nAhUd8uAKHdEdA1oQFjAKegQIAxAB&url=http%3A%2F%2Ffbd.beun.edu.tr%2Findex.php%2Fzkufbd%2Farticle%2Fdownload%2F53%2F100&usg=AOvVaw0l7fQAqr1Pealzx0MaZ6d0 ', 'measuring_assumption':True},
        'Philips K50102':        {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/philipsK50102/view', 'measuring_assumption':True},
        'Philips K50235':        {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://www.arpae.it/cms3/documenti/_cerca_doc/aria/rimini/rn_valloni_tesi_laurea_mar2010.pdf', 'measuring_assumption':True},
        'Philips PW9762/02':     {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://books.google.es/books?id=T5B9CAAAQBAJ&pg=PA415&lpg=PA415&dq=Philips+PW9762/02&source=bl&ots=cPN-nS_ZqB&sig=hPPtzoZLDIjgkvnmuhNb8mB0AC8&hl=en&sa=X&ved=0ahUKEwj0it6zuqzbAhUMCMAKHdI-AikQ6AEIKzAB#v=onepage&q=Philips%20PW9762%2F02&f=false', 'measuring_assumption':True},
        'Seres 2000G':           {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'1.0',  'documented_upper_limit_of_detection':'20000.0',                                                                          'documented_accuracy':'1.0%',        'documented_zero_drift':'1.0%/week', 'documented_span_drift':'1.0%/week',                                                     'documented_flow_rate':'0.5',     'instrument_manual_name':'Seres_2000G_Specs.pdf'},
        'SIR S-5012':            {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'Converter information not given in a manual or specs, but found online: http://www.ecomission.co.za/products/model_s_5012.htm'},
        'Tecan CLD 502':         {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://books.google.es/books?id=kyvwCAAAQBAJ&pg=PA113&lpg=PA113&dq=tecan+CLD+502&source=bl&ots=XVFDJIwcis&sig=ACfU3U0CGQjZuVEqtDTk2YX1jvf5J37tSg&hl=en&sa=X&ved=2ahUKEwjsu6aqqfzzAhULlRQKHUvHDsQQ6AF6BAgSEAM#v=onepage&q=tecan%20CLD%20502&f=false', 'measuring_assumption':True}, 
        'Teledyne 200':          {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/API200/view', 'measuring_assumption':True},
        'Teledyne 200A':         {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.4',  'documented_upper_limit_of_detection':'20000.0',                                       'documented_precision':'0.5%',     'documented_accuracy':'1.0%',        'documented_zero_drift':'0.5/day',   'documented_span_drift':'0.5%/7days',                                                    'documented_flow_rate':'0.5',     'instrument_manual_name':'Teledyne_200A_Manual.pdf'},
        'Teledyne 200AU':        {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.05', 'documented_upper_limit_of_detection':'2000.0',                                        'documented_precision':'0.5%>=50', 'documented_accuracy':'1.0%/0.1',    'documented_zero_drift':'0.1/day',   'documented_span_drift':'0.05/0.5%/7days',                                               'documented_flow_rate':'1.0',     'instrument_manual_name':'Teledyne_200AU_Manual.pdf'},
        'Teledyne 200E':         {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.4',  'documented_upper_limit_of_detection':'20000.0',                                       'documented_precision':'0.5%',     'documented_accuracy':'1.0%',        'documented_zero_drift':'0.5/day',   'documented_span_drift':'0.5%/7days',                                                    'documented_flow_rate':'0.5',     'instrument_manual_name':'Teledyne_200E_Manual.pdf'},
        'Teledyne 200EU':        {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.05', 'documented_upper_limit_of_detection':'2000.0',                                        'documented_precision':'0.5%>=50', 'documented_accuracy':'1.0%/0.1',    'documented_zero_drift':'0.1/day',   'documented_span_drift':'0.05/0.5%/day',                                                 'documented_flow_rate':'1.0',     'instrument_manual_name':'Teledyne_200EU_addendum.pdf'}, 
        'Teledyne 265E':         {'measured_parameters':['O3'],             'qa_accepted_parameters':['O3'], 'documented_lower_limit_of_detection':'0.3',  'documented_upper_limit_of_detection':'2000.0',                                        'documented_precision':'0.5%',     'documented_accuracy':'1.0%',        'documented_zero_drift':'0.5',       'documented_span_drift':'0.5%',                                                          'documented_flow_rate':'0.5',     'instrument_manual_name':'Teledyne_265E_Addendum.pdf'},
        'Teledyne ML2041':       {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://cambridgeshireinsight.org.uk/wp-content/uploads/2017/12/Cambs-City-Council-Air-Quality-Annual-Status-Report-2015.pdf', 'measuring_assumption':True},
        'Teledyne ML8440':       {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://www.umweltbundesamt.de/sites/default/files/medien/publikation/long/2702.pdf'},
        'Teledyne ML8440E':      {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but information is found about converter here: https://www.google.com/url?sa=t&rct=j&q=&esrc=s&source=web&cd=21&ved=0ahUKEwjLg9-In-raAhXFshQKHdzgBfE4FBAWCCgwAA&url=https%3A%2F%2Fnagoya.repo.nii.ac.jp%2F%3Faction%3Drepository_uri%26item_id%3D20273%26file_id%3D17%26file_no%3D1&usg=AOvVaw1jJ6Phm7rzGeqXxCppKjS7'},
        'Teledyne ML8840':       {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'1.0',  'documented_upper_limit_of_detection':'200.0',                                                                                                                                                                                                                                                                                 'instrument_further_details':'No manual or specifications available but information is found about converter here: https://books.google.es/books?id=OFfsCAAAQBAJ&pg=PA40&lpg=PA40&dq=Monitor+Labs+8840&source=bl&ots=tm9D_6gJ-t&sig=bwhco8GlW_p1yft9aESRbijGUN0&hl=en&sa=X&ved=0ahUKEwjU7KravKzbAhUjB8AKHZRJC-YQ6AEIYDAJ#v=onepage&q=Monitor%20Labs%208840&f=false ; details about limit of detection:https://books.google.es/books?id=G0EYUql3_98C&pg=PA283&lpg=PA283&dq=thermo+43a&source=bl&ots=k2RUJXlAVJ&sig=yIxG08AoSyuqedtS6dp-O7O04X4&hl=en&sa=X&ved=0ahUKEwj17qqB3OfbAhVCVRQKHanFDlo4ChDoAQgoMAA#v=onepage&q=thermo%2043a&f=false'},
        'Teledyne ML8841':       {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter:https://books.google.es/books?id=qmWXBwAAQBAJ&pg=PA88&lpg=PA88&dq=Monitor+Labs+8841&source=bl&ots=dBOJ1Rybm3&sig=VOhLyY-EPom-jDAcT7ST9qsQN7M&hl=en&sa=X&ved=0ahUKEwjB077wxKzbAhUjCMAKHXN_BE4Q6AEIVTAJ#v=onepage&q=Monitor%20Labs%208841&f=false', 'measuring_assumption':True},
        'Teledyne ML9840':       {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/ML9840/view',  'measuring_assumption':True},
        'Teledyne ML9841':       {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://www.arb.ca.gov/aaqm/qa/qa-manual/vol4/measurement_of_nox.pdf',              'measuring_assumption':True},
        'Teledyne ML8941A':      {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/ML8941A/view', 'measuring_assumption':True},
        'Teledyne ML9841A':      {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'20000.0', 'documented_uncertainty':'0.25',      'documented_precision':'0.5/1.0%',                                      'documented_zero_drift':'1.0/day',   'documented_span_drift':'1.0%/day',                     'documented_resolution':'0.001', 'documented_flow_rate':'0.64',    'instrument_further_details':'Teledyne_ML9841_Manual.pdf'},
        'Teledyne ML9841B':      {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.5',  'documented_upper_limit_of_detection':'20000.0', 'documented_uncertainty':'0.25',      'documented_precision':'0.5/1.0%',                                      'documented_zero_drift':'1.0/day',   'documented_span_drift':'1.0%/day',                     'documented_resolution':'0.001', 'documented_flow_rate':'0.64',    'instrument_further_details':'Teledyne_ML9841_Manual.pdf'},                                                                                                                                                                                                                                                                                                                                                                            
        'Teledyne ML9841T':      {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/ML9841T/view', 'measuring_assumption':True},
        'Teledyne T200':         {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.4',  'documented_upper_limit_of_detection':'20000.0',                                       'documented_precision':'0.5%>=50', 'documented_accuracy':'1.0%',        'documented_zero_drift':'0.5/day',   'documented_span_drift':'0.5%/day',                                                      'documented_flow_rate':'0.5',     'instrument_manual_name':'Teledyne_T200_Manual.pdf'},
        'Teledyne T200U':        {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.05', 'documented_upper_limit_of_detection':'2000.0',                                        'documented_precision':'0.5%>=5' , 'documented_accuracy':'1.0%/0.1',    'documented_zero_drift':'0.1/day',   'documented_span_drift':'0.05/0.5%/day',                                                 'documented_flow_rate':'1.1',     'instrument_manual_name':'Teledyne_T200U_Addendum.pdf'},
        'Teledyne T204-NOx':     {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.4',  'documented_upper_limit_of_detection':'20000.0',                                       'documented_precision':'0.5%>=50', 'documented_accuracy':'1.0%',        'documented_zero_drift':'0.5/day',   'documented_span_drift':'0.5%/day',                                                      'documented_flow_rate':'0.5',     'instrument_manual_name':'Teledyne_T204_Manual.pdf'},
        'Teledyne T265':         {'measured_parameters':['O3'],             'qa_accepted_parameters':['O3'], 'documented_lower_limit_of_detection':'0.3',  'documented_upper_limit_of_detection':'2000.0',                                        'documented_precision':'0.5%',     'documented_accuracy':'1.0%',        'documented_zero_drift':'0.5/day',   'documented_span_drift':'0.5%/day',                                                      'documented_flow_rate':'0.5',     'instrument_manual_name':'Teledyne_T265_Addendum.pdf'},
        'Thermo 14B':            {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_manual_name':'Thermo_14B_Operation.pdf'},
        'Thermo 14B/E':          {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_manual_name':'Thermo_14B/E_Operation.pdf'},
        'Thermo 14D/E':          {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://books.google.es/books?id=fXysxVKVzyQC&pg=PA137&lpg=PA137&dq=Thermo+14D/E&source=bl&ots=8ks5Vf7mrS&sig=r0-I08UWw6bVqUPPbIavs2GlIHk&hl=en&sa=X&ved=0ahUKEwiiot2juKzbAhXKKMAKHbE3D-AQ6AEINzAB#v=onepage&q=Thermo%2014D%2FE&f=false', 'measuring_assumption':True},
        'Thermo 42':             {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'Cannot find any information. AQS states it as continuous for NO2 which is almost always chemiluminescence with internal molybdenum converter. Assume this.', 'measuring_assumption':True},
        'Thermo 42C':            {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses chemiluminescence here, assume internal molybdenum converter: https://www.eufar.net/instruments/447', 'measuring_assumption':True}, 
        'Thermo 42C-HL':         {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'50.0', 'documented_upper_limit_of_detection':'5000000.0',                                                                        'documented_accuracy':'1.0%',        'documented_zero_drift':'50.0/day',  'documented_span_drift':'1.0%/day',                                                      'documented_flow_rate':'0.025',   'instrument_manual_name':'Thermo_42C-HL_Manual.pdf'},     
        'Thermo 42C-TL':         {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.05', 'documented_upper_limit_of_detection':'200.0',                                                                            'documented_accuracy':'1.0%',        'documented_zero_drift':'0.0',       'documented_span_drift':'1.0%/day',                                                      'documented_flow_rate':'1.0',     'instrument_manual_name':'Thermo_42C-TL_Manual.pdf'},
        'Thermo 42i':            {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.4',  'documented_upper_limit_of_detection':'100000.0',                                                                         'documented_accuracy':'1.0%',        'documented_zero_drift':'0.4/day',   'documented_span_drift':'1.0%/day',                                                      'documented_flow_rate':'0.6-0.8', 'instrument_manual_name':'Thermo_42i_Manual.pdf'},
        'Thermo 42i-LS':         {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'10.0', 'documented_upper_limit_of_detection':'500000.0',                                                                         'documented_accuracy':'1.0%',        'documented_zero_drift':'5.0/day',   'documented_span_drift':'1.0%/day',                                                      'documented_flow_rate':'0.025',   'instrument_manual_name':'Thermo_42i-LS_Manual.pdf'},
        'Thermo 42i-TL':         {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'], 'documented_lower_limit_of_detection':'0.05', 'documented_upper_limit_of_detection':'200.0',                                                                            'documented_accuracy':'1.0%',        'documented_zero_drift':'0.0',       'documented_span_drift':'1.0%/day',                                                      'documented_flow_rate':'1.0',     'instrument_manual_name':'Thermo_42i-TL_Manual.pdf'},
        'Thermo 42S':            {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but information is found about converter here: https://agupubs.onlinelibrary.wiley.com/doi/pdf/10.1029/98JD01202'},
        'Thermo 42W':            {'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO'],                                                                                                                                                                                                                                                                                                                                                                              'instrument_further_details':'No manual or specifications available but information is found about converter here: http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/thermo42w/view'}

    }
},

#------------------------------------
#chemiluminescence (external molybdenum converter)

#Chemiluminescence is the emission of light (luminescence), as the result of a chemical reaction. Chemiluminescence occurs as a result of the reaction of NO with ozone (NO+O3 --> NO2*+O2).
#The return to a fundamental electronic state of the excited NO2* molecules is made by luminous radiation in a 600-3000 nm spectrum (NO2* --> NO2 + hv), which can be measured.
#This method is designed specifically to directly measure NO (and NOy indirectly).

#Typical measurement operation for NO (and NOy indirectly)
#1. Sample air is drawn into the reaction cell via two separate (alternating) paths; the NO and NOy channels.
#2. NO in the first path (NO channel) reacts with ozone (from ozone generator) according to the following reaction: NO+O3 --> NO2*+O2.
#3. The excited NO2 molecules quickly returns to the ground state, releasing the excess energy. This release takes the form of a quantum of light (hv). The distribution of wavelengths for these quanta range between 600 and 3000 nm, with a peak at about 1200 nm (NO2* --> NO2 +hv (1200nm))
#4. Luminescence from the reaction is detected using a photomultiplier tube (PMT). Photons enter the PMT and strike a negatively charged photo cathode causing it to emit electrons. These electrons are accelerated by an applied high voltage and multiplied through a sequence of similar acceleration steps (dynodes) until a useable current signal is generated. The more light present (in this case photons given off by the chemiluminescent reaction described above), the more current is produced. Therefore the more NO present in the reaction cell the more current is produced by the PMT.
#5. Before entering the PMT, light is passed through a high-pass optical filter only transparent to wavelengths of light above 645nm. The narrowness of this band of sensitivity allows avoids the potential of extraneous light and radiation that might interfere with the measurement (e.g. some oxides of sulfur can also be chemiluminescent emitters when in contact with O3 but give off light at much shorter wavelengths (usually around 260nm to 480nm)).
#6. All things being constant (temperature, pressure, amount of ozone present, etc.), the relationship between the amount of NO present in the reaction cell and the amount of light emitted from the reaction is very linear. If more NO is present, more IR light is produced. 
#7. In addition, sometimes the excited NO2 collides with other gaseous molecules in the reaction cell chamber or even the molecules of the reaction cell walls and transfers its excess energy to this collision partner (represented by M) without emitting any light at all. In fact, by far the largest portion of the excited NO2 returns to the ground state this way, leaving only a few percent yield of usable chemiluminescence (NO2* + M --> NO2 + M). The probability of a collision between the NO2* molecule and a collision partner M increases proportionally with the reaction cell pressure. This non-radiating collision with the NO2* molecules is usually referred to as third body quenching. Even under the best conditions only about 20% of the NO2 that is formed by the key chemiluminescence reaction is in the excited state. In order to maximize chemiluminescence, the reaction cell is maintained at reduced pressure (thereby reducing the amount of available collision partners) and is supplied with a large, constant excess of ozone (about 3000-5000 ppm) from the internal ozone generator.
#8. The second path (NOy channel) before entering the instrument enters an externally mounted heated molybdenum converter (at 315degC), with less than 10cm tubing between the converter and sample entry. The heated molybdenum reacts with NOy in the sample gas and produces a NO gas and a variety of molybdenum (3NOy + Mo --> 3NO + MoO3 (at 315degC)). Minimising the transit time and surface contact area between the sample inlet and converter enables the conversion of labile components of NOy.
#9. Once the NOy in the sample gas has been converted to NO, it is routed to the reaction cell where it undergoes the chemiluminescence reaction and is measured by the PMT.
#10. By converting the NOy in the sample gas into NO, the analyser indirectly measures total NOy content of the sample gas (i.e. the NO present + the converted NO2 present). 
#11. By switching the sample gas stream in and out of the molybdenum converter every 6 - 10 seconds, the analyser is able to quasi-continuously measure both the NO and the total NOy content.

#Known interferences: Water vapour (above 20 ppmv), 3rd body quenching (CO2, SOx), other species undergoing chemiluminescence with O3 (SOx).

'chemiluminescence (external molybdenum converter)':{'abbreviation':'CL(EMC)', 'sampling_type':'low volume continuous', 'measured_parameters':['HNO3','NH3','NO','NOy'], 'qa_accepted_parameters':['HNO3','NH3','NO','NOy'], 
    'instruments':{
        'Ecotech EC9843':    {'documented_lower_limit_of_detection':'0.05', 'documented_upper_limit_of_detection':'2000.0', 'documented_precision':'0.05/0.5%', 'documented_accuracy':'1.0%',     'documented_zero_drift':'0.1/day', 'documented_span_drift':'1.0%/day',      'documented_flow_rate':'0.64', 'instrument_manual_name':'Ecotech_EC9843_Specs.pdf, Ecotech_EC9843_Manual.pdf'},
        'Teledyne 200EU-NOy':{'documented_lower_limit_of_detection':'0.05', 'documented_upper_limit_of_detection':'2000.0', 'documented_precision':'0.5%>=50',  'documented_accuracy':'1.0%/0.1', 'documented_zero_drift':'0.1/day', 'documented_span_drift':'0.05/0.5%/day', 'documented_flow_rate':'1.0',  'instrument_manual_name':'Teledyne_200EU-NOy_addendum.pdf'},
        'Thermo 42C-NOy':    {'documented_lower_limit_of_detection':'0.05', 'documented_upper_limit_of_detection':'200.0',                                      'documented_accuracy':'1.0%',     'documented_zero_drift':'0.0',     'documented_span_drift':'1.0%/day',      'documented_flow_rate':'1.0',  'instrument_manual_name':'Thermo_42C-NOy_Manual.pdf'},
        'Thermo 42i-NOy':    {'documented_lower_limit_of_detection':'0.05', 'documented_upper_limit_of_detection':'1000.0',                                     'documented_accuracy':'1.0%',     'documented_zero_drift':'0.0',     'documented_span_drift':'1.0%/day',      'documented_flow_rate':'1.0',  'instrument_manual_name':'Thermo_42i-NOy_Manual.pdf'}
    }
},

#------------------------------------
#chemiluminescence (internal photolytic converter) (CL(IPC))

#Chemiluminescence is the emission of light (luminescence), as the result of a chemical reaction. Chemiluminescence occurs as a result of the reaction of NO with ozone (NO+O3 --> NO2*+O2).
#The return to a fundamental electronic state of the excited NO2* molecules is made by luminous radiation in a 600-3000 nm spectrum (NO2* --> NO2 + hv), which can be measured.
#This method is designed specifically to directly measure NO (and NOx/NO2 indirectly). 

#Typical measurement operation for NO (and NOx/NO2 indirectly):
#1. Sample air is drawn into the reaction cell via two separate (alternating) paths; the NO and NOx channels.
#2. NO in the first path (NO channel) reacts with ozone (from ozone generator) according to the following reaction: NO+O3 --> NO2*+O2.
#3. The excited NO2 molecules quickly returns to the ground state, releasing the excess energy. This release takes the form of a quantum of light (hv). The distribution of wavelengths for these quanta range between 600 and 3000 nm, with a peak at about 1200 nm (NO2* --> NO2 +hv (1200nm))
#4. Luminescence from the reaction is detected using a photomultiplier tube (PMT). Photons enter the PMT and strike a negatively charged photo cathode causing it to emit electrons. These electrons are accelerated by an applied high voltage and multiplied through a sequence of similar acceleration steps (dynodes) until a useable current signal is generated. The more light present (in this case photons given off by the chemiluminescent reaction described above), the more current is produced. Therefore the more NO present in the reaction cell the more current is produced by the PMT.
#5. Before entering the PMT, light is passed through a high-pass optical filter only transparent to wavelengths of light above 645nm. The narrowness of this band of sensitivity allows avoids the potential of extraneous light and radiation that might interfere with the measurement (e.g. some oxides of sulfur can also be chemiluminescent emitters when in contact with O3 but give off light at much shorter wavelengths (usually around 260nm to 480nm)).
#6. All things being constant (temperature, pressure, amount of ozone present, etc.), the relationship between the amount of NO present in the reaction cell and the amount of light emitted from the reaction is very linear. If more NO is present, more IR light is produced. 
#7. In addition, sometimes the excited NO2 collides with other gaseous molecules in the reaction cell chamber or even the molecules of the reaction cell walls and transfers its excess energy to this collision partner (represented by M) without emitting any light at all. In fact, by far the largest portion of the excited NO2 returns to the ground state this way, leaving only a few percent yield of usable chemiluminescence (NO2* + M --> NO2 + M). The probability of a collision between the NO2* molecule and a collision partner M increases proportionally with the reaction cell pressure. This non-radiating collision with the NO2* molecules is usually referred to as third body quenching. Even under the best conditions only about 20% of the NO2 that is formed by the key chemiluminescence reaction is in the excited state. In order to maximize chemiluminescence, the reaction cell is maintained at reduced pressure (thereby reducing the amount of available collision partners) and is supplied with a large, constant excess of ozone (about 3000-5000 ppm) from the internal ozone generator.
#8. The second path (NOx channel) travels through a delay loop and a photolytic converter. As the sample gas passes through the converter chamber it is exposed to blue light at specific wavelengths (350-420 nm) from a photolytic light source (blue LEDs, metal halide). This selectively converts NO2 to NO with negligible radiant heating or interference from other gases (NO2 +hv (350-420nm) --> NO + O).
#9. Once the NO2 in the sample gas has been converted to NO, it is routed to the reaction cell where it undergoes the chemiluminescence reaction and is measured by the PMT.
#10. By converting the NO2 in the sample gas into NO, the analyser indirectly measures total NOx content of the sample gas (i.e. the NO present + the converted NO2 present). 
#11. By switching the sample gas stream in and out of the photolytic converter every 6 - 10 seconds, the analyser is able to quasi-continuously measure both the NO and the total NOx content.
#12. The NO2 concentration is not directly measured but calculated by subtracting the known NO content of the sample gas from the known NOx content.

#Known interferences: Water vapour (above 20 ppmv), 3rd body quenching (CO2, SOx), thermal decomposition of PAN to NO2 within the photolysis cell

'chemiluminescence (internal photolytic converter)':{'abbreviation':'CL(IPC)', 'sampling_type':'low volume continuous', 'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO','NO2','NOx'], 
    'instruments':{
        'Ecophysics CraNOx':{'documented_lower_limit_of_detection':'0.025',                               'documented_upper_limit_of_detection':'1000.0',                                    'documented_accuracy':'1.0%',                                                                            'documented_flow_rate':'2.7', 'instrument_manual_name':'Ecophysics_CraNOx_Specs.pdf'},
        'Teledyne T200P':   {'documented_lower_limit_of_detection':'0.4',                                 'documented_upper_limit_of_detection':'4000.0', 'documented_precision':'0.5%>=50', 'documented_accuracy':'1.0%',     'documented_zero_drift':'0.5/day', 'documented_span_drift':'0.5%/day', 'documented_flow_rate':'0.5', 'instrument_manual_name':'Teledyne_T200P_Manual.pdf'},
        'Teledyne T200UP':  {'documented_lower_limit_of_detection':{'NO':'0.05','NO2':'0.1','NOx':'0.1'}, 'documented_upper_limit_of_detection':'2000.0', 'documented_precision':'0.5%>=5',  'documented_accuracy':'1.0%/0.1', 'documented_zero_drift':'0.1/day', 'documented_span_drift':'0.5%/day', 'documented_flow_rate':'1.1', 'instrument_manual_name':'Teledyne_T200UP_addendum.pdf'}
    }
},

#------------------------------------
#chemiluminescence (internal molybdenum and quartz converters) (CL(IMQC))

#Chemiluminescence is the emission of light (luminescence), as the result of a chemical reaction. Chemiluminescence occurs as a result of the reaction of NO with ozone (NO+O3 --> NO2*+O2).
#The return to a fundamental electronic state of the excited NO2* molecules is made by luminous radiation in a 600-3000 nm spectrum (NO2* --> NO2 + hv), which can be measured.
#This method is designed specifically to directly measure NO (and NH3/NOx/NO2 indirectly). 

#Typical measurement operation for NO (and NH3/NOx/NO2 indirectly)
#1. Sample air is drawn into the reaction cell via three separate (alternating) paths; the NO, TNx and NOx channels.
#2. NO in the first path (NO channel) reacts with ozone (from ozone generator) according to the following reaction: NO+O3 --> NO2*+O2.
#3. The excited NO2 molecules quickly returns to the ground state, releasing the excess energy. This release takes the form of a quantum of light (hv). The distribution of wavelengths for these quanta range between 600 and 3000 nm, with a peak at about 1200 nm (NO2* --> NO2 +hv (1200nm))
#4. Luminescence from the reaction is detected using a photomultiplier tube (PMT). Photons enter the PMT and strike a negatively charged photo cathode causing it to emit electrons. These electrons are accelerated by an applied high voltage and multiplied through a sequence of similar acceleration steps (dynodes) until a useable current signal is generated. The more light present (in this case photons given off by the chemiluminescent reaction described above), the more current is produced. Therefore the more NO present in the reaction cell the more current is produced by the PMT.
#5. Before entering the PMT, light is passed through a high-pass optical filter only transparent to wavelengths of light above 645nm. The narrowness of this band of sensitivity allows avoids the potential of extraneous light and radiation that might interfere with the measurement (e.g. some oxides of sulfur can also be chemiluminescent emitters when in contact with O3 but give off light at much shorter wavelengths (usually around 260nm to 480nm)).
#6. All things being constant (temperature, pressure, amount of ozone present, etc.), the relationship between the amount of NO present in the reaction cell and the amount of light emitted from the reaction is very linear. If more NO is present, more IR light is produced. 
#7. In addition, sometimes the excited NO2 collides with other gaseous molecules in the reaction cell chamber or even the molecules of the reaction cell walls and transfers its excess energy to this collision partner (represented by M) without emitting any light at all. In fact, by far the largest portion of the excited NO2 returns to the ground state this way, leaving only a few percent yield of usable chemiluminescence (NO2* + M --> NO2 + M). The probability of a collision between the NO2* molecule and a collision partner M increases proportionally with the reaction cell pressure. This non-radiating collision with the NO2* molecules is usually referred to as third body quenching. Even under the best conditions only about 20% of the NO2 that is formed by the key chemiluminescence reaction is in the excited state. In order to maximize chemiluminescence, the reaction cell is maintained at reduced pressure (thereby reducing the amount of available collision partners) and is supplied with a large, constant excess of ozone (about 3000-5000 ppm) from the internal ozone generator.
#8. In the second path (TNx channel), sample air is passed through an internal heated quartz converter (at 980 degC). The heated quartz reacts with NH3 and NO2 in the sample gas and produces a NO gas (4NH3 + 5O2 --> 4NO + 6H2O).
#9. Once the NO2 and NH3 in the sample gas has been converted to NO, it is routed to the reaction cell where it undergoes the chemiluminescence reaction and is measured by the PMT.
#10. By converting the NO2 and NH3 in the sample gas into NO, the analyser indirectly measures total TNx content of the sample gas. 
#11. The third path (NOx channel) travels through a delay loop and a heated molybdenum converter (at 315degC). The heated molybdenum reacts with NO2 in the sample gas and produces a NO gas and a variety of molybdenum. (xNO2 + yMo --> xNO + MyOz (at 315degC)).
#12. Once the NO2 in the sample gas has been converted to NO, it is routed to the reaction cell where it undergoes the chemiluminescence reaction and is measured by the PMT.
#13. By converting the NO2 in the sample gas into NO, the analyser measures total NOx content of the sample gas (i.e. the NO present + the converted NO2 present). 
#14. By switching the sample gas stream between the 3 streams frequently, the analyser is able to quasi-continuously measure both the NO, the total NOx content and TNx.
#15. NO2 is indirectly calculated by subtracting the known NO content of the sample gas from the known NOx content.
#16. NH3 is indirectly calculated by subtracting the known NOx content of the sample gas from the known TNx content.

#Known interferences: Water vapour (above 20 ppmv), 3rd body quenching (CO2, SOx), other NOy species converted to NO by molybdenum converter (PAN, ethyl nitrate, ethyl nitrite, HONO, HNO3, methyl nitrate, n-propyl nitrate, n-butyl nitrate, nitrocresol, NH3), other species undergoing chemiluminescence with O3 (SOx).

#Not QA accepted measurement technique for NO2/NOx due to bias in measuring NO2, due to molybdenum converters.

'chemiluminescence (internal molybdenum and quartz converters)':{'abbreviation':'CL(IMQC)', 'sampling_type':'low volume continuous', 'measured_parameters':['HNO3','NH3','NO','NO2','NOx','NOy'], 'qa_accepted_parameters':['HNO3','NH3','NO','NOy'], 
    'instruments':{
        'Ecotech EC9842': {'documented_lower_limit_of_detection':'0.5', 'documented_upper_limit_of_detection':'2000.0', 'documented_precision':'0.5%', 'documented_accuracy':'1.0%', 'documented_zero_drift':'1.0/day', 'documented_span_drift':'0.5%/day', 'documented_resolution':'0.1',  'documented_flow_rate':'0.000355', 'instrument_manual_name':'Ecotech_EC9842_Specs.pdf'},
        'Ecotech EC9842T':{'documented_lower_limit_of_detection':'0.4',                                                 'documented_precision':'2.0%',                               'documented_zero_drift':'0.1/day', 'documented_span_drift':'1.0%/day',                                                                    'instrument_further_details':'Details found online:http://aep.alberta.ca/air/legislation-and-policy/air-monitoring-directive/documents/AmbientAirMonitoringSpecifications-Sep2014.pdf'}
    }
},

#------------------------------------
#chemiluminescence (internal molybdenum converter and external quartz converter) (CL(IMC-EQC))

#Chemiluminescence is the emission of light (luminescence), as the result of a chemical reaction. Chemiluminescence occurs as a result of the reaction of NO with ozone (NO+O3 --> NO2*+O2).
#The return to a fundamental electronic state of the excited NO2* molecules is made by luminous radiation in a 600-3000 nm spectrum (NO2* --> NO2 + hv), which can be measured.
#This method is designed specifically to directly measure NO (and NH3/NOx/NO2 indirectly). 

#Typical measurement operation for NO (and NH3/NOx/NO2 indirectly)
#1. Sample air is drawn into the reaction cell via three separate (alternating) paths; the NO, TNx and NOx channels.
#2. NO in the first path (NO channel) reacts with ozone (from ozone generator) according to the following reaction: NO+O3 --> NO2*+O2.
#3. The excited NO2 molecules quickly returns to the ground state, releasing the excess energy. This release takes the form of a quantum of light (hv). The distribution of wavelengths for these quanta range between 600 and 3000 nm, with a peak at about 1200 nm (NO2* --> NO2 +hv (1200nm))
#4. Luminescence from the reaction is detected using a photomultiplier tube (PMT). Photons enter the PMT and strike a negatively charged photo cathode causing it to emit electrons. These electrons are accelerated by an applied high voltage and multiplied through a sequence of similar acceleration steps (dynodes) until a useable current signal is generated. The more light present (in this case photons given off by the chemiluminescent reaction described above), the more current is produced. Therefore the more NO present in the reaction cell the more current is produced by the PMT.
#5. Before entering the PMT, light is passed through a high-pass optical filter only transparent to wavelengths of light above 645nm. The narrowness of this band of sensitivity allows avoids the potential of extraneous light and radiation that might interfere with the measurement (e.g. some oxides of sulfur can also be chemiluminescent emitters when in contact with O3 but give off light at much shorter wavelengths (usually around 260nm to 480nm)).
#6. All things being constant (temperature, pressure, amount of ozone present, etc.), the relationship between the amount of NO present in the reaction cell and the amount of light emitted from the reaction is very linear. If more NO is present, more IR light is produced. 
#7. In addition, sometimes the excited NO2 collides with other gaseous molecules in the reaction cell chamber or even the molecules of the reaction cell walls and transfers its excess energy to this collision partner (represented by M) without emitting any light at all. In fact, by far the largest portion of the excited NO2 returns to the ground state this way, leaving only a few percent yield of usable chemiluminescence (NO2* + M --> NO2 + M). The probability of a collision between the NO2* molecule and a collision partner M increases proportionally with the reaction cell pressure. This non-radiating collision with the NO2* molecules is usually referred to as third body quenching. Even under the best conditions only about 20% of the NO2 that is formed by the key chemiluminescence reaction is in the excited state. In order to maximize chemiluminescence, the reaction cell is maintained at reduced pressure (thereby reducing the amount of available collision partners) and is supplied with a large, constant excess of ozone (about 3000-5000 ppm) from the internal ozone generator.
#8. In the second path (TNx channel), sample air is passed through an externally mounted heated quartz converter (at 980 degC). The heated quartz reacts with NH3 and NO2 in the sample gas and produces a NO gas (4NH3 + 5O2 --> 4NO + 6H2O).
#9. Once the NO2 and NH3 in the sample gas has been converted to NO, it is routed to the reaction cell where it undergoes the chemiluminescence reaction and is measured by the PMT.
#10. By converting the NO2 and NH3 in the sample gas into NO, the analyser indirectly measures total TNx content of the sample gas. 
#11. The third path (NOx channel) travels through a delay loop and a heated molybdenum converter (at 315degC). The heated molybdenum reacts with NO2 in the sample gas and produces a NO gas and a variety of molybdenum. (xNO2 + yMo --> xNO + MyOz (at 315degC)).
#12. Once the NO2 in the sample gas has been converted to NO, it is routed to the reaction cell where it undergoes the chemiluminescence reaction and is measured by the PMT.
#13. By converting the NO2 in the sample gas into NO, the analyser measures total NOx content of the sample gas (i.e. the NO present + the converted NO2 present). 
#14. By switching the sample gas stream between the 3 streams frequently, the analyser is able to quasi-continuously measure both the NO, the total NOx content and TNx.
#15. NO2 is indirectly calculated by subtracting the known NO content of the sample gas from the known NOx content.
#16. NH3 is indirectly calculated by subtracting the known NOx content of the sample gas from the known TNx content.

#Known interferences: Water vapour (above 20 ppmv), 3rd body quenching (CO2, SOx), other NOy species converted to NO by molybdenum converter (PAN, ethyl nitrate, ethyl nitrite, HONO, HNO3, methyl nitrate, n-propyl nitrate, n-butyl nitrate, nitrocresol, NH3), other species undergoing chemiluminescence with O3 (SOx).

#Not QA accepted measurement technique for NO2/NOx due to bias in measuring NO2, due to molybdenum converters.

'chemiluminescence (internal molybdenum converter and external quartz converter)':{'abbreviation':'CL(IMC-EQC)', 'sampling_type':'low volume continuous', 'measured_parameters':['HNO3','NH3','NO','NO2','NOx','NOy'], 'qa_accepted_parameters':['HNO3','NH3','NO','NOy'], 
    'instruments':{
        'Environnement SA AC32e-CNH3':{'measured_parameters':['NH3','NO','NO2','NOx'], 'qa_accepted_parameters':['NH3','NO'], 'documented_lower_limit_of_detection':{'NO':'0.2','NO2':'0.2','NOx':'0.2','NH3':'1.0'}, 'documented_upper_limit_of_detection':{'NO':'10000.0','NO2':'10000.0','NOx':'10000.0','NH3':'1000.0'}, 'documented_precision':'1.0%', 'documented_accuracy':'1.0%', 'documented_zero_drift':'1.0/day', 'documented_span_drift':'1.0/7days', 'documented_flow_rate':'0.66', 'instrument_manual_name':'EnvironnementSA_AC32e-CNH3_Specs.pdf'}
    }
},  
  
#------------------------------------
#chemiluminescence (internal molybdenum and stainless steel converters) (CL(IMSC))

#Chemiluminescence is the emission of light (luminescence), as the result of a chemical reaction. Chemiluminescence occurs as a result of the reaction of NO with ozone (NO+O3 --> NO2*+O2).
#The return to a fundamental electronic state of the excited NO2* molecules is made by luminous radiation in a 600-3000 nm spectrum (NO2* --> NO2 + hv), which can be measured.
#This method is designed specifically to directly measure NO (and NH3/NOx/NO2 indirectly). 

#Typical measurement operation for NO (and NH3/NOx/NO2 indirectly)
#1. Sample air is drawn into the reaction cell via three separate (alternating) paths; the NO, TNx and NOx channels.
#2. NO in the first path (NO channel) reacts with ozone (from ozone generator) according to the following reaction: NO+O3 --> NO2*+O2.
#3. The excited NO2 molecules quickly returns to the ground state, releasing the excess energy. This release takes the form of a quantum of light (hv). The distribution of wavelengths for these quanta range between 600 and 3000 nm, with a peak at about 1200 nm (NO2* --> NO2 +hv (1200nm))
#4. Luminescence from the reaction is detected using a photomultiplier tube (PMT). Photons enter the PMT and strike a negatively charged photo cathode causing it to emit electrons. These electrons are accelerated by an applied high voltage and multiplied through a sequence of similar acceleration steps (dynodes) until a useable current signal is generated. The more light present (in this case photons given off by the chemiluminescent reaction described above), the more current is produced. Therefore the more NO present in the reaction cell the more current is produced by the PMT.
#5. Before entering the PMT, light is passed through a high-pass optical filter only transparent to wavelengths of light above 645nm. The narrowness of this band of sensitivity allows avoids the potential of extraneous light and radiation that might interfere with the measurement (e.g. some oxides of sulfur can also be chemiluminescent emitters when in contact with O3 but give off light at much shorter wavelengths (usually around 260nm to 480nm)).
#6. All things being constant (temperature, pressure, amount of ozone present, etc.), the relationship between the amount of NO present in the reaction cell and the amount of light emitted from the reaction is very linear. If more NO is present, more IR light is produced. 
#7. In addition, sometimes the excited NO2 collides with other gaseous molecules in the reaction cell chamber or even the molecules of the reaction cell walls and transfers its excess energy to this collision partner (represented by M) without emitting any light at all. In fact, by far the largest portion of the excited NO2 returns to the ground state this way, leaving only a few percent yield of usable chemiluminescence (NO2* + M --> NO2 + M). The probability of a collision between the NO2* molecule and a collision partner M increases proportionally with the reaction cell pressure. This non-radiating collision with the NO2* molecules is usually referred to as third body quenching. Even under the best conditions only about 20% of the NO2 that is formed by the key chemiluminescence reaction is in the excited state. In order to maximize chemiluminescence, the reaction cell is maintained at reduced pressure (thereby reducing the amount of available collision partners) and is supplied with a large, constant excess of ozone (about 3000-5000 ppm) from the internal ozone generator.
#8. In the second path (TNx channel), sample air is first drawn through an internal heated stainless steel converter (at 825 degC). The heated stainless steel reacts with NH3 and NO2 in the sample gas and produces a NO gas (4NH3 + 5O2 --> 4NO + 6H2O).
#9. Once the NO2 and NH3 in the sample gas has been converted to NO, it is routed to the reaction cell where it undergoes the chemiluminescence reaction and is measured by the PMT.
#10. By converting the NO2 and NH3 in the sample gas into NO, the analyser indirectly measures total TNx content of the sample gas. 
#11. The third path (NOx channel) travels through a delay loop and a heated molybdenum converter (at 315degC). The heated molybdenum reacts with NO2 in the sample gas and produces a NO gas and a variety of molybdenum. (xNO2 + yMo --> xNO + MyOz (at 315degC)).
#12. Once the NO2 in the sample gas has been converted to NO, it is routed to the reaction cell where it undergoes the chemiluminescence reaction and is measured by the PMT.
#13. By converting the NO2 in the sample gas into NO, the analyser measures total NOx content of the sample gas (i.e. the NO present + the converted NO2 present). 
#14. By switching the sample gas stream between the 3 streams frequently, the analyser is able to quasi-continuously measure both the NO, the total NOx content and TNx.
#15. NO2 is indirectly calculated by subtracting the known NO content of the sample gas from the known NOx content.
#16. NH3 is indirectly calculated by subtracting the known NOx content of the sample gas from the known TNx content.

#Known interferences: Water vapour (above 20 ppmv), 3rd body quenching (CO2, SOx), other NOy species converted to NO by molybdenum converter (PAN, ethyl nitrate, ethyl nitrite, HONO, HNO3, methyl nitrate, n-propyl nitrate, n-butyl nitrate, nitrocresol, NH3), other species undergoing chemiluminescence with O3 (SOx).

#Not QA accepted measurement technique for NO2/NOx due to bias in measuring NO2, due to molybdenum converters.

'chemiluminescence (internal molybdenum and stainless steel converters)':{'abbreviation':'CL(IMSC)', 'sampling_type':'low volume continuous', 'measured_parameters':['HNO3','NH3','NO','NO2','NOx','NOy'], 'qa_accepted_parameters':['HNO3','NH3','NO','NOy'], 
    'instruments':{
        'Teledyne 201A':{'documented_lower_limit_of_detection':'1.0', 'documented_upper_limit_of_detection':'2000.0', 'documented_precision':'1.0%', 'documented_accuracy':'2.0%', 'documented_zero_drift':'2.0/day', 'documented_span_drift':'1.0%/week', 'documented_flow_rate':'1.0', 'instrument_manual_name':'Teledyne_201A_Manual.pdf'}
    }
},

#------------------------------------
#chemiluminescence (internal molybdenum converter and external stainless steel converter) (CL(IMC-ESC))

#Chemiluminescence is the emission of light (luminescence), as the result of a chemical reaction. Chemiluminescence occurs as a result of the reaction of NO with ozone (NO+O3 --> NO2*+O2).
#The return to a fundamental electronic state of the excited NO2* molecules is made by luminous radiation in a 600-3000 nm spectrum (NO2* --> NO2 + hv), which can be measured.
#This method is designed specifically to directly measure NO (and NH3/NOx/NO2 indirectly). 

#Typical measurement operation for NO (and NH3/NOx/NO2 indirectly)
#1. Sample air is drawn into the reaction cell via three separate (alternating) paths; the NO, TNx and NOx channels.
#2. NO in the first path (NO channel) reacts with ozone (from ozone generator) according to the following reaction: NO+O3 --> NO2*+O2.
#3. The excited NO2 molecules quickly returns to the ground state, releasing the excess energy. This release takes the form of a quantum of light (hv). The distribution of wavelengths for these quanta range between 600 and 3000 nm, with a peak at about 1200 nm (NO2* --> NO2 +hv (1200nm))
#4. Luminescence from the reaction is detected using a photomultiplier tube (PMT). Photons enter the PMT and strike a negatively charged photo cathode causing it to emit electrons. These electrons are accelerated by an applied high voltage and multiplied through a sequence of similar acceleration steps (dynodes) until a useable current signal is generated. The more light present (in this case photons given off by the chemiluminescent reaction described above), the more current is produced. Therefore the more NO present in the reaction cell the more current is produced by the PMT.
#5. Before entering the PMT, light is passed through a high-pass optical filter only transparent to wavelengths of light above 645nm. The narrowness of this band of sensitivity allows avoids the potential of extraneous light and radiation that might interfere with the measurement (e.g. some oxides of sulfur can also be chemiluminescent emitters when in contact with O3 but give off light at much shorter wavelengths (usually around 260nm to 480nm)).
#6. All things being constant (temperature, pressure, amount of ozone present, etc.), the relationship between the amount of NO present in the reaction cell and the amount of light emitted from the reaction is very linear. If more NO is present, more IR light is produced. 
#7. In addition, sometimes the excited NO2 collides with other gaseous molecules in the reaction cell chamber or even the molecules of the reaction cell walls and transfers its excess energy to this collision partner (represented by M) without emitting any light at all. In fact, by far the largest portion of the excited NO2 returns to the ground state this way, leaving only a few percent yield of usable chemiluminescence (NO2* + M --> NO2 + M). The probability of a collision between the NO2* molecule and a collision partner M increases proportionally with the reaction cell pressure. This non-radiating collision with the NO2* molecules is usually referred to as third body quenching. Even under the best conditions only about 20% of the NO2 that is formed by the key chemiluminescence reaction is in the excited state. In order to maximize chemiluminescence, the reaction cell is maintained at reduced pressure (thereby reducing the amount of available collision partners) and is supplied with a large, constant excess of ozone (about 3000-5000 ppm) from the internal ozone generator.
#8. In the second path (TNx channel), sample air is first drawn through an externally mounted heated stainless steel converter (at 750 degC). The heated stainless steel reacts with NH3 and NO2 in the sample gas and produces a NO gas (4NH3 + 5O2 --> 4NO + 6H2O).
#9. Once the NO2 and NH3 in the sample gas has been converted to NO, it is routed to the reaction cell where it undergoes the chemiluminescence reaction and is measured by the PMT.
#10. By converting the NO2 and NH3 in the sample gas into NO, the analyser indirectly measures total TNx content of the sample gas. 
#11. The third path (NOx channel) travels through a delay loop and a heated molybdenum converter (at 315degC). The heated molybdenum reacts with NO2 in the sample gas and produces a NO gas and a variety of molybdenum. (xNO2 + yMo --> xNO + MyOz (at 315degC)).
#12. Once the NO2 in the sample gas has been converted to NO, it is routed to the reaction cell where it undergoes the chemiluminescence reaction and is measured by the PMT.
#13. By converting the NO2 in the sample gas into NO, the analyser measures total NOx content of the sample gas (i.e. the NO present + the converted NO2 present). 
#14. By switching the sample gas stream between the 3 streams frequently, the analyser is able to quasi-continuously measure both the NO, the total NOx content and TNx.
#15. NO2 is indirectly calculated by subtracting the known NO content of the sample gas from the known NOx content.
#16. NH3 is indirectly calculated by subtracting the known NOx content of the sample gas from the known TNx content.

#Known interferences: Water vapour (above 20 ppmv), 3rd body quenching (CO2, SOx), other NOy species converted to NO by molybdenum converter (PAN, ethyl nitrate, ethyl nitrite, HONO, HNO3, methyl nitrate, n-propyl nitrate, n-butyl nitrate, nitrocresol, NH3), other species undergoing chemiluminescence with O3 (SOx).

#Not QA accepted measurement technique for NO2/NOx due to bias in measuring NO2, due to molybdenum converters.

'chemiluminescence (internal molybdenum converter and external stainless steel converter)':{'abbreviation':'CL(IMC-ESC)', 'sampling_type':'low volume continuous', 'measured_parameters':['HNO3','NH3','NO','NO2','NOx','NOy'], 'qa_accepted_parameters':['HNO3','NH3','NO','NOy'], 
    'instruments':{
        'Teledyne 201E':  {'measured_parameters':['NH3','NO','NO2','NOx'], 'qa_accepted_parameters':['NH3','NO'], 'documented_lower_limit_of_detection':'1.0', 'documented_upper_limit_of_detection':'2000.0',   'documented_accuracy':{'NH3':'2.0%','NO':'1.0%','NO2':'1.0%','NOx':'1.0%'}, 'documented_zero_drift':'2.0/day', 'documented_span_drift':'1.0%/day', 'documented_flow_rate':'0.5-1.0', 'instrument_manual_name':'Teledyne_201E_Manual.pdf'},
        'Thermo 17C':     {                                                                                                                                                                                                                                                                                                                                                                           'instrument_further_details':'Converter information found here:https://archive.epa.gov/nrmrl/archive-etv/web/pdf/01_vs_thermo.pdf'},
        'Thermo 17i':     {                                                                                       'documented_lower_limit_of_detection':'1.0', 'documented_upper_limit_of_detection':'100000.0', 'documented_accuracy':'1.0%',                                               'documented_zero_drift':'1.0/day', 'documented_span_drift':'1.0%/day', 'documented_flow_rate':'0.6',     'instrument_manual_name':'Thermo_17i_Manual.pdf'}
    }
},

#------------------------------------
#gas chromatography - sulphur chemiluminescence (GC-SC)

#Chemiluminescence is the emission of light (luminescence), as the result of a chemical reaction. Chemiluminescence occurs as a result of the reaction of SO and ozone (SO+O3 --> SO2*+O2).
#The return to a fundamental electronic state of the excited SO2* molecules is made by luminous radiation in a specific spectrum (SO2* --> SO2 + hv), which can be measured (by a photomultiplier tube).
#The concentration of sample total sulphur is directly proportional to the intensity of light emitted.
#This method can be mixed with gas chromatography (GC) to allow to determination of specific sulphur compounds (i.e. SO2).
#The gas sample is passed through a GC column before being ultimately measured by the photomultiplier detector.

#Gas chromatography (GC) is a method used for separating and analysing compounds that can be vaporized without decomposition.
#A sample solution is injected into a instrument, entering a gas stream which transports the sample (mobile phase) into a separation tube known as the "column".
#The mobile phase is a carrier gas, usually an inert gas such as helium or an unreactive gas such as nitrogen. Helium remains the most commonly used carrier gas in about 90% of instruments although hydrogen is preferred for improved separations.
#The column consists of a microscopic layer of liquid or polymer on an inert solid support a microscopic layer of liquid or polymer on an inert solid support (stationary phase), inside a piece of glass or metal tubing, placed inside a piece of glass or metal tubing.
#Once inside the column, the gaseous compounds being analysed interact with the walls of the column coated with a stationary phase. This causes each compound to elute at a different time, known as the retention time of the compound. 
#The comparison of retention times is what gives GC its analytical usefulness.
#If greater separation of compounds is required, multiple distinct columns can be used for this purpose.

#See details of method for SO2: http://www.nrcresearchpress.com/doi/pdf/10.1139/v84-073, http://www.interline.nl/media/1000035/specifiekzwavelchromatograph.pdf'

'gas chromatography - sulphur chemiluminescence':{'abbreviation':'GC-SC', 'sampling_type':'low volume continuous', 'measured_parameters':['SO2','H2S'], 'qa_accepted_parameters':['SO2','H2S'], 
    'instruments':{}
},

#------------------------------------
#flame photometric detection (FPD)

#Many elements give characteristic emission when burned in flame. Absorption of energy from the flame allows a ground state atom or molecule to reach an excited state. 
#The atom/molecule may return to the ground state through emission of light (luminescence), which can be subsequently measured (by a photomultiplier tube). This process is a chemiluminescent process (where luminescence occurs as the result of a chemical reaction).
#The concentration of the sample gas is directly proportional to the intensity of light emitted.
#This method has been applied for the measurement of sulphur and potassium containing species (i.e. SO2) and as a detector 
#For specific measurement of solely SO2, the sample gas must be scrubbed of other sulphur species prior to measurement, and the photomultiplier detector measures emission centred near 394nm.

#Known interferences: Other sulphur compounds

#See details of method for SO2: https://nepis.epa.gov/Exe/ZyNET.exe/9101A5EI.txt?ZyActionD=ZyDocument&Client=EPA&Index=1976%20Thru%201980&Docs=&Query=&Time=&EndTime=&SearchMethod=1&TocRestrict=n&Toc=&TocEntry=&QField=&QFieldYear=&QFieldMonth=&QFieldDay=&UseQField=&IntQFieldOp=0&ExtQFieldOp=0&XmlQuery=&File=D%3A%5CZYFILES%5CINDEX%20DATA%5C76THRU80%5CTXT%5C00000026%5C9101A5EI.txt&User=ANONYMOUS&Password=anonymous&SortMethod=h%7C-&MaximumDocuments=1&FuzzyDegree=0&ImageQuality=r75g8/r75g8/x150y150g16/i425&Display=hpfr&DefSeekPage=x&SearchBack=ZyActionL&Back=ZyActionS&BackDesc=Results%20page&MaximumPages=1&ZyEntry=22
#See for potassium: http://hiq.linde-gas.com/en/analytical_methods/gas_chromatography/flame_photometric_detector.html

'flame photometric detection':{'abbreviation':'FPD', 'sampling_type':'low volume continuous', 'measured_parameters':['SO2','H2S','K+','SO4--','SO4--_NSS','SO4--_SS','PM10_K+','PM10_SO4--','PM10_SO4--_NSS','PM10_SO4--_SS','PM2.5_K+','PM2.5_SO4--','PM2.5_SO4--_NSS','PM2.5_SO4--_SS','PM1_K+','PM1_SO4--','PM1_SO4--_NSS','PM1_SO4--_SS'], 'qa_accepted_parameters':['SO2','H2S','K+','SO4--','SO4--_NSS','SO4--_SS','PM10_K+','PM10_SO4--','PM10_SO4--_NSS','PM10_SO4--_SS','PM2.5_K+','PM2.5_SO4--','PM2.5_SO4--_NSS','PM2.5_SO4--_SS','PM1_K+','PM1_SO4--','PM1_SO4--_NSS','PM1_SO4--_SS'],  
    'instruments':{
        'Bendix 8302':    {'instrument_further_details':'No manual or specifications available but confirmation it is uses flame photometric detection here:https://in.booksc.me/book/12890878/314c1d'},
        'Bendix 8303':    {'instrument_further_details':'No manual or specifications available but confirmation it is uses flame photometric detection here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA249&lpg=PA249&dq=monitor+labs+8450&source=bl&ots=-uIdNzc1_l&sig=LqRK0OiodCMg5Sow1wlXQgIVLx0&hl=en&sa=X&ved=0ahUKEwjCy8eNvOfbAhVBuhQKHZTJBUoQ6AEIWDAG#v=onepage&q=monitor%20labs%208450&f=false'},
        'Meloy SA185-2A': {'instrument_further_details':'No manual or specifications available but confirmation it is uses flame photometric detection here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA249&lpg=PA249&dq=monitor+labs+8450&source=bl&ots=-uIdNzc1_l&sig=LqRK0OiodCMg5Sow1wlXQgIVLx0&hl=en&sa=X&ved=0ahUKEwjCy8eNvOfbAhVBuhQKHZTJBUoQ6AEIWDAG#v=onepage&q=monitor%20labs%208450&f=false'},
        'Meloy SA285E':   {'instrument_further_details':'No manual or specifications available but confirmation it is uses flame photometric detection here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA249&lpg=PA249&dq=monitor+labs+8450&source=bl&ots=-uIdNzc1_l&sig=LqRK0OiodCMg5Sow1wlXQgIVLx0&hl=en&sa=X&ved=0ahUKEwjCy8eNvOfbAhVBuhQKHZTJBUoQ6AEIWDAG#v=onepage&q=monitor%20labs%208450&f=false'},
        'Teledyne ML8450':{'instrument_further_details':'No manual or specifications available but confirmation it is uses flame photometric detection here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA249&lpg=PA249&dq=monitor+labs+8450&source=bl&ots=-uIdNzc1_l&sig=LqRK0OiodCMg5Sow1wlXQgIVLx0&hl=en&sa=X&ved=0ahUKEwjCy8eNvOfbAhVBuhQKHZTJBUoQ6AEIWDAG#v=onepage&q=monitor%20labs%208450&f=false'}
    }
},

#------------------------------------
#flame ionisation detection (FID)

#This method is based on the principle of the generation of an electrical current that is proportional to the rate of ion formation, dependent on the concentrations of species in the sample gas.
#The method is typically the standard detection method for column separated hydrocarbons, however the method is also sensitive to almost all compounds, mostly combustible ones. 
#There are, however, a few compounds to which the method has very little, if any, sensitivity, including: O2, N2, SO2, NO, N2O, NO2, NH3, CO, CO2, and H2O.
#When used as a standalone method (without prior chromatography) it is non-specific for mixed gases in the sample.

#Typical measurement operation for hydrocarbons:
#1. The sample gas is introduced into a hydrogen flame inside the flame ionisation detector.
#2. Any hydrocarbons in the sample will produce ions when they are burnt. 
#3. Ions are detected using a metal collector which is biased with a high DC voltage. 
#4. The current across this collector is thus proportional to the rate of ionisation which in turn depends upon the concentration of hydrocarbons in the sample gas.

'flame ionisation detection':{'abbreviation':'FID', 'sampling_type':'low volume continuous', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['VOC','HC'], 
    'instruments':{
        'Thermo 55C':{'instrument_further_details':'No manual or specifications available but confirmation it is uses flame ionisation detection here: https://www.oregon.gov/deq/aq/cao/Documents/CWM-111119SourceTestPlan.pdf'}
    }
},

#------------------------------------
#selective combustion - flame ionisation detection (SC-FID)

#This method combines flame ionisation detection with selctive combustion to enable to measurement of specific hydrocarbons in a mixed gaseous sample.
#This is most commonly employed for the measurement of CH4, but can be done for other hydrocarbons also. 

#Typical measurement operation for hydrocarbons:
#1. The sample gas is split into 2 lines
#2. On the first line, the sample gas is introduced into a hydrogen flame inside the flame ionisation detector (reference measurement).
#2. Any hydrocarbons in the sample will produce ions when they are burnt. 
#3. Ions are detected using a metal collector which is biased with a high DC voltage. 
#4. The current across this collector is thus proportional to the rate of ionisation which in turn depends upon the concentration of hydrocarbons in the sample gas.
#5. On the second line, the sample gas is first passed though a selective catalytic combustion unit, leaving only a specific hydrocarbon in the sample (oxidising the others), for example CH4
#6. This is then measured using FID, and thus gives the concentration of the isolated hydrocarbon
#7. If the isolated hydrocarbon measured on the second line is CH4, then line1 - line2 will give the NMHC concentration

'selective combustion - flame ionisation detection':{'abbreviation':'SC-FID', 'sampling_type':'low volume continuous', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 
    'instruments':{
        'Horiba APHA350':{                                                                                                                                                                    'instrument_further_details':'Details about instrument found here: https://www.umweltbundesamt.de/sites/default/files/medien/publikation/long/2702.pdf'}, 
        'Horiba APHA360':{'documented_precision':'1.0%', 'documented_accuracy':'1.0%',                                     'documented_span_drift':'0.5%/week', 'documented_flow_rate':'0.5', 'instrument_manual_name':'Horiba_APMA360_Specs.pdf'}, 
        'Horiba APHA370':{'documented_precision':'2.0%', 'documented_accuracy':'1.0%', 'documented_zero_drift':'2.0%/day', 'documented_span_drift':'2.0%/day',  'documented_flow_rate':'0.9', 'instrument_manual_name':'Horiba_APHA370_Manual.pdf'} 
    }
},

#------------------------------------
#conductimetry (CD)

#Involves the absorption of a specific gas species in deionized water, to produce an acid, which can be detected by a conductivity cell.
#Through the use of the deionized water reagent, measurements are susceptible to interferences from CO2, salt aerosols, acid mists and basic gases. 
#Addition of hydrogen peroxide to the deionized water minimises interference (through reduced solubility of CO2 within the water).
#This method been used extensively to measure SO2.

#Known interferences: Any gas that can form or remove ions (NO2, NH3, HCl, Cl2 are the worst known interferants).

#Not QA accepted measurement technique for SO2 due to dominance of other more accurate techniques in measuring this species

#See details of method for SO2:https://books.google.es/books?id=DiRevS__HUkC&pg=PA6&lpg=PA6&dq=SO2+conductimetric&source=bl&ots=wTME59BaGo&sig=D8wxHHP99GihBGlBJv1DXCUxcis&hl=en&sa=X&ved=0ahUKEwjmloGg7-fbAhXDzRQKHdtIAwAQ6AEIZjAH#v=onepage&q=SO2%20conductimetric&f=false'

'conductimetry':{'abbreviation':'CD', 'sampling_type':'low volume continuous', 'sample_preparation_type':'reagent reaction', 'measured_parameters':['NH3','HNO3','HCl','SO2','H2S'], 'qa_accepted_parameters':['NH3','HNO3','HCl'], 
    'instruments':{
        'Asarco 500':{                                                      'measured_parameters':['SO2'], 'qa_accepted_parameters':[], 'instrument_further_details':'cannot find any information. AQS states it as conductance. Assume conductimetry as detailed in general method reference below', 'measuring_assumption':True},
        'ECN AMOR':  {'sample_preparation_type':'denuder;reagent reaction', 'measured_parameters':['NH3'],                              'instrument_further_details':'See reference here:https://www.sciencedirect.com/science/article/pii/096016869390280C'}
    }
},

#------------------------------------
#second derivative spectrophotometry (SDS)

#Spectrophotometry is a method based on Beer-Lambert's law, where the absorption of light transmitted through the medium is directly proportional to the medium concentration.
#The main disadvantage and limitation of the spectrophotometry is its low selectivity. The measurement of absorbance is burden by interferences derived from others components of sample.
#Derivative spectrophotometry involves plotting the first, second or higher order derivatives of a spectrum with respect to wavelength. 
#Usually this is obtained by a microprocessor connected in series with a spectrophotometer, which computes the derivative with respect to time as the spectrum is scanned at constant speed.
#The "true" wavelength derivative is linearly related to the time derivative recorded, the magnitude of which is directly affected by scanning speed and spectral band width.
#The derivative process provides two general advantages: first, an effective enhancement of resolution, which can be useful to separate two or more components with overlapping spectra; second, a discrimination in favour of the sharpest features of a spectrum, used to eliminate interferences by broad band constituents. However, this process results in a decrease in the signal to noise ratio.
#Both advantages and disadvantages increase with the derivative order. Generally, the second derivative is more useful than the first ones.

#Not QA accepted measurement technique for NO/NO2/SO2 due to dominance of other more accurate techniques in measuring these species

#See details about method for NO,NO2,NH3,SO2:https://www.osapublishing.org/as/abstract.cfm?uri=as-53-11-1352

'second derivative spectrophotometry':{'abbreviation':'SDS', 'sampling_type':'low volume continuous', 'measured_parameters':['NO','NO2','NH3','HNO3','HCl','SO2','H2S'], 'qa_accepted_parameters':['NH3','HNO3','HCl'], 
    'instruments':{
        'Ecotech SM1000':{'measured_parameters':['SO2'], 'qa_accepted_parameters':[], 'instrument_further_details':'No manual or specifications available but confirmation it is uses second derivative spectrophotometry here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA249&lpg=PA249&dq=monitor+labs+8450&source=bl&ots=-uIdNzc1_l&sig=LqRK0OiodCMg5Sow1wlXQgIVLx0&hl=en&sa=X&ved=0ahUKEwjCy8eNvOfbAhVBuhQKHZTJBUoQ6AEIWDAG#v=onepage&q=monitor%20labs%208450&f=false'}
    }
},

#------------------------------------
#coulometry (CM)

#Coulometry determines the amount of matter transformed during an electrolysis reaction by measuring the amount of electricity (in coulombs) consumed or produced. 
#It can be used for precision measurements of charge, and the amperes even used to have a coulometric definition.
#Typical coulometric methods use solutions of potassium bromide (halide) and bromine (halogen) in dilute sulphuric acid.  
#The concentration of bromine is measured as a voltage between an indicator electrode and a bromide reference electrode.
#When the measured species is present, the bromine concentration is decreased through reaction. A feedback system then compensates by generator bromine at a generator electrode.
#The current required to regenerate the bromine is directly proportional to the concentration of the measured species.
#The method is extremely sensitive to interference through reaction with other undesired species (typically NO, NO2, O3, sulphur species, Cl2).
#Other species are typically removed through pre-filters.
#This method been used to measure NO2 and SO2.

#Known interferences: Reaction with other gases, typically: NO, NO2, O3, sulphur species, Cl2.

#Not QA accepted measurement technique for O3, NO2 or NOx due to dominance of other more accurate techniques in measuring these species

#See details about method for NO2: https://www.tandfonline.com/doi/pdf/10.1080/00022470.1962.10468113; and for SO2: https://www.tandfonline.com/doi/pdf/10.1080/00022470.1968.10469104
#See details about method for O3: https://pubs.acs.org/doi/abs/10.1021/es00028a012

'coulometry':{'abbreviation':'CM', 'sampling_type':'low volume continuous', 'sample_preparation_type':'reagent reaction', 'measured_parameters':['O3','NO','NO2','NOx','CO','SO2','H2S'], 'qa_accepted_parameters':[], 
    'instruments':{
        'Philips PW9700':{'measured_parameters':['SO2'], 'documented_lower_limit_of_detection':'4.0', 'instrument_further_details':'No manual or specifications available but confirmation it is uses coulometry here, and gives detection limit:https://books.google.es/books?id=KGQyBwAAQBAJ&pg=PT423&lpg=PT423&dq=PHILIPS+PW9700&source=bl&ots=lbloftODx4&sig=uBDyugqVzVHAhH2K21mDH7q2wYc&hl=en&sa=X&ved=0ahUKEwjxgYGZwOfbAhWHORQKHWEiBCAQ6AEIMTAB#v=onepage&q=PHILIPS%20PW9700&f=false'},
        'Philips PW9755':{'measured_parameters':['SO2'],                                              'instrument_further_details':'No manual or specifications available but confirmation it is uses coulometry here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA249&lpg=PA249&dq=monitor+labs+8450&source=bl&ots=-uIdNzc1_l&sig=LqRK0OiodCMg5Sow1wlXQgIVLx0&hl=en&sa=X&ved=0ahUKEwjCy8eNvOfbAhVBuhQKHZTJBUoQ6AEIWDAG#v=onepage&q=monitor%20labs%208450&f=false'}
    }
},

#------------------------------------
#polarography (PO)

#An electrochemical technique where the cell current is measured as a function of time and as a function of the potential between the indicator and reference electrodes. 
#The working electrode is a dropping mercury (or other liquid conductor) electrode and unstirred solutions are used. 
#The potential is varied using pulses of linearly increasing amplitude (one pulse during each drop lifetime) and the current is sampled before and after each voltage pulse.

#Not QA accepted measurement technique for NO/NO2/SO2 due to dominance of other more accurate techniques in measuring these species

#See details about method for NO, NO2 and SO2: https://www.sciencedirect.com/science/article/pii/0039914079800041

'polarography':{'abbreviation':'PO', 'sampling_type':'injection', 'measured_parameters':['NO','NO2','SO2','H2S'], 'qa_accepted_parameters':[], 
    'instruments':{}
},

#------------------------------------
#capillary electrophoresis (CE)

#Capillary electrophoresis (CE) is a family of electrokinetic separation methods performed in submillimeter diameter capillaries and in micro and nanofluidic channels.
#In capillary electrophoresis methods, analytes migrate through electrolyte solutions under the influence of an electric field. 
#Analytes can be separated according to ionic mobility and/or partitioning into an alternate phase via non-covalent interactions. 
#Additionally, analytes may be concentrated or "focused" by means of gradients in conductivity and pH.

#Not QA accepted measurement technique for SO2 due to dominance of other more accurate techniques in measuring this species
#See details about method for SO2: https://journals.sagepub.com/doi/abs/10.1177/1082013213494900

#See details for ions here: https://www.waters.com/waters/library.htm?locale=en_AU&lid=1822645&cid=511436 and https://www.isterre.fr/english/platforms-services-1235/platform-synthesis-and-characterization/synthesis-and-solid-liquid-interactions/article/capillary-electrophoresis-waters.html

'capillary electrophoresis':{'abbreviation':'CE', 'sampling_type':'injection', 'measured_parameters':['SO2','H2S','NH3','HNO3','HCl','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'qa_accepted_parameters':['NH3','HNO3','HCl','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 
    'instruments':{
        'Waters Alliance':{'instrument_further_details':'No manual or specifications available but confirmation it is uses capillary electrophoresis here: https://www.isterre.fr/english/platforms-services-1235/geochemistry-mineralogy-platform/article/capillary-electrophoresis-waters.html'}}
},

#------------------------------------
#ultraviolet fluorescence (UVF)

#Fluorescence is the emission of light (luminescence) by a substance that has absorbed light or other electromagnetic radiation (excitation). In most cases, the emitted light has a longer wavelength, and therefore lower energy, than the absorbed radiation.
#Ultraviolet fluorescence specifically refers to the process of a species being excited by ultraviolet light (10nm to 400nm) (species + hv (UV) --> species*).
#The return to a fundamental electronic state of the excited species is made by luminous radiation on a longer wavelength spectrum (species* --> species + hv).
#SO2 is efficiently excited to an excited state (SO2*) by UV light between 190 nm-230 nm, subsequently fluorescing light at approximately 330nm (SO2* --> SO2 +hv(330nm)), which can be measured.

#Typical measurement operation for SO2:
#1. The sample is drawn into the sample bulkhead (typically the sample flows through a hydrocarbon “kicker,” which removes hydrocarbons from the sample by forcing the hydrocarbon molecules to permeate through the tube wall).
#2. The sample then flows into the measurement cell, where pulsating UV light from a light source (typically mercury or zinc vapour lamp) between 190nm-230nm excites the SO2 molecules by the reaction: SO2 + hv (190nm-230nm) --> SO2* . A vacuum diode, UV detector that converts UV light to a DC current is used to measure the intensity of the excitation UV source lamp.
#3. Typically a band pass filter limits the wavelength of the UV source light to approximately 214 nm, and therefore the excitation reaction becomes: SO2 + hv (214nm) --> SO2* 
#4. The amount SO2 converted to SO2* in the sample chamber is dependent on the average intensity of the UV light (Ia) and not its peak intensity because the intensity of UV light is not constant in every part of the sample chamber. Some of the photons are absorbed by the SO2 as the light travels through the sample gas. The equation for defining the average intensity of the UV light (Ia) is: Ia = I0 [1 − exp(− ax(SO2))], Where: I0 = Intensity of the excitation UV light, a = The absorption coefficient of SO2 (a constant), SO2 = Concentration of SO2 in the sample chamber, x = The distance between the UV source and the SO2 molecule(s) being affected (path length). 
#5. SO2* then fluoresces light at a longer (lower energy) wavelength centred at 330nm: SO2* --> SO2 + hv (330nm).
#6. The amount of detectable UV given off by the decay of the SO2* is affected by the rate at which this reaction occurs (k). F = k(SO2*), Where: F = the amount of fluorescent light given off, k = The rate at which the SO2* decays into SO2, SO2* = Amount of excited-state SO2 in the sample chamber. Therefore: (SO2*) -kf --> SO2 + hv (330nm)
#7. The function (k) is affected by the temperature of the gas. The warmer the gas, the faster the individual molecules decay back into their ground state and the more photons of UV light are given off per unit of time. Given that the absorption rate of SO2 (a) is constant, the amount of fluorescence (F) is a result of: a) The amount of SO2* created which is affected by the variable factors: concentration of SO2; intensity of UV light (I0); path length of the UV light(x) and b) The amount of fluorescent light created which is affected by the variable factors: the amount of SO2* present and the rate of decay (k) which changes based on the temperature of the gas.
#8. When and the intensity of the light (I0) is known; path length of excited light is short (x); the temperature of the gas is known and compensated for so that the rate of SO2*decay is constant (k). and; no interfering conditions are present (such as interfering gases or stray light); the amount of fluorescent light emitted (F) is directly related to the concentration of the SO2 in the Sample Chamber.
#9. A Photo Multiplier Tube (PMT) detects the UV given off by the SO2* decay (330 nm) and outputs an analog signal. Several focusing lenses and optical filters ensure that both the PMT and source lamp vacuum diode are exposed to an optimum amount of only the right wavelengths of UV. To further assure that the PMT only detects light given off by decaying SO2* the pathway of the excitation UV and field of view of the PMT are perpendicular to each other and the inside surfaces of the sample chamber are coated with a layer of black teflon that absorbs stray light.
#10. The net result of the careful instrumental precision is any variation in UV fluorescence can be directly attributed to changes in the concentration of SO2 in the sample gas.

#Known interferences: 3rd body quenching (NO, CO2, O2, H2O), light pollution, UV absorption by ozone, other species undergoing ultraviolet fluorescence (poly-nuclear aromatics, NO)

'ultraviolet fluorescence':{'abbreviation':'UVF', 'sampling_type':'low volume continuous', 'measured_parameters':['SO2','H2S'], 'qa_accepted_parameters':['SO2','H2S'], 
    'instruments':{
        'Airpointer-SO2':        {'documented_lower_limit_of_detection':'0.5',   'documented_upper_limit_of_detection':'10000.0',   'documented_precision':'1.0/1.0%',    'documented_accuracy':'1.0%>=100.0',                  'documented_zero_drift':'1.0/day',                      'documented_span_drift':'1.0%>=100/day',                                  'documented_flow_rate':'0.5',       'instrument_manual_name':'Airpointer_Specs.pdf'},
        'Beckman 953':           {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA249&lpg=PA249&dq=monitor+labs+8450&source=bl&ots=-uIdNzc1_l&sig=LqRK0OiodCMg5Sow1wlXQgIVLx0&hl=en&sa=X&ved=0ahUKEwjCy8eNvOfbAhVBuhQKHZTJBUoQ6AEIWDAG#v=onepage&q=monitor%20labs%208450&f=false'},  
        'Dasibi 4108':           {'documented_lower_limit_of_detection':'1.0',   'documented_upper_limit_of_detection':'2000.0',    'documented_precision':'1.0',                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here, and details about limit of detection:https://books.google.es/books?id=ZNvqCAAAQBAJ&pg=PA155&lpg=PA155&dq=Dasibi+4108&source=bl&ots=rxhgzzOLrY&sig=W2zmIMXsFjAcNUXGLgTCYViACR0&hl=en&sa=X&ved=0ahUKEwj8xOXN2ufbAhXDchQKHRVWBksQ6AEIVjAI#v=onepage&q=Dasibi%204108&f=false'},
        'DKK-TOA GFS-32':        {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:http://itepsrv1.itep.nau.edu/itep_course_downloads/DAI%20resources/old%20resource%20files/Reference%20and%20Equivalent%20Methods%201-2007.pdf'},
        'DKK-TOA GFS-112E':      {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:https://www.federalregister.gov/documents/2000/01/18/00-1083/ambient-air-monitoring-reference-and-equivalent-methods-designation-of-a-new-equivalent-method-for'},
        'DKK-TOA GFS-312E':      {'documented_lower_limit_of_detection':'0.2',   'documented_upper_limit_of_detection':'1000.0',                                                                                                'documented_zero_drift':'0.5/day',                      'documented_span_drift':'1.0%/day',                                                                           'instrument_manual_name':'DKK-TOA_GFS-312E_Specs.pdf'},
        'Ecotech AM2020':        {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA249&lpg=PA249&dq=monitor+labs+8450&source=bl&ots=-uIdNzc1_l&sig=LqRK0OiodCMg5Sow1wlXQgIVLx0&hl=en&sa=X&ved=0ahUKEwjCy8eNvOfbAhVBuhQKHZTJBUoQ6AEIWDAG#v=onepage&q=monitor%20labs%208450&f=false'},  
        'Ecotech EC9850':        {'documented_lower_limit_of_detection':'0.5',   'documented_upper_limit_of_detection':'20000.0',   'documented_precision':'0.5/1.0%',    'documented_accuracy':'1.0%',                         'documented_zero_drift':'1.0/day',                      'documented_span_drift':'0.5%/day',      'documented_resolution':'0.001', 'documented_flow_rate':'0.5',       'instrument_manual_name':'Ecotech_EC9850_Manual.pdf'},
        'Ecotech EC9850T':       {'documented_lower_limit_of_detection':'0.2',   'documented_upper_limit_of_detection':'2000.0',    'documented_precision':'0.5/0.5%',                                                          'documented_zero_drift':'0.2/day',                      'documented_span_drift':'0.5%/day',                                       'documented_flow_rate':'0.5',       'instrument_manual_name':'Ecotech_EC9850T_Specs.pdf'},
        'Ecotech Serinus50':     {'documented_lower_limit_of_detection':'0.3',   'documented_upper_limit_of_detection':'20000.0',   'documented_precision':'0.5/0.15%',   'documented_accuracy':'1.0%',                         'documented_zero_drift':'0.5/day',                      'documented_span_drift':'1.0%/day',                                       'documented_flow_rate':'0.75',      'instrument_manual_name':'Ecotech_Serinus50_Manual.pdf'},
        'Envea AF22e':           {'documented_lower_limit_of_detection':'0.4',   'documented_upper_limit_of_detection':'10000.0',                                         'documented_accuracy':'1.0%',                         'documented_zero_drift':'1.0/day',                      'documented_span_drift':'0.5%/day',                                       'documented_flow_rate':'0.333',     'instrument_manual_name':'Envea_AF22e_Specs.pdf'},     
        'Environnement SA AF20M':{                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:https://books.google.es/books?id=sZLqCAAAQBAJ&pg=PA73&lpg=PA73&dq=environnement+af20m&source=bl&ots=OOOfZQbtkW&sig=ACfU3U0bdsNJH5j2XhFzrkF2tnkb6e2HKQ&hl=en&sa=X&ved=2ahUKEwjAidHQtf_fAhXD8uAKHcYOBRYQ6AEwAnoECAgQAQ#v=onepage&q=environnement%20af20m&f=false'},
        'Environnement SA AF21M':{                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2586140/'},
        'Environnement SA AF22M':{'documented_lower_limit_of_detection':'0.4',   'documented_upper_limit_of_detection':'10000.0',                                         'documented_accuracy':'1.0%',                         'documented_zero_drift':'0.5/day',                      'documented_span_drift':'0.5%/day',                                       'documented_flow_rate':'0.3',       'instrument_manual_name':'Environnement_SA_AF22M_Specs.pdf'},
        'Horiba APSA300':        {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:https://books.google.co.uk/books?id=-rDwCAAAQBAJ&pg=PA121&lpg=PA121&dq=APSA-300+SO2&source=bl&ots=AnmNzO9vQI&sig=ACfU3U0L-QiSGScTD5rDtfIGRDZa1BF0Sg&hl=en&sa=X&ved=2ahUKEwj68aSO8v_zAhVeEWMBHQuIBUEQ6AF6BAgTEAM#v=onepage&q=APSA-300%20SO2&f=false'},
        'Horiba APSA350':        {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:https://books.google.es/books?id=qmWXBwAAQBAJ&pg=PA88&lpg=PA88&dq=apsa350&source=bl&ots=dBQKXVwgi2&sig=ACfU3U033XGNb0kOfeN06k0HDr7Oot-Imw&hl=en&sa=X&ved=2ahUKEwi0jbGJsP_fAhWC2OAKHTYyCU44ChDoATAAegQIABAB#v=onepage&q=apsa350&f=false'},
        'Horiba APSA360':        {'documented_lower_limit_of_detection':'0.5',   'documented_upper_limit_of_detection':'10000.0',   'documented_precision':'1.0%',        'documented_accuracy':'1.0%',                         'documented_zero_drift':'0.5/day',                      'documented_span_drift':'0.5/day',                                        'documented_flow_rate':'0.8',       'instrument_manual_name':'Horiba_APMA360_Specs.pdf'},
        'Horiba APSA365':        {'documented_lower_limit_of_detection':'0.1',   'documented_upper_limit_of_detection':'100.0',     'documented_precision':'2.0%',        'documented_accuracy':'2.0%',                         'documented_zero_drift':'2.0%<20.0/day 1.0%>=20.0/day', 'documented_span_drift':'2.0%/day',                                       'documented_flow_rate':'0.8',       'instrument_further_details':'Details found online:http://www.biowell.sk/produkt/podimis/365so2.html'},
        'Horiba APSA370':        {'documented_lower_limit_of_detection':'0.5',   'documented_upper_limit_of_detection':'10000.0',   'documented_precision':'1.0%',        'documented_accuracy':'1.0%',                         'documented_zero_drift':'0.5/day',                      'documented_span_drift':'0.5/day',                                        'documented_flow_rate':'0.7',       'instrument_manual_name':'Horiba_APSA370_Specs.pdf'},
        'MCV 64FUV':             {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'cannot find any information. EIONET states it as automatic for SO2 which is almost always ultraviolet fluorescencer. Assume this.', 'measuring_assumption':True},  
        'Meloy SA700':           {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA249&lpg=PA249&dq=monitor+labs+8450&source=bl&ots=-uIdNzc1_l&sig=LqRK0OiodCMg5Sow1wlXQgIVLx0&hl=en&sa=X&ved=0ahUKEwjCy8eNvOfbAhVBuhQKHZTJBUoQ6AEIWDAG#v=onepage&q=monitor%20labs%208450&f=false'},  
        'Philips K50033':        {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:https://www.google.com/url?sa=t&rct=j&q=&esrc=s&source=web&cd=4&ved=2ahUKEwiaspe-19jnAhWI4IUKHVswA1oQFjADegQICBAB&url=http%3A%2F%2Ffbd.beun.edu.tr%2Findex.php%2Fzkufbd%2Farticle%2Fdownload%2F53%2F100&usg=AOvVaw0l7fQAqr1Pealzx0MaZ6d0'},
        'Philips K50206/00':     {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:https://pdf.sciencedirectassets.com/277910/1-s2.0-S1876610214X00032/1-s2.0-S1876610214000745/main.pdf?X-Amz-Security-Token=IQoJb3JpZ2luX2VjELX%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FwEaCXVzLWVhc3QtMSJHMEUCIQCHD7TbrA0pGiTZ0F1yTQNWNKTRV8wDIzKAliNRM9MGrQIgKRH3nSIu%2Fb3f2feAIwkay5DxYtP0mB0uT1LlRkEBi4Iq%2BgMIXhAEGgwwNTkwMDM1NDY4NjUiDPRn9A4cq5O8L8dhwirXA4qKOPYgQb%2BFEhOxqj6i%2Fhf3MIQYD0xKUHImNbwCxb7VBVxFSh46qMQ8HufqiIxSPPQGcMFRQ27PCDd3WGi0Qse053q8vx7zMBwgAy3xX0jWWKwfwF74JWx0o6TIulxoEjetJywZLtjxygDhrRNUvgdmW6mJ29C80TFOzudlS%2FP73YJLAgkoS%2F7%2BQ4XrXOUGMNf%2FsV5%2B6lpYFJNyfeZ4ma5p7KdFRkTE3kD%2B1L%2Fl%2FHZWmnRB0WVkjPSPubgsoPaF%2FvEzqG6Rz8Jzn9WJ%2FSFljLMzQOm%2B3z6rk5Y3s7DivqxEDNWCNDYFgLnxuiCXUeh7PCxoZ2tyQg3%2BOMlTI96nx2HSp24B9txv%2BURLBJ7a3W8rYNZ0XGJ7GA6HU1K3uUns%2FmSK1CPXPmKNt2kOa5D65zvOjY%2FHpjnXLIyLOdbiJzi4b%2F5fS0XUa3n0GBn%2FnsTc3YZsqB7rvm%2FtsdmJMyqDc2Gy3nmZ8BU0L4i%2FiFzsvwdinbaAs4bOSo2GZDJSBgjkhEIvE6ww4V4QGfDvkcVeKqc85x%2BxBk4DSUzi7DE7G7A4XZSfvI81n9%2FExbtsRvsdjy5aLG6zPXp1RUEu%2F6BgimRSnJ5gQK5%2FrqqOzVRtCxKCrvmhHDJ74TDSq4%2BMBjqlAV8liNT1G81EJOD0K7MybECILa1HzEKLor5d7dopFy2TZ24XmOokzgLOEGhPY1zaHqWoKzDyY0nJiIDDuvs2oY9hw6sQgZoE6OIRZXlBQP%2FrU0q8wMmMm6i%2FlvFaz5y8dZsCiggEyrTbHTzMFcRv6vzmO7GWnuthvk9d00k%2Fuvd7U3kIXsU7rwF7GsMPe%2BrkY3QLlHkPn83JUZ7MVQv8MSr1GgsAEw%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20211104T135204Z&X-Amz-SignedHeaders=host&X-Amz-Expires=300&X-Amz-Credential=ASIAQ3PHCVTY2ZEZUA2D%2F20211104%2Fus-east-1%2Fs3%2Faws4_request&X-Amz-Signature=411af053b6470203f95e0d6e41b0e26586bf4b68b60637f65f60ef0b5ef7b052&hash=3484d1dabf3788181c39f021bddb513640ad955227ceb405b51e859348331426&host=68042c943591013ac2b2430a89b270f6af2c76d8dfd086a07176afe7c76c2c61&pii=S1876610214000745&tid=spdf-4e792382-28a3-4408-8d2c-e7d78fc31a79&sid=fca986874c986748ed0adba-f4e4c04dd527gxrqb&type=client'},
        'Sabio 6020':            {'documented_lower_limit_of_detection':'0.3',   'documented_upper_limit_of_detection':'2000.0',    'documented_precision':'1.0%',        'documented_accuracy':'1.0%',                         'documented_zero_drift':'3.0/day',                      'documented_span_drift':'1.0%/day',                                       'documented_flow_rate':'0.4-0.8',   'instrument_manual_name':'Sabio_6020_Specs.pdf'},
        'Seres SF2000G':         {'documented_lower_limit_of_detection':'1.0',   'documented_upper_limit_of_detection':'20000.0',                                         'documented_accuracy':'1.0%',                         'documented_zero_drift':'2.0/week',                     'documented_span_drift':'1.0%/month',                                     'documented_flow_rate':'0.5-0.833', 'instrument_manual_name':'Seres_SF2000G_Specs.pdf'},
        'SIR S-5001':            {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:http://www.eufar.net/instruments/423'},
        'Teledyne 100':          {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'cannot find any information. AQS states it as ultraviolet fluorescence. Assume this', 'measuring_assumption':True},
        'Teledyne 100A':         {'documented_lower_limit_of_detection':'0.4',   'documented_upper_limit_of_detection':'20000.0',   'documented_precision':'0.5%',        'documented_accuracy':'1.0%',                         'documented_zero_drift':'0.5/day',                      'documented_span_drift':'0.5%/7days',                                     'documented_flow_rate':'0.65',      'instrument_manual_name':'Teledyne_100A_Manual.pdf'},
        'Teledyne 100AH':        {'documented_lower_limit_of_detection':'100.0', 'documented_upper_limit_of_detection':'5000000.0', 'documented_precision':'0.5%',        'documented_accuracy':'1.0%',                         'documented_zero_drift':'1000.0/day',                   'documented_span_drift':'0.5%/7days',                                     'documented_flow_rate':'0.65',      'instrument_manual_name':'Teledyne_100AH_Manual.pdf'},
        'Teledyne 100E':         {'documented_lower_limit_of_detection':'0.4',   'documented_upper_limit_of_detection':'20000.0',   'documented_precision':'0.5%>=50.0',  'documented_accuracy':'1.0%',                         'documented_zero_drift':'0.5/day',                      'documented_span_drift':'0.5%/day',                                       'documented_flow_rate':'0.65',      'instrument_manual_name':'Teledyne_100E_Manual.pdf'},
        'Teledyne 100EH':        {'documented_lower_limit_of_detection':'100.0', 'documented_upper_limit_of_detection':'5000000.0', 'documented_precision':'0.5%',        'documented_accuracy':'1.0%',                         'documented_zero_drift':'1000.0/day',                   'documented_span_drift':'0.5%/7days',                                     'documented_flow_rate':'0.7',       'instrument_manual_name':'Teledyne_100EH_Addendum.pdf'},
        'Teledyne 100EU':        {'documented_lower_limit_of_detection':'0.05',  'documented_upper_limit_of_detection':'20000.0',   'documented_precision':'0.5%',        'documented_accuracy':'1.0%',                         'documented_zero_drift':'0.2/day',                      'documented_span_drift':'0.5%/7days',                                     'documented_flow_rate':'0.65',      'instrument_manual_name':'Teledyne_100EU_Addendum.pdf'},
        'Teledyne 101':          {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/API101/view?facet=HTML+Representation'},      
        'Teledyne 101A':         {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:http://pagina.jccm.es/medioambiente/rvca/pdfs/INFORME_UNIDAD-MOVIL-MORA_2_15%20.pdf'},
        'Teledyne T100':         {'documented_lower_limit_of_detection':'0.4',   'documented_upper_limit_of_detection':'20000.0',   'documented_precision':'0.5%>=50.0',  'documented_accuracy':'1.0%',                         'documented_zero_drift':'0.5/day',                      'documented_span_drift':'0.5%/day',                                       'documented_flow_rate':'0.65',      'instrument_manual_name':'Teledyne_T100_Manual.pdf'},
        'Teledyne T100H':        {'documented_lower_limit_of_detection':'200.0', 'documented_upper_limit_of_detection':'5000000.0', 'documented_precision':'0.5%>=10000', 'documented_accuracy':'1.0%',                         'documented_zero_drift':'1000.0/day',                   'documented_span_drift':'0.5%/day',                                       'documented_flow_rate':'0.7',       'instrument_manual_name':'Teledyne_T100H_Addendum.pdf'},
        'Teledyne T100U':        {'documented_lower_limit_of_detection':'0.05',  'documented_upper_limit_of_detection':'20000.0',   'documented_precision':'0.5%',        'documented_accuracy':'1.0%',                         'documented_zero_drift':'0.2/day',                      'documented_span_drift':'0.5%/day',                                       'documented_flow_rate':'0.65',      'instrument_manual_name':'Teledyne_T100U_Addendum.pdf'},
        'Teledyne T101':         {'documented_lower_limit_of_detection':'0.4',   'documented_upper_limit_of_detection':'20000.0',   'documented_precision':'0.5%>=50.0',  'documented_accuracy':'1.0%',                         'documented_zero_drift':'0.5/day',                      'documented_span_drift':'0.5%/day',                                       'documented_flow_rate':'0.65',      'instrument_manual_name':'Teledyne_T101_Manual.pdf'},
        'Teledyne ML8850':       {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:https://books.google.es/books?id=BODqCAAAQBAJ&pg=PA137&lpg=PA137&dq=MONITOR+LABS+8850&source=bl&ots=TUXOyv5p5D&sig=Ult9_8r_m8xwlzmZIHNu3a3JaMs&hl=en&sa=X&ved=0ahUKEwjsu52E2ufbAhXBvhQKHVNNCrkQ6AEIdzAM#v=onepage&q=MONITOR%20LABS%208850&f=false'},
        'Teledyne ML8850M':      {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/ML8850M/view'},
        'Teledyne ML8850S':      {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:https://www.der.wa.gov.au/images/documents/your-environment/air/publications/perth-photochemical-smog-study-1996.pdf'},
        'Teledyne ML9850':       {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:http://www.epa.ie/licences/lic_eDMS/090151b28005fa5c.pdf'},
        'Teledyne ML9850B':      {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:http://www.epa.ie/licences/lic_eDMS/090151b28005fa5c.pdf'},
        'Thermo 43':             {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA249&lpg=PA249&dq=monitor+labs+8450&source=bl&ots=-uIdNzc1_l&sig=LqRK0OiodCMg5Sow1wlXQgIVLx0&hl=en&sa=X&ved=0ahUKEwjCy8eNvOfbAhVBuhQKHZTJBUoQ6AEIWDAG#v=onepage&q=monitor%20labs%208450&f=false'},  
        'Thermo 43A':            {'documented_lower_limit_of_detection':'1.0',   'documented_upper_limit_of_detection':'100.0',                                                                                                                                                                                                                                                                       'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here, and details about limit of detection:https://books.google.es/books?id=G0EYUql3_98C&pg=PA283&lpg=PA283&dq=thermo+43a&source=bl&ots=k2RUJXlAVJ&sig=yIxG08AoSyuqedtS6dp-O7O04X4&hl=en&sa=X&ved=0ahUKEwj17qqB3OfbAhVCVRQKHanFDlo4ChDoAQgoMAA#v=onepage&q=thermo%2043a&f=false'},
        'Thermo 43B':            {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here:http://www.southwestohioair.org/UserFiles/Servers/Server_3788196/File/EnvironmentalServices/AirQuality/Monitoring/AirMonitoringNetwork16-17.pdf'},
        'Thermo 43C':            {'documented_lower_limit_of_detection':'0.5',   'documented_upper_limit_of_detection':'100000.0',  'documented_precision':'1.0/1.0%',    'documented_accuracy':'1.0%',                         'documented_zero_drift':'1.0/day',                      'documented_span_drift':'1.0%/day',                                       'documented_flow_rate':'0.5',       'instrument_manual_name':'Thermo_43C_Specs.pdf'},
        'Thermo 43C-TLE':        {'documented_lower_limit_of_detection':'0.05',  'documented_upper_limit_of_detection':'1000.0',    'documented_precision':'0.2/1.0%',                                                          'documented_zero_drift':'0.2/day',                      'documented_span_drift':'1.0%/day',                                       'documented_flow_rate':'0.5',       'instrument_manual_name':'Thermo_43C-TLE_Specs.pdf'},
        'Thermo 43H':            {'documented_lower_limit_of_detection':'10.0',  'documented_upper_limit_of_detection':'20000.0',                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here, and details about detection limits: https://www.yumpu.com/fr/document/read/7309580/catalogue-cleanair-europe/36'},
        'Thermo 43i':            {'documented_lower_limit_of_detection':'0.5',   'documented_upper_limit_of_detection':'100000.0',                                        'documented_accuracy':'1.0%',                         'documented_zero_drift':'1.0/day',                      'documented_span_drift':'1.0%/day',                                       'documented_flow_rate':'0.5',       'instrument_manual_name':'Thermo_43i_Manual.pdf'},
        'Thermo 43i-TLE':        {'documented_lower_limit_of_detection':'0.05',  'documented_upper_limit_of_detection':'1000.0',                                          'documented_accuracy':'1.0%',                         'documented_zero_drift':'0.2/day',                      'documented_span_drift':'1.0%/day',                                       'documented_flow_rate':'0.5',       'instrument_manual_name':'Thermo_43i-TLE_Manual.pdf'},
        'Thermo 43S':            {                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here: https://books.google.es/books?id=JqpwDwAAQBAJ&pg=PA232&lpg=PA232&dq=thermo+43s&source=bl&ots=h9MazeIX98&sig=ACfU3U0F3I9zCn-qI-pM7Np-yN5w7AxRGg&hl=en&sa=X&ved=2ahUKEwiwh9ri1f7fAhXh8eAKHfd9CxYQ6AEwCXoECAkQAQ#v=onepage&q=thermo%2043s&f=false'},
        'Thermo 43W':            {'documented_lower_limit_of_detection':'1.0',   'documented_upper_limit_of_detection':'750.0',                                                                                                                                                                                                                                                                       'instrument_further_details':'No manual or specifications available but confirmation it is uses ultraviolet fluorescence here, and details about limit of detection:https://books.google.es/books?id=G0EYUql3_98C&pg=PA283&lpg=PA283&dq=thermo+43a&source=bl&ots=k2RUJXlAVJ&sig=yIxG08AoSyuqedtS6dp-O7O04X4&hl=en&sa=X&ved=0ahUKEwj17qqB3OfbAhVCVRQKHanFDlo4ChDoAQgoMAA#v=onepage&q=thermo%2043a&f=false'},
        'Thermo 450i':           {'documented_lower_limit_of_detection':'1.5',   'documented_upper_limit_of_detection':'100000.0',                                        'documented_accuracy':'1.0%<100000.0 5.0%>=100000.0', 'documented_zero_drift':'1.0/day',                      'documented_span_drift':'1.0%/day',                                       'documented_flow_rate':'1.0',       'instrument_manual_name':'Thermo_450i_Manual.pdf'}
    }
},

#------------------------------------
#laser-induced fluorescence (LIF)

#Fluorescence is the emission of light (luminescence) by a substance that has absorbed light or other electromagnetic radiation (excitation). In most cases, the emitted light has a longer wavelength, and therefore lower energy, than the absorbed radiation.
#Laser-induced fluorescence (LIF) is a method where fluorescence is induced by an atom or molecule being excited  by the absorption of laser light.
#The emission light is detected using a photomutiplier tube. The amount of a species present in the sample is proportional to the intensity of the emission light of a specific wavelength and quantification is possible by using calibration data.
#The LIF method has been used mostly in the measurement of hydroxyl and hydroperoxyl radicals, but is also used for measurement of NO/NO2.

#Known Interferences: Laser power modulation (laser excitation could generate OH radicals biasing the measurement). 3rd body quenching.

#See details of method for NO/NO2 here:https://agupubs.onlinelibrary.wiley.com/doi/full/10.1029/1999GL900015 and here:http://etheses.whiterose.ac.uk/11668/1/Amy%20Foulds_Final%20MSc%20Thesis.pdf

'laser-induced fluorescence':{'abbreviation':'LIF', 'sampling_type':'low volume continuous', 'measured_parameters':['NO','NO2','NOx'], 'qa_accepted_parameters':['NO','NO2','NOx'], 
    'instruments':{}
},

#------------------------------------
#vacuum ultraviolet resonance fluorescence (VURF)

#Fluorescence is the emission of light (luminescence) by a substance that has absorbed light or other electromagnetic radiation (excitation). In most cases, the emitted light has a longer wavelength, and therefore lower energy, than the absorbed radiation.
#Vacuum Ultraviolet Resonance Fluorescence is a method where a reasonance lamp excited by reasonance fluorescence discharge (in combiantion with an optical filter) produces photons in the ultraviolet which react with a sample gas, inducing fluorescence in the vacuum ultraviolet (UV radiaion 10nm-200nm), subsequently detected by a photomultiplier tube.
#This method can be employed for CO, with the reasonance lamp emitting UV light between 145nm-151nm, with fluorescence occurring between 160nm-190nm.

#Known interferences: Water vapour, drifts in lamp intensity, continuum raman scattering by O2. 

'vacuum ultraviolet resonance fluorescence':{'abbreviation':'VURF', 'sampling_type':'low volume continuous', 'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 
    'instruments':{
        'Aerolaser AL5001':{'documented_lower_limit_of_detection':'0.8', 'documented_upper_limit_of_detection':'100000.0', 'measuring_instrument_manual_name':'Aerolaser_AL5001_Specs.pdf'},
        'Aerolaser AL5002':{'documented_lower_limit_of_detection':'0.8', 'documented_upper_limit_of_detection':'100000.0', 'measuring_instrument_manual_name':'Aerolaser_AL5002_Specs.pdf'}
    }
},

#------------------------------------
#cavity ringdown spectroscopy (CRDS)

#Based on absorption spectroscopy, Cavity Ringdown Spectroscopy works by attuning light rays to the unique molecular fingerprint of the sample species. 
#By measuring the time it takes the light to fade or "ring-down", you receive an accurate molecular count in milliseconds. 
#The time of light decay, in essence, provides an exact, non-invasive, and rapid means to detect contaminants in the air, in gases, and even in the breath.
#The method is typically employed to measure CO and other greenhouse gases (e.g CO2, CH4, H2O).

#Typical measurement operation:
#1. A Continuous Wave (CW) diode laser emits a directed beam of light energy through an ultra-high reflective mirror into the absorption cell (cavity).
#2. The light reflects back and forth between two ultra-high reflective mirrors multiple times, up to a total path length of 100 kilometers.
#3. Once the photodiode detector “sees” a preset level of light energy, the light source is shuttered or diverted from the cavity.
#4. On each successive pass, a small amount of light or ring-down signal emits through the second mirror and is sensed by the light detector.
#5. Once the light "rings down", the detector achieves a point of zero light energy in milliseconds, and the measurement is complete.    
#6. The computer-controlled system tunes the laser off the absorption peak for the sample species to determine the τ empty value, equivalent to a zero baseline correction. It tunes back to the absorption peak to determine the τ value, dependent on the sample species concentration.
#7. The concentration of the sample species is directly calculated using Beer’s Law. The measured value constitutes an absolute measurement and is unaffected by losses outside the ring-down cavity.

#Known interferences: Water vapour, CO2 and particulates.

'cavity ringdown spectroscopy':{'abbreviation':'CRDS', 'sampling_type':'low volume continuous', 'measured_parameters':['NO','NO2','SO2','CO','CO2','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO','H20','PAN','NH3','HNO3','HCl','HF','H2S'], 'qa_accepted_parameters':['NO','NO2','SO2','CO','CO2','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO','H20','PAN','NH3','HNO3','HCl','HF','H2S'], 
    'instruments':{
        'Picarro Envirosense 3000i':{'measured_parameters':['CO2','CH4','H2O'],      'qa_accepted_parameters':['CO2','CH4','H2O'],                                                                                                                                                                                                                                                                          'documented_precision':{'CO2':'200.0','CH4':'1.0','H2O':'100000.0'},       'documented_span_drift':{'CO2':'150.0/day','CH4':'1.0/day', 'H2O':np.NaN},                                                        'instrument_further_details':'No manual, but some specifications stated here:https://gml.noaa.gov/co2conference/posters_pdf/hartnett_picarro.pdf'},
        'Picarro G1301':            {'measured_parameters':['CO2','CH4','H2O'],      'qa_accepted_parameters':['CO2','CH4','H2O'],                                                                                                                                                                                                                                                                                                                                                                                                                                                       'documented_flow_rate':'0.4',   'instrument_manual_name':'Picarro_G1301_Manual.pdf'},
        'Picarro G2132-i':          {'measured_parameters':['CO2','CH4','H2O'],      'qa_accepted_parameters':['CO2','CH4','H2O'], 'documented_lower_limit_of_detection':{'CO2':'200000.0','CH4':'1200.0','H2O':np.NaN}, 'documented_upper_limit_of_detection':{'CO2':'2000000.0','CH4':'1500000.0','H2O':np.NaN},                                                                                          'documented_precision':{'CO2':'1000.0','CH4':'5.0','H2O':'100000.0'},                                                                                                        'documented_flow_rate':'0.025', 'instrument_manual_name':'Picarro_G2132-i_Specs.pdf'},
        'Picarro G2301':            {'measured_parameters':['CO2','CH4','H2O'],      'qa_accepted_parameters':['CO2','CH4','H2O'],                                                                                       'documented_upper_limit_of_detection':{'CO2':'1000000.0','CH4':'20000.0','H2O':np.NaN},                                                                                            'documented_precision':{'CO2':'25.0','CH4':'0.22','H2O':'30000.0'},        'documented_span_drift':{'CO2':'120.0/day','CH4':'1.0/day','H2O':'100000.0/day'},                 'documented_flow_rate':'0.4',   'instrument_manual_name':'Picarro_G2301_Specs.pdf'},
        'Picarro G2302':            {'measured_parameters':['CO','CO2','H2O'],       'qa_accepted_parameters':['CO','CO2','H2O'],                                                                                        'documented_upper_limit_of_detection':{'CO':'5000.0','CO2':'1000000.0','H2O':np.NaN},                 'documented_uncertainty':{'CO':'2.0','CO2':np.NaN,'H2O':np.NaN},             'documented_precision':{'CO':'2.0','CO2':'50.0','H2O':'50000.0'},          'documented_span_drift':{'CO':'15.0/day','CO2':'150.0/day','H2O':'100000.0/day'},                 'documented_flow_rate':'0.4',   'instrument_manual_name':'Picarro_G2302_Specs.pdf'},
        'Picarro G2303':            {'measured_parameters':['CO'],                   'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    'instrument_further_details':'EIONET reports instrumnent as a CRDS instrument but cannot find other documentation, assume it is a CRDS instrument as other Picarro instruments are: http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/picarroG2303/view', 'measuring_assumption':True},
        'Picarro G2401':            {'measured_parameters':['CO','CO2','CH4','H2O'], 'qa_accepted_parameters':['CO','CO2','CH4','H2O'],                                                                                  'documented_upper_limit_of_detection':{'CO':'5000.0','CO2':'1000000.0','CH4':'20000.0','H2O':np.NaN}, 'documented_uncertainty':{'CO':'2.0','CO2':'50.0','CH4':'1.0','H2O':np.NaN}, 'documented_precision':{'CO':'1.0','CO2':'10.0','CH4':'0.3','H2O':np.NaN}, 'documented_span_drift':{'CO':'10.0/day','CO2':'100.0/day','CH4':'1.0/day','H2O':'100000.0/day'}, 'documented_flow_rate':'0.4',   'instrument_manual_name':'Picarro_G2401_Specs.pdf'}
    }
},

#------------------------------------
#off-axis integrated cavity output spectroscopy (OA-ICOS)

#CRDS has practical limitations that make it nonideal for many applications. Specifically, the resonant wavelength coupling requires subnanometer optomechanical precision and stability. 
#Thus field service is impractical. In addition, resonant CRDS systems require extreme laser wavelength monitoring and control. 
#These challenges can, at times, be overcome in the laboratory but with a trade-off of increased instrument cost, complexity, and robustness.
#OA-ICOS, originally developed in cooperation with scientists at Harvard University (Cambridge, MA), was created in order to overcome these limitations, while still maintaining extremely high sensitivity and precision. 
#OA-ICOS uses off-axis, nonresonant alignment of the laser beam to the cavity. 
#This beam alignment or trajectory is not unique and is easy to achieve, making OA-ICOS instruments orders of magnitude more robust and less sensitive to thermal changes and vibrations. As a result, OA-ICOS instruments are relatively simple to manufacture and service, inexpensive to build, and robust. 
#The OA-ICOS measurement combines the advantages of CRDS with conventional high-resolution laser absorption spectroscopy by providing direct absorption measurements with optical pathlengths thousands of meters long.
#Unlike CRDS, which enables at most only limited wavelength sampling (discrete jumps between wavelengths), OA-ICOS easily allows continuous scanning of the laser wavelength and thus enables recording of fully resolved absorption spectra. 
#This is a critical advantage that permits OA-ICOS to provide accurate measurements of complex and/or contaminated samples without cross-sensitivity. 
#Moreover, OA-ICOS can be operated at any wavelength from the UV through the IR. As a result, instruments can target the strongest absorption lines, allowing sensitivity below 1 ppt for some gases. Furthermore, because it is an intensity measurement, only OA-ICOS can deliver several decades of linear dynamic range.

'off-axis integrated cavity output spectroscopy':{'abbreviation':'OA-ICOS', 'sampling_type':'low volume continuous', 'measured_parameters':['NO','NO2','SO2','CO','CO2','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO','H20','PAN','NH3','HNO3','HCl','HF','H2S'], 'qa_accepted_parameters':['NO','NO2','SO2','CO','CO2','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO','H20','PAN','NH3','HNO3','HCl','HF','H2S'], 
    'instruments':{
        'Los Gatos CO':    {'measured_parameters':['CO'],       'qa_accepted_parameters':['CO'],       'documented_lower_limit_of_detection':'1.0', 'documented_upper_limit_of_detection':'4000.0',   'documented_precision':'0.05',                               'documented_span_drift':'1.0/day', 'instrument_manual_name':'LosGatosCO_Specs.pdf'},
        'Los Gatos GGA':   {'measured_parameters':['CH4'],      'qa_accepted_parameters':['CH4'],      'documented_lower_limit_of_detection':'1.0', 'documented_upper_limit_of_detection':'100000.0', 'documented_precision':'0.3',  'documented_accuracy':'1.0%', 'documented_span_drift':'5.0/day', 'instrument_manual_name':'LosGatosGGA_Specs.pdf'},   
        'Los Gatos N2O/CO':{'measured_parameters':['CO','N2O'], 'qa_accepted_parameters':['CO','N2O'], 'documented_lower_limit_of_detection':'1.0', 'documented_upper_limit_of_detection':'4000.0',   'documented_precision':'0.3',                                'documented_span_drift':'1.0/day', 'instrument_manual_name':'LosGatosN20CO_Specs.pdf'}
    }
},

#------------------------------------
#tunable diode laser absorption spectroscopy (TDLAS)

#TDLAS measures the wavelength-dependent absorption of light through a gas medium. 
#As the name implies, the technique usually employs a tunable-wavelength diode laser as the light source.  
#When the wavelength of light matches one absorption line of a gas species present in the sample, the photodetector records a reduction in the light intensity. 
#When the gas concentration is ultra-low, the induced change in transmission becomes extremely small and difficult to detect sufficiently fast. 
#The sensitivity of TDLAS is significantly enhanced by modulating the current of the laser, which leads to a modulation of the wavelength and of the light intensity. 
#The information about the absorption response is then recovered by demodulating the signal from the photodetector at the modulating frequency and at its second-order harmonic.
#The advantage of TDLAS over other techniques for concentration measurement is its ability to achieve very low detection limits
#Apart from concentration, it is also possible to determine the temperature, pressure, velocity and mass flux of the gas under observation

'tunable diode laser absorption spectroscopy':{'abbreviation':'TDLAS', 'sampling_type':'low volume continuous', 'measured_parameters':['NO','NO2','SO2','CO','CO2','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO','H20','PAN','NH3','HNO3','HCl','HF','H2S'], 'qa_accepted_parameters':['NO','NO2','SO2','CO','CO2','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO','H20','PAN','NH3','HNO3','HCl','HF','H2S'], 
    'instruments':{}
},

#------------------------------------
#cavity attenuated phase shift spectroscopy (CAPS)

#Operates on the principle that a specific species efficiently absorbs light at a known wavelength. This is the case for NO2, at 450nm.
#The degree to which the light is absorbed by a specific species is directly related to the species concentration as described by the Beer-Lambert Law (A = εLC; A = Absorbance (mol litres-1), ε = Molar absorptivity (litres mol-1 cm-1), L = mean optical path length of cell, C = species concentration).
#Emitted light in an optical cell is reflected back and forth between two mirrors, building intensity and running a very long path length. The long path length extends the “time” or “life” of the photon, thus providing ample time to measure absorbance when a species is present.
#The method is typically employed for direct measurement of NO2.

#Typical measurement operation for NO2:
#1. The sample is drawn into the sample bulkhead, before being passed into the measurement cell. 
#2. Light is emitted from a blue ultraviolet (UV) light emitting diode (LED) centered at 450 nm (a prominent absorption band for NO2). 
#3. The measurement cell contains high reflectivity mirrors located at either end to provide an extensive optical path length. The optical cell resides in a temperature controlled oven. The oven raises the ambient temperature of the sample gas to 45 degrees Celsius. This mitigates the formation of moisture on the surfaces of the mirrors while also minimiSing changes in the absorption coefficient due to temperature fluctuations.
#4. The CAPS method is unique in that it applies the fundamental optical absorption law in the frequency domain, rather than using relative changes in light intensity as the primary signal. 
#5. UV light from the modulating high intensity LED enters a near confocal optical cell through the rear of mirror A. The intensity of the light, as observed by a vacuum photodiode detector, which is also modulating at a slightly different frequency, located behind Mirror B, builds exponentially in the cell while the LED is ON. The opposite is true when the LED is OFF. Because both mirrors are highly reflective at 450 nm, a prominent absorption band for NO2, the light takes a considerable amount of time to plateau in the absence of the absorbing gas. 
#6. When NO2 is present, the mean path length traveled by the light is significantly reduced. This has two effects on the observed intensity as measured by the detector: a) The light plateau intensity level is lower, b) The light intensity plateaus sooner. Thus, an observed phase shift from the modulating LED is detected. The phase shift is largest when measuring zero air and decreases when NO2 is present.
#7. Both the LED and the Detector are modulated ON and OFF such that the observed signal has a much lower frequency, equal to the difference between the modulated frequencies and is referred to as a beat frequency. The system hardware and software take advantage of this, as it makes it easier to post process the signal using a micro controller. The technique is known as heterodyning.
#8. The instrument translates the phase shift from the presence of absorbing gas into a concentration measurement. Typical absorption techniques of other analysers take a reference and measure value of the light intensity “level” in order to derive concentration and compensate for source drift. Using the CAPS technique the amount of phase shift remains constant for a given concentration, even if the LED drifts over time. The measurement approach offers many advantages over traditional chemiluminescence analysers, such as faster response (single gas stream), lower noise at span and more importantly greater specificity.

#Known interferences: direct spectral interference with photochemically produced 1,2-dicarbonyl species (e.g., glyoxal, methylglyoxal)

'cavity attenuated phase shift spectroscopy':{'abbreviation':'CAPS', 'sampling_type':'low volume continuous', 'measured_parameters':['NO','NO2'], 'qa_accepted_parameters':['NO','NO2'], 
    'instruments':{
        'Environnement SA AS32M':{'documented_lower_limit_of_detection':'0.1',  'documented_upper_limit_of_detection':'1000.0', 'documented_precision':'0.05',    'documented_accuracy':'4.0%', 'documented_zero_drift':'0.75/day', 'documented_span_drift':'1.5/day',                                'instrument_manual_name':'EnvironnementSA_AS32M_Specs.pdf'},
        'Teledyne N500':         {'documented_lower_limit_of_detection':'0.1',  'documented_upper_limit_of_detection':'1000.0', 'documented_precision':'0.5%>=5', 'documented_accuracy':'1.0%', 'documented_zero_drift':'0.2/day',  'documented_span_drift':'0.5%/day', 'documented_flow_rate':'1.0', 'instrument_manual_name':'Teledyne_N500_Specs.pdf'},
        'Teledyne T500U':        {'documented_lower_limit_of_detection':'0.04', 'documented_upper_limit_of_detection':'1000.0', 'documented_precision':'0.5%>=5', 'documented_accuracy':'1.0%', 'documented_zero_drift':'0.1/day',  'documented_span_drift':'0.5%/day', 'documented_flow_rate':'0.9', 'instrument_manual_name':'Teledyne_T500U_Manual.pdf'}
    }
},

#------------------------------------
#differential optical absorption spectroscopy (DOAS)

#The basic principle used in Differential optical absorption spectroscopy (DOAS) is absorption spectroscopy. DOAS allows the quantitative determination of multiple atmospheric trace gas concentrations by recording and evaluating the characteristic absorption structures (lines or bands) of the trace gas molecules along an absorption path of known length in the open atmosphere, following the Beer-Lambert law (I = Io e−KLC; K = molecular absorption coefficient at STP, L = optical path length of cell, C = species concentration , I = light intensity of sample gas, Io = light intensity of sample without measured species (reference gas) ).
#The wavelength of light where a distinct absorption peak occurs is determined for analyte. A wavelength on either side of the absorption peak is next determined. The intensity of a light source at wavelength is measured and then the intensity is measured again after the light passes through the analyte. The difference of the intensities is proportional to the concentration to the analyte.
#DOAS is a long path measuring technique. Measurements can be made in an optical pathway from 1 to 10 kilometers.
#The method measures the average concentration of a species along the path length, not for any single molecule.
#In the real atmosphere, multiple effects contribute to the overall attenuation of the light. In particular, aerosols and clouds scatter light and thereby reduce the intensity of the direct beam while increasing intensities measured in other directions. Also, there rarely is only one absorber relevant at a given wavelength and this also needs to be accounted for. 
#The solution to this problem lies in the use of measurements at several wavelengths. Each molecule has a characteristic absorption spectrum (its spectral fingerprint) and therefore, simultaneous measurements at different wavelengths enable the separation of the contributions of the different absorbers. This is what DOAS does.
#Scattering by aerosols also needs accounting for. Their extinction cross-sections can be approximated by power laws (λ**-4 for Rayleigh scattering and λ**-1..0 for Mie scattering), the coefficients of which can be determined in the fit. The basic principle behind the separation of aerosol extinction and trace gas absorption is that the latter are identified using those parts of their absorption cross-sections that vary rapidly with wavelength. 
#The more slowly varying parts of the absorption can not be separated from extinction by aerosols which is why the absorption cross-sections are often high pass filtered before use in a DOAS retrieval.
#In addition to their effect on the spectral distribution of the intensity measured, aerosols can also have a significant impact on the light path of scattered light. This has to be modelled explicitly when computing the airmass factors for the light path correction.
#The DOAS technique is characterised by the following: 
#(a) measuring the transmitted light intensity over a relatively (compared to the width of an absorption band) broad spectral interval; 
#(b) high‐pass filtering of the spectra to obtain a differential absorption signal and eliminating broad‐band extinction processes such as Rayleigh and Mie scattering (RS and MS); 
#(c) quantitative determination of trace column densities by matching the observed spectral signatures to prerecorded (reference) spectra by, for instance, least‐squares methods.
#DOAS instruments are often divided into two main groups: passive and active ones. The active DOAS system such as longpath(LP)-systems and cavity-enhanced(CE) DOAS systems have their own light-source, whereas passive ones use the sun as their light source, e.g. MAX(Multi-axial)-DOAS. Also the moon can be used for night-time DOAS measurements, but here usually direct light measurements need to be done instead of scattered light measurements as it is the case for passive DOAS systems such as the MAX-DOAS.

#Typical measurement operation:
#1. The positioning of the emitter and the receiver define the monitoring path. The light source in the emitter is a high-pressure xenon lamp. This type of light source radiates an almost smooth spectrum ranging from approximately 200 nm up to 500 nm, and from about 1200 to 3000 nm. Within these ranges a number of gaseous substances show specific absorption spectra. The lamp spectrum is however not perfectly smooth, but the remaining "hills" in the spectral output are being taken care of in the evaluation.
#2. The emitted light beam is directed towards the receiver, and on its way the intensity is affected by scattering and absorption in molecules and particles.
#3. To ascertain the initial intensity (Io) of the light source across the path, there are typically two approaches: (a) Use of measurements at different light paths. If it is possible to take two measurements with the same initial intensity but at different light path lengths L, the dependence on the initial intensity cancels if one looks at the ratio of the two measurements. This approach is e.g. used in zenith-sky DOAS, where a measurement at low sun (long light path through the atmosphere) is analysed with a measurement taken at high sun (short light path). (b) Use of measurements at different wavelengths. In this approach, one exploits the fact that many molecules have structured spectra and light at different wavelengths experiences different absorption strengths. If the initial intensity does not vary with wavelength, one can look at the ratio of two measurements at different wavelengths to remove the dependence on Io. This approach is often used in long-path measurements which employ a lamp. It also is one reason to call this method differential as changes in absorption are used.
#4. From the receiver the captured light is led via an opto-fibre to the analyser. The function of the fibre is only to avoid exposing the opto-analyser to dust, high humidity, temperature variations, etc.
#5. When the light reaches the analyser, it enters a spectrometer. Inside the spectrometer, a grating refracts the light into its wavelength components. The refracted light is then projected onto a rapid scanning slit in front of a photo-multiplier detector or an infrared sensitive diode, where a selected part of the spectrum is detected. The scanning slit device makes it possible to record all wavelengths separately.
#6. As the grating is moveable, any chosen part of the spectrum can be detected. The wavelength window can thus be optimised for a certain component, with respect to parameters such as sensitivity and interfering pollutants. Approximately 100 scans per second are recorded.
#7. The current from the detector is converted into digital signals by a 12 bit analogue-to-digital converter, and the signal is stored and accumulated in a multi-channel register. The detected spectrum is typically 40 nm wide in the UV range and approximately twice as wide in the IR range. Each scan is digitised into 1000 points.
#8. Each pollutant is monitored during a time period entered by the operator. When the data accumulation is finished, the evaluation process is started. At the same time the next data accumulation period starts.

#Known interferences: Interference by miscellaneous atmospheric constituents. Heavy rain and fog, and even high humidity. Atmospheric turbulence, such as that from thermal-induced effects, can distort reflections. Anything that interrupts the path of the laser will cause some interference (i.e., animals, cars, planes, etc.). 
#Can also measure size-distributed PM: https://www.scirp.org/journal/paperinformation.aspx?paperid=44237

'differential optical absorption spectroscopy':{'abbreviation':'DOAS', 'sampling_type':'remote', 'measured_parameters':['O3','NO','NO2','NOx','CO','SO2','NH3','HNO3','PAN','HCl','HF','H2S','HCHO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','PM10','PM2.5','PM1'], 'qa_accepted_parameters':['O3','NO','NO2','NOx','CO','SO2','NH3','HNO3','PAN','HCl','HF','H2S','HCHO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','PM10','PM2.5','PM1'], 
    'instruments':{
        'Environnement SA SANOA':{'measured_parameters':['NO','NO2','NOx','O3','SO2','NH3'],     'qa_accepted_parameters':['NO','NO2','NOx','O3','SO2','NH3'],     'documented_lower_limit_of_detection':{'NO':'1.5','NO2':'0.6','NOx':'1.5','O3':'0.6','SO2':'0.2','NH3':'3.5'}, 'documented_accuracy':'1.0%',                                                                                                                         'instrument_manual_name':'EnvironnementSA_SANOA_Specs.pdf'},
        'miniDOAS':              {'measured_parameters':['NO','SO2','NH3'],                      'qa_accepted_parameters':['NO','SO2','NH3'],                      'documented_lower_limit_of_detection':{'NO':'0.641','SO2':'0.3','NH3':'0.282'},                                                                                                                                                                                      'instrument_further_details':'Details found online: https://amt.copernicus.org/articles/9/2721/2016/amt-9-2721-2016.pdf'},
        'OPSIS AR500':           {                                                                                                                                                                                                                                                                                                                                                                                                      'instrument_manual_name':'OPSIS_AR500_Manual.pdf'},
        'OPSIS AR550':           {                                                                                                                                                                                                                                                                                                                                                                                                      'instrument_manual_name':'OPSIS_AR500_Manual.pdf'},
        'OPSIS AR600':           {                                                                                                                                                                                                                                                                                                                                                                                                      'instrument_manual_name':'OPSIS_AR500_Manual.pdf'},
        'OPSIS AR650':           {                                                                                                                                                                                                                                                                                                                                                                                                      'instrument_manual_name':'OPSIS_AR500_Manual.pdf'},
        'Thermo DOAS 2000':      {'measured_parameters':['O3','NO2','SO2','HCHO','BENZENE','TOLUENE'], 'qa_accepted_parameters':['O3','NO2','SO2','HCHO','BENZENE','TOLUENE'],                                                                                                                                              'documented_precision':{'O3':'4.8','NO2':'4.1','SO2':'0.6','HCHO':'1.4','BENZENE':np.NaN,'TOLUENE':np.NaN}, 'instrument_further_details':'Details found online:https://doi.pangaea.de/10.1594/PANGAEA.884310'}
    }
},
#------------------------------------
#electrochemical membrane diffusion (EMD)

#Methodology developed specifically for measurement of NH3: https://www.sciencedirect.com/science/article/pii/S0003267003012650?via%3Dihub
#Gas is sampled in a sampler comprising two opposite channels separated by a gas permeable, water repellent polypropylene membrane. 
#Subsequently, the acid sample solution is pumped into a selector where an alkaline solution is added to ionize all sampled ambient acid gasses, resulting in an enhanced selectivity. 
#In the selector, the ammonia can diffuse through a second membrane into a purified water stream where an electrolyte conductivity sensor quantifies the resulting ammonium concentration. 
#The realized system is shown to be selective enough not to be influenced by normal ambient carbon dioxide concentrations. 
#Experiments with a gas flow of 3 ml/min, containing ammonia concentrations ranging from 9.8 to 0.3 ppm in a nitrogen carrier flow, into a 15 μl/min sample solution flow and finally into a 5 μl/min purified water stream have been carried out and show that the system is sensitive to ammonia concentration below 1 ppmv.

#Known interferences: CO2

#See details of method here:https://www.sciencedirect.com/science/article/pii/S0003267003012650?via%3Dihub

'electrochemical membrane diffusion': {'abbreviation':'EMD', 'sampling_type':'low volume continuous', 'measured_parameters':['NH3','HNO3'], 'qa_accepted_parameters':['NH3','HNO3'], 
    'instruments':{}
},

#------------------------------------
#photoacoustic spectroscopy (PS)

#Gaseous samples are continuously drawn through a measurement cell, where they are interrogated spectroscopically with the output radiation of a carbon dioxide/tunable quantum cascade laser. 
#This is used in conjunction with a highly sensitive acoustic detector to measure trace gas concentrations. Laser tuning is accomplished using proprietary algorithms that do not require the use of a spectrometer or other wavelength measurement device. 
#The laser is tuned to a known gaseous absorption line and passed through a cell containing the gas sample. If the trace gas species is present, the gas sample will be slightly heated. This heating can be measured very accurately and linearly by microphones in the cell and the amplitude of the electrical signal from the microphones correlate with the trace gas concentration. 
#If there is no trace gas present, there will be no signal from the microphone. This is, therefore, the highly desirable "zero-background measurement". 
#Traditional spectroscopic methods measure the total power of the laser beam with and without a sample in the path. For very weak attenuations that result from trace gas concentrations, this process requires taking the difference of two large numbers, each with a finite uncertainty, to compute a small real number. This complication is avoided using this method.
#This method has been used to measure NH3.

'photoacoustic spectroscopy':{'abbreviation':'PS', 'sampling_type':'low volume continuous', 'measured_parameters':['NH3','HNO3'], 'qa_accepted_parameters':['NH3','HNO3'], 
    'instruments':{
        'Pranalytica Nitrolux100': {'documented_lower_limit_of_detection':'0.1', 'documented_upper_limit_of_detection':'300.0',  'documented_precision':'0.1/10%', 'documented_accuracy':'0.1/10%', 'documented_zero_drift':'0.1/week', 'documented_span_drift':'10.0%/month', 'documented_flow_rate':'1.6', 'instrument_manual_name':'Pranalytica_Nitrolux_Specs.pdf,Pranalytica_Nitrolux100_Specs.pdf'},
        'Pranalytica Nitrolux200': {'documented_lower_limit_of_detection':'0.2', 'documented_upper_limit_of_detection':'500.0',  'documented_precision':'0.2/10%', 'documented_accuracy':'0.2/10%', 'documented_zero_drift':'0.2/week', 'documented_span_drift':'10.0%/month', 'documented_flow_rate':'1.6', 'instrument_manual_name':'Pranalytica_Nitrolux_Specs.pdf',},
        'Pranalytica Nitrolux1000':{'documented_lower_limit_of_detection':'1.0', 'documented_upper_limit_of_detection':'2000.0', 'documented_precision':'1.0/10%', 'documented_accuracy':'1.0/10%', 'documented_zero_drift':'1.0/week', 'documented_span_drift':'10.0%/month', 'documented_flow_rate':'1.6', 'instrument_manual_name':'Pranalytica_Nitrolux_Specs.pdf'}
    }
},

#------------------------------------
#non-dispersive infrared absorption (luft) (NDIR-L)

#Operates on the principle that a specific species efficiently absorbs IR light at a known wavelength in the IR range. This is the case for CO, at 4.7um
#The degree to which the IR light is absorbed by a specific species is directly related to the species concentration as described by the Beer-Lambert Law (I/Io = e−KLC; K = molecular absorption coefficient at STP, L = optical path length of cell, C = species concentration , I = light intensity of sample gas, Io = light intensity of sample without measured species (reference gas) ).

#Typical measurement operation for CO:
#1. The sample is drawn into the through the sample bulkhead
#2. IR radiation from a IR source (at 4.7 um) is passed through a rotating shutter creating a series of pulses.
#3. These pulses are split simultaneously into 2 measurement cells. One cell contains the sample gas, termed the sample stream (I), and cell contains a non-absorbing gas at the specific wavelength (i.e N2), termed the reference stream (Io).
#4. Both the pulses then exit the cells and fall on a IR photo-detector. 
#5. The intensities of 4.7 um light transmitted through the sample (I) and CO-free reference beams (Io) are related to the concentration (C) of CO in the sample gas stream according to the Beer-Lambert Law.
#6. The analyser’s microprocessor calculates the absolute concentration in molecules cm-3, and is then converted to a mixing ratio using in-instrument measured sample temperature and pressure.

#Known Interferences:
#Water vapour and CO2

'non-dispersive infrared absorption (luft)':{'abbreviation':'NDIR-L', 'sampling_type':'low volume continuous', 'measured_parameters':['CO','CH4'], 'qa_accepted_parameters':['CO','CH4'], 
    'instruments':{
        'Beckman 866':      {'measured_parameters':['CO'],                                    'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (luft) here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA248&lpg=PA248&dq=HORIBA+AQM-10&source=bl&ots=-uIdMsd0Yg&sig=U34lJHqy1fXno_AwV36rZIKCwuE&hl=en&sa=X&ved=0ahUKEwjvlL2IxuLbAhUJthQKHZCQB90Q6AEIQjAC#v=onepage&q=HORIBA%20AQM-10&f=false'},
        'Bendix 8501-5CA':  {'measured_parameters':['CO'],                                    'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (luft) here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA248&lpg=PA248&dq=HORIBA+AQM-10&source=bl&ots=-uIdMsd0Yg&sig=U34lJHqy1fXno_AwV36rZIKCwuE&hl=en&sa=X&ved=0ahUKEwjvlL2IxuLbAhUJthQKHZCQB90Q6AEIQjAC#v=onepage&q=HORIBA%20AQM-10&f=false'},
        'Fuji ZRC':         {'measured_parameters':['CO'],                                    'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption here (assume luft):https://library.wmo.int/doc_num.php?explnum_id=10622', 'measuring_assumption':True},
        'Horiba 300E':      {'measured_parameters':['CO'],                                    'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (luft) here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA248&lpg=PA248&dq=HORIBA+AQM-10&source=bl&ots=-uIdMsd0Yg&sig=U34lJHqy1fXno_AwV36rZIKCwuE&hl=en&sa=X&ved=0ahUKEwjvlL2IxuLbAhUJthQKHZCQB90Q6AEIQjAC#v=onepage&q=HORIBA%20AQM-10&f=false'},
        'Horiba 300SE':     {'measured_parameters':['CO'],                                    'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (luft) here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA248&lpg=PA248&dq=HORIBA+AQM-10&source=bl&ots=-uIdMsd0Yg&sig=U34lJHqy1fXno_AwV36rZIKCwuE&hl=en&sa=X&ved=0ahUKEwjvlL2IxuLbAhUJthQKHZCQB90Q6AEIQjAC#v=onepage&q=HORIBA%20AQM-10&f=false'},
        'Horiba AQM10':     {'measured_parameters':['CO'],                                    'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (luft) here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA248&lpg=PA248&dq=HORIBA+AQM-10&source=bl&ots=-uIdMsd0Yg&sig=U34lJHqy1fXno_AwV36rZIKCwuE&hl=en&sa=X&ved=0ahUKEwjvlL2IxuLbAhUJthQKHZCQB90Q6AEIQjAC#v=onepage&q=HORIBA%20AQM-10&f=false'},
        'Horiba AQM11':     {'measured_parameters':['CO'],                                    'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (luft) here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA248&lpg=PA248&dq=HORIBA+AQM-10&source=bl&ots=-uIdMsd0Yg&sig=U34lJHqy1fXno_AwV36rZIKCwuE&hl=en&sa=X&ved=0ahUKEwjvlL2IxuLbAhUJthQKHZCQB90Q6AEIQjAC#v=onepage&q=HORIBA%20AQM-10&f=false'},
        'Horiba AQM12':     {'measured_parameters':['CO'],                                    'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (luft) here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA248&lpg=PA248&dq=HORIBA+AQM-10&source=bl&ots=-uIdMsd0Yg&sig=U34lJHqy1fXno_AwV36rZIKCwuE&hl=en&sa=X&ved=0ahUKEwjvlL2IxuLbAhUJthQKHZCQB90Q6AEIQjAC#v=onepage&q=HORIBA%20AQM-10&f=false'},
        'Massachusetts CO1':{'measured_parameters':['CO'],                                    'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'cannot find any information. AQS states it as non-dispersive infrared absorption. Assume this is correct, and assume luft type', 'measuring_assumption':True},
        'MSA 202S':         {'measured_parameters':['CO'],                                    'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (luft) here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA248&lpg=PA248&dq=HORIBA+AQM-10&source=bl&ots=-uIdMsd0Yg&sig=U34lJHqy1fXno_AwV36rZIKCwuE&hl=en&sa=X&ved=0ahUKEwjvlL2IxuLbAhUJthQKHZCQB90Q6AEIQjAC#v=onepage&q=HORIBA%20AQM-10&f=false'}, 
        'Signal 418':       {'measured_parameters':['CO','CH4','CO2','NH3','NO','SO2','H2O'], 'qa_accepted_parameters':['CO','CH4','CO2','NH3','NO','SO2','H2O'], 'documented_lower_limit_of_detection':{'CO':'1000.0','CH4':'2000.0','CO2':'300.0','NH3':'1000.0','NO':'5000.0','SO2':'2000.0','H2O':'5000.0'}, 'documented_precision':'0.1%', 'documented_accuracy':'0.5%', 'documented_zero_drift':'0.5%/day', 'documented_span_drift':'0.5%/day', 'documented_flow_rate':'0.2-2.0', 'instrument_manual_name':'Signal_418_Specs.pdf'},
        'Teledyne ML8310':  {'measured_parameters':['CO'],                                    'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (luft) here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA248&lpg=PA248&dq=HORIBA+AQM-10&source=bl&ots=-uIdMsd0Yg&sig=U34lJHqy1fXno_AwV36rZIKCwuE&hl=en&sa=X&ved=0ahUKEwjvlL2IxuLbAhUJthQKHZCQB90Q6AEIQjAC#v=onepage&q=HORIBA%20AQM-10&f=false'},
        'Teledyne ML8830':  {'measured_parameters':['CO'],                                    'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption here (assume luft):https://books.google.es/books?id=7eL1Rd8hB-0C&pg=PA418&lpg=PA418&dq=monitor+labs+8830&source=bl&ots=GlI29t_UB9&sig=uqlixhqiWdhaQf8BvZ5GQmu_1gw&hl=en&sa=X&ved=0ahUKEwi2rNOCweLbAhVDrRQKHVf7AsoQ6AEIVDAF#v=onepage&q=monitor%20labs%208830&f=false', 'measuring_assumption':True},
        'Teledyne ML8831':  {'measured_parameters':['CO'],                                    'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption here (assume luft):http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/ML8831/view', 'measuring_assumption':True}
    }
},
#------------------------------------
#non-dispersive infrared absorption (gas-filter correlation) (NDIR-GFC)

#Operates on the principle that a specific species efficiently absorbs IR light at a known wavelength in the IR range. This is the case for CO, at 4.7um
#The degree to which the IR light is absorbed by a specific species is directly related to the species concentration as described by the Beer-Lambert Law (I/Io = e−KLC; K = molecular absorption coefficient at STP, L = optical path length of cell, C = species concentration , I = light intensity of sample gas, Io = light intensity of sample without measured species (reference gas) ).
#The methodology also applies the gas-filter correlation technique. This is applied as water vapour and CO2 absorb light at 4.7 um as well as CO, and ensures measurements are specific for CO.

#Typical measurement operation for CO:
#1. The sample is drawn into the through the sample bulkhead
#2. IR radiation from a IR source (at 4.7 um) is passed through a multi-pass cell length with sample gas. The sample cell uses mirrors at each end to reflect the IR beam back and forth through the sample gas a number of times. The total length that the reflected light travels is directly related to the intended sensitivity of the instrument. The lower the concentrations the instrument is designed to detect, the longer the light path must be in order to create detectable levels of attenuation.
#A gas-filter correlation wheel is combined with this system. This wheel contains three parts to increase measurement accuracy: CO, N2 and the mask:
#    The CO window contains saturation (40 %) of CO which acts as a reference beam – absorbing a known amount of light.
#    The N2 window, containing 100 % N2, does not absorb IR at 4.7 um at all and is used during normal CO measurement.
#    The mask totally blocks the light source and is used to determine background signals, the strength of other signals relative to each other and the background.
#3. The IR source is passed through the rotating filter wheel before entering the multi-pass cell: 
#    When the light beam is intercepted by the CO portion of the wheel, the carbon monoxide, which is at relatively high concentration, absorbs all wavelengths that are co-specific, creating and emanating light beam that is “CO blind”. This “optically scrubbed” portion of the beam is designated the Reference beam (Io).
#    The nitrogen-intercepted portion of the beam, which is “CO sensitive”, is designated the sample beam (I). 
#    In the absence of CO no attenuation of the Io and I portion of the beam will occur.
#    Species other than CO will cause an equal attenuation of both I and Io portions of the beam. 
#    If CO is present in the air being sampled, then the beam portion generated by the CO side of the wheel will experience no attenuation, but the beam portion generated by the N2 portion of the wheel will be attenuated to the degree dictated by the level of CO concentration.
#    When the light beam is intercepted by the mask, this is labelled the "dark portion". This provides a zero light reference point to compensate for the "dark current" of the detector.
#4. A band pass filter is fitted on the exit of the sample cell, allowing only 4.7 um wavelength light to pass.
#5. The infrared radiation then exits the optical cell and falls on an IR photo-detector. 
#6. The intensities of 4.7 um light transmitted through the sample (I) and CO-free reference beams (Io) are related to the concentration (C) of CO in the sample gas stream according to the Beer-Lambert Law.
#7. The analyser's microprocessor records the imbalance between the I and Io beams portions and performs a data linearisation, calculating the absolute concentration in molecules cm-3, and is then converted to a mixing ratio using in-instrument measured sample temperature and pressure.

#Known Interferences:
#Water vapour and CO2

'non-dispersive infrared absorption (gas-filter correlation)':{'abbreviation':'NDIR-GFC', 'sampling_type':'low volume continuous', 'measured_parameters':['CO','CH4'], 'qa_accepted_parameters':['CO','CH4'],
    'instruments':{
        'Airpointer-CO':         {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'40.0',  'documented_upper_limit_of_detection':'1000000.0',  'documented_precision':'100.0',     'documented_accuracy':'1.0%<1000000.0',                 'documented_zero_drift':'100.0/day',  'documented_span_drift':'1.0%>=10000.0/day', 'documented_flow_rate':'0.5',     'instrument_manual_name':'Airpointer_Specs.pdf'},
        'Dasibi 3003':           {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (gas-filter correlation) here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA248&lpg=PA248&dq=HORIBA+AQM-10&source=bl&ots=-uIdMsd0Yg&sig=U34lJHqy1fXno_AwV36rZIKCwuE&hl=en&sa=X&ved=0ahUKEwjvlL2IxuLbAhUJthQKHZCQB90Q6AEIQjAC#v=onepage&q=HORIBA%20AQM-10&f=false'},
        'Dasibi 3008':           {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'100.0', 'documented_upper_limit_of_detection':'1000000.0',  'documented_precision':'100.0',     'documented_accuracy':'1.0%',                           'documented_zero_drift':'200.0/day',  'documented_span_drift':'1.0%/day',          'documented_flow_rate':'1.0',     'instrument_manual_name':'Dasibi_3008_Manual.pdf'},
        'DKK-TOA GFC-311E':      {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'400.0', 'documented_upper_limit_of_detection':'100000.0',                                                                                               'documented_zero_drift':'100.0/day',  'documented_span_drift':'2.0%/day',                                            'instrument_manual_name':'DKK-TOA_GFC-311E_Specs.pdf'},
        'Ecotech EC9830T':       {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'20.0',  'documented_upper_limit_of_detection':'20000.0',    'documented_precision':'20/0.1%',   'documented_accuracy':'1.0%',                           'documented_zero_drift':'50.0/day',   'documented_span_drift':'0.5%/day',          'documented_flow_rate':'1.0',     'instrument_manual_name':'Ecotech_EC9830T_Specs.pdf'},
        'Ecotech Serinus30':     {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'40.0',  'documented_upper_limit_of_detection':'200000.0',   'documented_precision':'20/0.1%',   'documented_accuracy':'1.0%',                           'documented_zero_drift':'100.0/day',  'documented_span_drift':'0.5%/day',          'documented_flow_rate':'1.0',     'instrument_manual_name':'Ecotech_Serinus30_Manual.pdf'},
        'Environnement SA CO10M':{'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':':No manual or specifications available but confirmation it is uses non-dispersive infrared absorption here (assume gas-filter correlation, as latest version of instrument uses it):http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/enviroCO10M/view', 'measuring_assumption':True},
        'Environnement SA CO11M':{'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':':No manual or specifications available but confirmation it is uses non-dispersive infrared absorption here (assume gas-filter correlation, as next version of instrument uses it):http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf', 'measuring_assumption':True},      
        'Environnement SA CO12e':{'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'50.0',  'documented_upper_limit_of_detection':'300000.0',                                       'documented_accuracy':'1.0%',                           'documented_zero_drift':'200.0/week', 'documented_span_drift':'0.5%/week',         'documented_flow_rate':'1.0',     'instrument_manual_name':'EnvironnementSA_CO12e_Specs.pdf'},   
        'Environnement SA CO12M':{'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'40.0',  'documented_upper_limit_of_detection':'200000.0',                                       'documented_accuracy':'1.0%',                           'documented_zero_drift':'500.0/day',  'documented_span_drift':'0.5%/day',          'documented_flow_rate':'1.0',     'instrument_manual_name':'EnvironnementSA_CO12M_Specs.pdf'},
        'Maihak Unor 6N':        {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption here (assume gas-filter correlation):https://books.google.es/books?id=ZDb3koqwApMC&pg=PA468&lpg=PA468&dq=maihak+unor+6n&source=bl&ots=EP38JON-AR&sig=H6eppHv6VGHfbrPaqjOBREST0mE&hl=en&sa=X&ved=2ahUKEwiLmbzS1O_fAhXiShUIHQBqArsQ6AEwC3oECAEQAQ#v=onepage&q=maihak%20unor%206n&f=false', 'measuring_assumption':True}, 
        'Maihak Unor 610':       {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption here (assume gas-filter correlation):https://books.google.es/books?id=0YV9hxFUP1QC&pg=PA229&lpg=PA229&dq=MAIHAK+UNOR+610&source=bl&ots=KEbg6PRHgo&sig=ACfU3U0WxRqWX8UDk8K5HrTnSXQI5QxkSA&hl=en&sa=X&ved=2ahUKEwiMg6LQh7XwAhWrRRUIHVYbBp4Q6AEwEXoECA8QAw#v=onepage&q=MAIHAK%20UNOR%20610&f=false', 'measuring_assumption':True}, 
        'Philips K50093':        {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption here (assume gas-filter correlation):https://dergipark.org.tr/tr/download/article-file/1329767', 'measuring_assumption':True}, 
        'Philips K50109':        {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (gas-filter correlation):https://www.isiaq.org/docs/papers/472.pdf'}, 
        'SIR S-5006':            {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'40.0',  'documented_upper_limit_of_detection':'200000.0',   'documented_precision':'100.0',     'documented_accuracy':'1.0%',                           'documented_zero_drift':'10.0/day',   'documented_span_drift':'0.5%/day',          'documented_flow_rate':'1.0',     'instrument_manual_name':'SIR_S-5006_Specs.pdf'},
        'Teledyne 300':          {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption here (assume gas-filter correlation, as other instruments with very similar codes use it):http://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf', 'measuring_assumption':True}, 
        'Teledyne 300E':         {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'40.0',  'documented_upper_limit_of_detection':'1000000.0',  'documented_precision':'0.5%/200',  'documented_accuracy':'1.0%',                           'documented_zero_drift':'100.0/day',  'documented_span_drift':'100.0/0.5%/day',    'documented_flow_rate':'0.8',     'instrument_manual_name':'Teledyne_300E_Manual.pdf'},
        'Teledyne 300EM':        {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'200.0', 'documented_upper_limit_of_detection':'5000000.0',  'documented_precision':'1.0%/1000', 'documented_accuracy':'1.0%',                           'documented_zero_drift':'500.0/day',  'documented_span_drift':'500.0',             'documented_flow_rate':'0.8',     'instrument_manual_name':'Teledyne_300E_Manual.pdf'},
        'Teledyne 300EU':        {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'20.0',  'documented_upper_limit_of_detection':'100000.0',   'documented_precision':'0.5%',      'documented_accuracy':'1.0%',                           'documented_zero_drift':'20.0/day',   'documented_span_drift':'0.5%/day',          'documented_flow_rate':'1.8',     'instrument_manual_name':'Teledyne_300EU_Addendum.pdf'},
        'Teledyne ML8930':       {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption here (assume gas-filter correlation): http://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/ML8930/view?facet=HTML+Representation', 'measuring_assumption':True},        
        'Teledyne ML9830':       {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (gas-filter correlation) here:https://twobtech.com/pdf/211%20FEM%20Modification_June2018.pdf'},    
        'Teledyne ML9830B':      {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (gas-filter correlation) here:https://twobtech.com/pdf/211%20FEM%20Modification_June2018.pdf'},    
        'Teledyne T300':         {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'40.0',  'documented_upper_limit_of_detection':'1000000.0',  'documented_precision':'0.5%/200',  'documented_accuracy':'1.0%',                           'documented_zero_drift':'100.0/day',  'documented_span_drift':'0.5%/day',          'documented_flow_rate':'0.8',     'instrument_manual_name':'Teledyne_T300_Manual.pdf'},
        'Teledyne T300M':        {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'200.0', 'documented_upper_limit_of_detection':'1000000.0',  'documented_precision':'1.0%/1000', 'documented_accuracy':'1.0%',                           'documented_zero_drift':'500.0/day',  'documented_span_drift':'500.0/day',         'documented_flow_rate':'0.8',     'instrument_manual_name':'Teledyne_T300_Manual.pdf'},
        'Teledyne T300U':        {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'20.0',  'documented_upper_limit_of_detection':'100000.0',   'documented_precision':'0.5%',      'documented_accuracy':'1.0%',                           'documented_zero_drift':'20.0/day',   'documented_span_drift':'0.5%>=500.0/day',   'documented_flow_rate':'1.8',     'instrument_manual_name':'Teledyne_T300U_Addendum.pdf'},
        'Thermo 48':             {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (gas-filter correlation) here:https://books.google.es/books?id=yCDGFlP3toUC&pg=PA248&lpg=PA248&dq=HORIBA+AQM-10&source=bl&ots=-uIdMsd0Yg&sig=U34lJHqy1fXno_AwV36rZIKCwuE&hl=en&sa=X&ved=0ahUKEwjvlL2IxuLbAhUJthQKHZCQB90Q6AEIQjAC#v=onepage&q=HORIBA%20AQM-10&f=false'},
        'Thermo 48C':            {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'40.0',  'documented_upper_limit_of_detection':'10000000.0', 'documented_precision':'100.0',     'documented_accuracy':'1.0%',                           'documented_zero_drift':'100.0/day',  'documented_span_drift':'1.0%/day',          'documented_flow_rate':'0.5-2.0', 'instrument_manual_name':'Thermo_48C_Specs.pdf'},
        'Thermo 48C-TL':         {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (gas-filter correlation) here:http://www.broward.org/Environment/AirQuality/AirMonitoring/Pages/AmbientAirMonitoringInstruments.aspx'},
        'Thermo 48i':            {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'40.0',  'documented_upper_limit_of_detection':'10000000.0',                                     'documented_accuracy':'1.0%<1000000.0 2.5%>=1000000.0', 'documented_zero_drift':'100.0/day',  'documented_span_drift':'1.0%/day',          'documented_flow_rate':'1.0',     'instrument_manual_name':'Thermo_48i_Manual.pdf'},
        'Thermo 48i-TLE':        {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'40.0',  'documented_upper_limit_of_detection':'1000000.0',                                      'documented_accuracy':'1.0%/40.0',                      'documented_zero_drift':'100.0/day',  'documented_span_drift':'1.0%/day',          'documented_flow_rate':'0.5',     'instrument_manual_name':'Thermo_48i-TLE_Manual.pdf'},
        'Thermo 48s':            {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (gas-filter correlation) here:https://csl.noaa.gov/groups/csl7/measurements/2000TexAQS/LaPorte/datainfo.html'},
        'Thermo 48W':            {'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (gas-filter correlation) here:https://www.rivm.nl/bibliotheek/rapporten/725601007.pdf'}
    }
},

#------------------------------------
#non-dispersive infrared absorption (cross-flow modulation) (NDIR-CFM)

#Operates on the principle that a specific species efficiently absorbs IR light at a known wavelength in the IR range. This is the case for CO, at 4.7um.
#The degree to which the IR light is absorbed by a specific species is directly related to the species concentration as described by the Beer-Lambert Law (I/Io = e−KLC; K = molecular absorption coefficient at STP, L = optical path length of cell, C = species concentration , I = light intensity of sample gas, Io = light intensity of sample without measured species (reference gas) ).
#The methodology also applies the cross-flow modulation technique. This is applied as water vapour and CO2 absorb light at 4.7 um as well as CO, and ensures measurements are specific for CO.

#Typical measurement operation for CO:
#1. The sample is drawn into the through the sample bulkhead
#2. IR radiation from a IR source (at 4.7 um) is passed through a measurement cell.
#3. Fixed amounts of the sample gas (I) and the reference gas (Io) are injected alternately into the measurement cell (cross-flow modulation). This has the advantage over the gas-filter correlation technique that, in principle, when analyzing small amounts of gas there is no generation of zero-drift. An additional advantage is that the elimination of rotary sectors precludes the need for optical adjustment. These features assure greatly improved stability over long periods of measurement. The reference gas is generated by purging the sample through an oxidation process, where an oxidizing catalyst burns the CO to CO2. These features eliminate interference from other elements, resulting in highly accurate measurements.
#4. The infrared radiation then exits the optical cell and falls on an IR photo-detector. 
#5. The intensities of 4.7 um light transmitted through the sample (I) and CO-free reference beams (Io) are related to the concentration (C) of CO in the sample gas stream according to the Beer-Lambert Law.
#6. The analyzer’s microprocessor calculates the absolute concentration in molecules cm-3, and is then converted to a mixing ratio using in-instrument measured sample temperature and pressure.

#Known Interferences:
#Water vapour and CO2

'non-dispersive infrared absorption (cross-flow modulation)':{'abbreviation':'NDIR-CFM', 'sampling_type':'low volume continuous', 'measured_parameters':['CO','CH4'], 'qa_accepted_parameters':['CO','CH4'], 
    'instruments':{
        'Horiba APMA300':{'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption here (assume cross-flow modulation): https://books.google.es/books?id=qmWXBwAAQBAJ&pg=PA88&lpg=PA88&dq=horiba+apna+300&source=bl&ots=dBQJ0TEai1&sig=plLm6SAUIZwuIhy37JzI8nZXccQ&hl=en&sa=X&ved=2ahUKEwihxOS9uvDfAhWGnhQKHYzoAPcQ6AEwCHoECAIQAQ#v=onepage&q=horiba%20apna%20300&f=false', 'measuring_assumption':True},
        'Horiba APMA350':{'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'],                                                                                                                                                                                                                                                                    'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption (cross-flow modulation) here: http://www.wseas.us/e-library/conferences/2009/timisoara/SSE2/SSE2-11.pdf'},
        'Horiba APMA360':{'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'20.0', 'documented_upper_limit_of_detection':'100000.0', 'documented_precision':'1.0%', 'documented_accuracy':'1.0%', 'documented_zero_drift':'20.0/day', 'documented_span_drift':'20.0/day', 'documented_flow_rate':'1.5', 'instrument_manual_name':'Horiba_APMA360_Specs.pdf'}, 
        'Horiba APMA370':{'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 'documented_lower_limit_of_detection':'20.0', 'documented_upper_limit_of_detection':'100000.0', 'documented_precision':'1.0%', 'documented_accuracy':'1.0%', 'documented_zero_drift':'20.0/day', 'documented_span_drift':'20.0/day', 'documented_flow_rate':'1.5', 'instrument_manual_name':'Horiba_APMA370_Manual.pdf'},
        'Horiba GA-360S':{                                                                                                                                                                                                                                                                                                                                  'instrument_further_details':'No manual or specifications available but confirmation it is uses non-dispersive infrared absorption here (assume cross-flow modulation): https://acp.copernicus.org/articles/8/3867/2008/acp-8-3867-2008.pdf', 'measuring_assumption':True}

    }
},

#------------------------------------
#dual isotope fluorescence (DIF)  

#This method is an adaption of non-dispersive infrared absorption for the measurement of CO.
#A broadband of infrared source is used to stimulate fluorescence of C16O and C18O, contained in a sealed cell.
#The stimulated radiation is passed first through a filter that alternately passes radiation from only one of the isotopic sources and is then passed through the sample cell to a solid state detector.
#Because almost all atmospheric CO consists of the C16O isotope, only radiation produced by the C16O source will be absorbed by the CO in the sample.
#The radiation produced by the C18O source is used as a reference beam to account for interference with H2O, or other interferants.

#Known Interferences:
#Water vapour and CO2

#See further details: https://books.google.es/books?id=V0oEzMp7axAC&pg=PA385&lpg=PA385&dq=dual+isotope+fluorescence+carbon+monoxide&source=bl&ots=azm5b16AqC&sig=pDZOFxw9LqZxQMsW8wOSxKv1YSE&hl=en&sa=X&ved=0ahUKEwiR7o2iyOLbAhVEtRQKHSFECGEQ6AEINjAB#v=onepage&q=dual%20isotope%20fluorescence%20carbon%20monoxide&f=false

'dual isotope fluorescence':{'abbreviation':'DIF', 'sampling_type':'low volume continuous', 'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 
    'instruments':{}
},

#------------------------------------
#fourier transform infrared spectroscopy (FTIR)

#Fourier Transform Infrared Spectroscopy (FTIR) is an analytical technique used to identify organic (and in some cases inorganic) materials. 
#This technique measures the absorption of infrared radiation by the sample material versus wavelength. 
#The infrared absorption bands identify molecular components and structures.
#When a material is irradiated with infrared radiation, absorbed IR radiation usually excites molecules into a higher vibrational state. 
#The wavelength of light absorbed by a particular molecule is a function of the energy difference between the at-rest and excited vibrational states. 
#The wavelengths that are absorbed by the sample are characteristic of its molecular structure.
#The FTIR spectrometer uses an interferometer to modulate the wavelength from a broadband infrared source. 
#A detector measures the intensity of transmitted or reflected light as a function of its wavelength. 
#The signal obtained from the detector is an interferogram, which must be analyzed with a computer using Fourier transforms to obtain a single-beam infrared spectrum. 
#The FTIR spectra are usually presented as plots of intensity versus wavenumber (in cm-1). Wavenumber is the reciprocal of the wavelength. 
#The intensity can be plotted as the percentage of light transmittance or absorbance at each wavenumber.

'fourier transform infrared spectroscopy':{'abbreviation':'FTIR', 'sampling_type':'low volume continuous', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 
    'instruments':{
        'Ecotech Spectronus':{'measured_parameters':['CO2','CH4','N2O','CO'], 'qa_accepted_parameters':['CO2','CH4','N2O','CO'], 'documented_precision':{'CO2':'40.0','CH4':'0.2','N2O':'0.06','CO':'0.2'}, 'documented_accuracy':'0.1%', 'documented_flow_rate':'1.0-1.5', 'instrument_manual_name':'Ecotech_Spectronus_Manual.pdf'}
    }
},

#------------------------------------
#gas chromatography (GC)

#Gas chromatography (GC) is a method used for separating and analysing compounds that can be vaporized without decomposition.
#A sample solution is injected into a instrument, entering a gas stream which transports the sample (mobile phase) into a separation tube known as the "column".
#The mobile phase is a carrier gas, usually an inert gas such as helium or an unreactive gas such as nitrogen. Helium remains the most commonly used carrier gas in about 90% of instruments although hydrogen is preferred for improved separations.
#The column consists of a microscopic layer of liquid or polymer on an inert solid support a microscopic layer of liquid or polymer on an inert solid support (stationary phase), inside a piece of glass or metal tubing, placed inside a piece of glass or metal tubing.
#Once inside the column, the gaseous compounds being analysed interact with the walls of the column coated with a stationary phase. This causes each compound to elute at a different time, known as the retention time of the compound. 
#The comparison of retention times is what gives GC its analytical usefulness.
#If greater separation of compounds is required, multiple distinct columns can be used for this purpose.

#Almost all methods involving GC are typically paired with a detector of some description to measure the GC separated components.
#These detectors are various in number, and are outlined in detail below.
#Typically GC-detector analysis is used for the separation of short lived hydrocarbon species. 

#Typically sample processing is undertaken before the passing of samples into the GC. 
#Typically this emcompasses the preconcentration, or passing samples through sorbent traps.

#If no other information is given, it is assumed all GC-detector analysis is done via injection and do not operate online (i.e. not continuously).

#------------------------------------
#gas chromatography - electron capture detection (GC-ECD)

#The Electron Capture Detector (ECD) is a selective detector for electro-negative compounds, especially halogens.
#A beta-ray radio active source which can ionize the carrier gas is located in the detector. 
#A current is produced between two electrodes in the detector supplied with a potential difference and this is monitored as a continuous background current. 
#When there are electro-negative components present in the carrier gas, the background current is reduced because these components capture electrons.
#This change in the background current is registered as 'holes' in the background. 
#Since it concerns very small holes, the background current should not be too high. 
#Therefore the (linear) dynamic range of the ECD is often limited.

'gas chromatography - electron capture detection': {'abbreviation':'GC-ECD', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','PAN','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','PAN','HCHO'], 
    'instruments':{}
},

#------------------------------------
#gas chromatography - flame ionisation detection (GC-FID)

#Flame ionisation detection operates in the exact same way as it does as a standalone method, only in this instance coupled to a GC.
#It is based on the principle of the generation of an electrical current that is proportional to the rate of ion formation, dependent on the concentrations of species in the sample gas.

#A limitation of GC-FID is that the retention time is merely characteristic of a compound and certainly not adjudicative or confirmatory of the specificity of that compound.
#This means a peak at a given retention time is not a unique qualitative measure (to the exclusion of every other compound in the universe).

'gas chromatography - flame ionisation detection':{'abbreviation':'GC-FID', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 
    'instruments':{
        'Nira Venus 301':{'sampling_type':'continuous injection', 'measured_parameters':['CH4','NMVOC','VOC','NMHC','HC'], 'qa_accepted_parameters':['CH4','NMVOC','VOC','NMHC','HC'], 'documented_lower_limit_of_detection':'200.0', 'documented_upper_limit_of_detection':'1000000.0', 'documented_precision':'1.0%', 'documented_accuracy':'1.0%', 'documented_span_drift':'1.0%/day', 'documented_flow_rate':'0.35', 'instrument_manual_name':'NiraVenus301_Specs.pdf'}
    }
},

#------------------------------------
#gas chromatography - photoionisation detection (GC-PID)

#Gas Chromatography - Photoionisation Detection (GC-PID) is a technique used to analyse a wide range of aromatic hydrocarbons and other organic compounds. 
#A typical application is the analysis of hydrocarbon pollution of water. The PID uses ultraviolet light to ionize the components exiting the column. 
#The ions are collected by electrodes and the current generated measures the concentration. 
#As common with other GC techniques, a carrier gas is required with low Water and Oxygen impurities since Water and Oxygen can interact with the stationary phase and cause significant problems such as high baseline noise and column bleed in the output gas chromatogram which both reduces the analyser sensitivity and decreases column lifetime.  
#Helium and Nitrogen are normally used. 
#A make-up gas is used in GC-PID at the exit of the column to increase the flow rate into the detector. 
#The routine calibration of the analyser using a calibration mixture is common.

'gas chromatography - photoionisation detection':{'abbreviation':'GC-PID', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 
    'instruments':{}
},

#------------------------------------
#gas chromatography - mercuric oxide reduction detection (GC-HgO) 

#Mercuric oxide (HgO) reduction is a detection method available for hydrogen, carbon monoxide, and other reducing gases.
#It uses mercuric oxide reaction tube and a mercury lamp in a heated UV detector cell. 
#The reaction tube is heated to 260-300oC. 
#Located immediately downstream of the reaction tube, the UV detector cell is heated to 170degC. 
#The UV detector cell is equipped with a mercury lamp and a UV photodiode. When a reducing gas such as carbon monoxide elutes from the GC column, it reacts with the HgO to form gaseous mercury vapor, which is then swept into the UV cell. 
#The gaseous mercury absorbs the UV light from the mercury lamp as it flows through the cell. The change in transmittance is converted by the data system into an absorbance output, which is linearly proportional to the amount of reducing gas.
#More information here: https://srigc.com/cn/downloads/41/RGD.pdf

'gas chromatography - mercuric oxide reduction detection':{'abbreviation':'GC-HgO', 'sampling_type':'injection', 'measured_parameters':['CO'], 'qa_accepted_parameters':['CO'], 
    'instruments':{}
},

#------------------------------------
#gas chromatography - dual flame ionisation detection (GC-DFID)

#To limit the main limitation of GC-FID, i.e. lack of specificity, it is possible to add another GC column and analyse the sample concurrently on both columns.
#This can be done by adding a y-splitter that will take the single injection made into the injector port and divide the sample into two different pathways with one part of the sample going to one column for analysis and the second part going to another column for analysis.
#The two different columns are given two different stationary phases, which provides different separation of the analytes (both in terms of retention time and possibly even eluting order). 
#This different eluting order and different retention times only minimises, but does not entirely eliminate the possibility of co-elution as again, the resolving (separating) power of the method is only determined by one non-unique measure which is the two retention times. 
#Even though there is a change in the eluting order potentially and the retention times are different based upon the stationary phase composition, again, remains the basic limitation of GC-FID remains in that a retention time is merely characteristic of the analyte, but is certainly not confirmatory.
#With dual detectors, the concentration of the analyte(s) of interest is determined from two calibration curves, and can serve as quantitative confirmation.

'gas chromatography - dual flame ionisation detection':{'abbreviation':'GC-DFID', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 
    'instruments':{}
},

#------------------------------------
#gas chromatography - fourier transform infrared spectroscopy (GC-FTIR)

#gas chromatography separation followed by fourier transform infrared spectroscopy

'gas chromatography - fourier transform infrared spectroscopy':{'abbreviation':'GC-FTIR', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 
    'instruments':{}
},

#------------------------------------
#gas chromatography - mass spectrometry (GC-MS)

#One of the most common detector types is a mass spectrometer.
#Mass spectrometry is the study of matter through the formation of gas-phase ions (with or without fragmentation) that are detected and characterised by their mass to charge ratios (m/z) and relative abundances. 
#The technique basically studies the effect of ionizing energy on molecules. It depends upon chemical reactions in the gas phase in which sample molecules are consumed during the formation of ionic and neutral species.
#A mass spectrometer generates multiple ions from the sample under investigation, it then separates them according to their specific mass-to-charge ratio (m/z), and then records the relative abundance of each ion type.
#The first step in the mass spectrometric analysis of compounds is the production of gas phase ions of the compound, basically by electron ionization. 
#This molecular ion undergoes fragmentation. Each primary product ion derived from the molecular ion, in turn, undergoes fragmentation, and so on. 
#The ions are separated in the mass spectrometer according to their mass-to-charge ratio, and are detected in proportion to their abundance. 
#A mass spectrum of the molecule is thus produced. It displays the result in the form of a plot of ion abundance versus mass-to-charge ratio. 
#Ions provide information concerning the nature and the structure of their precursor molecule. 
#In the spectrum of a pure compound, the molecular ion, if present, appears at the highest value of m/z (followed by ions containing heavier isotopes) and gives the molecular mass of the compound.

#There are a variety of different mass spectrometer types (almost never reported in observational data). 
#The most important difference among various mass spectrometers is the type of the analyser. 
#It is a fundamental characteristic of an instrument, and can not be modified. 
#The main analyzer types are: quadrupole, ion trap, time-of-flight (TOF), sector and Fourier-transform (FT) type instruments. 
#Some analysers are inherently capable for MS–MS (like ion trap or FT-MS), some others can be connected in tandem, offering MS–MS capabilities. 
#The most common analyzers (quadrupoles, sectors and in most respect ion traps) pass one ion (a given m / z ratio) at any one time, and filter out all other ions (which are lost for analysis).

#To obtain a mass spectrum, the analyser scans through the mass range of interest. 
#The consequence of this scanning is, that the sensitivity of detecting a given ion is much less compared to the case, when only a given mass is monitored (when the analyser stands still to pass only one ion). 
#These analyzers can be used in two ways: Either monitoring a given ion (with full sensitivity), or scanning the mass range (with loss of sensitivity) to obtain a mass spectrum. 
#Other analyzer types (TOF, FT-MS, in some respect ion traps and special sector instruments) can simultaneously detect all ions in a mass range, so there is no need for scanning. 
#In this case the full sensitivity is obtained irrespectively if one ion or a full mass spectrum is studied. 
#At present quadrupole-type instruments are most widespread. These are simple to use, cheap and rugged, good for basic GC–MS studies and for quantitation.

#In many instances the phrase 'mass selective detector' or 'MSD' is be used in place of mass spectrometer.
#However, there is no fundamental difference between mass selective detection and mass spectrometry; the name 'mass selective detection' is part of a marketing strategy, and implies a simple, relatively cheap instrument.
#Therefore all reported instances of MSD are grouped under mass spectrometry.

#Determination of Hg: https://www.atmos-meas-tech.net/9/2195/2016/amt-9-2195-2016.pdf
#Determination of Ni: https://www.longdom.org/open-access/gcms-determination-of-the-isotope-distribution-of-nickel-2157-7064-1000397.pdf
#Determination of OC: https://assets.thermofisher.com/TFS-Assets/CMD/Application-Notes/ANCCSGCTOCAN_0311.pdf

'gas chromatography - mass spectrometry':{'abbreviation':'GC-MS', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Hg-GEM','Hg-GOM','Hg-TGM','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 
    'instruments':{}
},

#------------------------------------
#pyrolysis - gas chromatography - mass spectrometry (Py-GC-MS)

#For the measurement of black carbon, GCMS can be used following pyrolysis.
#See more information here: https://www.sciencedirect.com/science/article/abs/pii/S014663800800185X and https://digital.csic.es/handle/10261/55429

'pyrolysis - gas chromatography - mass spectrometry':{'abbreviation':'Py-GC-MS', 'sampling_type':'injection', 'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC'], 'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC'], 
    'instruments':{}
},

#------------------------------------
#gas chromatography - direct temperature resolved mass spectrometry (GC-DTMS) 

#Direct Temperature Resolved Mass Spectrometry comprises of thermal depolymerization of compounds into smaller fragments that are either immediately analyzed by mass spectrometry.
#This technique has the advantage that compounds with high boiling points are prevented from condensation on 'cold' surfaces.
#The method is capable of detecting both extractable and non-extractable more complex organic fractions.
#It is typically used to give detailed information about a broad range of organic compounds such as lipids, waxes, terpenoids, polynuclear aromatic compounds, oligosaccharides, small peptides, protein fragments, and larger more condensed polymeric structures resulting from charred foodstuffs.

'gas chromatography - direct temperature resolved mass spectrometry':{'abbreviation':'GC-DTMS', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 
    'instruments':{}
},

#------------------------------------
#gas chromatography - mass spectrometry - flame ionisation detection (GC-MS-FID)

#In some instances, the simultaneous application of different detectors is extremely advantageous.
#For example, the combination of an a flame ionisation detector (which has a wide quantitation response range) with a mass spectrometer (which obtains qualitative information of unknown compounds) enables identification of volatile and semi-volatile compounds at trace levels.
#A splitter is applied after the GC column, splitting output to an FID and MS detector. 
#Using such a system enables users to obtain both a TIC (Total Ion Chromatogram) and FID chromatogram at the same time, giving the advantages of both detection types simultaneously.
#By acquiring abundant information at the same time, the detector splitting system saves both cost for analytical instruments and columns, and operators' time.

'gas chromatography - mass spectrometry - flame ionisation detection':{'abbreviation':'GC-MS-FID', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 
    'instruments':{}
},

#------------------------------------
#gas chromatography - mass spectrometry - photoionisation detection (GC-MS-PID)

#In the same way as GC-MS-FID, simultaneous detection using MS and PID detectors through splitting of GC output gives simultaneous advantage of both detection types. 

'gas chromatography - mass spectrometry - photoionisation detection':{'abbreviation':'GC-MS-PID', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 
    'instruments':{}
},

#------------------------------------
#gas chromatography - electron capture detection - photoionisation detection (GC-ECD-PID)

#In the same way as GC-MS-FID, simultaneous detection using ECD and PID detectors through splitting of GC output gives simultaneous advantage of both detection types. 

'gas chromatography - electron capture detection - photoionisation detection':{'abbreviation':'GC-ECD-PID', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','PAN','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','PAN','HCHO'], 
    'instruments':{}
},

#------------------------------------
#gas chromatography - flame ionisation detection - electron capture detection (GC-FID-ECD)

#In the same way as GC-MS-FID, simultaneous detection using FID and ECD detectors through splitting of GC output gives simultaneous advantage of both detection types. 

'gas chromatography - flame ionisation detection - electron capture detection':{'abbreviation':'GC-FID-ECD', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','PAN','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','PAN','HCHO'], 
    'instruments':{}
},

#------------------------------------
#gas chromatography - flame ionisation detection - photoionisation detection (GC-FID-PID)

#In the same way as GC-MS-FID, simultaneous detection using FID and PID detectors through splitting of GC output gives simultaneous advantage of both detection types. 

'gas chromatography - flame ionisation detection - photoionisation detection':{'abbreviation':'GC-FID-PID', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 
    'instruments':{}
},

#------------------------------------
#gas chromatography - fourier transform infrared spectroscopy - mass spectrometry (GC-FTIR-MS) 

#In the same way as GC-MS-FID, simultaneous detection using FTIR and MS detectors through splitting of GC output gives simultaneous advantage of both detection types. 

'gas chromatography - fourier transform infrared spectroscopy - mass spectrometry':{'abbreviation':'GC-FTIR-MS', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 
    'instruments':{}
},

#------------------------------------
#gas chromatography - unknown detection (GC-UNK)

'gas chromatography - unknown detection':{'abbreviation':'GC-UNK', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'],  
    'instruments':{}
},

#------------------------------------
#high performance liquid chromatography - mass spectrometry (HPLC-MS)

#Liquid chromatography (LC) is a separation process used to isolate the individual components of a mixture. 
#This process involves mass transfer of a sample through a polar mobile phase and non-polar stationary phase.
#The device is a column packed with the porous medium made of a granular solid material (i.e., stationary phase), such as polymers and silica, where the sample is injected and the solvent (i.e., mobile phase) passes to transport the sample.
#When a sample is injected, it is adsorbed on the stationary phase, and the solvent passes through the column to separate the compounds one by one, based on their relative affinity to the packing materials and the solvent. 
#The component with the most affinity to the stationary phase is the last to separate. This is because high affinity corresponds to more time to travel to the end of the column.

#High-performance liquid chromatography (HPLC), also known as high-pressure liquid chromatography, is an advanced type of LC. 
#HPLC is amenable to a wide range of applications, such as pharmaceuticals and food analysis. It is especially useful for low or non-volatile organic compounds, which cannot be handled with gas chromatography.
#The difference between traditional LC and HPLC is that the solvent in LC travels by the force of gravity. 
#In HPLC, the solvent travels under high pressure obtained by means of a pump to overcome the pressure drop in the packed column, which reduces the time of separation. 

#As with gas chromatography, liquid chromatography can be combined with mass spectrometry detection.
#Combining the two analytical methods reduces experimental error and improves accuracy. This technique is very useful in applications that involve a huge number of compounds, such as environmental effluents.

'high performance liquid chromatography - mass spectrometry':{'abbreviation':'HPLC-MS', 'sampling_type':'injection', 'measured_parameters':['HCHO','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'qa_accepted_parameters':['HCHO','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'],
    'instruments':{}
},

#------------------------------------
#high performance liquid chromatography - ultraviolet detection (HPLC-UV)

#The most commonly used detector in liquid chromatography is ultra violet absorption
#A UV detector uses a molecules ability to absorb ultraviolet light – detectors can use a single wavelength or variable wavelengths which can be more sensitive. 
#By passing UV light through the individual components of an eluting sample mix and measuring the amount of UV light absorbed by each component, you can determine the individual quantities of each component.

'high performance liquid chromatography - ultraviolet detection':{'abbreviation':'HPLC-UV', 'sampling_type':'injection', 'measured_parameters':['CH4','HCHO','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','NH4+','PM10_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'qa_accepted_parameters':['CH4','HCHO','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'],
    'instruments':{}
},

#------------------------------------
#high performance liquid chromatography - fluorescence detection (HPLC-FLD)

#Fluorescence detection can be a strong alternative to ultraviolet or other detectors for some compounds. 

'high performance liquid chromatography - fluorescence detection':{'abbreviation':'HPLC-FLD', 'sampling_type':'injection', 'measured_parameters':['HCHO','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'qa_accepted_parameters':['HCHO','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'],
    'instruments':{}
},

#------------------------------------
#high performance liquid chromatography - photodiode array detection (HPLC-PDA)

#Photodiode array detection uses the same principles of operation as a UV detector, however, the array of diodes enables simultaneous acquisition across a range of wavelengths, rather than just a single one.
#While a UV detector has only one sample-side light-receiving section, a DAD has multiple photodiode arrays to obtain information over a wide range of wavelengths at one time.

'high performance liquid chromatography - photodiode array detection':{'abbreviation':'HPLC-PDA', 'sampling_type':'injection', 'measured_parameters':['HCHO','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'qa_accepted_parameters':['HCHO','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'],
    'instruments':{}
},

#------------------------------------
#high performance liquid chromatography - mass spectrometry - fluorescence detection (HPLC-MS-FLD)

#Simultaneous detection using MS and FLD detectors through splitting of HPLC output gives simultaneous advantage of both detection types.

'high performance liquid chromatography - mass spectrometry - fluorescence detection':{'abbreviation':'HPLC-MS-FLD', 'sampling_type':'injection', 'measured_parameters':['HCHO','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'qa_accepted_parameters':['HCHO','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'],
    'instruments':{}
},

#------------------------------------
#high performance liquid chromatography - unknown detection (HPLC-UNK)

'high performance liquid chromatography - unknown detection':{'abbreviation':'HPLC-UNK', 'sampling_type':'injection', 'measured_parameters':['CH4','HCHO','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','NH4+','PM10_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'qa_accepted_parameters':['CH4','HCHO','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'],
    'instruments':{}
},

#------------------------------------
#proton transfer reaction

#An alternative to using gas chromatography for the separation of compounds is proton transfer reaction. 
#The main advantage of proton transfer reaction is that in most of cases it does not induce any significant fragmentation, and when it does only few fragments are produced. 
#Soft-ionisation results in only one or two characteristic ions, and the matrix of signals is much less complex than in other mass spectrometric techniques.
#Upon entering a PTR device, gas-phase molecules are ionised (by an ionisation source), transferring a proton (H3O+, O2+, NO+, or others) to the molecules of interest.
#For the proton transfer reaction to occur, the ionised molecule must have a higher proton affinity than the transferred proton (i.e. H3O+).
#Fast switching between reagent ions can also be used to separate isomeric compounds (molecules containing same atoms, but having different conformation), which have different reactivity with the various reagent ions.
#The ions produced in this reaction are subsequently detected by a mass analyser.
#Using PTR, comes with the advantage that ambient air can be directly sampled, without any prior sample processing (i.e. preconcentration). 

#------------------------------------
#proton transfer reaction - mass spectrometry (PTR-MS)

#Almost always PTR is coupled with a mass spectrometer for detection (PTR-MS).
#Proton Transfer Reaction - Mass Spectrometry characterises compounds according to their mass-to-charge ratio. 
#A great advantage of PTR-MS is that common air constituents (N2, O2) have low proton affinities and do not react, thus no diluting buffer gas is required.
#PTR-MS is the most widely used method for real-time monitoring of volatile organic compounds (VOCs) at low concentrations.

#Unless another specific detector is stated, where Proton Transfer Reaction is given as a methodology, it will be assumed that it uses Mass Spectrometry detection.

'proton transfer reaction - mass spectrometry':{'abbreviation':'PTR-MS', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 
    'instruments':{}
},

#------------------------------------
#proton transfer reaction - unknown detection (PTR-UNK)

'proton transfer reaction - unknown detection':{'abbreviation':'PTR-UNK', 'sampling_type':'injection', 'measured_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 'qa_accepted_parameters':['CO','CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','BaP','PM10_BaP','PM2.5_BaP','PM1_BaP','NMVOC','VOC','NMHC','HC','HCHO'], 
    'instruments':{}
},

#------------------------------------
#Passive methods (colorimetry, spectrophotometry, ion chromatography, continuous flow analysis, titration)

#It is not always possible to measure atmospheric composition using a automatic, continuous methodology. 
#Typically automated instrumentation is expensive, needs dedicated laboratory space, and requires regular measurement maintenance.
#There is a growing trend in the use of passive methodologies, which are much cheaper, require less maintenance and can be used in high volume. 

#These methods typically involve 2 or more sample steps. Typically these are:
#1. Air is sampled passively, and a desired compound(s) is first trapped (through either physical or chemical means) (passive phase).
#2. The trapped species concentration is determined in a laboratory, using a suitable methodology (typically done in batch) (analysis phase). 

#The in-lab analysis phase methodologies are almost always one of the following: 
#1. colorimetry
#2. spectrophotometry
#3. ion chromatography 
#4. continuous flow analysis
#5. titration

#---------------------
#colorimetry (CO)

#Colorimetry fundamentally refers to the determination of species concentrations based on the colour of a solution.
#Intrinsically, this is based on Beer-Lambert's law, according to which the absorption of light transmitted through the medium is directly proportional to the medium concentration.
#However, the way this colour is extrapolated to give a concentration, adds complexity to the definition of this method.
#In the most basic of senses the colour can be analysed manually, with the concentration directly extrapolated through a colour chart, or given through the length of colour stain on a glass tube from reagent reaction.
#The determination can be done in a more specific, automated fashion through the use of colorimeters.
#In a colorimeter, a beam of light with a specific wavelength is passed through a solution via a series of lenses, which navigate the colored light to the measuring device. 
#This analyses the color compared to an existing standard. A microprocessor then calculates the absorbance or percent transmittance.
#If the concentration of the solution is greater, more light will be absorbed, which can be identified by measuring the difference between the amount of light at its origin and that after passing the solution.
#Colorimeters are usually portable and use LED light sources and color filters. As a result, they operate at fixed wavelengths and can only accommodate tests that incorporate those wavelengths.   

#Unless additional information is given, make assumption colorimetry method uses instrumental injection of sample species, following chemical reaction with a reagent.

#Not QA accepted measurement technique for multiple species due to dominance of other more accurate techniques in measuring these species

'colorimetry':{'abbreviation':'CO', 'sampling_type':'injection', 'measured_parameters':['O3','NO','NO2','NOx','CO','SO2','NH3','HNO3','PAN','HCl','HF','H2S','HCHO','Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','BC','PM10_BC','PM2.5_BC','PM1_BC','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 'qa_accepted_parameters':['NH3','HNO3','PAN','HCl','HF','H2S','HCHO','Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','BC','PM10_BC','PM2.5_BC','PM1_BC','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 
    'instruments':{}
},
#---------------------
#spectrophotometry (SP)

#Spectrophotometry is also based on Beer-Lambert's law, where the absorption of light transmitted through the medium is directly proportional to the medium concentration.
#The main disadvantage and limitation of the spectrophotometry is its low selectivity. The measurement of absorbance is burden by interferences derived from others components of sample.
#A spectrophotometer is a photometer (a device for measuring light intensity) that can measure intensity as a function of the color, or more specifically, the wavelength of light.  
#Spectrophotometers are usually bench top instruments and use light sources that can produce a range of wavelengths, use monochromators to select for a desired wavelength. As a result, spectrophotometers can be used for a broad range of tests. 

#The distinction between spectrophotometry and colorimetry is not always clear. 
#Colorimeters and spectrophotometers both measure sample absorbance to determine analyte concentrations. 
#The main distinction is that a spectrophotometer can be set to measure % transmittance or absorbance over a wide range of wavelengths, whereas a colorimeter determines absorbance at specific, visible, wavelengths.

#Unless additional information is given, make assumption spectrophotometry method uses instrumental injection of sample species, following chemical reaction with a reagent.

#Not QA accepted measurement technique for multiple species due to dominance of other more accurate techniques in measuring these species

'spectrophotometry':{'abbreviation':'SP', 'sampling_type':'injection', 'measured_parameters':['O3','NO','NO2','NOx','CO','SO2','NH3','HNO3','PAN','HCl','HF','H2S','HCHO','Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','BC','PM10_BC','PM2.5_BC','PM1_BC','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 'qa_accepted_parameters':['NH3','HNO3','PAN','HCl','HF','H2S','HCHO','Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','BC','PM10_BC','PM2.5_BC','PM1_BC','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 
    'instruments':{ 
        'Perkin Elmer 18 UV/VIS':{'instrument_further_details':'No manual or specifications available but confirmation it is uses spectrophotometry here: https://www.recycledgoods.com/perkin-elmer-lambda-18-uv-vis-spectrophotometer/'},
        'Shimadzu UV-1201':      {'instrument_further_details':'No manual or specifications available but confirmation it is uses spectrophotometry here: http://www.medwow.com/med/spectrophotometer/shimadzu/uv-1201/12483.model-spec'},
        'Thermo Aquakem 200':    {'instrument_manual_name':'Thermo_Aquekem_200_250_600_Specs.pdf'},
        'Thermo Aquakem 250':    {'instrument_manual_name':'Thermo_Aquekem_200_250_600_Specs.pdf'},
        'Thermo Aquakem 600':    {'instrument_manual_name':'Thermo_Aquekem_200_250_600_Specs.pdf'},
    }
},
#---------------------
#ion chromatography (IC)

#Ion chromatography (or ion-exchange chromatography) is a chromatography process that separates ions and polar molecules based on their affinity to the ion exchanger. 
#It works on almost any kind of charged molecule—including large proteins, small nucleotides, and amino acids. However, ion chromatography must be done in conditions that are one unit away from the isoelectric point of a protein. 
#The two types of ion chromatography are anion-exchange and cation-exchange. 
#Cation-exchange chromatography is used when the molecule of interest is positively charged. The molecule is positively charged because the pH for chromatography is less than the pI. In this type of chromatography, the stationary phase is negatively charged and positively charged molecules are loaded to be attracted to it. 
#Anion-exchange chromatography is when the stationary phase is positively charged and negatively charged molecules (meaning that pH for chromatography is greater than the pI) are loaded to be attracted to it.

#Typical Operation:
#1. A sample of the mixture to be analyzed (analyte) is injected into a carrier fluid (the eluent).
#2. The combination is passed through a column containing a stationary fixed material (adsorbent).
#3. Compounds contained in the analyte are then partitioned between the stationary adsorbent and the moving eluent/analyte mixture.
#4. Different dissolved materials adhere to the adsorbent with different forces. The ones that adhere strongly are moved through the adsorbent more slowly as the eluent flows over them. As the eluent flows through the column the components of the analyte will move down the column at different speeds and therefore separate from one another.
#5. A detector is used to analyze the output at the end of the column. There are various detectors, typically they use an electrical conductivity detector.
#6. Each time analyte molecules/ions emerge from the chromatography column the detector generates a measurable signal which is usually printed out as a peak on the chromatogram.
#7. A suppressor is used to reduce the background conductance of the eluent and at the same time enhance the conductance of the sample ions.

#Not QA accepted measurement technique for NO/NO2/NOx/SO2 due to dominance of other more accurate techniques in measuring these species

#See details of method for NH3:https://www.atmos-meas-tech.net/7/2733/2014/amt-7-2733-2014.html'

'ion chromatography':{'abbreviation':'IC', 'sampling_type':'injection', 'measured_parameters':['NO','NO2','NOx','SO2','NH3','HNO3','PAN','HCl','HF','H2S','HCHO','Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 'qa_accepted_parameters':['NH3','HNO3','PAN','HCl','HF','H2S','HCHO','Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 
    'instruments':{
        'Dionex DX-120':             {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Dionex_DX-120_Manual.pdf'},
        'Dionex DX-300':             {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses ion chromatography here: https://shop.labexchange.com/en/dionex-dx-300-47283.html'}, 
        'Dionex DX-600':             {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses ion chromatography here: https://books.google.es/books?id=9Yq9BgAAQBAJ&pg=PA97&lpg=PA97&dq=ic+dionex+600&source=bl&ots=lUbV9Iwed7&sig=ACfU3U3gUbLCaUNYglSBZQeHrWmVITMVoA&hl=en&sa=X&ved=2ahUKEwigsonXt9roAhUVAWMBHafJAXwQ6AEwEHoECAkQMQ#v=onepage&q=ic%20dionex%20600&f=false'},
        'Dionex 2010i':              {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_further_details':'No manual or specifications available but confirmation it is uses ion chromatography here: https://books.google.es/books?id=fortCAAAQBAJ&pg=PA395&lpg=PA395&dq=dionex+2010i&source=bl&ots=qObKPXBKMt&sig=ACfU3U13WHkWJOW7eD9NSNHuggZFjQnGVA&hl=en&sa=X&ved=2ahUKEwiZ2obZiKLgAhW4BGMBHYu7DHMQ6AEwDXoECAEQAQ#v=onepage&q=dionex%202010i&f=false'},
        'Metrohm 881 compact IC pro':{                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Metrohm_881_compact_IC_pro.pdf'},
        'Metrohm MARGA':             {'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['HCl','HNO2','HNO3','NH3','SO2','Ca++','PM10_Ca++','PM2.5_Ca++','Cl-','PM10_Cl-','PM2.5_Cl-','K+','PM10_K+','PM2.5_K+','Mg++','PM10_Mg++','PM2.5_Mg++','Na+','PM10_Na+','PM2.5_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','NO3-','PM10_NO3-','PM2.5_NO3-','SO4--','PM10_SO4--','PM2.5_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'qa_accepted_parameters':['HCl','HNO2','HNO3','NH3','SO2','Ca++','PM10_Ca++','PM2.5_Ca++','Cl-','PM10_Cl-','PM2.5_Cl-','K+','PM10_K+','PM2.5_K+','Mg++','PM10_Mg++','PM2.5_Mg++','Na+','PM10_Na+','PM2.5_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','NO3-','PM10_NO3-','PM2.5_NO3-','SO4--','PM10_SO4--','PM2.5_SO4--'], 'documented_lower_limit_of_detection':{'HCl':np.NaN,'HNO2':np.NaN,'HNO3':np.NaN,'NH3':np.NaN,'SO2':np.NaN,'Ca++':'0.09','PM10_Ca++':'0.09','PM2.5_Ca++':'0.09','Cl-':'0.01','PM10_Cl-':'0.01','PM2.5_Cl-':'0.01','K+':'0.09','PM10_K+':'0.09','PM2.5_K+':'0.09','Mg++':'0.06','PM10_Mg++':'0.06','PM2.5_Mg++':'0.06','Na+':'0.05','PM10_Na+':'0.05','PM2.5_Na+':'0.05','NH4+':'0.05','PM10_NH4+':'0.05','PM2.5_NH4+':'0.05','NO3-':'0.05','PM10_NO3-':'0.05','PM2.5_NO3-':'0.05','SO4--':'0.04','PM10_SO4--':'0.04','PM2.5_SO4--':'0.04','SO4--_NSS':'0.04','PM10_SO4--_NSS':'0.04','PM2.5_SO4--_NSS':'0.04','SO4--_SS':'0.04','PM10_SO4--_SS':'0.04','PM2.5_SO4--_SS':'0.04'}, 'documented_flow_rate':'16.67', 'instrument_manual_name':'Metrohm_MARGA_Specs.pdf'},
        'Shimadzu CDD-10Avp':        {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Shimadzu CDD-10Avp-Asp_Manual.pdf'},                                 
        'Shimadzu CDD-10Asp':        {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Shimadzu CDD-10Avp-Asp_Manual.pdf'},                              
        'Thermo Dionex Aquion':      {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Thermo_Dionex_Aquion_Specs.pdf'},                              
        'Thermo ICS-1000':           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Thermo_ICS-1000_Manual.pdf'},
        'Thermo ICS-1100':           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Thermo_ICS-1100_Manual.pdf'},
        'Thermo ICS-2000':           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Thermo_ICS-2000_Manual.pdf'},
        'Thermo ICS-2100':           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Thermo_ICS-2100_Manual.pdf'},
        'Thermo ICS-3000':           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Thermo_ICS-3000_Manual.pdf'},
        'Thermo ICS-5000':           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Thermo_ICS-5000_Manual.pdf'},
        'Thermo ICS-6000':           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Thermo_ICS-6000_Specs.pdf'},
        'Waters Alliance':           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Waters_Alliance_IC_Specs.pdf'}                                                       
    }
},

#------------------------------------
#continuous flow analysis (CFA)

#In continuous flow analysis (CFA), a sample is injected into a flowing carrier solution passing rapidly through small-bore tubing. 
#The sample is mixed with a reagent, which reacts with the sample to develop a color and determine the sample concentration (typically by spectrophotometry).
#The use of carefully controlled flow conditions ensures that the color development reaction is reproducible, so that the color measurement need not wait until the reaction has gone to completion.
#There are several technologies of CFA such as segmented flow analysis (SFA), which uses turbulent flow conditions that allow for complete sample dispersion. 
#Other technologies include flow injected analysis (FIA) and sequential injection analysis, which uses laminar flow existing in the narrow-bore tubing to mix with the reagent to eliminate the need for air bubble partitioning.
#Most continuous flow analyzers depend on color reactions using a flow through photometer, however, also methods have been developed that use ISE, flame photometry, ICAP, fluorometry, and so forth.

'continuous flow analysis':{'abbreviation':'CFA', 'sampling_type':'injection', 'sample_preparation_type':'reagent reaction', 'measured_parameters':['NO','NO2','NOx','SO2','NH3','HNO3','PAN','HCl','HF','H2S','HCHO','Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','BC','PM10_BC','PM2.5_BC','PM1_BC','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 'qa_accepted_parameters':['NH3','HNO3','PAN','HCl','HF','H2S','HCHO','Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','BC','PM10_BC','PM2.5_BC','PM1_BC','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 
    'instruments':{
       'Bran+Luebbe Autoanalyzer 3':{'measured_parameters':['NH3','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','K+','PM10_K+','PM2.5_K+','PM1_K+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-'],                                                                                                                                                                                                                                                                                                'qa_accepted_parameters':['NH3','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','K+','PM10_K+','PM2.5_K+','PM1_K+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-'],                                                                                                                                                                                                                                                                                           'instrument_manual_name':'Bran+Luebbe_Autoanalyzer_3_Specs.pdf'},
       'Foss FIAstar 5000':         {'measured_parameters':['NH3','Al','PM10_Al','PM2.5_Al','PM1_Al','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-'],                                                                                                                                                                                                                                                     'qa_accepted_parameters':['NH3','Al','PM10_Al','PM2.5_Al','PM1_Al','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-'],                                                                                                                                                                                                                                                'instrument_manual_name':'Foss_FIAstar_5000_Specs.pdf'},
       'SEAL AA500':                {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                'instrument_manual_name':'SEAL_AA500_Specs.pdf'},
       'Thermo Gallery':            {'measured_parameters':['SO2','NH3','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'],'qa_accepted_parameters':['NH3','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'instrument_manual_name':'Thermo_Gallery_Specs.pdf'},  
       'Thermo Gallery Plus':       {'measured_parameters':['SO2','NH3','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'],'qa_accepted_parameters':['NH3','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'instrument_manual_name':'Thermo_Gallery_Specs.pdf'}  
    }
},


#---------------------
#titration (TI)

#Titration is a common laboratory method of quantitative chemical analysis that is used to determine the concentration of an identified analyte. 
#Since volume measurements play a key role in titration, it is also known as volumetric analysis. 
#A reagent, called the titrant or titrator is prepared as a standard solution. 
#A known concentration and volume of titrant reacts with a solution of analyte or titrand to determine concentration. 
#The volume of titrant reacted is called titration volume.
#Titration can be also done automatically, in a continuous process.

#Typical measurement operation:
#1. A typical titration begins with a beaker or Erlenmeyer flask containing a very precise volume of the analyte and a small amount of indicator (such as phenolphthalein) placed underneath a calibrated burette or chemistry pipetting syringe containing the titrant. 
#2. Small volumes of the titrant are then added to the analyte and indicator until the indicator changes color in reaction to the titrant saturation threshold, reflecting arrival at the endpoint of the titration. 
#3. Depending on the endpoint desired, single drops or less than a single drop of the titrant can make the difference between a permanent and temporary change in the indicator.    
#4. When the endpoint of the reaction is reached, the volume of reactant consumed is measured and used to calculate the concentration of analyte by: Ca = Ct*Vt*M / Va  ;where Ca is the concentration of the analyte, typically in molarity; Ct is the concentration of the titrant, typically in molarity; Vt is the volume of the titrant used, typically in liters; M is the mole ratio of the analyte and reactant from the balanced chemical equation; and Va is the volume of the analyte used, typically in liters.

#Unless additional information is given, make assumption titration uses instrumental injection of sample species, following chemical reaction with a reagent.

#Not QA accepted measurement technique for SO2 due to dominance of other more accurate techniques in measuring this species

#See more details about method for SO2:https://www3.epa.gov/ttnemc01/ctm/ctm013B.pdf

'titration':{'abbreviation':'TI', 'sampling_type':'injection', 'measured_parameters':['SO2'], 'qa_accepted_parameters':[], 
    'instruments':{}
},

#------------------------------------
#thermal reduction - chemiluminescence (TR-CL)

#Indirect method for the measurement of nitrate (NO3-) through the conversion of NO3- to NOy by thermal reduction, which is then measured by chemiluminescence, as a result of the reaction of NO with ozone (NO+O3 --> NO2*+O2).
#The nitrate is converted by drawing a continuous stream of sample across a hot reactive surface that reduces nitrate particles in the sample stream to NOy
#The concentrations of nitrate in the ambient air are then quantified by comparing the signal produced when aerosol-laden sample is drawn directly into the converter to a background signal that is produced when the sample stream is run through a high-efficiency particulate aerosol filter that removes the nitrate before conversion. 
#The difference in signal between the filtered and unfiltered sample can be attributed to the NOy that is formed from nitrate particles in the unfiltered sample stream. 
#By routinely switching between the filtered and unfiltered sample streams, the instrument readings can be continuously adjusted or corrected for changes in background signal that might be produced by traces of NOy or other interfering gases. 
#This frequent adjustment for changes in background signal improves the system stability, which in turn improves the limit of detection.

'thermal reduction - chemiluminescence':{'abbreviation':'TR-CL', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-'], 'qa_accepted_parameters':['NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-'], 
    'instruments':{
        'Rupprecht and Patashnick 8400N':{'measured_parameters':['PM2.5_NO3-'], 'qa_accepted_parameters':['PM2.5_NO3-'], 'documented_resolution':'0.2', 'instrument_further_details':'No manual or specifications available but confirmation it is uses thermal reduction - chemiluminescence here: https://lampx.tugraz.at/~tunnel2016/history/Tunnel_2002_CD/PDF/16-Rosenberger.pdf'}
    }
},

#------------------------------------
#thermal reduction - ultraviolet fluorescence (TR-UVF)

#Indirect method for the measurement of sulphate (SO4--) through the conversion of SO4-- to SO2 by thermal reduction, which is then measured by ultraviolet fluorescence.
#The sulphate is converted by drawing a continuous stream of sample across a hot reactive surface that reduces sulphate particles in the sample stream to sulfur dioxide gas. 
#The concentrations of sulphate in the ambient air are then quantified by comparing the signal produced when aerosol-laden sample is drawn directly into the converter to a background signal that is produced when the sample stream is run through a high-efficiency particulate aerosol filter that removes the sulfate before conversion. 
#The difference in signal between the filtered and unfiltered sample can be attributed to sulfur dioxide that is formed from sulfate particles in the unfiltered sample stream. 
#By routinely switching between the filtered and unfiltered sample streams, the instrument readings can be continuously adjusted or corrected for changes in background signal that might be produced by traces of SO2 or other interfering gases. 
#This frequent adjustment for changes in background signal improves the system stability, which in turn improves the limit of detection.

'thermal reduction - ultraviolet fluorescence':{'abbreviation':'TR-UVF', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'qa_accepted_parameters':['SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 
    'instruments':{
        'Rupprecht and Patashnick 8400S':{'measured_parameters':['PM2.5_SO4--','PM2.5_SO4--_NSS','PM2.5_SO4--_SS'], 'qa_accepted_parameters':['PM2.5_SO4--','PM2.5_SO4--_NSS','PM2.5_SO4--_SS'],                                                                                                                                                                                                    'documented_resolution':'0.2',                                   'instrument_further_details':'No manual or specifications available but confirmation it is uses thermal reduction - ultraviolet fluorescence here: https://lampx.tugraz.at/~tunnel2016/history/Tunnel_2002_CD/PDF/16-Rosenberger.pdf'},
        'Thermo 5020':                   {                                                                                                                                                       'documented_lower_limit_of_detection':'0.5', 'documented_upper_limit_of_detection':'50.0',                                                                    'documented_span_drift':'1.0%/week',                                'documented_flow_rate':'1.2',     'instrument_manual_name':'Thermo_5020_Specs.pdf'},
        'Thermo 5020i':                  {                                                                                                                                                       'documented_lower_limit_of_detection':'0.5', 'documented_upper_limit_of_detection':'4000.0', 'documented_accuracy':'1.0%', 'documented_zero_drift':'1.0/day', 'documented_span_drift':'1.0%/day',                                 'documented_flow_rate':'0.4-0.5', 'instrument_manual_name':'Thermo_5020i_Manual.pdf'}
    }
},

##------------------------------------
#aerosol mass spectrometry (AMS)

#This method is a variation of the Mass Spectrometry method to analyse gas-phase species, specific for aerosols.
#In Aerosol Mass Spectrometry, aerosol particles are – together with a small fraction (10−7) of a carrier gas – directed onto the tungsten vaporizer, typically heated to a temperature of 550–600 ◦C.
#Non-refractory material flash-vaporises, and the emerging vapor is electron-impact-ionised (70 eV) for subsequent analysis with a quadrupole or a time-of-flight mass spectrometer.
#In addition to the measurement of the aerosol beam (“beam-open” mass spectrum), also the instrument background is measured with the aerosol beam blocked (“beam-closed” mass spectrum). 
#The particle contribution is calculated from the difference of both measurements (“difference” spectrum), assuming that all particle components vaporize quickly compared to the beam-open–beam-closed cycle length. 
#For calculation of mass concentrations of species from the difference spectrum, each individual m/z is associated with one or several of these species.
#For the standard analysis of the unit mass resolution spectra, these associations are listed in the “frag table”, which needs to be adapted for special measurement situations by the user. 
#Measurements with the high-resolution time-of-flight AMS (HR-ToF-AMS) allow the quantification of individual signals in the spectra with certain elemental composition.
#For measurements of the continental or urban background aerosol the assumptions behind this procedure are typically well met and the AMS provides robust quantitative information on the sub-micron aerosol composition. 
#However, under certain conditions, e.g., when measuring close to anthropogenic sources or in the marine environment, the limitations of these assumptions are sometimes reached and the standard analysis could result in misinterpretation of the mass spectra.
#The method can be performed online (sampling ambient air) or offline (through injection). 
#It is assumed unless otherwise stated that instrumental sampling is done via injection.

'aerosol mass spectrometry':{'abbreviation':'AMS', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'qa_accepted_parameters':['Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 
    'instruments':{
        'Aerodyne CTOF AMS':       {'measured_parameters':['PM2.5_Cl-','PM1_Cl-','PM2.5_NO3-','PM1_NO3-','PM2.5_NH4+','PM1_NH4+','PM2.5_SO4--','PM1_SO4--'], 'qa_accepted_parameters':['PM2.5_Cl-','PM1_Cl-','PM2.5_NO3-','PM1_NO3-','PM2.5_NH4+','PM1_NH4+','PM2.5_SO4--','PM1_SO4--','PM2.5_SO4--_NSS','PM1_SO4--_NSS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'documented_lower_limit_of_detection':{'PM2.5_Cl-':np.NaN,'PM1_Cl-':np.NaN,'PM2.5_NO3-':'0.0012','PM1_NO3-':'0.0012','PM2.5_NH4+':'0.024','PM1_NH4+':'0.024','PM2.5_SO4--':'0.0024','PM1_SO4--':'0.0024','PM2.5_SO4--_NSS':'0.0024','PM1_SO4--_NSS':'0.0024','PM2.5_SO4--_SS':'0.0024','PM1_SO4--_SS':'0.0024'}, 'documented_flow_rate':'0.85',  'instrument_manual_name':'Aerodyne_AMS_Specs.pdf'},
        'Aerodyne HRTOF AMS':      {'measured_parameters':['PM2.5_Cl-','PM1_Cl-','PM2.5_NO3-','PM1_NO3-','PM2.5_NH4+','PM1_NH4+','PM2.5_SO4--','PM1_SO4--'], 'qa_accepted_parameters':['PM2.5_Cl-','PM1_Cl-','PM2.5_NO3-','PM1_NO3-','PM2.5_NH4+','PM1_NH4+','PM2.5_SO4--','PM1_SO4--','PM2.5_SO4--_NSS','PM1_SO4--_NSS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'documented_lower_limit_of_detection':{'PM2.5_Cl-':np.NaN,'PM1_Cl-':np.NaN,'PM2.5_NO3-':'0.0029','PM1_NO3-':'0.0029','PM2.5_NH4+':'0.058','PM1_NH4+':'0.058','PM2.5_SO4--':'0.0058','PM1_SO4--':'0.0058','PM2.5_SO4--_NSS':'0.0058','PM1_SO4--_NSS':'0.0058','PM2.5_SO4--_SS':'0.0058','PM1_SO4--_SS':'0.0058'}, 'documented_flow_rate':'0.85',  'instrument_manual_name':'Aerodyne_AMS_Specs.pdf'},
        'Aerodyne quadrupole ACSM':{'measured_parameters':['PM2.5_Cl-','PM1_Cl-','PM2.5_NO3-','PM1_NO3-','PM2.5_NH4+','PM1_NH4+','PM2.5_SO4--','PM1_SO4--'], 'qa_accepted_parameters':['PM2.5_Cl-','PM1_Cl-','PM2.5_NO3-','PM1_NO3-','PM2.5_NH4+','PM1_NH4+','PM2.5_SO4--','PM1_SO4--','PM2.5_SO4--_NSS','PM1_SO4--_NSS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'documented_lower_limit_of_detection':{'PM2.5_Cl-':'0.02','PM1_Cl-':'0.02','PM2.5_NO3-':'0.02','PM1_NO3-':'0.02','PM2.5_NH4+':'0.5','PM1_NH4+':'0.5','PM2.5_SO4--':'0.04','PM1_SO4--':'0.04','PM2.5_SO4--_NSS':'0.04','PM1_SO4--_NSS':'0.04','PM2.5_SO4--_SS':'0.04','PM1_SO4--_SS':'0.04'},                     'documented_flow_rate':'0.085', 'instrument_manual_name':'Aerodyne_quadrupole_ACSM_Specs.pdf'},
        'Aerodyne eTOF ACSM':      {'measured_parameters':['PM2.5_Cl-','PM1_Cl-','PM2.5_NO3-','PM1_NO3-','PM2.5_NH4+','PM1_NH4+','PM2.5_SO4--','PM1_SO4--'], 'qa_accepted_parameters':['PM2.5_Cl-','PM1_Cl-','PM2.5_NO3-','PM1_NO3-','PM2.5_NH4+','PM1_NH4+','PM2.5_SO4--','PM1_SO4--','PM2.5_SO4--_NSS','PM1_SO4--_NSS','PM2.5_SO4--_SS','PM1_SO4--_SS'], 'documented_lower_limit_of_detection':{'PM2.5_Cl-':'0.003','PM1_Cl-':'0.003','PM2.5_NO3-':'0.007','PM1_NO3-':'0.007','PM2.5_NH4+':'0.06','PM1_NH4+':'0.06','PM2.5_SO4--':'0.006','PM1_SO4--':'0.006','PM2.5_SO4--_NSS':'0.006','PM1_SO4--_NSS':'0.006','PM2.5_SO4--_SS':'0.006','PM1_SO4--_SS':'0.006'},         'documented_flow_rate':'0.085', 'instrument_manual_name':'Aerodyne_eTOF_ACSM_Specs.pdf'},    
    }
},
#------------------------------------
#gravimetry (GR)

#An air pump draws ambient air at a constant flow rate into a specially shaped inlet where particulate matter is separated into size fractions. 
#Particulate matter is then collected on a filter. Each filter is weighed before and after use, to determine the net mass gain due to collected matter. 
#The total volume of air filtered is known from the constant air flow, and the difference in filter weights is used to calculate the particulate matter concentration in micrograms per cubic meter (μg/m3) of air.

#Known Interference:
#Particulate matter may be lost during filter handling and weighing procedures, especially if filter is exposed to warming.
#Gaseous species may contaminate filters.
#Humidity and absorbed water may be difficult to control both during operations and when handling filters.
#Removing filters and transporting to a lab for analysis may affect results. 
#Meteorological conditions may affect flow rate.

'gravimetry':{'abbreviation':'GR', 'sampling_type':'manual', 'measured_parameters':['PM10','PM2.5','PM1'], 'qa_accepted_parameters':['PM10','PM2.5','PM1'], 
    'instruments':{}
},

#------------------------------------
#tapered element oscillating microbalance - gravimetry (TEOM-GR)

#The TEOM - Gravimetric method essentially involves a true “gravimetric” instrument that draws (then heats) ambient air through a filter at constant flow rate, continuously weighing the filter and calculating near real-time mass concentrations of particulate matter.
#The TEOM uses a hollow glass tube as a microbalance. Incoming particles are deposited on a filter at the tip of the tube, and the added mass causes a change in its oscillation frequency which is detected electronically. 
#The element is periodically cycled to return it to its natural frequency. TEOM devices operate continuously and do not need filter changes as frequently as high-volume air samplers.

#Typical operation:
#1. Air is drawn through a tapered glass element with a filter attached. 
#2. The element oscillates according to a characteristic frequency, that decreases as mass accumulates on the attached filter. 
#3. Measurement of the change in frequency converts to measurement of the accumulated mass.

'tapered element oscillating microbalance - gravimetry':{'abbreviation':'TEOM-GR', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['PM10','PM2.5','PM1'], 'qa_accepted_parameters':['PM10','PM2.5','PM1'], 
    'instruments':{
        'Thermo 1400':  {                                                                                                                                                                                                                                                                                                                'instrument_further_details':'No manual or specifications available but confirmation it is uses tapered element oscillating microbalance - gravimetry here: https://uk-air.defra.gov.uk/assets/documents/reports/empire/lsoman/LSO_manual_2012_18_R&P_TEOM1400.pdf'},
        'Thermo 1400a': {                                                                                                                                                                                                                                                                              'documented_flow_rate':'3.0',     'instrument_manual_name':'Thermo_1400a_Manual.pdf'},
        'Thermo 1400ab':{                                                                                   'documented_lower_limit_of_detection':'0.06', 'documented_upper_limit_of_detection':'5000000', 'documented_precision':'1.5', 'documented_accuracy':'0.75%', 'documented_resolution':'0.1', 'documented_flow_rate':'0.5-4.0', 'instrument_manual_name':'Thermo_1400ab_Specs.pdf'},
        'Thermo 1405':  {                                                                                                                                 'documented_upper_limit_of_detection':'1000000', 'documented_precision':'2.0', 'documented_accuracy':'0.75%', 'documented_resolution':'0.1', 'documented_flow_rate':'3.0',     'instrument_manual_name':'Thermo_1405_Specs.pdf'},
        'Thermo 1405D': {'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'],                                                                                                                                                                                            'documented_flow_rate':'3.0',     'instrument_manual_name':'Thermo_1405D_Manual.pdf'}
    }
},

#------------------------------------
#tapered element oscillating microbalance - filter dynamics measurement system - gravimetry (TEOM-FDMS-GR)

#The Filter Dynamics Measurement System (FDMS) is a hybrid of the TEOM system which is able to measure the semi-volatile fraction of airborne particulate matter. 
#This can be important in the study of primary and secondary PM and to ensure there are no losses of these semi volatile fractions due to sampling conditions. 
#TEOM instruments with FDMS alternate between a base cycle and a reference cycle, the latter of which measures the mass loss of the filter when clean air is passed through it, allowing the mass loss during the base cycle to be estimated.
#It is important that the air conditioning system not cycle over the same period as the TEOM instrument, because this can cause aliasing.

'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry':{'abbreviation':'TEOM-FDMS-GR', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['PM10','PM2.5','PM1'], 'qa_accepted_parameters':['PM10','PM2.5','PM1'], 
    'instruments':{
        'Thermo 1405-DF':{'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'], 'documented_upper_limit_of_detection':'1000000', 'documented_precision':'2.0', 'documented_resolution':'0.1', 'documented_flow_rate':'1.67-12.0', 'instrument_manual_name':'Thermo_1405-DF_Specs.pdf'},
        'Thermo 1405F':  {'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'],                                                                                                               'documented_flow_rate':'3.0',       'instrument_manual_name':'Thermo_1405F_Manual.pdf'}
    }
},

#------------------------------------
#quartz crystal microbalance - gravimetry (QCM-GR)

#The quartz crystal microbalance - gravimetry method is a very sensitive mass deposition sensor based on the piezoelectric properties of the quartz crystal.  
#Particles are deposited by inertial impaction (or another method) onto the surface of a quartz crystal.
#The natural resonant frequency of the crystal is monitored and compared with that of a clean reference crystal.
#As particle mass increases, the frequency decreases.

'quartz crystal microbalance - gravimetry':{'abbreviation':'QCM-GR', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['PM10','PM2.5','PM1'], 'qa_accepted_parameters':['PM10','PM2.5','PM1'], 
    'instruments':{}
},

#------------------------------------
#pressure drop tape sampling (PDTS)

#Pressure drop tape sampling measures particles by monitoring the pressure drop across a porous membrane filter (fluoropore).
#The method was originally developed at Harvard University (Babich et al., 1997).
#The pressure drop is linearly correlated to the particle mass deposited on the filter.
#The filter face veolocity is chosen such that pore obstruction by interception is the dominant cause of particle related pressure drop change over time.
#The monitor exposes a new segment of filter tape every 20-60 minutes for particle collection.

'pressure drop tape sampling':{'abbreviation':'PDTS', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['PM10','PM2.5','PM1'], 'qa_accepted_parameters':['PM10','PM2.5','PM1'], 
    'instruments':{
         'CAMMS':{'instrument_further_details':'No manual or specifications available but confirmation it is uses pressure drop tape sampling here: https://books.google.es/books?id=fXFbDCpB88sC&pg=SA3-PA15&lpg=SA3-PA15&dq=CAMMS+Mass+pressure+drop&source=bl&ots=zYpHrfTg-T&sig=ACfU3U0d2zQC9YMTfgYELcs6HQmwW0IVgA&hl=en&sa=X&ved=2ahUKEwiNw_bhrK72AhWa7rsIHRz5BgkQ6AF6BAgUEAM#v=onepage&q=CAMMS%20Mass%20pressure%20drop&f=false'}
    }
},

#------------------------------------
#beta-attenuation (BA)

#Beta particles (electrons with energies in the 0.01 to 0.1 MeV range) are attenuated according to an approximate exponential function when they pass through particulate deposits on a filter tape. 
#Automated samplers (analyzers) use a continuous filter tape, first measuring the attenuation by the unexposed tape, and then measuring the attenuation after the tape has passed through the ambient air flow. 
#The attenuation measurement converts to a measure of the mass on the filter, so that the filters do not require later laboratory analysis for the mass variable. For some devices, the beta particle source is 14C.

#Known Interference:
#Particulate matter may be lost due to filter tape advance and vibration, especially if filter is exposed to warming.
#Gaseous species may contaminate filters.
#Humidity and absorbed water may be difficult to control during operations. 
#Meteorological conditions may affect flow rate.
#Although on-site real-time mass measurement offers significant improvements over the filter removal and laboratory analysis process, the beta emission and detection process present additional on-site maintenance requirements.

'beta-attenuation':{'abbreviation':'BA', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['PM10','PM2.5','PM1'], 'qa_accepted_parameters':['PM10','PM2.5','PM1'], 
    'instruments':{
        'ADM 9000':                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: https://pubmed.ncbi.nlm.nih.gov/28094548/'},         
        'Airpointer-PMBA':            {'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'], 'documented_lower_limit_of_detection':'1.0', 'documented_upper_limit_of_detection':'1000000.0',                                                                                                                               'documented_resolution':'0.1', 'documented_zero_drift':'1.0/day',                                     'documented_flow_rate':'16.7',            'instrument_manual_name':'Airpointer_Specs.pdf'},
        'Andersen FH621-N':           {                                                                                   'documented_lower_limit_of_detection':'2.0', 'documented_upper_limit_of_detection':'10000',                                                                     'documented_accuracy':'2.0',                                                                                                                                                                                  'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: https://cdiac.ess-dive.lbl.gov/programs/NARSTO/Compendium_combined.pdf; see details about detection limits here: https://books.google.es/books?id=gDJEBAAAQBAJ&pg=PT323&lpg=PT323&dq=andersen+fh621-n&source=bl&ots=fU7hu5zKUU&sig=ACfU3U1TC8JhHedeoUMYXqfHdPIfrrN0rw&hl=en&sa=X&ved=2ahUKEwi--YrC_oPgAhVi5OAKHaf2CEcQ6AEwAnoECAcQAQ#v=onepage&q=andersen%20fh621-n&f=false'},
        'Dasibi 7001':                {                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: http://www.sistemapiemonte.it/ambiente/srqa/stazioni/pdf/151.pdf'},
        'DKK-TOA FPM-377':            {'measured_parameters':['PM2.5'],        'qa_accepted_parameters':['PM2.5'],                                                     'documented_upper_limit_of_detection':'5000',                                                                                                                                                                                                                                                                                    'instrument_further_details':'Information was found online:https://www.directindustry.com/prod/dkk-toa-corporation/product-13577-1326921.html'},
        'Eberline FH62-1':            {                                                                                                                                                                                                                                                   'documented_accuracy':'2.0',                                                                                                                                                                                  'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: http://adsabs.harvard.edu/abs/2012AtmEn..54...18Gf'},
        'Environnement SA MP101M':    {                                                                                   'documented_lower_limit_of_detection':'0.5', 'documented_upper_limit_of_detection':'10000',                                                                                                                                                                                                                                         'documented_flow_rate':'16.666667',       'instrument_manual_name':'Environnement_SA_MP101M_Specs.pdf'},
        'Environnement SA MPSI100':   {                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: https://eurekamag.com/pdf/008/008897460.pdf'},
        'FAI Swam 5a':                {                                                                                   'documented_lower_limit_of_detection':'1.0', 'documented_upper_limit_of_detection':'10000',                                                                                                   'documented_uncertainty':'1.0',                                                                                                       'documented_flow_rate':'13.33-41.6667',   'instrument_manual_name':'FAI_Swam_5a_specs.pdf', 'measuring_instrument_further_details':'See more details here: http://fai-instruments.net/swam-dual-channel-2/?lang=en'}, 
        'Horiba APDA350E':            {                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: https://books.google.es/books?id=wYUUzLCIMzsC&pg=PA112&lpg=PA112&dq=horiba+apda+350e&source=bl&ots=uvOvFFezdD&sig=ACfU3U041o7NZz92BUIgpBh1x0CLKwY5lQ&hl=en&sa=X&ved=2ahUKEwjfo6_x4ojgAhUFQhoKHeuAAlgQ6AEwCHoECAEQAQ#v=onepage&q=horiba%20apda%20350e&f=false'},
        'Horiba APDA351':             {                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: https://dd.eionet.europa.eu/vocabularyconcept/aq/measurementequipment/horibaAPDA351/view'},
        'Horiba APDA351E':            {                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: https://pdfs.semanticscholar.org/fc4c/3b522b04166381468138e590ac4b866759d3.pdf'},
        'Horiba APDA371':             {'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'], 'documented_lower_limit_of_detection':'4.8', 'documented_upper_limit_of_detection':'10000',                                                                                                                                   'documented_resolution':'0.1',                                                                        'documented_flow_rate':'16.666667',       'instrument_manual_name':'Horiba_APDA371_Specs.pdf'},
        'Horiba APDA3750A':           {'measured_parameters':['PM2.5'],        'qa_accepted_parameters':['PM2.5'],                                                     'documented_upper_limit_of_detection':'10000',                        'documented_precision':'2.0%',               'documented_accuracy':'5.0%',                                                                                                   'documented_span_drift':'3.0%/day', 'documented_flow_rate':'16.7',            'instrument_further_details':'Specifications here: https://www-horiba-com.translate.goog/jp/process-environmental/products-jp/ambient/details_09/apda-3750a-ambient-dust-monitor-pm2-5-19081/?_x_tr_sl=ja&_x_tr_tl=en&_x_tr_hl=en&_x_tr_pto=sc'},
        'Horiba FH621-R':             {                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: https://scindeks-clanci.ceon.rs/data/pdf/1451-9372/2008/1451-93720801005J.pdf'},
        'Kimoto 168S':                {                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: https://www.dri.edu/images/stories/editors/eafeditor/Watsonetal2010SCAQMDFugDustReport.pdf'},
        'Kimoto PM-712':              {'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'],                                              'documented_upper_limit_of_detection':{'PM10':'5000','PM2.5':'1000'}, 'documented_precision':'2.0%',               'documented_accuracy':'5.0%',                                                                'documented_zero_drift':'5.0/day', 'documented_span_drift':'3.0%/day', 'documented_flow_rate':'16.7',            'instrument_manual_name':'Kimoto_PM-712_Specs.pdf'},
        'Kimoto PM-717':              {                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: https://www.kimoto-electric.co.jp/topics/article/pm25test.html'},
        'Kimoto SPM-613D':            {'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'],                                                                                                                                                                                                                                                                                                                                                                               'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: https://www3.epa.gov/ttnamti1/files/2006conference/tisch.pdf'},
        'Kimoto SPM-613F':            {                                                                                   'documented_lower_limit_of_detection':'2.0', 'documented_upper_limit_of_detection':'1000',                         'documented_precision':'2.0%',               'documented_accuracy':'3.0%',                                                                'documented_zero_drift':'3.0/day', 'documented_span_drift':'2.0%/day', 'documented_flow_rate':'16.666667',       'instrument_manual_name':'Kimoto_SPM-613-F_Manual.pdf'},
        'Met One BAM1020':            {'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'], 'documented_lower_limit_of_detection':'4.8', 'documented_upper_limit_of_detection':'10000',                                                                                                   'documented_uncertainty':'1.0', 'documented_resolution':'0.24',                                                                       'documented_flow_rate':'16.7',            'instrument_manual_name':'Met_One_BAM1020_Manual.pdf'}, 
        'Met One BAM1022':            {                                                                                                                                                                                                                                                                                                                                                                                                                       'documented_flow_rate':'16.7',            'instrument_manual_name':'Met_One_BAM1022_Specs.pdf'}, 
        'Met One E-BAM':              {                                                                                                                                                                                                                                                                                                                                                                                                                       'documented_flow_rate':'16.7',            'instrument_manual_name':'Met_One_E-BAM_Specs.pdf'}, 
        'Met One E-BAM Plus':         {                                                                                                                                                                                                                                                                                                                                                                                                                       'documented_flow_rate':'16.7',            'instrument_manual_name':'Met_One_E-BAM_Plus_Specs.pdf'},         
        'OPSIS SM200':                {'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'], 'documented_lower_limit_of_detection':'0.5', 'documented_upper_limit_of_detection':'1000',                                                                                                                                                                                                                                          'documented_flow_rate':'16.666667',       'instrument_manual_name':'OPSIS_SM200_Specs'},
        'Sibata BAM102':              {                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: https://www.nrc.gov/docs/ML1104/ML110400420.pdf'},
        'Teledyne 602 BetaPlus':      {                                                                                   'documented_lower_limit_of_detection':'2.0', 'documented_upper_limit_of_detection':'2000',                                                                                                                                    'documented_resolution':'0.1',                                                                        'documented_flow_rate':'16.6667-41.6667', 'instrument_manual_name':'Teledyne_602_BetaPlus_Manual.pdf'},
        'Thermo 5014i':               {                                                                                   'documented_lower_limit_of_detection':'4.0', 'documented_upper_limit_of_detection':'10000',                        'documented_precision':'2.0<80.0 5.0>=80.0', 'documented_accuracy':'5.0%',                                 'documented_resolution':'0.1',                                                                        'documented_flow_rate':'16.666667',       'instrument_manual_name':'Thermo_5014i_Manual.pdf'},
        'Thermo FH62 C14':            {                                                                                   'documented_lower_limit_of_detection':'4.0', 'documented_upper_limit_of_detection':'10000',                        'documented_precision':'2.0',                                                                              'documented_resolution':'1.0',                                                                        'documented_flow_rate':'0.6',             'instrument_manual_name':'Thermo_FH62_C14_Specs.pdf'},
        'Thermo Andersen FH62 I-R':   {'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'], 'documented_lower_limit_of_detection':'3.0', 'documented_upper_limit_of_detection':'10000',                                                                     'documented_accuracy':'1.0',                                  'documented_resolution':'1.0',                                                                        'documented_flow_rate':'16.666667',       'instrument_further_details':'Instrument was called Eberline FH62 I-R previously, specifications for that instrument are here: https://www.breckland.gov.uk/media/1176/The-Detailed-Assessment-for-PM10-2004/pdf/The_Detailed_Assessment_for_PM10_2004.pdf'},
        'Verewa F701':                {'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'], 'documented_lower_limit_of_detection':'5.0', 'documented_upper_limit_of_detection':'10000',                                                                                                                                                                                                                                         'documented_flow_rate':'16.666667',       'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: http://www.ecomonitoring.pl/pylomierz%20-%20imisja.html'},
        'Verewa F701-20':             {'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'], 'documented_lower_limit_of_detection':'1.0', 'documented_upper_limit_of_detection':'10000',                                                                     'documented_accuracy':'2.0%',                                                                                                                                       'documented_flow_rate':'16.666667',       'instrument_manual_name':'Verewa_F701-20_Manual.pdf'},
        'Verewa F703':                {'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'],                                                                                                                                                                                                                                                                                                                                                                               'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: https://books.google.es/books?id=zx6YBwAAQBAJ&pg=SL13-PA29&lpg=SL13-PA29&dq=Verewa+F703&source=bl&ots=zS_cDXxd_x&sig=ACfU3U2QDSpJnu-ns_aORl3mN9x_SrGYeQ&hl=en&sa=X&ved=2ahUKEwjuhNf38IbgAhXxoXEKHeP6BDoQ6AEwC3oECAQQAQ#v=onepage&q=Verewa%20F703&f=false'},
        'Wedding and Associates Beta':{                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'instrument_further_details':'No manual or specifications available but confirmation it is uses beta-attenuation here: https://uk-air.defra.gov.uk/assets/documents/reports/empire/quarg/quarg_11.pdf'}
    }
},

#------------------------------------
#nephelometry (NP)

#Nephelometry is based on the bulk scattering of visible light by an air sample. It is a mature technology and provides reliable measurements of the spectral (aerosol) scattering coefficient of the air (unit Mm‐1).
#An integrating nephelometer combines the light from all scattering angles to produce a spectral scattering coefficient for the wavelength of light used.
#The signal obtained by integrating nephelometer measurements is closely proportional to the cosine weighted scattering function, which is equal to the scattering part of the Lambert‐Beer extinction coefficient.
#The scattering coefficient is not linearly proportional to aerosol mass concentration, but a more complex function of aerosol composition, shape, density, size distribution, and index of refraction. 
#Thus to convert the scattering coefficient to a mass concentration, assumptions are needed to be made regarding the type of aerosol present, and environmental factors (k-factors).
#K-factors are typically determined by the user by running the nephelometer next to an air sampling pump and comparing results.
#The correction factors necessary may have a significant seasonal and spatial variability (Chow et al. 2006); hence nephelometry is best suited in environments with a rather stable aerosol composition.

'nephelometry':{'abbreviation':'NP', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 
    'instruments':{
        'Ambilabs 2-WIN':                {'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'documented_lower_limit_of_detection':{'AERO_LIGHT_SCAT_COEFF_450nm':'0.3','AERO_LIGHT_SCAT_COEFF_525nm':'0.3','AERO_LIGHT_SCAT_COEFF_635nm':'0.3','PM10_AERO_LIGHT_SCAT_COEFF_450nm':'0.3','PM10_AERO_LIGHT_SCAT_COEFF_525nm':'0.3','PM10_AERO_LIGHT_SCAT_COEFF_635nm':'0.3','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm':'0.3','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm':'0.3','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm':'0.3','PM1_AERO_LIGHT_SCAT_COEFF_450nm':'0.3','PM1_AERO_LIGHT_SCAT_COEFF_525nm':'0.3','PM1_AERO_LIGHT_SCAT_COEFF_635nm':'0.3'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              'documented_upper_limit_of_detection':{'PM10':'100000.0','PM2.5':'100000.0','PM1':'100000.0','AERO_LIGHT_SCAT_COEFF_450nm':'20000.0','AERO_LIGHT_SCAT_COEFF_525nm':'20000.0','AERO_LIGHT_SCAT_COEFF_635nm':'20000.0','PM10_AERO_LIGHT_SCAT_COEFF_450nm':'20000.0','PM10_AERO_LIGHT_SCAT_COEFF_525nm':'20000.0','PM10_AERO_LIGHT_SCAT_COEFF_635nm':'20000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm':'20000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm':'20000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm':'20000.0','PM1_AERO_LIGHT_SCAT_COEFF_450nm':'20000.0','PM1_AERO_LIGHT_SCAT_COEFF_525nm':'20000.0','PM1_AERO_LIGHT_SCAT_COEFF_635nm':'20000.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             'documented_flow_rate':'5.0',    'instrument_manual_name':'Ambilabs_2-WIN_Specs.pdf'},                                                                                      
        'Ecotech 9003':                  {'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_470nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_630nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_470nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_630nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_470nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_630nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_470nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_630nm'],                                                                                                                                                                                        'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_470nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_630nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_470nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_630nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_470nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_630nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_470nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_630nm'],                                                                                                                                                                                        'documented_lower_limit_of_detection':{'AERO_LIGHT_SCAT_COEFF_450nm':'0.4','AERO_LIGHT_SCAT_COEFF_470nm':'0.4','AERO_LIGHT_SCAT_COEFF_520nm':'0.4','AERO_LIGHT_SCAT_COEFF_525nm':'0.4','AERO_LIGHT_SCAT_COEFF_630nm':'0.4','PM10_AERO_LIGHT_SCAT_COEFF_450nm':'0.4','PM10_AERO_LIGHT_SCAT_COEFF_470nm':'0.4','PM10_AERO_LIGHT_SCAT_COEFF_520nm':'0.4','PM10_AERO_LIGHT_SCAT_COEFF_525nm':'0.4','PM10_AERO_LIGHT_SCAT_COEFF_630nm':'0.4','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm':'0.4','PM2.5_AERO_LIGHT_SCAT_COEFF_470nm':'0.4','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm':'0.4','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm':'0.4','PM2.5_AERO_LIGHT_SCAT_COEFF_630nm':'0.4','PM1_AERO_LIGHT_SCAT_COEFF_450nm':'0.4','PM1_AERO_LIGHT_SCAT_COEFF_470nm':'0.4','PM1_AERO_LIGHT_SCAT_COEFF_520nm':'0.4','PM1_AERO_LIGHT_SCAT_COEFF_525nm':'0.4','PM1_AERO_LIGHT_SCAT_COEFF_630nm':'0.4'},                                                                                                                                                                                                                'documented_upper_limit_of_detection':{'AERO_LIGHT_SCAT_COEFF_450nm':'2000.0','AERO_LIGHT_SCAT_COEFF_470nm':'2000.0','AERO_LIGHT_SCAT_COEFF_520nm':'2000.0','AERO_LIGHT_SCAT_COEFF_525nm':'2000.0','AERO_LIGHT_SCAT_COEFF_630nm':'2000.0','PM10_AERO_LIGHT_SCAT_COEFF_450nm':'2000.0','PM10_AERO_LIGHT_SCAT_COEFF_470nm':'2000.0','PM10_AERO_LIGHT_SCAT_COEFF_520nm':'2000.0','PM10_AERO_LIGHT_SCAT_COEFF_525nm':'2000.0','PM10_AERO_LIGHT_SCAT_COEFF_630nm':'2000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm':'2000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_470nm':'2000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm':'2000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm':'2000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_630nm':'2000.0','PM1_AERO_LIGHT_SCAT_COEFF_450nm':'2000.0','PM1_AERO_LIGHT_SCAT_COEFF_470nm':'2000.0','PM1_AERO_LIGHT_SCAT_COEFF_520nm':'2000.0','PM1_AERO_LIGHT_SCAT_COEFF_525nm':'2000.0','PM1_AERO_LIGHT_SCAT_COEFF_630nm':'2000.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Ecotech_9003_Specs.pdf', 'instrument_further_details':'Manual suggests measurement at 520nm and optional measurement at 470nm and 630nm upon request. However, have reported wavelengths at 450nm and 525nm also.'},                                                                                    
        'Ecotech Aurora 1000':           {'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'documented_lower_limit_of_detection':{'AERO_LIGHT_SCAT_COEFF_450nm':'0.25','AERO_LIGHT_SCAT_COEFF_525nm':'0.25','AERO_LIGHT_SCAT_COEFF_635nm':'0.25','PM10_AERO_LIGHT_SCAT_COEFF_450nm':'0.25','PM10_AERO_LIGHT_SCAT_COEFF_525nm':'0.25','PM10_AERO_LIGHT_SCAT_COEFF_635nm':'0.25','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm':'0.25','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm':'0.25','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm':'0.25','PM1_AERO_LIGHT_SCAT_COEFF_450nm':'0.25','PM1_AERO_LIGHT_SCAT_COEFF_525nm':'0.25','PM1_AERO_LIGHT_SCAT_COEFF_635nm':'0.25'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  'documented_upper_limit_of_detection':{'AERO_LIGHT_SCAT_COEFF_450nm':'2000.0','AERO_LIGHT_SCAT_COEFF_525nm':'2000.0','AERO_LIGHT_SCAT_COEFF_635nm':'2000.0','PM10_AERO_LIGHT_SCAT_COEFF_450nm':'2000.0','PM10_AERO_LIGHT_SCAT_COEFF_525nm':'2000.0','PM10_AERO_LIGHT_SCAT_COEFF_635nm':'2000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm':'2000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm':'2000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm':'2000.0','PM1_AERO_LIGHT_SCAT_COEFF_450nm':'2000.0','PM1_AERO_LIGHT_SCAT_COEFF_525nm':'2000.0','PM1_AERO_LIGHT_SCAT_COEFF_635nm':'2000.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               'documented_flow_rate':'5.0',    'instrument_manual_name':'Ecotech_Aurora_1000_Manual.pdf'},                                                                                      
        'Ecotech Aurora 3000':           {'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm'], 'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm'], 'documented_lower_limit_of_detection':{'AERO_LIGHT_BACKSCAT_COEFF_450nm':'0.1','AERO_LIGHT_SCAT_COEFF_450nm':'0.1','AERO_LIGHT_BACKSCAT_COEFF_525nm':'0.1','AERO_LIGHT_SCAT_COEFF_525nm':'0.1','AERO_LIGHT_BACKSCAT_COEFF_635nm':'0.1','AERO_LIGHT_SCAT_COEFF_635nm':'0.1','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm':'0.1','PM10_AERO_LIGHT_SCAT_COEFF_450nm':'0.1','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm':'0.1','PM10_AERO_LIGHT_SCAT_COEFF_525nm':'0.1','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm':'0.1','PM10_AERO_LIGHT_SCAT_COEFF_635nm':'0.1','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm':'0.1','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm':'0.1','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm':'0.1','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm':'0.1','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm':'0.1','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm':'0.1','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm':'0.1','PM1_AERO_LIGHT_SCAT_COEFF_450nm':'0.1','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm':'0.1','PM1_AERO_LIGHT_SCAT_COEFF_525nm':'0.1','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm':'0.1','PM1_AERO_LIGHT_SCAT_COEFF_635nm':'0.1'}, 'documented_upper_limit_of_detection':{'AERO_LIGHT_BACKSCAT_COEFF_450nm':'20000.0','AERO_LIGHT_SCAT_COEFF_450nm':'20000.0','AERO_LIGHT_BACKSCAT_COEFF_525nm':'20000.0','AERO_LIGHT_SCAT_COEFF_525nm':'20000.0','AERO_LIGHT_BACKSCAT_COEFF_635nm':'20000.0','AERO_LIGHT_SCAT_COEFF_635nm':'20000.0','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm':'20000.0','PM10_AERO_LIGHT_SCAT_COEFF_450nm':'20000.0','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm':'20000.0','PM10_AERO_LIGHT_SCAT_COEFF_525nm':'20000.0','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm':'20000.0','PM10_AERO_LIGHT_SCAT_COEFF_635nm':'20000.0','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm':'20000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm':'20000.0','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm':'20000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm':'20000.0','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm':'20000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm':'20000.0','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm':'20000.0','PM1_AERO_LIGHT_SCAT_COEFF_450nm':'20000.0','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm':'20000.0','PM1_AERO_LIGHT_SCAT_COEFF_525nm':'20000.0','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm':'20000.0','PM1_AERO_LIGHT_SCAT_COEFF_635nm':'20000.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'documented_flow_rate':'5.0',    'instrument_manual_name':'Ecotech_Aurora_3000_Manual.pdf'},
        'Ecotech Aurora 4000':           {'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm'], 'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm'], 'documented_lower_limit_of_detection':{'AERO_LIGHT_BACKSCAT_COEFF_450nm':'0.3','AERO_LIGHT_SCAT_COEFF_450nm':'0.3','AERO_LIGHT_BACKSCAT_COEFF_525nm':'0.3','AERO_LIGHT_SCAT_COEFF_525nm':'0.3','AERO_LIGHT_BACKSCAT_COEFF_635nm':'0.3','AERO_LIGHT_SCAT_COEFF_635nm':'0.3','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm':'0.3','PM10_AERO_LIGHT_SCAT_COEFF_450nm':'0.3','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm':'0.3','PM10_AERO_LIGHT_SCAT_COEFF_525nm':'0.3','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm':'0.3','PM10_AERO_LIGHT_SCAT_COEFF_635nm':'0.1','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm':'0.3','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm':'0.3','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm':'0.3','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm':'0.3','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm':'0.3','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm':'0.3','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm':'0.3','PM1_AERO_LIGHT_SCAT_COEFF_450nm':'0.3','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm':'0.3','PM1_AERO_LIGHT_SCAT_COEFF_525nm':'0.3','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm':'0.3','PM1_AERO_LIGHT_SCAT_COEFF_635nm':'0.3'}, 'documented_upper_limit_of_detection':{'AERO_LIGHT_BACKSCAT_COEFF_450nm':'2000.0','AERO_LIGHT_SCAT_COEFF_450nm':'2000.0','AERO_LIGHT_BACKSCAT_COEFF_525nm':'2000.0','AERO_LIGHT_SCAT_COEFF_525nm':'2000.0','AERO_LIGHT_BACKSCAT_COEFF_635nm':'2000.0','AERO_LIGHT_SCAT_COEFF_635nm':'2000.0','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm':'2000.0','PM10_AERO_LIGHT_SCAT_COEFF_450nm':'2000.0','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm':'2000.0','PM10_AERO_LIGHT_SCAT_COEFF_525nm':'2000.0','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm':'2000.0','PM10_AERO_LIGHT_SCAT_COEFF_635nm':'2000.0','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm':'2000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm':'2000.0','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm':'2000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm':'2000.0','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm':'2000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm':'2000.0','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm':'2000.0','PM1_AERO_LIGHT_SCAT_COEFF_450nm':'2000.0','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm':'2000.0','PM1_AERO_LIGHT_SCAT_COEFF_525nm':'2000.0','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm':'2000.0','PM1_AERO_LIGHT_SCAT_COEFF_635nm':'2000.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              'documented_flow_rate':'5.0',    'instrument_manual_name':'Ecotech_Aurora_4000_Manual.pdf'},                                                                                                                                                                                                                                 
        'Meteorology Research Inc. 4-W': {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               'instrument_further_details':'No manual or specifications available', 'measuring_assumption':True},
        'Meteorology Research Inc. 1550':{                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               'instrument_further_details':'No manual or specifications available but confirmation it is uses nephelometry here: https://www.rti.org/sites/default/files/resources/bk-0003-1109-chapter09.pdf'},
        'MS Electron 3W-02':             {'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm'], 'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             'instrument_further_details':'No manual or specifications available but confirmation it is uses nephelometry here: http://carg.atmos.washington.edu/sys/research/CLAMS/TabCLAMS.pdf'},
        'Optec NGN-2':                   {'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    'documented_lower_limit_of_detection':{'AERO_LIGHT_SCAT_COEFF_550nm':'10.0','PM10_AERO_LIGHT_SCAT_COEFF_550nm':'10.0','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm':'10.0','PM1_AERO_LIGHT_SCAT_COEFF_550nm':'10.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        'documented_upper_limit_of_detection':{'AERO_LIGHT_SCAT_COEFF_550nm':'24000.0','PM10_AERO_LIGHT_SCAT_COEFF_550nm':'24000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm':'24000.0','PM1_AERO_LIGHT_SCAT_COEFF_550nm':'24000.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'documented_accuracy':{'AERO_LIGHT_SCAT_COEFF_550nm':'10.0%','PM10_AERO_LIGHT_SCAT_COEFF_550nm':'10.0%','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm':'10.0%','PM1_AERO_LIGHT_SCAT_COEFF_550nm':'10.0%'},                                  'instrument_manual_name':'Optec_NGN-2_Manual.pdf'}, 
        'Radiance Research 903':         {'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_SCAT_COEFF_530nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_SCAT_COEFF_530nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             'documented_lower_limit_of_detection':{'AERO_LIGHT_SCAT_COEFF_530nm':'1.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'documented_upper_limit_of_detection':{'AERO_LIGHT_SCAT_COEFF_530nm':'1000.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'instrument_manual_name':'Radiance_Research_903_Manual.pdf'},                                                                                    
        'TSI 3563':                      {'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm'], 'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    'documented_upper_limit_of_detection':{'AERO_LIGHT_BACKSCAT_COEFF_450nm':'20000.0','AERO_LIGHT_SCAT_COEFF_450nm':'20000.0','AERO_LIGHT_BACKSCAT_COEFF_550nm':'20000.0','AERO_LIGHT_SCAT_COEFF_550nm':'20000.0','AERO_LIGHT_BACKSCAT_COEFF_700nm':'20000.0','AERO_LIGHT_SCAT_COEFF_700nm':'20000.0','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm':'20000.0','PM10_AERO_LIGHT_SCAT_COEFF_450nm':'20000.0','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm':'20000.0','PM10_AERO_LIGHT_SCAT_COEFF_550nm':'20000.0','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm':'20000.0','PM10_AERO_LIGHT_SCAT_COEFF_700nm':'20000.0','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm':'20000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm':'20000.0','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm':'20000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm':'20000.0','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm':'20000.0','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm':'20000.0','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm':'20000.0','PM1_AERO_LIGHT_SCAT_COEFF_450nm':'20000.0','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm':'20000.0','PM1_AERO_LIGHT_SCAT_COEFF_550nm':'20000.0','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm':'20000.0','PM1_AERO_LIGHT_SCAT_COEFF_700nm':'20000.0'}, 'documented_resolution':{'AERO_LIGHT_BACKSCAT_COEFF_450nm':'0.3','AERO_LIGHT_SCAT_COEFF_450nm':'0.3','AERO_LIGHT_BACKSCAT_COEFF_550nm':'0.2','AERO_LIGHT_SCAT_COEFF_550nm':'0.2','AERO_LIGHT_BACKSCAT_COEFF_700nm':'0.2','AERO_LIGHT_SCAT_COEFF_700nm':'0.2','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm':'0.3','PM10_AERO_LIGHT_SCAT_COEFF_450nm':'0.3','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm':'0.2','PM10_AERO_LIGHT_SCAT_COEFF_550nm':'0.2','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm':'0.2','PM10_AERO_LIGHT_SCAT_COEFF_700nm':'0.2','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm':'0.3','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm':'0.3','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm':'0.2','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm':'0.2','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm':'0.2','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm':'0.2','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm':'0.3','PM1_AERO_LIGHT_SCAT_COEFF_450nm':'0.3','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm':'0.2','PM1_AERO_LIGHT_SCAT_COEFF_550nm':'0.2','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm':'0.2','PM1_AERO_LIGHT_SCAT_COEFF_700nm':'0.2'},                                                                                                                                                                                                 'documented_flow_rate':'20-200', 'instrument_manual_name':'TSI_3563_Specs.pdf'}
    }
},

#------------------------------------
#nephelometry - laser spectrometry (NP-LS)

#In order to accurately calculate k-factors to convert the aerosol scattering coeffient to a mass concentration, a laser spectrometry can be used simultaneously to measure the size and number concentration of airborne particles.
#The spectrometry works by using intensity of light scattered from a laser to measure the particle size, at mutiple defined size bins simultaneously.

'nephelometry - laser spectrometry':{'abbreviation':'NP-LS', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 
    'instruments':{
        'Airpointer-PM':{'documented_lower_limit_of_detection':'1.0', 'documented_upper_limit_of_detection':'1000.0', 'documented_precision':'1.0', 'documented_zero_drift':'1.0/day', 'documented_span_drift':'1.0%/day', 'documented_flow_rate':'2.0', 'instrument_manual_name':'Airpointer_Specs.pdf'}
    }
},

#------------------------------------
#light scattering photometry (LSP)
        
#Light scattering photometers do not integrate the light from all scattered angles, but only view a portion of the available scattering angles. 
#Like nephelometry, it produces a bulk scattering coefficient of an air sample with similar time resolution and range. 
#In general they are calibrated with a standard test aerosol such as Arizona Test Dust, but typically these test aerosols have significantly different optical properties from real ambient air aerosols.
#Again, therefore, assumptions about the type of aerosol present, and environmental factors, are needed to convert the scattering coefficient to a mass concentration, and uncertainties tend to be higher than for nephelometers.
#See here for distinction between nephelometry and light scattering photometry: https://db-airmontech.jrc.ec.europa.eu/download/PM_Mass_MMTI_OpticalDevices.pdf

'light scattering photometry':{'abbreviation':'LSP', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 
    'instruments':{
        'Comde Derenda APM-2':{'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'],                                                                                                                                                                                                                                                                         'instrument_manual_name':'Comde_Derenda_APM-2_Specs.pdf'},
        'Horiba APDA372':     {                                                                                                                                 'documented_upper_limit_of_detection':'10000.0',                                                                                                                                      'documented_flow_rate':'4.8',       'instrument_manual_name':'Horiba_APDA372_Specs.pdf'},
        'Unitec LSPM10':      {'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'],                                               'documented_upper_limit_of_detection':'20000.0', 'documented_accuracy':'1.0%', 'documented_zero_drift':'0.1%/day', 'documented_span_drift':'0.1%/day', 'documented_resolution':'0.1', 'documented_flow_rate':'15.0-40.0', 'instrument_manual_name':'Unitec_LSPM10_Manual.pdf'}
    }
},

#------------------------------------
#optical particle counter (OPC)

#Optical particle counters differ from nephelometry and light scattering photometry in being based on measuring the light scattered by individual particles. 
#The light scattered by individual particles is picked up as (typically) a discrete signal, so that the size of each particle can be gauged from the intensity of the signal, while the number concentration of particles can be gauged from the number of signals and the air flow rate through the light beam volume. 
#In general, therefore, a size spectrum of particles will be obtained, so that information about specific size fractions, such as PM10 and PM2.5, can be obtained simultaneously. 
#These can then be converted to mass concentrations by making assumptions about the particle density.
#Uncertainties arise from the determination of the size of the volume of air being examined by the laser, scattering from more than one particle at the same time (coincidence), and the conversions from intensity to size and then to mass. 
#Nevertheless, performance can be comparable with other automatic methods for PM10 and PM2.5.

'optical particle counter':{'abbreviation':'OPC', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 
    'instruments':{
        'Grimm 107':       {                                                                                                                                                             'documented_flow_rate':'1.2',  'instrument_manual_name':'Grimm_107_Specs.pdf', 'measuring_instrument_further_details':'Portable version of Grimm 180.'},
        'Grimm 180':       {                                                                                                                                                             'documented_flow_rate':'1.2',  'instrument_manual_name':'Grimm_180_Manual.pdf'},
        'Grimm 190':       {                                                                                                                                                                                            'instrument_further_details':'No manual or specifications available but confirmation it is uses optical particle counter here: https://core.ac.uk/download/pdf/52132153.pdf'},
        'Grimm 365':       {                                                                                                                                                             'documented_flow_rate':'1.2',  'instrument_manual_name':'Grimm_365_Specs.pdf'},
        'Palas Fidas 200': {                                              'documented_upper_limit_of_detection':'10000.0',                                                               'documented_flow_rate':'5.0',  'instrument_manual_name':'Palas_Fidas_200_Specs.pdf'},
        'Palas Fidas 200E':{                                              'documented_upper_limit_of_detection':'10000.0',                                                               'documented_flow_rate':'5.0',  'instrument_manual_name':'Palas_Fidas_200_Specs.pdf'},
        'Palas Fidas 200S':{                                              'documented_upper_limit_of_detection':'10000.0',                                                               'documented_flow_rate':'5.0',  'instrument_manual_name':'Palas_Fidas_200_Specs.pdf'},
        'Teledyne 640':    {'documented_lower_limit_of_detection':'0.1',  'documented_upper_limit_of_detection':'10000.0', 'documented_precision':'0.5',                                 'documented_flow_rate':'5.0',  'instrument_manual_name':'Teledyne_640_Manual.pdf'},
        'Teledyne 640x':   {'documented_lower_limit_of_detection':'0.1',  'documented_upper_limit_of_detection':'10000.0', 'documented_precision':'0.5',                                 'documented_flow_rate':'16.7', 'instrument_manual_name':'Teledyne_640x_Manual.pdf'},
        'Turnkey Osiris':  {'documented_lower_limit_of_detection':'0.01', 'documented_upper_limit_of_detection':'6000.0',                                 'documented_resolution':'0.1', 'documented_flow_rate':'0.6',  'instrument_manual_name':'Turnkey_Osiris_Specs.pdf'}
        
    }
},

#------------------------------------
#beta-attenuation - nephelometry (BA-NP)

#Hybrid of beta-attenuation and nephelometry methods enhancing the overall performance.
#Has accuracy of the beta-attenuation method with high time resolution of a nephelometer.
#see https://www3.epa.gov/ttnamti1/files/2006conference/goohssharp.pdf for more information

'beta-attenuation - nephelometry':{'abbreviation':'BA-NP', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 
    'instruments':{
        'Thermo 5030 SHARP':{'measured_parameters':['PM10','PM2.5'], 'qa_accepted_parameters':['PM10','PM2.5'], 'documented_lower_limit_of_detection':'0.5', 'documented_upper_limit_of_detection':'10000', 'documented_precision':'2.0<80.0 5.0>=80.0', 'documented_accuracy':'5.0%', 'documented_span_drift':'0.02%/day', 'documented_flow_rate':'16.666667', 'instrument_manual_name':'Thermo_5030_SHARP_Specs.pdf'}
    }
},

#------------------------------------
#differential mobility particle sizer (DMPS)

#Optical techniques are limited in being unable to detect particles smaller than 50-100nm, and are also susceptible to sizing errors resulting from variations in particle shape and refractive index. 
#Differential mobility techniques are able to measure much smaller particles, down to 2.5nm and are not sensitive to differences in refractive index, but are not able to achieve the same temporal resolution as optical instruments, and are also sensitive to variations in particle shape.
#The Differential Mobility Particle Sizer couples a differential mobility analyzer (DMA), which classifies charged particles according to their mobility in an electric field, and a condensation particle counter (CPC) to count particles of a specific mobility. 
#In the atmosphere most aerosol are charged, often with multiple elementary charges, in order for classification by electrical mobility to be meaningful in terms of particle size, the aerosol must be given a known charge distribution. 
#This is done by exposing the particles to a stream of beta radiation, resulting in a charge distribution known as a Fuchs distribution. 
#This distribution has known fractions of the aerosol population which are singly, doubly and triply charged etc. 
#When viewing the output from a DMA set to select a single electrical mobility with an optical particle counter, multiply charged peaks can clearly be seen. 
#Clearly when performing a scan, the presence of multiply charged particles has the potential to affect the results. 
#This is generally dealt with by physically limiting the maximum particle size which may enter the DMA (this is usually done with an impactor) to just above the maximum size which is to be scanned. 
#Thus there can be no multiply charged particles present in the largest size bins. In turn multiply charged particles may be removed from smaller size bins based on the known charge distribution and the measured number concentration of these larger particles.

'differential mobility particle sizer':{'abbreviation':'DMPS', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 
    'instruments':{},
},

#------------------------------------
#scanning mobility particle sizer (SMPS)

#The Scanning Mobility Particle Sizer is based on the principal of the mobility of a charged particle in an electric field. 
#Particles entering the system are neutralized (using a radioactive source) such that they have a Fuchs equilibrium charge distribution. 
#They then enter a Differential Mobility Analyser (DMA) where the aerosol is classified according to electrical mobility, with only particles of a narrow range of mobility exiting through the output slit. 
#This monodisperse distribution then goes to a Condensation Particle Counter which determines the particle concentration at that size. 
#The DMA consists of a cylinder, with a negatively charged rod at the center, the main flow through the DMA is particle free 'sheath' air. 
#It is important that this flow is laminar. The particle flow is injected at the outside edge of the DMA, particles with a positive charge move accross the sheath flow towards the central rod, at a rate determined by their electrical mobility. 
#Particles of a given mobility exit through the sample slit at the top of the DMA, while all other particles exit with the exhaust flow. The size of particle exiting through the slit being determined by the particles size, charge, central rod voltage, and flow within the DMA. 
#By (in the case of the SMPS) exponentially scanning the voltage on the central rod, a full particle size distribution is built up. 
#SMPS is a variation of DMPS, with some advantages. A major advantage is it’s ease of portability and significantly lower power consumption. It also can produce aerosol spectra with a much higher temporal resolution, though the sampling statistics suffer especially if fast scans are carried out under clean ambient conditions. 

'scanning mobility particle sizer':{'abbreviation':'SMPS', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 'qa_accepted_parameters':['PM10','PM2.5','PM1','AERO_LIGHT_BACKSCAT_COEFF_450nm','AERO_LIGHT_BACKSCAT_COEFF_520nm','AERO_LIGHT_BACKSCAT_COEFF_525nm','AERO_LIGHT_BACKSCAT_COEFF_530nm','AERO_LIGHT_BACKSCAT_COEFF_532nm','AERO_LIGHT_BACKSCAT_COEFF_550nm','AERO_LIGHT_BACKSCAT_COEFF_635nm','AERO_LIGHT_BACKSCAT_COEFF_700nm','AERO_LIGHT_BACKSCAT_COEFF_850nm','AERO_LIGHT_SCAT_COEFF_450nm','AERO_LIGHT_SCAT_COEFF_520nm','AERO_LIGHT_SCAT_COEFF_525nm','AERO_LIGHT_SCAT_COEFF_530nm','AERO_LIGHT_SCAT_COEFF_532nm','AERO_LIGHT_SCAT_COEFF_550nm','AERO_LIGHT_SCAT_COEFF_635nm','AERO_LIGHT_SCAT_COEFF_700nm','AERO_LIGHT_SCAT_COEFF_850nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM10_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM10_AERO_LIGHT_SCAT_COEFF_450nm','PM10_AERO_LIGHT_SCAT_COEFF_520nm','PM10_AERO_LIGHT_SCAT_COEFF_525nm','PM10_AERO_LIGHT_SCAT_COEFF_530nm','PM10_AERO_LIGHT_SCAT_COEFF_532nm','PM10_AERO_LIGHT_SCAT_COEFF_550nm','PM10_AERO_LIGHT_SCAT_COEFF_635nm','PM10_AERO_LIGHT_SCAT_COEFF_700nm','PM10_AERO_LIGHT_SCAT_COEFF_850nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM2.5_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM2.5_AERO_LIGHT_SCAT_COEFF_450nm','PM2.5_AERO_LIGHT_SCAT_COEFF_520nm','PM2.5_AERO_LIGHT_SCAT_COEFF_525nm','PM2.5_AERO_LIGHT_SCAT_COEFF_530nm','PM2.5_AERO_LIGHT_SCAT_COEFF_532nm','PM2.5_AERO_LIGHT_SCAT_COEFF_550nm','PM2.5_AERO_LIGHT_SCAT_COEFF_635nm','PM2.5_AERO_LIGHT_SCAT_COEFF_700nm','PM2.5_AERO_LIGHT_SCAT_COEFF_850nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_450nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_520nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_525nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_530nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_532nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_550nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_635nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_700nm','PM1_AERO_LIGHT_BACKSCAT_COEFF_850nm','PM1_AERO_LIGHT_SCAT_COEFF_450nm','PM1_AERO_LIGHT_SCAT_COEFF_520nm','PM1_AERO_LIGHT_SCAT_COEFF_525nm','PM1_AERO_LIGHT_SCAT_COEFF_530nm','PM1_AERO_LIGHT_SCAT_COEFF_532nm','PM1_AERO_LIGHT_SCAT_COEFF_550nm','PM1_AERO_LIGHT_SCAT_COEFF_635nm','PM1_AERO_LIGHT_SCAT_COEFF_700nm','PM1_AERO_LIGHT_SCAT_COEFF_850nm'], 
    'instruments':{
        'TROPOS':{'instrument_manual_name':'TROPOS_Manual.pdf'}}
},

#------------------------------------
#thermal analysis (TA)

#Thermal analysis (TA) differentiates between organic carbon (OC) and light-absorbing carbon (LAC) by oxidising collected samples to CO2 at an intermediate temperature and at a high final burn temperature. 
#Because thermal analysis does not yield a true value of LAC, the refractory component it measures takes a different name: elemental carbon (EC). (This has no relationship to a chemist’s definition of “elemental.”) 
#Thermal methods rely on the assumption that organic carbon is volatile, and the strongly light absorbing carbon is stable or refractory at elevated temperatures. 
#It assumes that LAC in a sample is unchanged until exposed to both elevated temperatures and an oxidant, so that carbon driven off at low temperatures or in the absence of oxygen must be OC. 
#The method physically determines OC, EC and total carbon (TC) by measurement of CO2 by infrared detection.
#By heating in stages it is possible to attribute carbon broken down at lower temperatures to OC, and those at higher temperatures to EC.
#Ideally, TA would be a separation-and-detection technique similar to chromatography, but failures of separation have hampered its interpretation.
#Some OC undergoes pyrolysis or charring in the absence of oxygen, forming pyrolyzed carbon (“charring”) instead of volatilizing (Huntzicker et al., 1982). 
#TA applies no correction for this.

'thermal analysis':{'abbreviation':'TA', 'sampling_type':'injection', 'measured_parameters':['C','PM10_C','PM2.5_C','PM1_C','EC','PM10_EC','PM2.5_EC','PM1_EC','OC','PM10_OC','PM2.5_OC','PM1_OC'], 'qa_accepted_parameters':['C','PM10_C','PM2.5_C','PM1_C','EC','PM10_EC','PM2.5_EC','PM1_EC','OC','PM10_OC','PM2.5_OC','PM1_OC'], 
    'instruments':{
        'Rupprecht and Patashnick 5400':{'sampling_type':'continuous injection', 'sample_preparation_type':'filter', 'documented_flow_rate':'16.7', 'instrument_further_details':'No manual or specifications available but confirmation it is uses thermal-optical analysi here: https://archive.epa.gov/nrmrl/archive-etv/web/pdf/01_vr_rupp_5400c.pdf'}        
    }
},

#------------------------------------
#thermal-optical analysis - unknown protocol (TOA-UNK)

#Thermal-optical analysis (TOA) is the most widely used method to measure organic and elemental carbon in the scientific community. 
#It uses the same principals as thermal analysis but builds upon them by using laser monitoring to correct for the transformation of non-absorbing carbon into pyrolytic carbon that absorbs light.
#Additionally, samples are oxidised at various steps on a temperature ramp (rather than just 2), to provide more information for OC/EC split.
#The sample usually becomes blacker as charring occurs in an inert atmosphere, and the blackness decreases as the pyrolyzed carbon burns later in the analysis. 
#This change is detected by monitoring either transmittance or reflectance (sometimes both). When the filter’s transmittance or reflectance returns to the original value, it is assumed that the remaining material is EC. 
#This point is colloquially termed the OC/EC split.
#There have been wide discourse regarding the OC/EC split definition, with different protocols having their own temperature ramps and step durations.  
#There are currently 3 well established protocols for the handling of the apportionment between EC and OC:
# NIOSH 5040 
# IMPROVE-A 
# EUSAAR2 

#TOA instruments can be operated following any of these protocols, therefore for clarity, a separate method is given referencing each of these these protocols
#https://www.sciencedirect.com/science/article/pii/S1309104215302415#bib0020
#https://www.sciencedirect.com/science/article/pii/S1352231018300578#bib3

'thermal-optical analysis - unknown protocol':{'abbreviation':'TOA-UNK', 'sampling_type':'injection', 'measured_parameters':['C','PM10_C','PM2.5_C','PM1_C','EC','PM10_EC','PM2.5_EC','PM1_EC','OC','PM10_OC','PM2.5_OC','PM1_OC'], 'qa_accepted_parameters':['C','PM10_C','PM2.5_C','PM1_C','EC','PM10_EC','PM2.5_EC','PM1_EC','OC','PM10_OC','PM2.5_OC','PM1_OC'], 
    'instruments':{
        'DRI Model 2001':     {'instrument_manual_name':'DRI_Model_2001_Manual.pdf'},
        'Sunset OCEC Model 4':{'instrument_manual_name':'Sunset_OCEC_4_Specs.pdf'},
        'Sunset OCEC Model 5':{'instrument_manual_name':'Sunset_OCEC_5l_Manual.pdf', 'instrument_further_details':'See more details here:http://www.sunlab.com/wp-content/uploads/Lab-Instrument-brochure.pdf'}
    }
},

#------------------------------------
#thermal-optical analysis - NIOSH 5040 (TOA-N)

'thermal-optical analysis - NIOSH 5040':{'abbreviation':'TOA-N', 'sampling_type':'injection', 'measured_parameters':['C','PM10_C','PM2.5_C','PM1_C','EC','PM10_EC','PM2.5_EC','PM1_EC','OC','PM10_OC','PM2.5_OC','PM1_OC'], 'qa_accepted_parameters':['C','PM10_C','PM2.5_C','PM1_C','EC','PM10_EC','PM2.5_EC','PM1_EC','OC','PM10_OC','PM2.5_OC','PM1_OC'], 
    'instruments':{
        'DRI Model 2001':     {'instrument_manual_name':'DRI_Model_2001_Manual.pdf'},
        'Sunset OCEC Model 4':{'instrument_manual_name':'Sunset_OCEC_4_Specs.pdf'},
        'Sunset OCEC Model 5':{'instrument_manual_name':'Sunset_OCEC_5l_Manual.pdf', 'instrument_further_details':'See more details here:http://www.sunlab.com/wp-content/uploads/Lab-Instrument-brochure.pdf'}
    }
},

#------------------------------------
#thermal-optical analysis - IMPROVE-A (TOA-I)

'thermal-optical analysis - IMPROVE-A':{'abbreviation':'TOA-I', 'sampling_type':'injection', 'measured_parameters':['C','PM10_C','PM2.5_C','PM1_C','EC','PM10_EC','PM2.5_EC','PM1_EC','OC','PM10_OC','PM2.5_OC','PM1_OC'], 'qa_accepted_parameters':['C','PM10_C','PM2.5_C','PM1_C','EC','PM10_EC','PM2.5_EC','PM1_EC','OC','PM10_OC','PM2.5_OC','PM1_OC'], 
    'instruments':{
        'DRI Model 2001':     {'instrument_manual_name':'DRI_Model_2001_Manual.pdf'},
        'Sunset OCEC Model 4':{'instrument_manual_name':'Sunset_OCEC_4_Specs.pdf'},
        'Sunset OCEC Model 5':{'instrument_manual_name':'Sunset_OCEC_5l_Manual.pdf', 'instrument_further_details':'See more details here:http://www.sunlab.com/wp-content/uploads/Lab-Instrument-brochure.pdf'}
    }
},

#------------------------------------
#thermal-optical analysis - EUSAAR2 (TOA-E)

'thermal-optical analysis - EUSAAR2':{'abbreviation':'TOA-E', 'sampling_type':'injection', 'measured_parameters':['C','PM10_C','PM2.5_C','PM1_C','EC','PM10_EC','PM2.5_EC','PM1_EC','OC','PM10_OC','PM2.5_OC','PM1_OC'], 'qa_accepted_parameters':['C','PM10_C','PM2.5_C','PM1_C','EC','PM10_EC','PM2.5_EC','PM1_EC','OC','PM10_OC','PM2.5_OC','PM1_OC'], 
    'instruments':{
        'DRI Model 2001':     {'instrument_manual_name':'DRI_Model_2001_Manual.pdf'},
        'Sunset OCEC Model 4':{'instrument_manual_name':'Sunset_OCEC_4_Specs.pdf'},
        'Sunset OCEC Model 5':{'instrument_manual_name':'Sunset_OCEC_5l_Manual.pdf', 'instrument_further_details':'See more details here:http://www.sunlab.com/wp-content/uploads/Lab-Instrument-brochure.pdf'}
    }
},

#------------------------------------
#aethalometer (ATH)

#The aethalometer is an instrument designed for the measurement of black carbon deposit collected continuously deposited on quartz fiber filter (designed by Magee Scientific).
#The 'aethalometer' is the dominant market instrument (designed by Magee Scientific), of which there are multiple versions. 
#It measures the concentration of optically absorbing (‘black’) suspended particulates in a gas colloid stream; commonly visualised as smoke or haze, often seen in ambient air under polluted conditions. 
#The word aethalometer is derived from the Classical Greek verb ‘aethaloun’, meaning ‘to blacken with soot’.
#The gas stream (frequently ambient air) passes through a filter material which traps the suspended particulates, creating a deposit of increasing density. 
#A light beam projected through the deposit is attenuated by those particles which are absorbing (‘black’) rather than scattering (‘white’).
#Measurements are made at successive regular time intervals. The increase in attenuation from one measurement to the next is proportional to the increase in the density of optically absorbing material on the filter: which, in turn, is proportional to the concentration of the material in the sampled air stream. 
#The sample is collected as a spot on a roll of filter tape. 
#When the density of the deposit spot reaches a pre-set limit, the tape advances to a fresh spot and the measurements continue. 
#Measurement of the sample gas flow rate and knowledge of the instrument’s optical and mechanical characteristics permit a calculation of the average concentration of absorbing particles in the gas stream during the sampling period. 
#Aethalometers may operate on time-base periods as rapid as 1 second, providing quasi-real-time data. Comparison of aethalometer data with other physical and chemical analyses allows the output to be expressed as a concentration of black carbon.
#More information here: https://www.tandfonline.com/doi/full/10.1080/027868290901972

'aethalometer':{'abbreviation':'ATH', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_467nm','AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_522nm','AERO_ABS_COEFF_525nm','AERO_ABS_COEFF_528nm','AERO_ABS_COEFF_530nm','AERO_ABS_COEFF_550nm','AERO_ABS_COEFF_565nm','AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_637nm','AERO_ABS_COEFF_652nm','AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_670nm','AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_467nm','PM10_AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_522nm','PM10_AERO_ABS_COEFF_525nm','PM10_AERO_ABS_COEFF_528nm','PM10_AERO_ABS_COEFF_530nm','PM10_AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_565nm','PM10_AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_637nm','PM10_AERO_ABS_COEFF_652nm','PM10_AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_670nm','PM10_AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_467nm','PM2.5_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_522nm','PM2.5_AERO_ABS_COEFF_525nm','PM2.5_AERO_ABS_COEFF_528nm','PM2.5_AERO_ABS_COEFF_530nm','PM2.5_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_565nm','PM2.5_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_637nm','PM2.5_AERO_ABS_COEFF_652nm','PM2.5_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_670nm','PM2.5_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_467nm','PM1_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_522nm','PM1_AERO_ABS_COEFF_525nm','PM1_AERO_ABS_COEFF_528nm','PM1_AERO_ABS_COEFF_530nm','PM1_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_565nm','PM1_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_637nm','PM1_AERO_ABS_COEFF_652nm','PM1_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_670nm','PM1_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_950nm'], 'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_467nm','AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_522nm','AERO_ABS_COEFF_525nm','AERO_ABS_COEFF_528nm','AERO_ABS_COEFF_530nm','AERO_ABS_COEFF_550nm','AERO_ABS_COEFF_565nm','AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_637nm','AERO_ABS_COEFF_652nm','AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_670nm','AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_467nm','PM10_AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_522nm','PM10_AERO_ABS_COEFF_525nm','PM10_AERO_ABS_COEFF_528nm','PM10_AERO_ABS_COEFF_530nm','PM10_AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_565nm','PM10_AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_637nm','PM10_AERO_ABS_COEFF_652nm','PM10_AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_670nm','PM10_AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_467nm','PM2.5_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_522nm','PM2.5_AERO_ABS_COEFF_525nm','PM2.5_AERO_ABS_COEFF_528nm','PM2.5_AERO_ABS_COEFF_530nm','PM2.5_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_565nm','PM2.5_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_637nm','PM2.5_AERO_ABS_COEFF_652nm','PM2.5_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_670nm','PM2.5_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_467nm','PM1_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_522nm','PM1_AERO_ABS_COEFF_525nm','PM1_AERO_ABS_COEFF_528nm','PM1_AERO_ABS_COEFF_530nm','PM1_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_565nm','PM1_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_637nm','PM1_AERO_ABS_COEFF_652nm','PM1_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_670nm','PM1_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_950nm'], 
    'instruments':{
        'Magee AE8':   {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                         'instrument_further_details':'No manual or specifications available but confirmation instrument is aethalometer here: https://acp.copernicus.org/preprints/acp-2017-1117/acp-2017-1117-manuscript-version4.pdf  Instrument is boradband, peaking at 830nm for EC: https://aaqr.org/articles/aaqr-15-05-simts-0358.pdf'},
        'Magee AE16':  {'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'instrument_further_details':'No manual or specifications available but confirmation instrument is aethalometer here: https://www.psi.ch/sites/default/files/import/lac/ProjectAddonCatcosOperationsEN/Aethalometer_book_2005.07.02.pdf'},
        'Magee AE22':  {'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                'documented_resolution':{'BC':'0.1','PM10_BC':'0.1','PM2.5_BC':'0.1','PM1_BC':'0.1'},     'documented_flow_rate':'2.0-6.0', 'instrument_manual_name':'Magee_AE31_Specs.pdf'},
        'Magee AE30':  {'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_950nm'], 'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_950nm'],                                                                                                                                                                                                                                                                                                                                                     'instrument_further_details':'No manual or specifications available but confirmation instrument is aethalometer here: https://books.google.es/books?id=5btgEAAAQBAJ&pg=PA50&lpg=PA50&dq=magee+ae30&source=bl&ots=KYjCcR8cpP&sig=ACfU3U2a5-F3Pk0v2zEwLoG6ICBzir-egg&hl=en&sa=X&ved=2ahUKEwjphuPjraeAAxVQgf0HHbFtAoUQ6AF6BAgFEAM#v=onepage&q=magee%20ae30&f=false'},   
        'Magee AE31':  {'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_950nm'], 'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_950nm'],                                                                                                                                                                                                                         'documented_resolution':{'BC':'0.1','PM10_BC':'0.1','PM2.5_BC':'0.1','PM1_BC':'0.1'},     'documented_flow_rate':'2.0-6.0', 'instrument_manual_name':'Magee_AE31_Specs.pdf'},
        'Magee AE33':  {'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_950nm'], 'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_950nm'], 'documented_lower_limit_of_detection':{'BC':'0.005','PM10_BC':'0.005','PM2.5_BC':'0.005','PM1_BC':'0.005'}, 'documented_upper_limit_of_detection':{'BC':'100.0','PM10_BC':'100.0','PM2.5_BC':'100.0','PM1_BC':'100.0'}, 'documented_resolution':{'BC':'0.03','PM10_BC':'0.03','PM2.5_BC':'0.03','PM1_BC':'0.03'}, 'documented_flow_rate':'2.0-5.0', 'instrument_manual_name':'Magee_AE33_Specs.pdf'},
        'Magee AE42-2':{'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                'documented_resolution':{'BC':'0.1','PM10_BC':'0.1','PM2.5_BC':'0.1','PM1_BC':'0.1'},     'documented_flow_rate':'2.0-5.0', 'instrument_manual_name':'Magee_AE42_Specs.pdf'},                                                                                                                                           
        'Magee AE42-7':{'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_950nm'], 'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_950nm'],                                                                                                                                                                                                                         'documented_resolution':{'BC':'0.1','PM10_BC':'0.1','PM2.5_BC':'0.1','PM1_BC':'0.1'},     'documented_flow_rate':'2.0-5.0', 'instrument_manual_name':'Magee_AE42_Specs.pdf'}
    }
},

#------------------------------------
#multi angle absorption photometer (MAAP)

#The MAAP instrument (introduced with the Thermo-Scientific model 5012) was specifically developed to reduce the uncertainties in black carbon measurements caused by aerosol scattering.
#The MAAP measures black carbon mass loadings based on aerosol optical absorption. In common with the Aethalometer this measurement is made on particles collected on a filter substrate, however there are some important differences in how the instruments operate, as well as with the algorithms used to extract the absorption coefficient and mass loading.
#With the MAAP the optical absorption coefficient of aerosol collected on a filter is determined by radiative transfer considerations which include multiple scattering effects and absorption enhancement due to reflections from the filter. 
#This calculation is based on the transmitted and reflected phase functions which are defined by directly measured values of transmission, direct and diffuse back scattering. 
#Full details of the algorithms used are given by Petzold et al (2002 and 2004). 
#Lab work showed that the diffuse back scatter component was a strong function of the fraction of scattering aerosol, being decreased (relative to the filter) for low scattering fractions and increased for high scattering fractions. 
#This lab work also informed the choice of angles for the detectors used.

#See here for more details: http://www.cas.manchester.ac.uk/restools/instruments/aerosol/maap/

#See references:
#Petzold A, Kramer H, and Schönlinner M, Continuous Measurement of Atmospheric Black Carbon Using a Multi-angle Absorption Photometer. Environmental Science and Pollution Research, Special Issue 4, 78-82, 2002.
#Petzold A and Schönlinner M, The Multi-angle absorption photometer – A new method for the measurement of aerosol light absorption and atmospheric black carbon. Journal of Aerosol Science, 35, 421-441, 2004.

'multi angle absorption photometer':{'abbreviation':'MAAP', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_467nm','AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_522nm','AERO_ABS_COEFF_525nm','AERO_ABS_COEFF_528nm','AERO_ABS_COEFF_530nm','AERO_ABS_COEFF_550nm','AERO_ABS_COEFF_565nm','AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_637nm','AERO_ABS_COEFF_652nm','AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_670nm','AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_467nm','PM10_AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_522nm','PM10_AERO_ABS_COEFF_525nm','PM10_AERO_ABS_COEFF_528nm','PM10_AERO_ABS_COEFF_530nm','PM10_AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_565nm','PM10_AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_637nm','PM10_AERO_ABS_COEFF_652nm','PM10_AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_670nm','PM10_AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_467nm','PM2.5_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_522nm','PM2.5_AERO_ABS_COEFF_525nm','PM2.5_AERO_ABS_COEFF_528nm','PM2.5_AERO_ABS_COEFF_530nm','PM2.5_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_565nm','PM2.5_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_637nm','PM2.5_AERO_ABS_COEFF_652nm','PM2.5_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_670nm','PM2.5_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_467nm','PM1_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_522nm','PM1_AERO_ABS_COEFF_525nm','PM1_AERO_ABS_COEFF_528nm','PM1_AERO_ABS_COEFF_530nm','PM1_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_565nm','PM1_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_637nm','PM1_AERO_ABS_COEFF_652nm','PM1_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_670nm','PM1_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_950nm'], 'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_467nm','AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_522nm','AERO_ABS_COEFF_525nm','AERO_ABS_COEFF_528nm','AERO_ABS_COEFF_530nm','AERO_ABS_COEFF_550nm','AERO_ABS_COEFF_565nm','AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_637nm','AERO_ABS_COEFF_652nm','AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_670nm','AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_467nm','PM10_AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_522nm','PM10_AERO_ABS_COEFF_525nm','PM10_AERO_ABS_COEFF_528nm','PM10_AERO_ABS_COEFF_530nm','PM10_AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_565nm','PM10_AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_637nm','PM10_AERO_ABS_COEFF_652nm','PM10_AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_670nm','PM10_AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_467nm','PM2.5_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_522nm','PM2.5_AERO_ABS_COEFF_525nm','PM2.5_AERO_ABS_COEFF_528nm','PM2.5_AERO_ABS_COEFF_530nm','PM2.5_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_565nm','PM2.5_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_637nm','PM2.5_AERO_ABS_COEFF_652nm','PM2.5_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_670nm','PM2.5_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_467nm','PM1_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_522nm','PM1_AERO_ABS_COEFF_525nm','PM1_AERO_ABS_COEFF_528nm','PM1_AERO_ABS_COEFF_530nm','PM1_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_565nm','PM1_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_637nm','PM1_AERO_ABS_COEFF_652nm','PM1_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_670nm','PM1_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_950nm'], 
    'instruments':{
        'Thermo 5012':{'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_670nm','PM10_AERO_ABS_COEFF_670nm','PM2.5_AERO_ABS_COEFF_670nm','PM1_AERO_ABS_COEFF_670nm'], 'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_670nm','PM10_AERO_ABS_COEFF_670nm','PM2.5_AERO_ABS_COEFF_670nm','PM1_AERO_ABS_COEFF_670nm'], 'documented_lower_limit_of_detection':{'BC':'0.02','PM10_BC':'0.02','PM2.5_BC':'0.02','PM1_BC':'0.02','AERO_ABS_COEFF_670nm':'0.13','PM10_AERO_ABS_COEFF_670nm':'0.13','PM2.5_AERO_ABS_COEFF_670nm':'0.13','PM1_AERO_ABS_COEFF_670nm':'0.13'}, 'documented_upper_limit_of_detection':{'BC':'180.0','PM10_BC':'180.0','PM2.5_BC':'180.0','PM1_BC':'180.0'}, 'documented_flow_rate':'16.7', 'instrument_manual_name':'Thermo_5012_Manual.pdf'}
    }
},

#------------------------------------
#particulate soot absorption photometer (PSAP)

#The PSAP instrument (manufactured by Radiance Research Inc.) measures the change in optical transmission of a filter as aerosol is deposited on the filter over time.
#The PSAP measures black carbon mass loadings based on aerosol optical absorption. In common with the Aethalometer this measurement is made on particles collected on a filter substrate, however there are some important differences in how the instruments operate, as well as with the algorithms used to extract the absorption coefficient and mass loading.
#There are currently 2 instrument types, 1 measuring at 1 wavelength (525nm), and one at 3 wavelengths (467, 530 and 660nm). 
#The change in transmission from one measurement to the next is related to the optical absorption coefficient (bap m-1) of the aerosol deposited in this time. 
#In turn this can be related to black carbon mass by dividing by the specific absorption for black carbon at the wavelength of the instrument. 
#Comparisons between the PSAP aerosol optical absorption measurement and reference techniques presented by Bond et al (1999), show that the PSAP also responds to pure scattering particles and external mixtures of scattering and absorbing particles. 
#This is because back scatter from scattering particles on the filter reduces the transmitted intensity. Bond et al show that the apparent absorption coefficient of purely scattering particles as measured by the PSAP is about 2% of their scattering coefficient. 
#Additionally even when the PSAP measurement is corrected for errors in flow rate, collection spot area and the presence of scattering particles, the same series of experiments showed that PSAP optical absorption coefficients were about 20% higher than the reference measurement for the type of aerosol used. 
#Thus PSAP data must be carefully calibrated and corrected based on measurements of aerosol scattering coefficient.

#See reference: Bond TC, Anderson TL, Campbell D, Calibration and intercomparison of filter-based measurements of visible light absorption by aerosols. Aerosol Science and Technology 30, 582–600, 1999

'particulate soot absorption photometer':{'abbreviation':'PSAP', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_467nm','AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_522nm','AERO_ABS_COEFF_525nm','AERO_ABS_COEFF_528nm','AERO_ABS_COEFF_530nm','AERO_ABS_COEFF_550nm','AERO_ABS_COEFF_565nm','AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_637nm','AERO_ABS_COEFF_652nm','AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_670nm','AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_467nm','PM10_AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_522nm','PM10_AERO_ABS_COEFF_525nm','PM10_AERO_ABS_COEFF_528nm','PM10_AERO_ABS_COEFF_530nm','PM10_AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_565nm','PM10_AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_637nm','PM10_AERO_ABS_COEFF_652nm','PM10_AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_670nm','PM10_AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_467nm','PM2.5_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_522nm','PM2.5_AERO_ABS_COEFF_525nm','PM2.5_AERO_ABS_COEFF_528nm','PM2.5_AERO_ABS_COEFF_530nm','PM2.5_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_565nm','PM2.5_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_637nm','PM2.5_AERO_ABS_COEFF_652nm','PM2.5_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_670nm','PM2.5_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_467nm','PM1_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_522nm','PM1_AERO_ABS_COEFF_525nm','PM1_AERO_ABS_COEFF_528nm','PM1_AERO_ABS_COEFF_530nm','PM1_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_565nm','PM1_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_637nm','PM1_AERO_ABS_COEFF_652nm','PM1_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_670nm','PM1_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_950nm'], 'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_467nm','AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_522nm','AERO_ABS_COEFF_525nm','AERO_ABS_COEFF_528nm','AERO_ABS_COEFF_530nm','AERO_ABS_COEFF_550nm','AERO_ABS_COEFF_565nm','AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_637nm','AERO_ABS_COEFF_652nm','AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_670nm','AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_467nm','PM10_AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_522nm','PM10_AERO_ABS_COEFF_525nm','PM10_AERO_ABS_COEFF_528nm','PM10_AERO_ABS_COEFF_530nm','PM10_AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_565nm','PM10_AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_637nm','PM10_AERO_ABS_COEFF_652nm','PM10_AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_670nm','PM10_AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_467nm','PM2.5_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_522nm','PM2.5_AERO_ABS_COEFF_525nm','PM2.5_AERO_ABS_COEFF_528nm','PM2.5_AERO_ABS_COEFF_530nm','PM2.5_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_565nm','PM2.5_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_637nm','PM2.5_AERO_ABS_COEFF_652nm','PM2.5_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_670nm','PM2.5_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_467nm','PM1_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_522nm','PM1_AERO_ABS_COEFF_525nm','PM1_AERO_ABS_COEFF_528nm','PM1_AERO_ABS_COEFF_530nm','PM1_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_565nm','PM1_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_637nm','PM1_AERO_ABS_COEFF_652nm','PM1_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_670nm','PM1_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_950nm'],
    'instruments':{
        'ITM Custom':      {'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_525nm','PM10_AERO_ABS_COEFF_525nm','PM2.5_AERO_ABS_COEFF_525nm','PM1_AERO_ABS_COEFF_525nm','AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_550nm'],                                                                                                                                                                                                                       'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_525nm','PM10_AERO_ABS_COEFF_525nm','PM2.5_AERO_ABS_COEFF_525nm','PM1_AERO_ABS_COEFF_525nm','AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_550nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  'instrument_further_details':'No manual or specifications available but confirmation instrument is particulate soot absorption photometer here: https://amt.copernicus.org/articles/4/245/2011/amt-4-245-2011.pdf'},                                        
        'MISU PSAP-1W':    {'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_522nm','PM10_AERO_ABS_COEFF_522nm','PM2.5_AERO_ABS_COEFF_522nm','PM1_AERO_ABS_COEFF_522nm'],                                                                                                                                                                                                                                                                                                                                  'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_522nm','PM10_AERO_ABS_COEFF_522nm','PM2.5_AERO_ABS_COEFF_522nm','PM1_AERO_ABS_COEFF_522nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             'instrument_further_details':'No manual or specifications available but confirmation instrument is particulate soot absorption photometer here: https://amt.copernicus.org/articles/4/245/2011/amt-4-245-2011.pdf'},
        'Radiance PSAP-1W':{'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_522nm','PM10_AERO_ABS_COEFF_522nm','PM2.5_AERO_ABS_COEFF_522nm','PM1_AERO_ABS_COEFF_522nm','AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_550nm','AERO_ABS_COEFF_565nm','PM10_AERO_ABS_COEFF_565nm','PM2.5_AERO_ABS_COEFF_565nm','PM1_AERO_ABS_COEFF_565nm'],                                                                                                            'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_522nm','PM10_AERO_ABS_COEFF_522nm','PM2.5_AERO_ABS_COEFF_522nm','PM1_AERO_ABS_COEFF_522nm','AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_550nm','AERO_ABS_COEFF_565nm','PM10_AERO_ABS_COEFF_565nm','PM2.5_AERO_ABS_COEFF_565nm','PM1_AERO_ABS_COEFF_565nm'],                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'instrument_further_details':'No manual or specifications available but confirmation instrument is particulate soot absorption photometer here: https://gml.noaa.gov/aero/instrumentation/psap.html'},
        'Radiance PSAP-3W':{'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_467nm','PM10_AERO_ABS_COEFF_467nm','PM2.5_AERO_ABS_COEFF_467nm','PM1_AERO_ABS_COEFF_467nm','AERO_ABS_COEFF_530nm','PM10_AERO_ABS_COEFF_530nm','PM2.5_AERO_ABS_COEFF_530nm','PM1_AERO_ABS_COEFF_530nm','AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_550nm','AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_660nm'], 'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_467nm','PM10_AERO_ABS_COEFF_467nm','PM2.5_AERO_ABS_COEFF_467nm','PM1_AERO_ABS_COEFF_467nm','AERO_ABS_COEFF_530nm','PM10_AERO_ABS_COEFF_530nm','PM2.5_AERO_ABS_COEFF_530nm','PM1_AERO_ABS_COEFF_530nm','AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_550nm','AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_660nm'], 'documented_resolution':{'AERO_ABS_COEFF_467nm':'0.3','PM10_AERO_ABS_COEFF_467nm':'0.3','PM2.5_AERO_ABS_COEFF_467nm':'0.3','PM1_AERO_ABS_COEFF_467nm':'0.3','AERO_ABS_COEFF_530nm':'0.3','PM10_AERO_ABS_COEFF_530nm':'0.3','PM2.5_AERO_ABS_COEFF_530nm':'0.3','PM1_AERO_ABS_COEFF_530nm':'0.3','AERO_ABS_COEFF_550nm':'0.3','PM10_AERO_ABS_COEFF_550nm':'0.3','PM2.5_AERO_ABS_COEFF_550nm':'0.3','PM1_AERO_ABS_COEFF_550nm':'0.3','AERO_ABS_COEFF_660nm':'0.3','PM10_AERO_ABS_COEFF_660nm':'0.3','PM2.5_AERO_ABS_COEFF_660nm':'0.3','PM1_AERO_ABS_COEFF_660nm':'0.3'}, 'documented_precision':{'AERO_ABS_COEFF_467nm':'0.1','PM10_AERO_ABS_COEFF_467nm':'0.1','PM2.5_AERO_ABS_COEFF_467nm':'0.1','PM1_AERO_ABS_COEFF_467nm':'0.1','AERO_ABS_COEFF_530nm':'0.1','PM10_AERO_ABS_COEFF_530nm':'0.1','PM2.5_AERO_ABS_COEFF_530nm':'0.1','PM1_AERO_ABS_COEFF_530nm':'0.1','AERO_ABS_COEFF_550nm':'0.1','PM10_AERO_ABS_COEFF_550nm':'0.1','PM2.5_AERO_ABS_COEFF_550nm':'0.1','PM1_AERO_ABS_COEFF_550nm':'0.1','AERO_ABS_COEFF_660nm':'0.1','PM10_AERO_ABS_COEFF_660nm':'0.1','PM2.5_AERO_ABS_COEFF_660nm':'0.1','PM1_AERO_ABS_COEFF_660nm':'0.1'}, 'documented_flow_rate':'1.0', 'instrument_manual_name':'Radiance_PSAP-3W_Manual.pdf'}
    }
},

#------------------------------------
#continuous light absorption photometer (CLAP)

#The Continuous Light Absorption Photometer (CLAP) is a NOAA/GMD developed, filter-based method that measures light absorption by particles at three wavelengths (467, 528, 652 nm). 
#The CLAP is functionally comparable to a widely-used commercial instrument, the Particle/Soot Absorption Photometer (PSAP), but does not suffer from the same operating limitations as the PSAP (data loss from overloaded filters, unheated optics block, internal black box processing of data). 
#The choice of a functionally comparable design to the PSAP allowed measurements from the CLAP to use the well-established PSAP correction schemes that account for known imperfections in all filter-based absorption measurements. 
#Ogren et al. (2017) describes the design, operation and uncertainties of the CLAP and presents comparisons with the PSAP at several NFAN sites.
#The primary advantages of the CLAP, compared to the PSAP, are that it is one-tenth the size, can sample for roughly eight times as long before the operator needs to change the filter, and is temperature-stabilized to reduce sensitivity to changes in room temperature. 
#In addition, the computer software running on the internal microprocessor is completely open on the CLAP. These features make the CLAP much better suited for long-term monitoring applications, and currently NOAA-built CLAPs are deployed at 23 stations around the globe.
#During the initial field deployment of the NOAA-built CLAPs, they were run in parallel with PSAPs for over a year at multiple stations to confirm that the CLAPs and PSAPs give comparable results. Subsequently, NOAA has retired the PSAPs from service at its monitoring stations.

'continuous light absorption photometer':{'abbreviation':'CLAP', 'sampling_type':'low volume continuous', 'sample_preparation_type':'filter', 'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_467nm','AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_522nm','AERO_ABS_COEFF_525nm','AERO_ABS_COEFF_528nm','AERO_ABS_COEFF_530nm','AERO_ABS_COEFF_550nm','AERO_ABS_COEFF_565nm','AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_637nm','AERO_ABS_COEFF_652nm','AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_670nm','AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_467nm','PM10_AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_522nm','PM10_AERO_ABS_COEFF_525nm','PM10_AERO_ABS_COEFF_528nm','PM10_AERO_ABS_COEFF_530nm','PM10_AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_565nm','PM10_AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_637nm','PM10_AERO_ABS_COEFF_652nm','PM10_AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_670nm','PM10_AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_467nm','PM2.5_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_522nm','PM2.5_AERO_ABS_COEFF_525nm','PM2.5_AERO_ABS_COEFF_528nm','PM2.5_AERO_ABS_COEFF_530nm','PM2.5_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_565nm','PM2.5_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_637nm','PM2.5_AERO_ABS_COEFF_652nm','PM2.5_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_670nm','PM2.5_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_467nm','PM1_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_522nm','PM1_AERO_ABS_COEFF_525nm','PM1_AERO_ABS_COEFF_528nm','PM1_AERO_ABS_COEFF_530nm','PM1_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_565nm','PM1_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_637nm','PM1_AERO_ABS_COEFF_652nm','PM1_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_670nm','PM1_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_950nm'], 'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_370nm','AERO_ABS_COEFF_467nm','AERO_ABS_COEFF_470nm','AERO_ABS_COEFF_520nm','AERO_ABS_COEFF_522nm','AERO_ABS_COEFF_525nm','AERO_ABS_COEFF_528nm','AERO_ABS_COEFF_530nm','AERO_ABS_COEFF_550nm','AERO_ABS_COEFF_565nm','AERO_ABS_COEFF_590nm','AERO_ABS_COEFF_637nm','AERO_ABS_COEFF_652nm','AERO_ABS_COEFF_660nm','AERO_ABS_COEFF_670nm','AERO_ABS_COEFF_880nm','AERO_ABS_COEFF_950nm','PM10_AERO_ABS_COEFF_370nm','PM10_AERO_ABS_COEFF_467nm','PM10_AERO_ABS_COEFF_470nm','PM10_AERO_ABS_COEFF_520nm','PM10_AERO_ABS_COEFF_522nm','PM10_AERO_ABS_COEFF_525nm','PM10_AERO_ABS_COEFF_528nm','PM10_AERO_ABS_COEFF_530nm','PM10_AERO_ABS_COEFF_550nm','PM10_AERO_ABS_COEFF_565nm','PM10_AERO_ABS_COEFF_590nm','PM10_AERO_ABS_COEFF_637nm','PM10_AERO_ABS_COEFF_652nm','PM10_AERO_ABS_COEFF_660nm','PM10_AERO_ABS_COEFF_670nm','PM10_AERO_ABS_COEFF_880nm','PM10_AERO_ABS_COEFF_950nm','PM2.5_AERO_ABS_COEFF_370nm','PM2.5_AERO_ABS_COEFF_467nm','PM2.5_AERO_ABS_COEFF_470nm','PM2.5_AERO_ABS_COEFF_520nm','PM2.5_AERO_ABS_COEFF_522nm','PM2.5_AERO_ABS_COEFF_525nm','PM2.5_AERO_ABS_COEFF_528nm','PM2.5_AERO_ABS_COEFF_530nm','PM2.5_AERO_ABS_COEFF_550nm','PM2.5_AERO_ABS_COEFF_565nm','PM2.5_AERO_ABS_COEFF_590nm','PM2.5_AERO_ABS_COEFF_637nm','PM2.5_AERO_ABS_COEFF_652nm','PM2.5_AERO_ABS_COEFF_660nm','PM2.5_AERO_ABS_COEFF_670nm','PM2.5_AERO_ABS_COEFF_880nm','PM2.5_AERO_ABS_COEFF_950nm','PM1_AERO_ABS_COEFF_370nm','PM1_AERO_ABS_COEFF_467nm','PM1_AERO_ABS_COEFF_470nm','PM1_AERO_ABS_COEFF_520nm','PM1_AERO_ABS_COEFF_522nm','PM1_AERO_ABS_COEFF_525nm','PM1_AERO_ABS_COEFF_528nm','PM1_AERO_ABS_COEFF_530nm','PM1_AERO_ABS_COEFF_550nm','PM1_AERO_ABS_COEFF_565nm','PM1_AERO_ABS_COEFF_590nm','PM1_AERO_ABS_COEFF_637nm','PM1_AERO_ABS_COEFF_652nm','PM1_AERO_ABS_COEFF_660nm','PM1_AERO_ABS_COEFF_670nm','PM1_AERO_ABS_COEFF_880nm','PM1_AERO_ABS_COEFF_950nm'],
    'instruments':{
        'NOAA CLAP':{'measured_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_467nm','PM10_AERO_ABS_COEFF_467nm','PM2.5_AERO_ABS_COEFF_467nm','PM1_AERO_ABS_COEFF_467nm','AERO_ABS_COEFF_528nm','PM10_AERO_ABS_COEFF_528nm','PM2.5_AERO_ABS_COEFF_528nm','PM1_AERO_ABS_COEFF_528nm','AERO_ABS_COEFF_652nm','PM10_AERO_ABS_COEFF_652nm','PM2.5_AERO_ABS_COEFF_652nm','PM1_AERO_ABS_COEFF_652nm'], 'qa_accepted_parameters':['BC','PM10_BC','PM2.5_BC','PM1_BC','AERO_ABS_COEFF_467nm','PM10_AERO_ABS_COEFF_467nm','PM2.5_AERO_ABS_COEFF_467nm','PM1_AERO_ABS_COEFF_467nm','AERO_ABS_COEFF_528nm','PM10_AERO_ABS_COEFF_528nm','PM2.5_AERO_ABS_COEFF_528nm','PM1_AERO_ABS_COEFF_528nm','AERO_ABS_COEFF_652nm','PM10_AERO_ABS_COEFF_652nm','PM2.5_AERO_ABS_COEFF_652nm','PM1_AERO_ABS_COEFF_652nm'], 'documented_lower_limit_of_detection':{'AERO_ABS_COEFF_467nm':'0.02','PM10_AERO_ABS_COEFF_467nm':'0.02','PM2.5_AERO_ABS_COEFF_467nm':'0.02','PM1_AERO_ABS_COEFF_467nm':'0.02','AERO_ABS_COEFF_528nm':'0.02','PM10_AERO_ABS_COEFF_528nm':'0.02','PM2.5_AERO_ABS_COEFF_528nm':'0.02','PM1_AERO_ABS_COEFF_528nm':'0.02','AERO_ABS_COEFF_652nm':'0.02','PM10_AERO_ABS_COEFF_652nm':'0.02','PM2.5_AERO_ABS_COEFF_652nm':'0.02','PM1_AERO_ABS_COEFF_652nm':'0.02'}, 'documented_precision':{'AERO_ABS_COEFF_467nm':'4.0%','PM10_AERO_ABS_COEFF_467nm':'4.0%','PM2.5_AERO_ABS_COEFF_467nm':'4.0%','PM1_AERO_ABS_COEFF_467nm':'4.0%','AERO_ABS_COEFF_528nm':'4.0%','PM10_AERO_ABS_COEFF_528nm':'4.0%','PM2.5_AERO_ABS_COEFF_528nm':'4.0%','PM1_AERO_ABS_COEFF_528nm':'4.0%','AERO_ABS_COEFF_652nm':'4.0%','PM10_AERO_ABS_COEFF_652nm':'4.0%','PM2.5_AERO_ABS_COEFF_652nm':'4.0%','PM1_AERO_ABS_COEFF_652nm':'4.0%'}, 'documented_flow_rate':'1.0', 'instrument_further_details':'Specifications given in paper: https://amt.copernicus.org/articles/10/4805/2017/amt-10-4805-2017.pdf'}
    }
},

#------------------------------------
#flame atomic absorption spectroscopy (F-AAS)

#Atomic absorption spectroscopy (AAS) detects elements in either liquid or solid samples through the application of characteristic wavelengths of electromagnetic radiation from a light source. 
#Individual elements will absorb wavelengths differently, and these absorbances are measured against standards. In effect, AAS takes advantage of the different radiation wavelengths that are absorbed by different atoms.
#In AAS, analytes are first atomized so that their characteristic wavelengths are emitted and recorded. Then, during excitation, electrons move up one energy level in their respective atoms when those atoms absorb a specific energy.
#As electrons return to their original energy state, they emit energy in the form of light. 
#This light has a wavelength that is characteristic of the element. Depending on the light wavelenth and its intensity, specific elements can be detected and their concentrations measured.
#Applying the Beer-Lambert law directly in AA spectroscopy is difficult due to variations in the atomization efficiency from the sample matrix, and nonuniformity of concentration and path length of analyte atoms.
#Flame atomic absorption spectrometry (F-AAS) is where the sample is first atomised using flames, principally a air-acetylene flame with a temperature of about 2300 °C and a nitrous oxide system (N2O)-acetylene flame with a temperature of about 2700 °C. 
#The latter flame, in addition, offers a more reducing environment, being ideally suited for analytes with high affinity to oxygen.
#Liquid or dissolved samples are typically used with flame atomizers. 
#The sample solution is aspirated by a pneumatic analytical nebulizer, transformed into an aerosol, which is introduced into a spray chamber, where it is mixed with the flame gases and conditioned in a way that only the finest aerosol droplets (< 10 μm) enter the flame. 
#This conditioning process is responsible that only about 5% of the aspirated sample solution reaches the flame, but it also guarantees a relatively high freedom from interference.
#See here for more information: https://sites.chem.colostate.edu/diverdi/C433/experiments/iron%20analysis/references/aas%20methods%20-%20agilent.pdf

'flame atomic absorption spectroscopy':{'abbreviation':'F-AAS', 'sampling_type':'injection', 'measured_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 'qa_accepted_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','Na+','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 
    'instruments':{
        'GBC 932':                  {'instrument_further_details':'No manual or specifications available but confirmation it is uses flame atomic absorption spectroscopy here: http://www.speciation.net/Database/Instruments/GBC-Scientific-Equipment-Ltd/932-Plus--Atomic-Absorption-Spectrometer-;i3193'},
        'Perkin Elmer AAnalyst 600':{'instrument_manual_name':'Perkin_Elmer_AAnalyst_600_800_Specs.pdf'},
        'Perkin Elmer AAnalyst 700':{'instrument_manual_name':'Perkin_Elmer_AAnalyst_700_Specs.pdf'},
        'Perkin Elmer AAnalyst 800':{'instrument_manual_name':'Perkin_Elmer_AAnalyst_600_800_Specs.pdf'},
        'Varian SpectrAA 50':       {'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 55':       {'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 110':      {'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 220':      {'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 880':      {'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian 240 AA':            {'instrument_manual_name':'Varian_240_AA_Specs.pdf'}
    }
},

#------------------------------------
#graphite furnace atomic absorption spectroscopy (GF-AAS)

#Graphite furnace atomic absorption spectroscopy (GF-AAS) (also known as Electrothermal Atomic Absorption Spectroscopy (ET-AAS)) is a type of spectrometry that uses a graphite-coated furnace to vaporize the sample. 
#In GF-AAS, samples are deposited in a small graphite or pyrolytic carbon coated graphite tube, which can then be heated to vaporize and atomize the analyte. 
#The main advantages of the graphite furnace comparing to flame atomic absorption are the following:
#-The detection limits for the graphite furnace fall in the ppb range for most elements
#-Interference problems are minimized with the development of improved instrumentation
#-The graphite furnace can determine most elements measurable by flame atomic absorption in a wide variety of matrices.

'graphite furnace atomic absorption spectroscopy':{'abbreviation':'GF-AAS', 'sampling_type':'injection', 'measured_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 'qa_accepted_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 
    'instruments':{
        'Perkin Elmer AAnalyst 600': {'instrument_manual_name':'Perkin_Elmer_AAnalyst_600_800_Specs.pdf'},
        'Perkin Elmer AAnalyst 700': {'instrument_manual_name':'Perkin_Elmer_AAnalyst_700_Specs.pdf'},
        'Perkin Elmer AAnalyst 800': {'instrument_manual_name':'Perkin_Elmer_AAnalyst_600_800_Specs.pdf'},
        'Perkin Elmer 4110 ZL':      {'instrument_further_details':'No manual or specifications available but confirmation it is uses graphite furnace atomic absorption spectroscopy here: https://www.labx.com/product/perkin-elmer-4100zl-atomic-absorption'},
        'Unicam Solar 989 QZ':       {'instrument_further_details':'No manual or specifications available but confirmation it is uses graphite furnace atomic absorption spectroscopy here: https://www.sedgeochem.uni-bremen.de/aas.html'},
        'Varian SpectrAA 50':        {'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 55':        {'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 110':       {'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 220':       {'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 400 GTA-96':{'instrument_further_details':'No manual or specifications available but confirmation it is uses graphite furnace atomic absorption spectroscopy here: http://www.labwrench.com/?equipment.view/equipmentNo/20158/Varian/SpectrAA-400/'},
        'Varian SpectrAA 880':       {'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 880z':      {'instrument_further_details':'No manual or specifications available but confirmation it is uses graphite furnace atomic absorption spectroscopy here: http://www.medwow.com/med/atomic-absorption/varian/spectraa-880z/2927.model-spec'}, 
        'Varian 240 AA':             {'instrument_manual_name':'Varian_240_AA_Specs.pdf'}
        }
},

#------------------------------------
#cold vapour atomic absorption spectroscopy (CV-AAS)

#Cold vapour atomic absorption spectroscopy (CV-AAS) is a method limited to the determination of mercury, due to it being the only metallic element to have a large enough vapor pressure at ambient temperature. 
#Because of this, it has an important use in determining organic mercury compounds in samples and their distribution in the environment. 
#The method initiates by converting mercury into Hg2+ by oxidation from nitric and sulfuric acids, followed by a reduction of Hg2+ with tin(II) chloride. 
#The mercury, is then swept into a long-pass absorption tube by bubbling a stream of inert gas through the reaction mixture. 
#The concentration is determined by measuring the absorbance of this gas at 253.7 nm. Detection limits for this technique are in the parts-per-billion range making it an excellent mercury detection atomization method.
#See here for determination of Cd using this method also: https://www.ncbi.nlm.nih.gov/pubmed/19071344

'cold vapour atomic absorption spectroscopy':{'abbreviation':'CV-AAS', 'sampling_type':'injection', 'measured_parameters':['Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg'], 'qa_accepted_parameters':['Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg'], 
    'instruments':{
        'Envea UT-3000':      {'sampling_type':'continuous injection', 'sample_preparation_type':'filter;preconcentration', 'instrument_manual_name':'Envea_UT-3000_Specs.pdf'},
        'Varian SpectrAA 50': {                                                                                             'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 55': {                                                                                             'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 110':{                                                                                             'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 220':{                                                                                             'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 880':{                                                                                             'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian 240 AA':      {                                                                                             'instrument_manual_name':'Varian_240_AA_Specs.pdf'}
    }
},

#------------------------------------
#hydride generation atomic absorption spectroscopy (HG-AAS)

#The hydride generation atomic absorption spectroscopy (HG-AAS) method is specialised for the measurement of specific elements. 
#The technique provides a means of introducing samples containing arsenic, antimony, selenium, bismuth, and lead into an atomizer in the gas phase. 
#With these elements, hydride atomization enhances detection limits by a factor of 10 to 100 compared to alternative methods. 
#Hydride generation occurs by adding an acidified aqueous solution of the sample to a 1% aqueous solution of sodium borohydride, all of which is contained in a glass vessel. 
#The volatile hydride generated by the reaction that occurs is swept into the atomization chamber by an inert gas, where it undergoes decomposition. 
#This process forms an atomized form of the analyte, which can then be measured by absorption or emission spectrometry.
#See more information here: https://www.sciencedirect.com/topics/chemistry/hydride-generation

'hydride generation atomic absorption spectroscopy':{'abbreviation':'HG-AAS', 'sampling_type':'injection', 'measured_parameters':['As','PM10_As','PM2.5_As','PM1_As','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se'], 'qa_accepted_parameters':['As','PM10_As','PM2.5_As','PM1_As','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se'], 
    'instruments':{
        'Varian SpectrAA 50': {'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 55': {'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 110':{'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 220':{'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian SpectrAA 880':{'instrument_manual_name':'Varian_SpectrAA_50_55_110_220_Specs.pdf'},
        'Varian 240 AA':      {'instrument_manual_name':'Varian_240_AA_Specs.pdf'}
    }
},

#------------------------------------
#flame atomic emission spectroscopy (F-AES)

#Atomic emission spectroscopy (AES) is a method of chemical analysis that uses the intensity of light emitted from a flame, plasma, arc, or spark at a particular wavelength to determine the quantity of an element in a sample. 
#The wavelength of the atomic spectral line gives the identity of the element while the intensity of the emitted light is proportional to the number of atoms of the element.

#AAS and AES are similar methods but the differences are as follows:
#-Light Source: In AAS, a monochromatic light source is used to provide energy for the excitation of electrons. In the case of AES, it is a flame that is often used.
#-Atomization: In AAS, there is a separate chamber for atomization of the sample. However, in AES, atomization takes place step by step upon the introduction of the sample to the flame.
#-Principle of operation: In AAS, when monochromatic light is bombarded through the sample the atoms absorb energy, and the extent of absorption is recorded. In AES, the sample which gets atomised in the flame then absorbs the energy through the electrons which get excited. Later this energy is released upon the relaxation of the atoms and is measured by the instrument as the emitted energy.

#In flame atomic emission spectroscopy (F-AES) a sample of a material (analyte) is brought into the flame as a gas, sprayed solution, or directly inserted into the flame by use of a small loop of wire, usually platinum. 
#The heat from the flame evaporates the solvent and breaks chemical bonds to create free atoms. The thermal energy also excites the atoms into excited electronic states that subsequently emit light when they return to the ground electronic state. 
#Each element emits light at a characteristic wavelength, which is dispersed by a grating or prism and detected in the spectrometer.

'flame atomic emission spectroscopy':{'abbreviation':'F-AES', 'sampling_type':'injection', 'measured_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 'qa_accepted_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 
    'instruments':{}
},

#------------------------------------
#inductively coupled plasma atomic emission spectroscopy (ICP-AES)

#Inductively coupled plasma atomic emission spectroscopy (ICP-AES), also referred to as inductively coupled plasma optical emission spectrometry (ICP-OES), is an analytical technique used for the detection of chemical elements. 
#It is a type of emission spectroscopy that uses the inductively coupled plasma to produce excited atoms and ions that emit electromagnetic radiation at wavelengths characteristic of a particular element. 
#It is a flame technique with a flame temperature in a range from 6000 to 10,000 K. 
#The intensity of this emission is indicative of the concentration of the element within the sample.

'inductively coupled plasma atomic emission spectroscopy':{'abbreviation':'ICP-AES', 'sampling_type':'injection', 'measured_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 'qa_accepted_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 
    'instruments':{
        'Agilent 5100':                     {'instrument_manual_name':'Agilent_5100_Specs.pdf'},
        'Perkin Elmer Optima 7300':         {'instrument_manual_name':'PerkinElmerOptima7300_Specs.pdf'},
        'Thermo IRIS Intrepid II':          {'instrument_further_details':'No manual or specifications available but confirmation it is uses inductively coupled plasma atomic emission spectroscopy here: http://www.speciation.net/Database/Instruments/Thermo-Scientific/IRIS-Intrepid-II-;i123'},
        'Thermo Jarrell Ash Atomscan 16':   {'instrument_further_details':'No manual or specifications available but confirmation it is uses inductively coupled plasma atomic emission spectroscopy here: https://caeonline.com/buy/spectrometers/thermo-jarrell-ash-baird-atomscan-16/9115986'},      
        'Thermo Jarrell Ash IRIS Advantage':{'instrument_further_details':'No manual or specifications available but confirmation it is uses inductively coupled plasma atomic emission spectroscopy here: https://www.spectralabsci.com/equipment/thermo-jarrell-ash-iris-advantage-inductively-coupled-plasma-optical-emission-spectrometer/'},
        'Varian Vista-Pro':                 {'instrument_manual_name':'Varian_Vista-Pro_Specs.pdf'}
    }
},

#------------------------------------
#cold vapour atomic fluorescence spectroscopy (CV-AFS)

#Cold vapour atomic fluorescence spectroscopy (CV-AFS) is a subset of the analytical technique known as atomic fluorescence spectroscopy (AFS).
#Used in the measurement of trace amounts of volatile heavy metals such as mercury, cold vapour AFS makes use of the unique characteristic of mercury that allows vapour measurement at room temperature. 
#Free mercury atoms in a carrier gas are excited by a collimated ultraviolet light source at a wavelength of 253.7 nanometres. 
#The excited atoms re-radiate their absorbed energy (fluoresce) at this same wavelength. Unlike the directional excitation source, the fluorescence is omnidirectional and may thus be detected using a photomultiplier tube or UV photodiode.
#Gold coated traps may be used to collect mercury in ambient air or other media. 
#The traps are then heated, releasing the mercury from the gold while passing argon through the cartridge. 
#This preconcentrates the mercury, increasing sensitivity, and also transfers the mercury into an inert gas.
#See more information for measurement of Hg here: https://ec.europa.eu/environment/air/pdf/pp_mercury4.pdf

'cold vapour atomic fluorescence spectroscopy':{'abbreviation':'CV-AFS', 'sampling_type':'injection', 'measured_parameters':['Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg'], 'qa_accepted_parameters':['Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg'], 
    'instruments':{
        'Tekran 2537A':          {'sampling_type':'continuous injection', 'sample_preparation_type':'preconcentration',                'measured_parameters':['Hg-TGM'],                                                      'qa_accepted_parameters':['Hg-TGM'],                                                      'documented_lower_limit_of_detection':{'Hg-TGM':'0.0001'}, 'documented_upper_limit_of_detection':{'Hg-TGM':'10.0'}, 'documented_flow_rate':'0.5-1.5', 'instrument_further_details':'Specifications here: https://speciation.net/Database/Instruments/Tekran-Instruments-Corporation/Model-2537A-Mercury-Vapor-Analyzer-;i306'},
        'Tekran 2537A-1130':     {'sampling_type':'continuous injection', 'sample_preparation_type':'denuder;filter;preconcentration', 'measured_parameters':['Hg-GEM','Hg-GOM','Hg-TGM'],                                    'qa_accepted_parameters':['Hg-GEM','Hg-GOM','Hg-TGM'],                                    'documented_lower_limit_of_detection':{'Hg-TGM':'0.0001'}, 'documented_upper_limit_of_detection':{'Hg-TGM':'10.0'}, 'documented_flow_rate':'0.5-1.5', 'instrument_manual_name':'Tekran1130_Specs.pdf'},
        'Tekran 2537A-1130-1135':{'sampling_type':'continuous injection', 'sample_preparation_type':'denuder;filter;preconcentration', 'measured_parameters':['Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg'], 'qa_accepted_parameters':['Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg'], 'documented_lower_limit_of_detection':{'Hg-TGM':'0.0001'}, 'documented_upper_limit_of_detection':{'Hg-TGM':'10.0'}, 'documented_flow_rate':'0.5-1.5', 'instrument_manual_name':'Tekran1135_Specs.pdf'},
        'Tekran 2537X':          {'sampling_type':'continuous injection', 'sample_preparation_type':'preconcentration',                'measured_parameters':['Hg-TGM'],                                                      'qa_accepted_parameters':['Hg-TGM'],                                                      'documented_lower_limit_of_detection':{'Hg-TGM':'0.0001'},                                                          'documented_flow_rate':'0.7-1.5', 'instrument_manual_name':'Tekran2537X_Specs.pdf'},
        'Tekran 2537X-1130':     {'sampling_type':'continuous injection', 'sample_preparation_type':'denuder;filter;preconcentration', 'measured_parameters':['Hg-GEM','Hg-GOM','Hg-TGM'],                                    'qa_accepted_parameters':['Hg-GEM','Hg-GOM','Hg-TGM'],                                    'documented_lower_limit_of_detection':{'Hg-TGM':'0.0001'},                                                          'documented_flow_rate':'0.7-1.5', 'instrument_manual_name':'Tekran1130_Specs.pdf'},
        'Tekran 2537X-1130-1135':{'sampling_type':'continuous injection', 'sample_preparation_type':'denuder;filter;preconcentration', 'measured_parameters':['Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg'], 'qa_accepted_parameters':['Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg'], 'documented_lower_limit_of_detection':{'Hg-TGM':'0.0001'},                                                          'documented_flow_rate':'10.0',    'instrument_manual_name':'Tekran1135_Specs.pdf'}
    }
},

#------------------------------------
#gas chromatography - cold vapour atomic fluorescence spectroscopy (GC-CV-AFS)

#Combination of gas chromatography and cold vapour atomic fluorescence spectroscopy for increased speciation

'gas chromatography - cold vapour atomic fluorescence spectroscopy':{'abbreviation':'GC-CV-AFS', 'sampling_type':'injection', 'measured_parameters':['Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg'], 'qa_accepted_parameters':['Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg'], 
    'instruments':{}
},

#------------------------------------
#inductively coupled plasma mass spectrometry (ICP-MS)

#Inductively coupled plasma mass spectrometry (ICP-MS) is a type of mass spectrometry which is capable of detecting metals and several non-metals at concentrations as low as one part per quadrillion on non-interfered low-background isotopes. 
#This is achieved by ionizing the sample with inductively coupled plasma and then using a mass spectrometer to separate and quantify those ions.
#Compared to atomic absorption spectroscopy, ICP-MS has greater speed, precision, and sensitivity. 
#However, compared with other types of mass spectrometry, such as thermal ionization mass spectrometry (TIMS) and glow discharge mass spectrometry (GD-MS), ICP-MS introduces many interfering species: argon from the plasma, component gases of air that leak through the cone orifices, and contamination from glassware and the cones.

'inductively coupled plasma mass spectrometry':{'abbreviation':'ICP-MS', 'sampling_type':'injection', 'measured_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','BC','PM10_BC','PM2.5_BC','PM1_BC','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 'qa_accepted_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 
    'instruments':{
        'Agilent 7500':            {'instrument_manual_name':'Agilent_7500_Specs.pdf'},
        'Agilent 7700':            {'instrument_manual_name':'Agilent_7700_Specs.pdf'},
        'Agilent 7700e':           {'instrument_manual_name':'Agilent_7700_Specs.pdf'},
        'Agilent 7700s':           {'instrument_manual_name':'Agilent_7700_Specs.pdf'},
        'Agilent 7700x':           {'instrument_manual_name':'Agilent_7700_Specs.pdf'},
        'Agilent 7800':            {'instrument_manual_name':'Agilent_7800_Specs.pdf'},
        'Agilent 7900':            {'instrument_manual_name':'Agilent_7900_Specs.pdf'},
        'Agilent 8900':            {'instrument_manual_name':'Agilent_8900_Specs.pdf'},
        'Perkin Elmer ELAN DRC II':{'instrument_manual_name':'Perkin_Elmer_ELAN_DRC_II_specs.pdf'},
        'Perkin Elmer ELAN 6000':  {'instrument_further_details':'No manual or specifications available but confirmation it is uses inductively coupled plasma mass spectrometry here: http://www.speciation.net/Database/Instruments/PerkinElmer-SCIEX/ELAN-6000-;i533'},
        'Thermo Finnigan ELEMENT2':{'instrument_manual_name':'Thermo_Finnigan_ELEMENT2_Specs.pdf'},
        'Thermo iCAP RQ':          {'instrument_manual_name':'Thermo_iCAP_RQ_Specs.pdf'},
        'Thermo XSERIES':          {'instrument_further_details':'No manual or specifications available but confirmation it is uses inductively coupled plasma mass spectrometry here: https://profcontrol.de/Thermo-Scientific-X-Series-ICP-MS-Quadrupole-Mass-Spectrometer_1'},
        'Thermo XSERIES 2':        {'instrument_manual_name':'Thermo_XSERIES_2_Specs.pdf'},
    }
},

#------------------------------------
#X-ray fluorescence spectroscopy (XRFS)

#X-Ray Fluorescence Spectroscopy (XRFS) is an analytical technique that uses the interaction of x-rays with a material to determine its elemental composition. XRF is suitable for solids, liquids and powders, and in most circumstances is non-destructive.
#There are two main XRF methodologies - energy dispersive (EDXRF) and wavelength dispersive (WDXRF). Each method has its own advantages and disadvantages.
#The sample is irradiated with x-rays which causes the emission of fluorescent x-rays to emerge from the sample.
#The range of detectable elements varies according to instrument configuration and set up, but typically EDXRF covers all elements from sodium (Na) to uranium (U), whilst WDXRF can extend this down to beryllium (Be). Concentrations can range from 100% down to ppm and in some cases sub-ppm levels. 
#Limits of detection depend upon the specific element and the sample matrix, but as a general rule, heavier elements will have better detection limits.
#The key components of a typical XRF spectrometer are: 1. Source of X-Rays used to irradiate the sample. 2. Sample. 3. Detection of the emitted fluorescent X-Rays. 4. The resulting XRF spectrum shows intensity of X-Rays (usually in counts per second) as a function of energy (usually in eV).

'X-ray fluorescence spectroscopy':{'abbreviation':'XRFS', 'sampling_type':'injection', 'measured_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','BC','PM10_BC','PM2.5_BC','PM1_BC','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 'qa_accepted_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 
    'instruments':{
        'Pall Xact 620': {'sampling_type':'continuous injection', 'sample_preparation_type':'filter', 'documented_flow_rate':'16.7', 'instrument_manual_name':'Pall_Xact_620_Specs.pdf'},
        'Pall Xact 625': {'sampling_type':'continuous injection', 'sample_preparation_type':'filter', 'documented_flow_rate':'16.7', 'instrument_manual_name':'Pall_Xact_625_Specs.pdf'},
        'Philips PW2400':{                                                                                                           'instrument_further_details':'No manual or specifications available but confirmation it is uses X-ray fluorescence spectroscopy here:http://sr-sc-8f00.unifr.ch/geoscience/geology/assets/files/Site%20Science%20de%20la%20terre/PDFLabo/PW2400.pdf'}
    }
},

#------------------------------------
#particle induced X-ray emission (PIXE)

#particle induced X-ray emission is a powerful and relatively simple analytical technique that can be used to identify and quantify trace elements typically ranging from Al to U. 
#After high energy ion bombardment, X-rays are emitted which can be measured.
#Sample irradiation is usually performed by means of protons or alpha particles produced by an accelerator.
#X-ray detection is usually done by energy dispersive semiconductor detectors such as Si(Li) or High Purity Germanium detectors. 
#Varies from X-ray fluorescence spectroscopy as x-ray emission is induced by a particle, not X-rays. 

'particle induced X-ray emission':{'abbreviation':'PIXE', 'sampling_type':'injection', 'measured_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','BC','PM10_BC','PM2.5_BC','PM1_BC','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','MSA','PM10_MSA','PM2.5_MSA','PM1_MSA','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','NH4NO3','PM10_NH4NO3','PM2.5_NH4NO3','PM1_NH4NO3','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 'qa_accepted_parameters':['Al','PM10_Al','PM2.5_Al','PM1_Al','As','PM10_As','PM2.5_As','PM1_As','C','PM10_C','PM2.5_C','PM1_C','Ca++','PM10_Ca++','PM2.5_Ca++','PM1_Ca++','Cd','PM10_Cd','PM2.5_Cd','PM1_Cd','Cl-','PM10_Cl-','PM2.5_Cl-','PM1_Cl-','Co','PM10_Co','PM2.5_Co','PM1_Co','Cr','PM10_Cr','PM2.5_Cr','PM1_Cr','Cu','PM10_Cu','PM2.5_Cu','PM1_Cu','Fe','PM10_Fe','PM2.5_Fe','PM1_Fe','Hg-GEM','Hg-GOM','Hg-TGM','Hg','PM10_Hg','PM2.5_Hg','PM1_Hg','K+','PM10_K+','PM2.5_K+','PM1_K+','Mg++','PM10_Mg++','PM2.5_Mg++','PM1_Mg++','Mn','PM10_Mn','PM2.5_Mn','PM1_Mn','Na+','PM10_Na+','PM2.5_Na+','PM1_Na+','NH4+','PM10_NH4+','PM2.5_NH4+','PM1_NH4+','Ni','PM10_Ni','PM2.5_Ni','PM1_Ni','NO3-','PM10_NO3-','PM2.5_NO3-','PM1_NO3-','OC','PM10_OC','PM2.5_OC','PM1_OC','Pb','PM10_Pb','PM2.5_Pb','PM1_Pb','Se','PM10_Se','PM2.5_Se','PM1_Se','SO4--','PM10_SO4--','PM2.5_SO4--','PM1_SO4--','SO4--_NSS','PM10_SO4--_NSS','PM2.5_SO4--_NSS','PM1_SO4--_NSS','SO4--_SS','PM10_SO4--_SS','PM2.5_SO4--_SS','PM1_SO4--_SS','V','PM10_V','PM2.5_V','PM1_V','Zn','PM10_Zn','PM2.5_Zn','PM1_Zn'], 
    'instruments':{}
},

#------------------------------------
#photometry - direct (P-D)

#remote photometry involves measurements of spectral sun/moon irradiance (direct) and sky radiances (sky) 
#sun photometry is a passive remote-sensing measurement technique in which mainly collimated light generally not scattered or absorbed by the atmosphere illuminates a photodiode detector and this light energy is converted to a digital signal. 
#The digital signal (V) measured by the instrument is proportional to the solar irradiance.
#A sensor head fitted with 25 cm collimators is attached to a 40 cm robot base which systematically points the sensor head at the Sun, the sky and the moon according to a preprogrammed routine.
#The measurements are made in eight spectral bands requiring approximately 10 seconds. 
#Eight interference filters at wavelengths of 340, 380, 440, 500, 670, 870, 940 and 1020 nm are located in a filter wheel which is rotated by a direct drive stepping motor. 
#The 940 nm channel is used for column water abundance determination. A preprogrammed sequence of measurements is taken by these instruments starting at an air mass of 7 in the morning and ending at an air mass of 7 in the evening.
#Optical depth is calculated from spectral extinction of direct beam radiation at each wavelength based on the Beer-Bouguer Law. Attenuation due to Rayleigh scatter, and absorption by ozone, and gaseous pollutants is estimated and removed to isolate the aerosol optical depth (AOD). 
#A sequence of three such measurements are taken 30 seconds apart creating a triplet observation per wavelength.
#During the large air mass periods direct sun measurements are made at 0.25 air mass intervals, while at smaller air masses the interval between measurements is typically 15 minutes. 
#The time variation of clouds is usually greater than that of aerosols causing an observable variation in the triplets that can be used to screen clouds in many cases. 
#Additionally the 15-minute interval allows a longer temporal frequency check for cloud contamination.

#The latest models can also perform nighttime measurements of the spectral lunar irradiance.
#See paper here for detailed description of the methodology verified for a specific instrument:
#https://www.atmos-meas-tech.net/9/631/2016/amt-9-631-2016.pdf

#See paper here for more details on the methodology applied at a specific station:
#https://www.atmos-meas-tech.net/10/3007/2017/amt-10-3007-2017.pdf

'photometry - direct':{'abbreviation':'P-D', 'sampling_type':'remote', 'measured_parameters':['AOD_380nm','AOD_440nm','AOD_500nm','AOD_500nm_COARSE','AOD_500nm_FINE','FINE_MODE_FRAC_500nm','AOD_550nm','AOD_675nm','AOD_870nm','AOD_1020nm','AE_440-870nm'], 'qa_accepted_parameters':['AOD_380nm','AOD_440nm','AOD_500nm','AOD_500nm_COARSE','AOD_500nm_FINE','FINE_MODE_FRAC_500nm','AOD_550nm','AOD_675nm','AOD_870nm','AOD_1020nm','AE_440-870nm'],
    'instruments':{
        'Cimel CE318-1':{                               'instrument_further_details':'No manual or specifications available but confirmation it is uses photometry - direct here: https://aeronet.gsfc.nasa.gov/new_web/system_descriptions_instrument.html'}, 
        'Cimel CE318-2':{                               'instrument_further_details':'No manual or specifications available but confirmation it is uses photometry - direct here: https://aeronet.gsfc.nasa.gov/new_web/system_descriptions_instrument.html'},
        'Cimel CE318-N':{'documented_precision':'0.1%', 'instrument_manual_name':'Cimel_CE318N_Manual.pdf'},
        'Cimel CE318-T':{'documented_precision':'0.1%', 'instrument_manual_name':'Cimel_CE318T_Manual.pdf'}
    }
},

#------------------------------------
#photometry - sky (P-S)

#In addition to the direct solar irradiance measurements remote photometric instruments makes with a field of view of 1.2 degrees, these instruments measure the sky radiance in four spectral bands (440, 670, 870 and 1020 nm) along the solar principal plane (i.e., at constant azimuth angle, with varied scattering angles) up to nine times a day and along the solar almucantar (i.e., at constant elevation angle, with varied azimuth angles) up to six times a day. 
#The approach is to acquire aureole and sky radiances observations through a large range of scattering angles from the sun through a constant aerosol profile to retrieve size distribution, phase function and aerosol optical depth. 
#More than eight almucantar sequences are made daily at an optical air mass of 4, 3, 2 and 1.7 both morning and afternoon. Sky radiance measurements are inverted with the Dubovik and Nakajima inversions to provide aerosol properties of size distribution and phase function over the particle size range of 0.1 to 5 um.
#The robot mounted sensor head is parked pointed near-nadir when idle to prevent contamination of the optical windows from rain and foreign particles.

'photometry - sky':{'abbreviation':'P-S', 'sampling_type':'remote', 'measured_parameters':['EXTAOD_440nm','EXTAOD_440nm_COARSE','EXTAOD_440nm_FINE','EXTAOD_675nm','EXTAOD_675nm_COARSE','EXTAOD_675nm_FINE','EXTAOD_870nm','EXTAOD_870nm_COARSE','EXTAOD_870nm_FINE','EXTAOD_1020nm','EXTAOD_1020nm_COARSE','EXTAOD_1020nm_FINE','EXTAE_440-870nm','ABSAOD_440nm','ABSAOD_675nm','ABSAOD_870nm','ABSAOD_1020nm','ABSAE_440-870nm','SSA_440nm','SSA_675nm','SSA_870nm','SSA_1020nm','ASY_440nm','ASY_440nm_COARSE','ASY_440nm_FINE','ASY_675nm','ASY_675nm_COARSE','ASY_675nm_FINE','ASY_870nm','ASY_870nm_COARSE','ASY_870nm_FINE','ASY_1020nm','ASY_1020nm_COARSE','ASY_1020nm_FINE','SPH','RIN_REAL_440nm','RIN_REAL_675nm','RIN_REAL_870nm','RIN_REAL_1020nm','RIN_IMAG_440nm','RIN_IMAG_675nm','RIN_IMAG_870nm','RIN_IMAG_1020nm','VCONC','VCONC_FINE','VCONC_COARSE','VCONC_BIN1','VCONC_BIN2','VCONC_BIN3','VCONC_BIN4','VCONC_BIN5','VCONC_BIN6','VCONC_BIN7','VCONC_BIN8','VCONC_BIN9','VCONC_BIN10','VCONC_BIN11','VCONC_BIN12','VCONC_BIN13','VCONC_BIN14','VCONC_BIN15','VCONC_BIN16','VCONC_BIN17','VCONC_BIN18','VCONC_BIN19','VCONC_BIN20','VCONC_BIN21','VCONC_BIN22'], 'qa_accepted_parameters':['EXTAOD_440nm','EXTAOD_440nm_COARSE','EXTAOD_440nm_FINE','EXTAOD_675nm','EXTAOD_675nm_COARSE','EXTAOD_675nm_FINE','EXTAOD_870nm','EXTAOD_870nm_COARSE','EXTAOD_870nm_FINE','EXTAOD_1020nm','EXTAOD_1020nm_COARSE','EXTAOD_1020nm_FINE','EXTAE_440-870nm','ABSAOD_440nm','ABSAOD_675nm','ABSAOD_870nm','ABSAOD_1020nm','ABSAE_440-870nm','SSA_440nm','SSA_675nm','SSA_870nm','SSA_1020nm','ASY_440nm','ASY_440nm_COARSE','ASY_440nm_FINE','ASY_675nm','ASY_675nm_COARSE','ASY_675nm_FINE','ASY_870nm','ASY_870nm_COARSE','ASY_870nm_FINE','ASY_1020nm','ASY_1020nm_COARSE','ASY_1020nm_FINE','SPH','RIN_REAL_440nm','RIN_REAL_675nm','RIN_REAL_870nm','RIN_REAL_1020nm','RIN_IMAG_440nm','RIN_IMAG_675nm','RIN_IMAG_870nm','RIN_IMAG_1020nm','VCONC','VCONC_FINE','VCONC_COARSE','VCONC_BIN1','VCONC_BIN2','VCONC_BIN3','VCONC_BIN4','VCONC_BIN5','VCONC_BIN6','VCONC_BIN7','VCONC_BIN8','VCONC_BIN9','VCONC_BIN10','VCONC_BIN11','VCONC_BIN12','VCONC_BIN13','VCONC_BIN14','VCONC_BIN15','VCONC_BIN16','VCONC_BIN17','VCONC_BIN18','VCONC_BIN19','VCONC_BIN20','VCONC_BIN21','VCONC_BIN22'],
    'instruments':{
        'Cimel CE318-1':{                               'instrument_further_details':'No manual or specifications available but confirmation it is uses photometry - direct here: https://aeronet.gsfc.nasa.gov/new_web/system_descriptions_instrument.html'}, 
        'Cimel CE318-2':{                               'instrument_further_details':'No manual or specifications available but confirmation it is uses photometry - direct here: https://aeronet.gsfc.nasa.gov/new_web/system_descriptions_instrument.html'},
        'Cimel CE318-N':{'documented_precision':'0.1%', 'instrument_manual_name':'Cimel_CE318N_Manual.pdf'},
        'Cimel CE318-T':{'documented_precision':'0.1%', 'instrument_manual_name':'Cimel_CE318T_Manual.pdf'}
    }
},

#------------------------------------
#unknown methodology

'unknown':{'abbreviation':'UNK', 'measured_parameters':[], 'qa_accepted_parameters':[], 
    'instruments':{}
}
}

###--------------------------------------------------------------------------------------------------###
#DEFINE STANDARD COUNTRY DEFINITION INFORMATION
###--------------------------------------------------------------------------------------------------###

#dictionary that stores country ISO3166-1 codes with associated country names

country_ISO3166_to_name = {
'AF': 'Afghanistan',
'AX': 'Åland Islands',
'AL': 'Albania',
'DZ': 'Algeria',
'AS': 'American Samoa',
'AD': 'Andorra',
'AO': 'Angola',
'AI': 'Anguilla',
'AQ': 'Antarctica',
'AG': 'Antigua and Barbuda',
'AR': 'Argentina',
'AM': 'Armenia',
'AW': 'Aruba',
'AU': 'Australia',
'AT': 'Austria',
'AZ': 'Azerbaijan',
'BS': 'Bahamas',
'BH': 'Bahrain',
'BD': 'Bangladesh',
'BB': 'Barbados',
'BY': 'Belarus',
'BE': 'Belgium',
'BZ': 'Belize',
'BJ': 'Benin',
'BM': 'Bermuda',
'BT': 'Bhutan',
'BO': 'Bolivia',
'BQ': 'Bonaire, Sint Eustatius and Saba',
'BA': 'Bosnia and Herzegovina',
'BW': 'Botswana',
'BV': 'Bouvet Island',
'BR': 'Brazil',
'IO': 'British Indian Ocean Territory',
'BN': 'Brunei',
'BG': 'Bulgaria',
'BF': 'Burkina Faso',
'BI': 'Burundi',
'KH': 'Cambodia',
'CM': 'Cameroon',
'CA': 'Canada',
'CV': 'Cape Verde',
'KY': 'Cayman Islands',
'CF': 'Central African Republic',
'TD': 'Chad',
'CL': 'Chile',
'CN': 'China',
'CX': 'Christmas Island',
'CC': 'Cocos Islands',
'CO': 'Colombia',
'KM': 'Comoros',
'CG': 'Republic of the Congo',
'CD': 'Democratic Republic of the Congo',
'CK': 'Cook Islands',
'CR': 'Costa Rica',
'CI': "Côte d'Ivoire",
'HR': 'Croatia',
'CU': 'Cuba',
'CW': 'Curaçao',
'CY': 'Cyprus',
'CZ': 'Czech Republic',
'DK': 'Denmark',
'DJ': 'Djibouti',
'DM': 'Dominica',
'DO': 'Dominican Republic',
'EC': 'Ecuador',
'EG': 'Egypt',
'SV': 'El Salvador',
'GQ': 'Equatorial Guinea',
'ER': 'Eritrea',
'EE': 'Estonia',
'ET': 'Ethiopia',
'FK': 'Falkland Islands',
'FO': 'Faroe Islands',
'FJ': 'Fiji',
'FI': 'Finland',
'FR': 'France',
'GF': 'French Guiana',
'PF': 'French Polynesia',
'TF': 'French Southern Territories',
'GA': 'Gabon',
'GM': 'Gambia',
'GE': 'Georgia',
'DE': 'Germany',
'GH': 'Ghana',
'GI': 'Gibraltar',
'GR': 'Greece',
'GL': 'Greenland',
'GD': 'Grenada',
'GP': 'Guadeloupe',
'GU': 'Guam',
'GT': 'Guatemala',
'GG': 'Guernsey',
'GN': 'Guinea',
'GW': 'Guinea-Bissau',
'GY': 'Guyana',
'HT': 'Haiti',
'HM': 'Heard Island and McDonald Islands',
'VA': 'Vatican',
'HN': 'Honduras',
'HK': 'Hong Kong',
'HU': 'Hungary',
'IS': 'Iceland',
'IN': 'India',
'ID': 'Indonesia',
'IR': 'Iran',
'IQ': 'Iraq',
'IE': 'Ireland',
'IM': 'Isle of Man',
'IL': 'Israel',
'IT': 'Italy',
'JM': 'Jamaica',
'JP': 'Japan',
'JE': 'Jersey',
'JO': 'Jordan',
'KZ': 'Kazakhstan',
'KE': 'Kenya',
'KI': 'Kiribati',
'KP': 'North Korea',
'KR': 'South Korea',
'KW': 'Kuwait',
'KG': 'Kyrgyzstan',
'LA': "Lao People's Democratic Republic",
'LV': 'Latvia',
'LB': 'Lebanon',
'LS': 'Lesotho',
'LR': 'Liberia',
'LY': 'Libya',
'LI': 'Liechtenstein',
'LT': 'Lithuania',
'LU': 'Luxembourg',
'MO': 'Macao',
'MK': 'North Macedonia',
'MG': 'Madagascar',
'MW': 'Malawi',
'MY': 'Malaysia',
'MV': 'Maldives',
'ML': 'Mali',
'MT': 'Malta',
'MH': 'Marshall Islands',
'MQ': 'Martinique',
'MR': 'Mauritania',
'MU': 'Mauritius',
'YT': 'Mayotte',
'MX': 'Mexico',
'FM': 'Micronesia',
'MD': 'Moldova',
'MC': 'Monaco',
'MN': 'Mongolia',
'ME': 'Montenegro',
'MS': 'Montserrat',
'MA': 'Morocco',
'MZ': 'Mozambique',
'MM': 'Myanmar',
'NA': 'Namibia',
'NR': 'Nauru',
'NP': 'Nepal',
'NL': 'Netherlands',
'NC': 'New Caledonia',
'NZ': 'New Zealand',
'NI': 'Nicaragua',
'NE': 'Niger',
'NG': 'Nigeria',
'NU': 'Niue',
'NF': 'Norfolk Island',
'MP': 'Northern Mariana Islands',
'NO': 'Norway',
'OM': 'Oman',
'PK': 'Pakistan',
'PW': 'Palau',
'PS': 'Palestinian Territory',
'PA': 'Panama',
'PG': 'Papua New Guinea',
'PY': 'Paraguay',
'PE': 'Peru',
'PH': 'Philippines',
'PN': 'Pitcairn',
'PL': 'Poland',
'PT': 'Portugal',
'PR': 'Puerto Rico',
'QA': 'Qatar',
'RE': 'Réunion',
'RO': 'Romania',
'RU': 'Russia',
'RW': 'Rwanda',
'BL': 'Saint Barthélemy',
'SH': 'Saint Helena, Ascension and Tristan da Cunha',
'KN': 'Saint Kitts and Nevis',
'LC': 'Saint Lucia',
'MF': 'Saint Martin',
'PM': 'Saint Pierre and Miquelon',
'VC': 'Saint Vincent and the Grenadines',
'WS': 'Samoa',
'SM': 'San Marino',
'ST': 'Sao Tome and Principe',
'SA': 'Saudi Arabia',
'SN': 'Senegal',
'RS': 'Serbia',
'SC': 'Seychelles',
'SL': 'Sierra Leone',
'SG': 'Singapore',
'SX': 'Sint Maarten',
'SK': 'Slovakia',
'SI': 'Slovenia',
'SB': 'Solomon Islands',
'SO': 'Somalia',
'ZA': 'South Africa',
'GS': 'South Georgia and the South Sandwich Islands',
'SS': 'South Sudan',
'ES': 'Spain',
'LK': 'Sri Lanka',
'SD': 'Sudan',
'SR': 'Suriname',
'SJ': 'Svalbard and Jan Mayen',
'SZ': 'Swaziland',
'SE': 'Sweden',
'CH': 'Switzerland',
'SY': 'Syria',
'TW': 'Taiwan',
'TJ': 'Tajikistan',
'TZ': 'Tanzania',
'TH': 'Thailand',
'TL': 'Timor-Leste',
'TG': 'Togo',
'TK': 'Tokelau',
'TO': 'Tonga',
'TT': 'Trinidad and Tobago',
'TN': 'Tunisia',
'TR': 'Turkey',
'TM': 'Turkmenistan',
'TC': 'Turks and Caicos Islands',
'TV': 'Tuvalu',
'UG': 'Uganda',
'UA': 'Ukraine',
'AE': 'United Arab Emirates',
'GB': 'United Kingdom',
'US': 'United States',
'UM': 'United States Minor Outlying Islands',
'UY': 'Uruguay',
'UZ': 'Uzbekistan',
'VU': 'Vanuatu',
'VE': 'Venezuela',
'VN': 'Vietnam',
'VG': 'British Virgin Islands',
'VI': 'U.S. Virgin Islands',
'WF': 'Wallis and Futuna',
'EH': 'Western Sahara',
'XK': 'Kosovo',
'YE': 'Yemen',
'ZM': 'Zambia',
'ZW': 'Zimbabwe',
'CS': 'Serbia and Montenegro'
}

#dictionary that stores country FIPS 10-4 codes with associated country names

country_FIPS104_to_name = {
'AN': 'Andorra',
'AE': 'United Arab Emirates',
'AF': 'Afghanistan',
'AC': 'Antigua and Barbuda',
'AV': 'Anguilla',
'AL': 'Albania',
'AM': 'Armenia',
'AO': 'Angola',
'AY': 'Antarctica',
'AR': 'Argentina',
'AQ': 'American Samoa',
'AU': 'Austria',
'AS': 'Australia',
'AA': 'Aruba',
'AJ': 'Azerbaijan',
'BK': 'Bosnia and Herzegovina',
'BB': 'Barbados',
'BG': 'Bangladesh',
'BE': 'Belgium',
'UV': 'Burkina Faso',
'BU': 'Bulgaria',
'BA': 'Bahrain',
'BY': 'Burundi',
'BN': 'Benin',
'TB': 'Saint Barthélemy',
'BD': 'Bermuda',
'BX': 'Brunei',
'BL': 'Bolivia',
'BR': 'Brazil',
'BF': 'Bahamas',
'BT': 'Bhutan',
'BV': 'Bouvet Island',
'BC': 'Botswana',
'BO': 'Belarus',
'BH': 'Belize',
'CA': 'Canada',
'CH': 'China',
'CK': 'Cocos Islands',
'CG': 'Democratic',
'CT': 'Central African Republic',
'CF': 'Republic of the Congo',
'CG': 'Democratic Republic of the Congo',
'SZ': 'Switzerland',
'IV': "Côte d'Ivoire",
'CW': 'Cook Islands',
'CI': 'Chile',
'CM': 'Cameroon',
'CO': 'Colombia',
'CS': 'Costa Rica',
'CU': 'Cuba',
'CV': 'Cape Verde',
'UC': 'Curaçao',
'KT': 'Christmas Island',
'CY': 'Cyprus',
'EZ': 'Czech Republic',
'GM': 'Germany',
'DJ': 'Djibouti',
'DA': 'Denmark',
'DO': 'Dominica',
'DR': 'Dominican Republic',
'AG': 'Algeria',
'EC': 'Ecuador',
'EN': 'Estonia',
'EG': 'Egypt',
'WI': 'Western Sahara',
'ER': 'Eritrea',
'SP': 'Spain',
'ET': 'Ethiopia',
'FI': 'Finland',
'FG': 'French Guiana',
'FJ': 'Fiji',
'FK': 'Falkland Islands',
'FM': 'Micronesia',
'FO': 'Faroe Islands',
'FR': 'France',
'FS': 'French Southern Territories',
'GB': 'Gabon',
'UK': 'United Kingdom',
'GJ': 'Grenada',
'GG': 'Georgia',
'FG': 'French',
'GK': 'Guernsey',
'GH': 'Ghana',
'GI': 'Gibraltar',
'GL': 'Greenland',
'GA': 'Gambia',
'GV': 'Guinea',
'GP': 'Guadeloupe',
'EK': 'Equatorial Guinea',
'GR': 'Greece',
'SX': 'South Georgia and the South Sandwich Islands',
'GT': 'Guatemala',
'GQ': 'Guam',
'PU': 'Guinea-Bissau',
'GY': 'Guyana',
'HK': 'Hong Kong',
'HM': 'Heard Island and McDonald Islands',
'HO': 'Honduras',
'HR': 'Croatia',
'HA': 'Haiti',
'HU': 'Hungary',
'ID': 'Indonesia',
'EI': 'Ireland',
'IS': 'Israel',
'IM': 'Isle of Man',
'IN': 'India',
'IO': 'British Indian Ocean Territory',
'IZ': 'Iraq',
'IR': 'Iran',
'IC': 'Iceland',
'IT': 'Italy',
'JE': 'Jersey',
'JM': 'Jamaica',
'JO': 'Jordan',
'JA': 'Japan',
'KE': 'Kenya',
'KG': 'Kyrgyzstan',
'CB': 'Cambodia',
'KR': 'Kiribati',
'CN': 'Comoros',
'SC': 'Saint Kitts and Nevis',
'KN': 'North Korea',
'KS': 'South Korea',
'KV': 'Kosovo',
'KU': 'Kuwait',
'CJ': 'Cayman Islands',
'KZ': 'Kazakhstan',
'LA': 'Laos',
'LE': 'Lebanon',
'ST': 'Saint Lucia',
'LS': 'Liechtenstein',
'CE': 'Sri Lanka',
'LI': 'Liberia',
'LT': 'Lesotho',
'LH': 'Lithuania',
'LU': 'Luxembourg',
'LG': 'Latvia',
'LY': 'Libya',
'MO': 'Morocco',
'MN': 'Monaco',
'MD': 'Moldova',
'MJ': 'Montenegro',
'RN': 'Saint Martin',
'MA': 'Madagascar',
'RM': 'Marshall Islands',
'MK': 'North Macedonia',
'ML': 'Mali',
'BM': 'Myanmar',
'MG': 'Mongolia',
'MC': 'Macao',
'CQ': 'Northern Mariana Islands',
'MB': 'Martinique',
'MR': 'Mauritania',
'MH': 'Montserrat',
'MT': 'Malta',
'MP': 'Mauritius',
'MV': 'Maldives',
'MI': 'Malawi',
'MX': 'Mexico',
'MY': 'Malaysia',
'MZ': 'Mozambique',
'WA': 'Namibia',
'NC': 'New Caledonia',
'NG': 'Niger',
'NF': 'Norfolk Island',
'NI': 'Nigeria',
'NU': 'Nicaragua',
'NL': 'Netherlands',
'NO': 'Norway',
'NP': 'Nepal',
'NR': 'Nauru',
'NE': 'Niue',
'NZ': 'New Zealand',
'MU': 'Oman',
'PM': 'Panama',
'PE': 'Peru',
'FP': 'French Polynesia',
'PP': 'Papua New Guinea',
'RP': 'Philippines',
'PK': 'Pakistan',
'PL': 'Poland',
'SB': 'Saint Pierre and Miquelon',
'PC': 'Pitcairn',
'RQ': 'Puerto Rico',
'WE': 'Palestinian Territory',
'PO': 'Portugal',
'PS': 'Palau',
'PA': 'Paraguay',
'QA': 'Qatar',
'RE': 'Réunion',
'RO': 'Romania',
'RI': 'Serbia',
'RS': 'Russia',
'RW': 'Rwanda',
'SA': 'Saudi Arabia',
'BP': 'Solomon Islands',
'SE': 'Seychelles',
'SU': 'Sudan',
'OD': 'South Sudan',
'SW': 'Sweden',
'SN': 'Singapore',
'SH': 'Saint Helena, Ascension and Tristan da Cunha',
'SI': 'Slovenia',
'SV': 'Svalbard and Jan Mayen',
'LO': 'Slovakia',
'SL': 'Sierra Leone',
'SM': 'San Marino',
'SG': 'Senegal',
'SO': 'Somalia',
'NS': 'Suriname',
'TP': 'Sao Tome and Principe',
'ES': 'El Salvador',
'NN': 'Sint Maarten',
'SY': 'Syria',
'WZ': 'Swaziland',
'TK': 'Turks and Caicos Islands',
'CD': 'Chad',
'FS': 'French',
'TO': 'Togo',
'TH': 'Thailand',
'TI': 'Tajikistan',
'TL': 'Tokelau',
'TT': 'Timor-Leste',
'TX': 'Turkmenistan',
'TS': 'Tunisia',
'TN': 'Tonga',
'TU': 'Turkey',
'TD': 'Trinidad and Tobago',
'TV': 'Tuvalu',
'TW': 'Taiwan',
'TZ': 'Tanzania',
'UP': 'Ukraine',
'UG': 'Uganda',
'UM': 'United States Minor Outlying Islands',
'US': 'United States',
'UY': 'Uruguay',
'UZ': 'Uzbekistan',
'VT': 'Vatican',
'VC': 'Saint Vincent and the Grenadines',
'VE': 'Venezuela',
'VI': 'British Virgin Islands',
'VQ': 'U.S. Virgin Islands',
'VM': 'Vietnam',
'NH': 'Vanuatu',
'WF': 'Wallis and Futuna',
'WS': 'Samoa',
'YM': 'Yemen',
'MF': 'Mayotte',
'SF': 'South Africa',
'ZA': 'Zambia',
'ZI': 'Zimbabwe',
'YI': 'Serbia and Montenegro'}
###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###
